/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "../status.h"
#include "../utils/iutils.h"
#include "commlink.h"
#include "serialport.h"
#include "../utils/xml/xml.h"

#include "bluetooth.h"

const wxString Bluetooth::ItemDesc = "Bluetooth";

/* Set to 1 to display every Bluetooth packet that are transmited or received. */
#define DISPLAY_BLUETOOTH_PACKETS 0

Bluetooth::Bluetooth() {
	mSerialPort = NULL;
	mPort = "COM1";
	mBaudRate = 57600;
	mConnected = false;
}

Bluetooth::~Bluetooth() {
	if(mConnected) Disconnect();
	if (mSerialPort) delete mSerialPort;
}

wxInt8 Bluetooth::Connect() {
	if(mConnected) return 0; // Already connected..
	wxInt8 ret;

	// Make sure to disconnect the serial port, to flush
	// and reset upon connection.
	if (mSerialPort) {
		mSerialPort->Disconnect(); 
		delete mSerialPort;
	}

	mSerialPort = new SerialPort(1, mBaudRate, mPort);
	
	if((ret = mSerialPort->Connect())) {
		StatusLineErr("Bluetooth - Couldn't connect to serial port.", ret);
		return ret;
	}

 	// RESET
	mBuffer[0] = 0x01;	// command packet designator
	mBuffer[1] = 0x03;	// RESET command
	mBuffer[2] = 0x0C;
	mBuffer[3] = 0x00;	// Parameters total length

	if((ret = mSerialPort->SendBuffer(mBuffer, 4))) {
		StatusLine("Bluetooth - Reset host (SendBuffer)");
		mSerialPort->Disconnect();
		return ret;
	}
	if((ret = ReceivePacket()))	{ // Wait for response
		StatusLine("Bluetooth - Reset host (ReceivePacket)");
		mSerialPort->Disconnect();
		return ret;
	}

	// SET_EVENT_MASK
	mBuffer[0]  = 0x01;	// command packet designator
	mBuffer[1]  = 0x01;	// SET_EVENT_MASK command
	mBuffer[2]  = 0x0C;
	mBuffer[3]  = 0x08;	// Parameters total length
	// Mask all event except the ones that are indicated
	mBuffer[4]  = 0x17; // Inquiry Complete, Inquiry Result, Connection_Complete, Disconnection_Complete
	mBuffer[5]  = 0x60;	// Command Complete and Command Status event (unmaskable)
	mBuffer[6]  = 0x04;	// Number Of Completed Packets event (unmaskable)
	mBuffer[7]  = 0x00;
	mBuffer[8]  = 0x00;
	mBuffer[9]  = 0x00;
	mBuffer[10] = 0x00;
	mBuffer[11] = 0x00;

	if((ret = mSerialPort->SendBuffer(mBuffer, 12))) {
		StatusLine("Bluetooth - Set event mask (SendBuffer)");
		mSerialPort->Disconnect();
		return ret;
	}
	if((ret = ReceivePacket()))	{ // Wait for response
		StatusLine("Bluetooth - Set event mask (ReceivePacket)");
		mSerialPort->Disconnect();
		return ret;
	}

	// Send Read BD Address command.
	mBuffer[0] = 0x01; // HCI command header
	mBuffer[1] = 0x09; // READ_BD_ADDRESS command
	mBuffer[2] = 0x10;
	mBuffer[3] = 0x00; // HCI parameters Length

	if((ret = mSerialPort->SendBuffer(mBuffer, 4))) {
		StatusLine("Bluetooth - Read local BD Address (SendBuffer)");
		mSerialPort->Disconnect();
		return ret;
	}
	if((ret = ReceivePacket()))	{ // Wait for response
		StatusLine("Bluetooth - Read local BD Address (ReceivePacket)");
		mSerialPort->Disconnect();
		return ret;
	}

	if (mBuffer[11] == 0x80) mROK = true;		// ROK BD Addresses are always 00:80:37:xx:xx:xx
	else mROK = false;
	mROK ? StatusLine("Bluetooth - Local BT module is a ROK") : StatusLine("Bluetooth - Local BT module is NOT a ROK");
	
	// INQUIRY (read the address of the first BT module responding
	if ((ret=Inquiry())) {
		StatusLine("Bluetooth - Inquiry failed");
		mSerialPort->Disconnect();
		return ret;
	}

	// CONNECT
	mBuffer[0]  = 0x01;	// command packet designator
	mBuffer[1]  = 0x05;	// CREATE_CONNECTION command for module
	mBuffer[2]  = 0x04;
	mBuffer[3]  = 0x0D;	// Parameters total length
	mBuffer[4]  = mModuleAddress[0];	// BD_ADDR to connect to.
	mBuffer[5]  = mModuleAddress[1];
	mBuffer[6]  = mModuleAddress[2];
	mBuffer[7]  = mModuleAddress[3];
	mBuffer[8]  = mModuleAddress[4];
	mBuffer[9]  = mModuleAddress[5];
	mBuffer[10] = 0x00;	// DM3 packets (123Bytes)
	mBuffer[11] = 0x04;
	mBuffer[12] = 0x00;	// Page scan repetition mode?
	mBuffer[13] = 0x00;	// Page scan mode (mandatory)
	mBuffer[14] = 0x00;	// 16 bit clock offset
	mBuffer[15] = 0x00;
	mBuffer[16] = 0x00;	// Do not accept role switch

	if((ret = mSerialPort->SendBuffer(mBuffer, 17))) {
		StatusLine("Bluetooth - Connect (SendBuffer)");
		mSerialPort->Disconnect();
		return ret;
	}
 	
	// Waiting for connection complete event (a different number of events arise depending on the local BT module)
	mBuffer[1] = 0;
	while (mBuffer[1] != 0x03) {	
		if ((ret = ReceivePacket())) {
			mSerialPort->Disconnect();
			return ret;
		}
	}
	if(mBuffer[3] != 0x00)	{ // Check if connection successfull
		StatusLine("Connection didn't succeed!");
		mSerialPort->Disconnect();
		return 4;
	}

	// Store the connection handle
	mConnectionHandleLow  = mBuffer[4];		
	mConnectionHandleHigh = mBuffer[5] & 0x0F;

#if DISPLAY_BLUETOOTH_PACKETS
	StatusLine(wxString::Format("Connection handle: 0x%02X%02X", mConnectionHandleHigh, mConnectionHandleLow));
#endif

	// successfully connected
	StatusLine("Bluetooth connection established.");
	mConnected = true;
		
	return 0;
}

wxInt8 Bluetooth::Disconnect() {
	if(!mConnected) return 0;
	mConnected = false;

	wxInt8 ret;

	mBuffer[0] = 0x01;	// Command packet designator
	mBuffer[1] = 0x06;	// DISCONNECT command for module
	mBuffer[2] = 0x04;
	mBuffer[3] = 0x03;	// Parameters total length
	mBuffer[4] = mConnectionHandleLow;
	mBuffer[5] = mConnectionHandleHigh;
	mBuffer[6] = 0x13;	// Disconnection reason: user

	mSerialPort->SendBuffer(mBuffer, 7);

	if((ret = ReceivePacket()))	// Wait for command status
		return ret;
	if((ret = ReceivePacket()))	// Wait for disconnection complete
		return ret;
	mSerialPort->Disconnect();
	return 0;
}

wxInt8 Bluetooth::SendBuffer(const wxUint8 *buffer, const wxUint16 length) {
	mBuffer[0] = 0x02;	//  DATA packet designator
	mBuffer[1] = mConnectionHandleLow;
	mBuffer[2] = mConnectionHandleHigh | 0x10;	// Add flags (BUG !? previously 0x20)
	mBuffer[3] = (wxByte)length;				// Data length low
	mBuffer[4] = 0x00;							// Data length high
	for(wxUint16 i=0; i<length; i++)
		mBuffer[5+i] = buffer[i];

	if(mSerialPort->SendBuffer(mBuffer, 5 + length))
		return 1;
	
	do {
		if (ReceivePacket()) {
			StatusLine("Bluetooth Send Buffer - Could not receive confirmation.");
			return 1;
		} 
		if (mBuffer[0] == 0x02) {
			StatusLine("Bluetooth Send Buffer - Unexpected data packet.");
			return 1;
		}
	   
		if (mBuffer[1] == 0x05) {	// It's a disconnection !
			StatusLine("Bluetooth Send Buffer - Disconnected.");
			return 3;
		}
	   
		if (mBuffer[0] == 0x04 && mBuffer[1] == 0x13) {	// NUMBER_OF_COMPLETED_PACKET_EVENT
			if (mBuffer[3] != 1 || mBuffer[4] != mConnectionHandleLow || mBuffer[5] != mConnectionHandleHigh || mBuffer[6] != 1 || mBuffer[7] != 0) {
                 StatusLine(wxString::Format("Bluetooth Send Buffer - Confirmation received - Bad results. # = %d. Conn = %X %X. #Res = %d.", mBuffer[3], mBuffer[4], mBuffer[5], mBuffer[6]));
                   return 1;               
			} else {
#if DISPLAY_BLUETOOTH_PACKETS
				 StatusLine("Bluetooth Send Buffer - Data successfully transmitted.");
#endif
                 break;
			}
		}
	} while(true);

	return 0;
}

wxInt8 Bluetooth::SendString(const wxString& string) {
	return 1;
}

wxInt8 Bluetooth::ReceiveBuffer(wxUint8 *buffer, const wxUint16 length) {
	wxInt8 ret;
	wxUint16 i=0, packetLength, recvLength=0;

	// It may happen (typically with ROK modules) that long data packets are subdivided into several smaller packets..
	while (recvLength < length) {
		mBuffer[0]=0x00;
		while(mBuffer[0] != 0x02) { // wait for data packet
			if((ret = ReceivePacket()))
				return ret;
		}
		packetLength = mBuffer[3] + (wxUint16)mBuffer[4] * 0xFF; 
		for(i=0; i<packetLength; i++)
			buffer[i+recvLength] = mBuffer[5+i];
		recvLength += packetLength;
		if (mROK) {	// this is the hack for ROK modules (OSE Crash Event)
			mBuffer[0]=0x00;
			while(mBuffer[0] != 0x04) { // wait for Crash Event
				if((ret = ReceivePacket()))
					return ret;
			}
		}
	}

	if(recvLength < length) {
		StatusLine(wxString::Format("Bluetooth - Expected %d bytes, got only %d!", length, recvLength));
		return -1;
	}

	return 0;
}

wxInt8 Bluetooth::ReceiveString(wxString& string) {
	return 1;
}

wxInt8 Bluetooth::ReceivePacket() {
#if DISPLAY_BLUETOOTH_PACKETS
	wxString s;
	wxUint16 i, length;
#endif
	if(mSerialPort->ReceiveBuffer(mBuffer, 1))
		return 1;
	
	// The following is not very clean, but necessary to deal with
	// obsolete bytes sent by Ericsson module.
	while(mBuffer[0] != 0x04 && mBuffer[0] != 0x02)
	{
#if DISPLAY_BLUETOOTH_PACKETS
	    StatusLine("Bluetooth ReceivePacket - Received obsolete byte.");
#endif
		if(mSerialPort->ReceiveBuffer(mBuffer, 1))
			return 2;
	}			

	if(mBuffer[0] == 0x04) {	// EVENT (0x04)
		// Read event code and total length of parameters
		if (mSerialPort->ReceiveBuffer(mBuffer+1, 2))
			return 3;
		// Read rest of packet depending on length in buffer[2]
		if (mSerialPort->ReceiveBuffer(mBuffer+3, mBuffer[2]))
			return 4;
#if DISPLAY_BLUETOOTH_PACKETS
		length = 3 + mBuffer[2];
		s = "Bluetooth received event: ";
		for (i=0; i<length; i++)
			s += wxString::Format("%02X ", mBuffer[i]);
		StatusLine(s);
#endif
	} else { 					// DATA (0x02)
		// Read connection handle and total length of parameters
		if(mSerialPort->ReceiveBuffer(mBuffer+1, 4))
			return 5;
		// Read rest of packet depending on length in buffer[3]
		if(mSerialPort->ReceiveBuffer(mBuffer+5, mBuffer[3]))
			return 6;
#if DISPLAY_BLUETOOTH_PACKETS
		length = 5 + mBuffer[3] + (wxUint16)mBuffer[4]*0xFF;
		s = "Bluetooth received data: ";
		for (i=0; i<length; i++)
			s += wxString::Format("%02X ", mBuffer[i]);
		StatusLine(s);
#endif
	}
	return 0;
}

wxInt8 Bluetooth::Inquiry() {
	wxInt8 ret;

	mBuffer[0] =0x01;	// HCI command packet designator
	mBuffer[1] =0x01;	// INQUIRY command for module
	mBuffer[2] =0x04;
	mBuffer[3] =0x05;	// Parameters total length
	mBuffer[4] =0x33;	// LAP access code (assigned)
	mBuffer[5] =0x8B;
	mBuffer[6] =0x9E;
	mBuffer[7] =0x0A;	// Max inquiry time: 13 seconds
	mBuffer[8] =0x01;	// Max number of responses

	mSerialPort->SendBuffer(mBuffer, 9);

	if((ret = ReceivePacket()))	{ // Wait for command status
		StatusLine("Bluetooth - No command status event after inquiry request.");
		return ret;
	}

	while((ret = ReceivePacket())) {	// Waits for inquiry results or inquiry completed
    	iSleep(100);
	}

	if(mBuffer[1]==0x02) {	// Read the inquiry result
		mModuleAddress[0] = mBuffer[4];	// Copy BD_ADDR of module responding
		mModuleAddress[1] = mBuffer[5];
		mModuleAddress[2] = mBuffer[6];
		mModuleAddress[3] = mBuffer[7];
		mModuleAddress[4] = mBuffer[8];
		mModuleAddress[5] = mBuffer[9];
		if(ReceivePacket())	{	// Wait for inquiry complete
			StatusLine("Bluetooth - No inquiry complete event.");
			return 1;
		}
		StatusLine("Got inquiry result!");
		StatusLine(wxString::Format("Module address: %02X:%02X:%02X:%02X:%02X:%02X", mModuleAddress[5], mModuleAddress[4], mModuleAddress[3], mModuleAddress[2], mModuleAddress[1], mModuleAddress[0]));
	} else {
		StatusLine("Bluetooth - Inquiry completed but no result.");
		return 2;	// Inquiry completed but no result
	}
	return 0;
}

void Bluetooth::Configure() {
	wxString portChoices[] = { "COM1", "COM2", "COM3", "COM4" };
	wxString result = wxGetSingleChoice("Please select the appropiate COM port.", "Configure BT serial port", 4, portChoices);
	if(!result.IsEmpty()) {
#ifdef WIN32
		mPort = result;
#else
		if(result == "COM1")
			mPort = "/dev/cua0";
		else if(result == "COM2")
			mPort = "/dev/cua1";
		else if(result == "COM3")
			mPort = "/dev/cua2";
		else if(result == "COM4")
			mPort = "/dev/cua3";
#endif
	}

	wxString choices[] = { "57600", "115200" };
	result = wxGetSingleChoice("Please select the appropiate baudrate.", "Configure BT serial port", 2, choices);
	if(!result.IsEmpty()) {
		if(result == "57600")
			mBaudRate = 57600;
		else if(result == "115200")
			mBaudRate = 115200;
	}
	
	StatusLine(wxString::Format("Bluetooth SerialPort - COM port set to %s.", mPort.c_str()));
	StatusLine(wxString::Format("Bluetooth SerialPort - Baudrate set to %d.", mBaudRate));
}

/* --- XML --- */
void Bluetooth::Load(XMLElement *myself, XMLLoader *loader) {
	mPort = (myself->GetAttribute("port")->GetValue());
	mBaudRate = myself->GetAttribute("baudrate")->GetIntValue();
}

XMLElement* Bluetooth::SaveSelf() {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
	result->AddAttribute(new XMLAttribute("port",mPort));
	result->AddAttribute(new XMLAttribute("baudrate",(wxInt32)mBaudRate));
	return result;
}

wxString Bluetooth::GetXMLName() {
	return GetShortXMLName();
}
