/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#ifdef WIN32
	#include <windows.h>
	#include <wx/msw/winundef.h>
#else
	#include <stdio.h>
	#include <fcntl.h>
	#include <termios.h>
	#include <unistd.h>
	#include <sys/ioctl.h>
#endif

#include "../status.h"
#include "../utils/xml/xml.h"
#include "commlink.h"

#include "serialport.h"


const wxString SerialPort::ItemDesc = "Serial Port";


SerialPort::SerialPort() {
	mBuffer = new char[SERIAL_BUFFERLEN];

#ifdef WIN32
	mPort = "COM1";
	mParity = NOPARITY;
	mByteSize = 8;

	mStopBits = TWOSTOPBITS;
#else
	mPort = "/dev/cua0";
	mStopBits = 2;

	mFileDesc = -1;
#endif

	mBaudRate = 57600;
}

SerialPort::SerialPort(wxByte stopBits, wxUint32 baudRate) {
	mBuffer = new char[SERIAL_BUFFERLEN];

#ifdef WIN32
	mPort = "COM1";
	mParity = NOPARITY;
	mByteSize = 8;

	if(stopBits == 2)
		mStopBits = TWOSTOPBITS;
	else
		mStopBits = ONESTOPBIT;
#else
	mPort = "/dev/cua0";
	mStopBits = stopBits;

	mFileDesc = -1;
#endif

	mBaudRate = baudRate;
}

SerialPort::SerialPort(wxByte stopBits, wxUint32 baudRate, wxString	port) {
	mBuffer = new char[SERIAL_BUFFERLEN];

#ifdef WIN32
	mPort = port;
	mParity = NOPARITY;
	mByteSize = 8;

	if(stopBits == 2)
		mStopBits = TWOSTOPBITS;
	else
		mStopBits = ONESTOPBIT;
#else
	if(port == "COM1")
		mPort = "/dev/cua0";
	else if(port == "COM2")
		mPort = "/dev/cua1";
	else if(port == "COM3")
		mPort = "/dev/cua2";
	else if(port == "COM4")
		mPort = "/dev/cua3";

	mStopBits = stopBits;

	mFileDesc = -1;
#endif

	mBaudRate = baudRate;
}

SerialPort::~SerialPort() {
	if(mConnected)
		Disconnect();
	delete [] mBuffer;
}

wxInt8 SerialPort::Connect() {
	if(mConnected) return 0; // Already connected..

#ifdef WIN32
	mHandle = CreateFile(mPort.c_str(),
                        GENERIC_READ | GENERIC_WRITE,
                        0,    /* comm devices must be opened w/exclusive-access */
                        NULL, /* no security attrs */
                        OPEN_EXISTING, /* comm devices must use OPEN_EXISTING */
                        0,    /* not overlapped I/O */
                        NULL  /* hTemplate must be NULL for comm devices */
                        );
	if(mHandle == INVALID_HANDLE_VALUE)
		return 1;
	// Now get the DCB properties of the port we just opened
	if(!GetCommState(mHandle,&mDcb)) {
		CloseHandle(mHandle);
		return 2;
	}

	// Fill mDcb with requested settings
	mDcb.BaudRate  =  mBaudRate;
	mDcb.ByteSize  =  mByteSize;
	mDcb.Parity    =  mParity;
	mDcb.StopBits  =  mStopBits;
	mDcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
    
	// Set the port with requested settings.
	if(!SetCommState(mHandle,&mDcb)) {
		CloseHandle(mHandle);
		return 3;
	}

	// Set the intial size of the transmit and receive queues.
	// Receive buffer to 32k, and the transmit buffer to 9k (just a default).
	if(!SetupComm(mHandle, 1024*32, 1024*9)) {
		CloseHandle(mHandle);
		return 4;
	}

	// Deliberately made rather slow to match Linux settings (which cannot be made faster).
	mTimeOuts.ReadIntervalTimeout         = 25;
	mTimeOuts.ReadTotalTimeoutMultiplier  = 0;		// 1
	mTimeOuts.ReadTotalTimeoutConstant    = 100;	// 15
	mTimeOuts.WriteTotalTimeoutMultiplier = 1;
	mTimeOuts.WriteTotalTimeoutConstant   = 250;

	// Original
/*     mTimeOuts.ReadIntervalTimeout         = 15;
    mTimeOuts.ReadTotalTimeoutMultiplier  = 1;
    mTimeOuts.ReadTotalTimeoutConstant    = 250;
    mTimeOuts.WriteTotalTimeoutMultiplier = 1;
    mTimeOuts.WriteTotalTimeoutConstant   = 250;
*/

/*	// From Blimp bluetooth app
    mTimeOuts.ReadIntervalTimeout         = 100;
    mTimeOuts.ReadTotalTimeoutMultiplier  = 0;
    mTimeOuts.ReadTotalTimeoutConstant    = 5000;
    mTimeOuts.WriteTotalTimeoutMultiplier = 10;
    mTimeOuts.WriteTotalTimeoutConstant   = 100;*/


/*
    // Specifies the maximum time, in [ms], allowed to elapse between
    // the arrival of two successive characters
    mTimeOuts.ReadIntervalTimeout         = 100;

    // Total time out =
    // ReadTotalTimeoutMultiplier x requested number of bytes + ReadTotalTimeoutConstant
    mTimeOuts.ReadTotalTimeoutMultiplier  = 0;
    mTimeOuts.ReadTotalTimeoutConstant    = 5000;
    mTimeOuts.WriteTotalTimeoutMultiplier = 10;
    mTimeOuts.WriteTotalTimeoutConstant   = 100;
*/

	if(!SetCommTimeouts(mHandle, &mTimeOuts)) {
		CloseHandle(mHandle);
		return 5;
	}
    
	/* Flush the content of the port. */
	//FlushFileBuffers(mHandle);
	char buffer;
	unsigned long read;
	do	{
		ReadFile(mHandle, &buffer, 1, &read, NULL);
	}
	while(read != 0);

#else /* Linux */

	int err;
	mFileDesc = open (mPort, O_RDWR | O_NOCTTY);
	mHandle = fdopen (mFileDesc, "a+");
	if(!mHandle)
		return 1;

	// Don't block on device configuration.
	int flags = fcntl(mFileDesc, F_GETFL, 0);
	err = fcntl(mFileDesc, F_SETFL, flags | O_NDELAY);
	if (err == -1)
		return 2;

	// Flush any garbage remaining on the port from previous operations.
	err = tcflush(mFileDesc, TCIOFLUSH);
	if (err == -1)
		return 3;

	struct termios tset;

	// Setup the terminal for ordinary I/O
	// e.g. as an ordinary modem or serial port printer.
	//
	// Default config: hangup on close.
	// if the HUPCL bit set, closing the port will hang up
	// the modem. (i.e. will lower the modem control lines)
	//
	// Default config: use hardware flow control
	// if CRTSCTS is set, use h/w flow control (IXON/OFF should not be set).
	// The best documentation for these calls can be found in
	// the Linux man pages for termios(3) and stty(1).
	//
	// Default config:
	// 8 bits, no parity
	tset.c_cflag = CREAD|CS8|CRTSCTS|HUPCL;

	switch (mStopBits) {
		case 2:
			tset.c_cflag |= CSTOPB;
			break;
		default:
			; // one stop bit is default
	}

	switch (mBaudRate) {
		case 230400:
			tset.c_cflag |= B230400;
			break;
		case 115200:
			tset.c_cflag |= B115200;
			break;
		case 57600:
			tset.c_cflag |= B57600;
			break;
		case 38400:
			tset.c_cflag |= B38400;
			break;
		case 19200:
			tset.c_cflag |= B19200;
			break;
		case 9600:
			tset.c_cflag |= B9600;
			break;
		case 4800:
			tset.c_cflag |= B4800;
			break;
		case 2400:
			tset.c_cflag |= B2400;
			break;
		case 1800:
			tset.c_cflag |= B1800;
			break;
		case 1200:
			tset.c_cflag |= B1200;
			break;
		case 600:
			tset.c_cflag |= B600;
			break;
		case 300:
			tset.c_cflag |= B300;
			break;
		default:
			return 4;
	}

	tset.c_cflag &= ~CRTSCTS;

	// Default config: ignore break, do not ignore parity
	tset.c_iflag = IGNBRK;

	// Default: no delay on carriage return, backspace, tab, etc.
	tset.c_oflag &= ~ (NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
	tset.c_oflag |= NL0|CR0|TAB0|BS0|VT0|FF0;

	// disable canonical processing
	tset.c_lflag &= ~ICANON;

	// disable echoing of input
	tset.c_lflag &= ~ECHO;

	tset.c_cc[VEOF]   = _POSIX_VDISABLE;
	tset.c_cc[VEOL]   = _POSIX_VDISABLE;
	tset.c_cc[VERASE] = _POSIX_VDISABLE;
	tset.c_cc[VINTR]  = _POSIX_VDISABLE;
	tset.c_cc[VKILL]  = _POSIX_VDISABLE;
	tset.c_cc[VQUIT]  = _POSIX_VDISABLE;
	tset.c_cc[VSUSP]  = _POSIX_VDISABLE;
	tset.c_cc[VSTART] = _POSIX_VDISABLE;
	tset.c_cc[VSTOP]  = _POSIX_VDISABLE;
	tset.c_cc[VMIN] = 0;
	tset.c_cc[VTIME] = 1;

	err = tcsetattr(mFileDesc, TCSANOW, &tset);
	if (err == -1)
		return 5;

	// allow commmunications to block
	flags = fcntl(mFileDesc, F_GETFL, 0);
	err = fcntl(mFileDesc, F_SETFL, flags & ~O_NDELAY);
	if (err == -1)
		return 6;

	// flush any unwritten, unread data
	tcflush(mFileDesc, TCIOFLUSH);
#endif /* End of Linux part */

	// Connected!
	mConnected = true;
	return 0;
}

wxInt8 SerialPort::Disconnect() {
	if(!mConnected)       // if already closed, return
		return 0;

#ifdef WIN32
	if(!CloseHandle(mHandle)) // non-zero on success
		return 1;
	mHandle = 0x0;
#else
	if (mFileDesc <= 0) return 1;
	if (mHandle == 0x0) return 2;

	tcflush(mFileDesc, TCIOFLUSH);
	fclose(mHandle);
	mHandle = 0x0;
	mFileDesc = -1;
#endif

	mConnected = false;
	return 0;
}

wxInt8 SerialPort::SendBuffer(const wxByte *buffer, const wxUint16 length) {
	if(!mConnected) return 1;

	if((length==0) || (buffer==NULL))
		return 2;
#ifdef WIN32
	unsigned long bogus;
	if(!WriteFile(mHandle, buffer, length, &bogus, NULL))
        return 3;
#else
	if(fwrite(buffer, sizeof(wxByte), length, mHandle) < length)
		return 3;
#endif

	return 0;
}

wxInt8 SerialPort::SendString(const wxString& string) {
	if(!mConnected) return 1;

	if(string.Len() == 0)
	return 2;
#ifdef WIN32
	unsigned long bogus;
	FlushFileBuffers(mHandle);
	if(!WriteFile(mHandle, string.c_str(), string.Len(), &bogus, NULL))
		return 3;
#else
	unsigned int length = string.Len();
	tcflush(mFileDesc, TCIOFLUSH);
	if(fwrite(string.c_str(), sizeof(wxByte), length, mHandle) < length)
		return 3;
#endif
	return 0;
}

wxInt8 SerialPort::ReceiveBuffer(wxByte *buffer, const wxUint16 length) {
	if(!mConnected) return 1;
	unsigned long bytesRead = 0, totalRead = 0;
	wxUint8 counter = 0;

	while(totalRead < length && counter < 50) {
#ifdef WIN32
		if(!ReadFile(mHandle, buffer + totalRead, length - totalRead, &bytesRead, NULL) || !bytesRead)
#else
		if(!(bytesRead = fread(buffer + totalRead, sizeof(wxByte), length - totalRead, mHandle)))
#endif
		counter++;
		totalRead += bytesRead;
	}

	if(totalRead < length) {
		StatusLine(wxString::Format("SerialPort - ReceiveBuffer - Read %ld bytes, expected %d.", totalRead, length));
		return 2;
	}
	return 0;
}

wxInt8 SerialPort::ReceiveString(wxString& string) {
	if(!mConnected) return 1;
	unsigned long bytesRead = 0, counter = 0;
	char *buffer = mBuffer;

	while(!bytesRead && counter < 100) {
#ifdef WIN32
		ReadFile(mHandle, buffer, 1, &bytesRead, NULL);
#else
		bytesRead = fread(buffer, sizeof(wxByte), 1, mHandle);
#endif
		counter++;
	}
	//StatusLine(wxString::Format("Counter is %d, read %d bytes.", counter, bytesRead));
	if(!bytesRead) return 2;
	buffer++;
	counter = 0;
	while(*(buffer-1) != '\n' && counter < 1000 && buffer - mBuffer < SERIAL_BUFFERLEN) {
#ifdef WIN32
		ReadFile(mHandle, buffer, 1, &bytesRead, NULL);
#else
		bytesRead = fread(buffer, sizeof(wxByte), 1, mHandle);
#endif
		if(!bytesRead)
			counter++;
		else
			buffer++;
	}
	//StatusLine(wxString::Format("Counter is %d, read %d bytes.", counter, (wxUint16)(buffer - (char*)mBuffer)));
	if(*(buffer-1) != '\n') {
		StatusLine(wxString::Format("SerialPort - ReceiveString - Couldn't receive string (no end of line)."));
		return 3;
	}
	buffer[0] = '\0';

	string = mBuffer;
	return 0;
}

wxInt8 SerialPort::Synchronize() {
	unsigned long bytesRead = 1;
	wxUint16 count = 0;
	char buff[1];

	// Drain input buffer
	while(bytesRead && count < 100) {
#ifdef WIN32
		ReadFile(mHandle, buff, 1, &bytesRead, NULL);
#else
		bytesRead = fread(buff, sizeof(wxByte), 1, mHandle);
#endif
		if(!bytesRead) count++;
	}

	// Flush output buffer
#ifdef WIN32
	PurgeComm(mHandle, PURGE_TXCLEAR | PURGE_RXCLEAR);
#else
	tcflush(mFileDesc, TCIOFLUSH);
#endif
	return 0;
}

void SerialPort::Configure() {
	wxString portChoices[] = { "COM1", "COM2", "COM3", "COM4" };
	wxString result = wxGetSingleChoice("Please select the appropiate COM port.", "Configure serial port", 4, portChoices);
	wxString port = "COM1";
	if(!result.IsEmpty()) {
#ifdef WIN32
		port = result;
#else
		if(result == "COM1")
			port = "/dev/cua0";
		else if(result == "COM2")
			port = "/dev/cua1";
		else if(result == "COM3")
			port = "/dev/cua2";
		else if(result == "COM4")
			port = "/dev/cua3";
#endif
	}
	SetCommPort(port);

	wxString choices[] = { "9600", "19200", "38400", "57600", "115200", "230400" };
	result = wxGetSingleChoice("Please select the appropiate baudrate.", "Configure serial port", 6, choices);
	wxUint32 rate = 57600;
	if(!result.IsEmpty()) {
		if(result == "9600")
			rate = 9600;
		else if(result == "19200")
			rate = 19200;
		else if(result == "38400")
			rate = 38400;
		else if(result == "57600")
			rate = 57600;
		else if(result == "115200")
			rate = 115200;
		else if(result == "230400")
			rate = 230400;
	}
	SetBaudRate(rate);
	
	StatusLine(wxString::Format("SerialPort - COM port set to %s.", port.c_str()));
	StatusLine(wxString::Format("SerialPort - Baudrate set to %d.", rate));
}

/* --- XML --- */
void SerialPort::Load(XMLElement *myself, XMLLoader *loader) {
	mPort = (myself->GetAttribute("port")->GetValue());
	mBaudRate = myself->GetAttribute("baudrate")->GetIntValue();
	mParity = myself->GetAttribute("parity")->GetIntValue();
	mByteSize = myself->GetAttribute("byte_size")->GetIntValue();
	mStopBits = myself->GetAttribute("stop_bits")->GetIntValue();
}

XMLElement* SerialPort::SaveSelf() {	
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
	result->AddAttribute(new XMLAttribute("port",mPort));
	result->AddAttribute(new XMLAttribute("baudrate",(wxInt32)mBaudRate));
	result->AddAttribute(new XMLAttribute("parity",mParity));
	result->AddAttribute(new XMLAttribute("byte_size",mByteSize));
	result->AddAttribute(new XMLAttribute("stop_bits",mStopBits));
	return result;
}

wxString SerialPort::GetXMLName() {
	return GetShortXMLName();
}
