/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/file.h>
#include <wx/filename.h>
#include <wx/textfile.h>
#include <wx/dir.h>

#include "modelsettings.h"
#include "manualmodel.h"
#include "manualview.h"
#include "evomodel.h"
#include "evoview.h"
#include "evosettingsdialog.h"
#include "ga/individual.h"
#include "ga/population.h"
#include "email.h"
#include "utils/smtp.h"
#include "register.h"

#include "comm/commlinkfactory.h"
#include "exp/experimentfactory.h"
#include "nn/neuralnetworkfactory.h"

#include "controller.h"

Controller::Controller() : wxFrame((wxFrame*) NULL, -1,
						"goevo 1.0", wxPoint(-1,-1), wxSize(600,450)) {
	
	/* Register modules. */
	Register::RegisterAll();
						
	/* Automatically redo layout on resize */
	SetAutoLayout(true);

	/* Create and set status bar */
	CreateStatusBar(2);
	SetStatusText("Bienvenue!", 0);
	SetStatusText("(c) 2002-2003 goevo developers team", 1);

	/* Create and set controls of this frame. */
	InitControls();

	/* Initialize some member variables */
	mOutputDirectory = wxGetCwd() + "/data";
	mModelSettings = new ModelSettings(this, mOutputDirectory);
	mIConfFileArray = NULL ;

	/* Create and set menu bar */
	InitMenuBar();
	mMenuBar->Enable(MENU_EVOL_STOP_BATCH, false);

	/* Initialize models and views (as far as necessary). */
	mEvoModel = new EvoModel(mModelSettings);
	mEvoView = new EvoView(this, mEvoModel);
	mEvoView->SetConstraints(GetViewConstraints());
	mEvoModel->SetView(mEvoView);

	mManualModel = NULL;
	mManualView = NULL;
	
	mShowView = true;

	/* Set Smtp host & port (if necessary) */
	Smtp::SetSmtpHost("localhost");
	Smtp::SetSmtpPort(25);
}

Controller::~Controller() { // see also Controller::OnClose
	delete mModelSettings;
	if(mManualView)
		mManualView->Destroy();
	if(mEvoView)
		mEvoView->Destroy();
}

/* ------------------------ STATUS LOG --------------------------- */

void Controller::StatusLine(const wxString& message) {
	mLogWindow->AppendText(message + "\n");
}

void Controller::OnStatusLine(wxCommandEvent &event) {
	mLogWindow->AppendText(event.GetString() + "\n");
}

void Controller::StatusLineErr(const wxString& message, wxInt8 errorCode) {
	mLogWindow->AppendText(message);
	mLogWindow->AppendText(" Error code: " + wxString::Format("%d",errorCode) + "\n");
}

void Controller::OnStatusLineErr(wxCommandEvent &event) {
	mLogWindow->AppendText(event.GetString());
	mLogWindow->AppendText(" Error code: " + wxString::Format("%d",(int)event.GetInt()) + "\n");
}

/* ------------------------ SENDMAIL --------------------------- */

void Controller::OnSendMail(wxCommandEvent &event) {
	wxInt8 ret;
	EmailMessage *emailMessage = (EmailMessage *)event.GetClientData();
	if((ret = Smtp::SendMail(emailMessage->from, emailMessage->to, emailMessage->subject, emailMessage->body)))
		StatusLineErr("Controller - SendMail failed, unable to send e-mail.", ret);
	delete emailMessage;
}

/* ------------------------ CONTROLS --------------------------- */

void Controller::InitControls() {
	/* Add status log window */
	wxLayoutConstraints *layout = new wxLayoutConstraints();
	layout->width.SameAs(this, wxWidth);
	layout->height.PercentOf(this, wxHeight, 25);
	layout->left.AsIs();
	layout->bottom.SameAs(this, wxHeight);
	mLogWindow = new wxTextCtrl(this, LOG_WINDOW, "", wxPoint(0,0), wxDefaultSize, wxTE_MULTILINE | wxTE_READONLY | wxTE_RICH);
	mLogWindow->SetConstraints(layout);

	/* Do layout now */
	Layout();
}

wxLayoutConstraints* Controller::GetViewConstraints() {
	wxLayoutConstraints *layout = new wxLayoutConstraints();
	layout->width.SameAs(this, wxWidth);
	layout->top.AsIs();
	layout->left.AsIs();
	layout->bottom.SameAs(mLogWindow, wxTop);
	return layout;
}

/* --------------------- CHANGE EVENTS -------------------- */

void Controller::OnRobotChange(int oldID) {

	// Remove old items.
	wxUint32 oldNum = ExperimentFactory::GetNumTypes(oldID);
	for(wxUint32 i=0; i<oldNum; i++)
		mExperimentMenu->Delete(MENU_OPTIONS_EXP + i);
	
	// Add new items.
	int robotID = mModelSettings->GetRobotID();
	for(wxUint32 i = 0; i < ExperimentFactory::GetNumTypes(robotID); i++)
		mExperimentMenu->Append(MENU_OPTIONS_EXP+i, ExperimentFactory::GetIndexedDesc(i, robotID), "", true);
	mExperimentMenu->Check(MENU_OPTIONS_EXP, true);	
		
	
	mExperimentMenu->UpdateUI();	
	mModelSettings->SetExperimentByIndex(0);
	StatusLine("Experiment " + ExperimentFactory::GetIndexedDesc(0, mModelSettings->GetRobotID()) + " selected.");
	
	mEvoModel->OnGenomeLengthChanged();
	mEvoModel->NotifyView();
}

void Controller::OnRobotLoad(int oldID) {

	// Remove old items.
	wxUint32 oldNum = ExperimentFactory::GetNumTypes(oldID);
	for(wxUint32 i=0; i<oldNum; i++)
		mExperimentMenu->Delete(MENU_OPTIONS_EXP + i);
	
	// Add new items.
	int robotID = mModelSettings->GetRobotID();
	for(wxUint32 i = 0; i < ExperimentFactory::GetNumTypes(robotID); i++)
		mExperimentMenu->Append(MENU_OPTIONS_EXP+i, ExperimentFactory::GetIndexedDesc(i, robotID), "", true);
	mExperimentMenu->Check(MENU_OPTIONS_EXP, true);

	mExperimentMenu->UpdateUI();

	mEvoModel->NotifyView();
}

/* ------------------------ MENU --------------------------- */

void Controller::InitMenuBar() {
	mMenuBar = new wxMenuBar();
	wxMenu *menu, *subMenu;

	// File menu
	menu = new wxMenu();
	menu->Append(MENU_FILE_NEW_POP, "&New population...", "Create a new population to evolve.");
	menu->Append(MENU_FILE_LOAD_POP, "Load &population...", "Load a population from disk.");
	menu->AppendSeparator();
	menu->Append(MENU_FILE_OUTPUT_DIR, "Set &data dir...", "Choose the default data directory (where all output will be saved).");
	menu->AppendSeparator();
	menu->Append(MENU_FILE_LOAD_SETTINGS, "&Load settings...", "Load all general settings.");
	menu->Append(MENU_FILE_SAVE_SETTINGS, "&Save settings...", "Save all general settings.");
	menu->AppendSeparator();
	menu->Append(MENU_FILE_LOAD_BATCH, "&Load batch file...", "Load and run a set of experiment.");
	menu->AppendSeparator();
	menu->Append(MENU_FILE_EXIT, "E&xit", "Au revoir...");
	mMenuBar->Append(menu, "&File");
	
	// Evolution menu
	menu = new wxMenu();
	menu->Append(MENU_EVOL_RUN, "&Run", "Perform an evolutionary run.");
	menu->Append(MENU_EVOL_PAUSE, "&Pause", "Pause evolution.");
	menu->Append(MENU_EVOL_STOP, "&Stop", "Stop evolution.");
	menu->Append(MENU_EVOL_STOP_BATCH, "&Stop Batch", "Stop all evolutions.");
	menu->AppendSeparator();
	menu->Append(MENU_EVOL_TEST_POP, "Test p&opulation...", "Test a saved population.");
	menu->Append(MENU_EVOL_TEST_IND, "&Test individual...", "Test a saved individual.");
	menu->AppendSeparator();
	menu->Append(MENU_EVOL_SETTINGS, "Se&ttings...", "Set evolution parameters.");
	mMenuBar->Append(menu, "&Evolution");

	// Control menu
	menu = new wxMenu();
	
		// Manual submenu
	subMenu = new wxMenu();
	subMenu->Append(MENU_CONTROL_MANUAL_START, "St&art", "Start manual control of the robot.");
	subMenu->Append(MENU_CONTROL_MANUAL_STOP, "S&top", "Stop manual control of the robot.");
	menu->Append(MENU_CONTROL_MANUAL_SUB, "&Manual", subMenu);

	
	mMenuBar->Append(menu, "&Control");


	// Observe menu
	menu = new wxMenu();	

	menu->Append(MENU_OBS_VIEW, "Show &view", "Enable/disable display of the main view.", true);
	menu->Check(MENU_OBS_VIEW, true);
	menu->AppendSeparator();
	menu->Append(MENU_OBS_SENSOR, "Log &sensor data", "Enable/disable logging of sensor data.", true);
	menu->Append(MENU_OBS_EXP, "Log &experiment data", "Enable/disable logging of experiment data.", true);
	menu->AppendSeparator();
	menu->Append(MENU_OBS_CLEAR_LOG, "&Clear status log", "Clear the status log window.");
	menu->Append(MENU_OBS_SAVE_LOG, "Save status log &as...", "Save the contents of the status log window.");
	mMenuBar->Append(menu, "&Observe");

	// Options menu
	menu = new wxMenu();

		// Robot submenu
	subMenu = new wxMenu();
	for(wxUint32 i = 0; i < RobotFactory::GetNumTypes(); i++)
		subMenu->Append(MENU_OPTIONS_ROBOT+i, RobotFactory::GetIndexedDesc(i), "", true);
	subMenu->Check(MENU_OPTIONS_ROBOT, true); // we check the first robot
	menu->Append(MENU_OPTIONS_ROBOT_SUB, "&Robot", subMenu);

		// CommLink submenu
	subMenu = new wxMenu();
	for(wxUint32 i = 0; i < CommLinkFactory::GetNumTypes(); i++)
		subMenu->Append(MENU_OPTIONS_COMM+i, CommLinkFactory::GetIndexedDesc(i), "", true);
	subMenu->Check(MENU_OPTIONS_COMM, true); // we check the first commlink
	menu->Append(MENU_OPTIONS_COMM_SUB, "&Communication", subMenu);
	menu->AppendSeparator();

		// Experiment submenu
	int robotID = mModelSettings->GetRobotID();
	mExperimentMenu = new wxMenu();
	for(wxUint32 i = 0; i < ExperimentFactory::GetNumTypes(robotID); i++)
		mExperimentMenu->Append(MENU_OPTIONS_EXP+i, ExperimentFactory::GetIndexedDesc(i, robotID), "", true);
	mExperimentMenu->Check(MENU_OPTIONS_EXP, true);
	menu->Append(MENU_OPTIONS_EXP_SUB, "&Experiment", mExperimentMenu);
	menu->AppendSeparator();

		// Neural network submenu
	subMenu = new wxMenu();
	subMenu ->Append(MENU_OPTIONS_NN, "&Perceptron", "", true);
	subMenu ->Append(MENU_OPTIONS_NN + 1, "Si&gmoid", "", true);
	subMenu ->Append(MENU_OPTIONS_NN + 2, "&Spiking", "", true);
	menu->Append(MENU_OPTIONS_NN_SUB, "&Neural network", subMenu);
	menu->Check(MENU_OPTIONS_NN, true);

	menu->Append(MENU_OPTIONS_ENCODING, "&Genome encoding...");
	
	
	mMenuBar->Append(menu, "&Options");

	// Help menu
	menu = new wxMenu();
	menu->Append(MENU_HELP_ABOUT, "&About...", "About goevo");
	mMenuBar->Append(menu, "&Help");

	// Set menu bar
	SetMenuBar(mMenuBar);
}

void Controller::FillMenu(wxMenu *menu, wxUint16 startId, const wxString *descs, wxUint8 num) {
	for(int i=0; i<num; i++)
		menu->Append(startId + i, descs[i], "", true);
	menu->Check(startId, true);
}

void Controller::CheckMenu(wxUint16 startId, wxUint8 num, wxUint16 checkId) {
	for(wxUint16 id=startId; id<startId+num; id++)
		if(id==checkId)
			mMenuBar->Check(id, true);
		else
			mMenuBar->Check(id, false);
}

void Controller::SetMenuMode(MenuMode mode) {
	mMenuMode = mode;
	switch (mMenuMode) {
		case MENU_MODE_NORMAL:
		{
			mMenuBar->EnableTop(0, true);
			mMenuBar->EnableTop(1, true);
			mMenuBar->EnableTop(2, true);
			mMenuBar->EnableTop(3, true);
			mMenuBar->EnableTop(4, true);
			mMenuBar->EnableTop(5, true);
			mMenuBar->Enable(MENU_OBS_SENSOR, true);
			mMenuBar->Enable(MENU_OBS_EXP, true);
			break;
		}
		case MENU_MODE_EVOLUTION:
		{
			mMenuBar->EnableTop(0, false);
			mMenuBar->EnableTop(1, true);
			mMenuBar->EnableTop(2, false);
			mMenuBar->EnableTop(3, true);
			mMenuBar->EnableTop(4, false);
			mMenuBar->EnableTop(5, true);		
			mMenuBar->EnableTop(5, true);
			mMenuBar->Enable(MENU_OBS_SENSOR, true);
			mMenuBar->Enable(MENU_OBS_EXP, true);
			break;
		}
		case MENU_MODE_MANUAL:
		{
			mMenuBar->EnableTop(0, false);
			mMenuBar->EnableTop(1, false);
			mMenuBar->EnableTop(2, true);
			mMenuBar->EnableTop(3, true);
			mMenuBar->EnableTop(4, false);
			mMenuBar->EnableTop(5, true);
			mMenuBar->Enable(MENU_OBS_SENSOR, true);
			mMenuBar->Enable(MENU_OBS_EXP, true);
			break;
		}
		case MENU_MODE_WAIT:
		{
			mMenuBar->EnableTop(0, false);
			mMenuBar->EnableTop(1, false);
			mMenuBar->EnableTop(2, false);
			mMenuBar->EnableTop(3, false);
			mMenuBar->EnableTop(4, false);
			mMenuBar->EnableTop(5, false);
			break;
		}
	}
}

/* ------------------------ EVENT TABLE --------------------------- */

BEGIN_EVENT_TABLE(Controller, wxFrame)
	EVT_CLOSE(Controller::OnClose)
	EVT_MENU(CMD_STATUS_LINE, Controller::OnStatusLine)
	EVT_MENU(CMD_STATUS_LINE_ERR, Controller::OnStatusLineErr)
	EVT_MENU(CMD_SENDMAIL, Controller::OnSendMail)
	EVT_MENU(MANUAL_STOPPED, Controller::OnManualStopped)
	EVT_MENU(EVOLUTION_STOPPED, Controller::OnEvolutionStopped)
	EVT_MENU(MENU_FILE_NEW_POP, Controller::OnMenuFileNewPop)
	EVT_MENU(MENU_FILE_LOAD_POP, Controller::OnMenuFileLoadPop)
	EVT_MENU(MENU_FILE_OUTPUT_DIR, Controller::OnMenuFileOutputDir)
	EVT_MENU(MENU_FILE_LOAD_SETTINGS, Controller::OnMenuFileLoadSettings)
	EVT_MENU(MENU_FILE_SAVE_SETTINGS, Controller::OnMenuFileSaveSettings)
	EVT_MENU(MENU_FILE_LOAD_BATCH, Controller::OnMenuFileLoadBatch)
	EVT_MENU(MENU_FILE_EXIT, Controller::OnMenuFileExit)
	EVT_MENU(MENU_EVOL_RUN, Controller::OnMenuEvolutionRun)
	EVT_MENU(MENU_EVOL_PAUSE, Controller::OnMenuEvolutionPause)
	EVT_MENU(MENU_EVOL_STOP, Controller::OnMenuEvolutionStop)
	EVT_MENU(MENU_EVOL_STOP_BATCH, Controller::OnMenuEvolutionStopBatch)
	EVT_MENU(MENU_EVOL_TEST_POP, Controller::OnMenuEvolutionTestPop)
	EVT_MENU(MENU_EVOL_TEST_IND, Controller::OnMenuEvolutionTestInd)
	EVT_MENU(MENU_EVOL_SETTINGS, Controller::OnMenuEvolutionSettings)
	EVT_MENU(MENU_CONTROL_MANUAL_START, Controller::OnMenuControlManualStart)
	EVT_MENU(MENU_CONTROL_MANUAL_STOP, Controller::OnMenuControlManualStop)
	EVT_MENU(MENU_OBS_VIEW, Controller::OnMenuObserveView)
	EVT_MENU(MENU_OBS_SENSOR, Controller::OnMenuObserveSensor)
	EVT_MENU(MENU_OBS_EXP, Controller::OnMenuObserveExperiment)
	EVT_MENU(MENU_OBS_CLEAR_LOG, Controller::OnMenuObserveClearLog)
	EVT_MENU(MENU_OBS_SAVE_LOG, Controller::OnMenuObserveSaveLogAs)
	EVT_MENU_RANGE(MENU_OPTIONS_ROBOT, MENU_OPTIONS_COMM_SUB - 1, Controller::OnMenuOptionsRobot)
	EVT_MENU_RANGE(MENU_OPTIONS_COMM, MENU_OPTIONS_EXP_SUB - 1, Controller::OnMenuOptionsCommLink)
	EVT_MENU_RANGE(MENU_OPTIONS_EXP, MENU_OPTIONS_NN_SUB - 1, Controller::OnMenuOptionsExperiment)
	EVT_MENU_RANGE(MENU_OPTIONS_NN, MENU_OPTIONS_ENCODING - 1, Controller::OnMenuOptionsNeuralNetwork)
	EVT_MENU(MENU_OPTIONS_ENCODING, Controller::OnMenuOptionsEncoding)
	EVT_MENU(MENU_HELP_ABOUT, Controller::OnMenuHelpAbout)
END_EVENT_TABLE()

/* ------------------------ EVENT HANDLERS --------------------------- */

void Controller::OnClose(wxCloseEvent &event) {
	if(mMenuMode != MENU_MODE_WAIT) {
		SetMenuMode(MENU_MODE_WAIT);
		mEvoModel->Stop();
		if(mEvoView)
			mEvoView->SetViewEnabled(false);
		delete mEvoModel;
		mEvoModel = NULL;
		if(mManualModel) {
			mManualModel->Stop();
			delete mManualModel;
			mManualModel = NULL;
		}
	}
	wxYield();
	Destroy();
}

void Controller::OnManualStopped(wxCommandEvent &event) {
	if(mManualView) {
		wxYield(); // process pending update events
		mEvoView->Enable(true);
		mManualView->Destroy();
		mManualView = NULL;
	}
	delete mManualModel;
	mManualModel = NULL;
	SetMenuMode(MENU_MODE_NORMAL);
	StatusLine("Manual control stopped.");
}

void Controller::OnEvolutionStopped(wxCommandEvent &event) {
	mEvoModel->CleanUp();
	SetMenuMode(MENU_MODE_NORMAL);

	NotifyExperimentFinished();
}

/* File menu handlers */

void Controller::OnMenuFileNewPop(wxCommandEvent &event) {
	wxUint16 size = mEvoModel->GetPopulation()->GetSize();
	long result;
	result = wxGetNumberFromUser("Please specify the size for the new population.", "Number of individuals:", "New population...", size, 1, 500);
	if(result > 0)
		mEvoModel->NewPopulation((wxUint16)result);
}

void Controller::OnMenuFileLoadPop(wxCommandEvent &event) {
	wxInt8 ret;
	wxFileDialog *fileDialog = new wxFileDialog(this, "Load population...", mOutputDirectory, "", "*.pop", wxOPEN);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString loadName = fileDialog->GetPath(); 
		ret = mEvoModel->LoadPopulation(loadName);
		if(ret)
			StatusLineErr("Couldn't load population from \"" + loadName + "\".", ret);
		else
			StatusLine("Loaded population from \"" + loadName + "\".");
	}
	fileDialog->Destroy();


}

void Controller::OnMenuFileOutputDir(wxCommandEvent &event) {
	wxDirDialog *dirDialog = new wxDirDialog(this, "Select output directory", mOutputDirectory,wxDD_NEW_DIR_BUTTON);
	if(dirDialog->ShowModal() == wxID_OK) {
		mOutputDirectory = dirDialog->GetPath();
		mModelSettings->SetOutputDirectory(mOutputDirectory);
		StatusLine("Output directory set to " + mOutputDirectory);
	} else 
		StatusLine("Output directory unchanged: " + mOutputDirectory);
	dirDialog->Destroy();
}

void Controller::OnMenuFileLoadSettings(wxCommandEvent &event) {
	wxFileDialog *fileDialog = new wxFileDialog(this, "Load settings...", mOutputDirectory, "", "*.iconf", wxOPEN);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString loadName = fileDialog->GetPath(); 

		LoadIConfFile( loadName );
	}
	fileDialog->Destroy();


}

void Controller::OnMenuFileSaveSettings(wxCommandEvent &event) {
	wxFileDialog *fileDialog = new wxFileDialog(this, "Save settings...", mOutputDirectory, "", "*.iconf", wxSAVE);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString saveName = fileDialog->GetPath(); 
		XMLSaver *saver = new XMLSaver(saveName);
		saver->SaveObject(mModelSettings);
		saver->SaveObject(mEvoModel);	
		saver->Close();
		delete saver;

		StatusLine("Saved settings to " + saveName);
	}
	fileDialog->Destroy();
}

void Controller::OnMenuFileLoadBatch(wxCommandEvent &event) {
	wxTextFile *batchFile;
	wxString iconfFileName = "";
	
	wxFileDialog *fileDialog = new wxFileDialog(this, "Load batch file...", mOutputDirectory, "", "*.ibatch", wxOPEN);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString loadName = fileDialog->GetPath();
		StatusLine("Reading batchFile : " + loadName);
		
		batchFile = new wxTextFile(loadName.c_str() );
		if(!batchFile->Open()) {
			StatusLine("Couldn't open file \"" + loadName + "\" for reading.");
		} else {
			mIConfFileArray = new wxArrayString();
			mIBatchLog = new wxFile( (loadName+".log").c_str() , wxFile::write_append);
			if(mIBatchLog->IsOpened()) {
				mIBatchLog->Write("Log started at " + wxDateTime::Now().Format() + "\n");
			}

			/* The batchFile is opened, now, we read it */
			for ( iconfFileName = batchFile->GetFirstLine(); !batchFile->Eof(); iconfFileName = batchFile->GetNextLine() )
			{
				/* End of line reached : we may have an entire fileName */
				if (!wxFile::Exists( iconfFileName.c_str() ) ) {
					StatusLine("The file \"" + iconfFileName + "\" doesn't seem to exist " ) ;
     			} else {
					/* The FileName is valid */
					mIConfFileArray->Add(iconfFileName) ;
					StatusLine("Added Experiment " + iconfFileName ) ;
					if(mIBatchLog->IsOpened()) {
						mIBatchLog->Write("Added Experiment " + iconfFileName + "\n");
					}
				}
			}
			/* Last Line ... */
			if (!wxFile::Exists( iconfFileName.c_str() ) ) {
				StatusLine("The file \"" + iconfFileName + "\" doesn't exist " ) ;
			}else {
				mIConfFileArray->Add( iconfFileName ) ;
				StatusLine("Added Experiment " + iconfFileName ) ;
				if(mIBatchLog->IsOpened()) {
					mIBatchLog->Write("Added Experiment " + iconfFileName + "\n");
				}
			}
		}	/* if !batchFile->IsOpened() */	
		StatusLine("End of batchFile" );
		mMenuBar->Enable(MENU_EVOL_STOP_BATCH, true);
		RunNextExperiment();

		delete batchFile ;
		batchFile = NULL ;
	}
	fileDialog->Destroy();

}

void Controller::LoadIConfFile(wxString iconfFileName) {
	XMLLoader *loader = new XMLLoader(iconfFileName);
	int oldID = mModelSettings->GetRobot()->GetID();
	loader->Load(NULL,mModelSettings);
	loader->Load(NULL,mEvoModel);
	OnRobotLoad(oldID);
	delete loader;

	CheckMenu(MENU_OPTIONS_NN, NeuralNetworkFactory::GetNumTypes(), (mEvoModel->GetNNFactory())->GetType() + MENU_OPTIONS_NN - 1);
	
	CheckMenu(MENU_OPTIONS_COMM, CommLinkFactory::GetNumTypes(),
		CommLinkFactory::GetIndexByID(mModelSettings->GetCommLink()->GetID()) + MENU_OPTIONS_COMM);
	CheckMenu(MENU_OPTIONS_ROBOT, RobotFactory::GetNumTypes(),
		RobotFactory::GetIndexByID(mModelSettings->GetRobotID()) + MENU_OPTIONS_ROBOT);
	
	CheckMenu(MENU_OPTIONS_EXP, ExperimentFactory::GetNumTypes(mModelSettings->GetRobotID()), 
		(ExperimentFactory::GetIndexByID(mModelSettings->GetExperiment()->GetID())) + MENU_OPTIONS_EXP);

	StatusLine("Loaded settings from " + iconfFileName);
}

void Controller::OnMenuFileExit(wxCommandEvent &event) {
	Close(false);	
}


/* Evolution menu handlers */

void Controller::OnMenuEvolutionRun(wxCommandEvent &event) {
	EvolutionRun();
}

void Controller::EvolutionRun() {
	wxInt8 ret;

	if(mEvoModel->IsPaused()) {
		if((ret = mEvoModel->Resume()))
			StatusLineErr("Couldn't resume evolution.", ret);
		else
			StatusLine("Resumed evolution.");
	} else {
		if((ret = mEvoModel->Start()))
			StatusLineErr("Couldn't launch evolution.", ret);
		else
			StatusLine("Started evolution.");
	}
	SetMenuMode(MENU_MODE_EVOLUTION);
}

void Controller::NotifyExperimentFinished() {
	if ( mIConfFileArray)
	{
		mIConfFileArray->Remove( 0 , 1);
		RunNextExperiment();
	}
}

void Controller::RunNextExperiment() {
	if (!mIConfFileArray->IsEmpty() )
	{
		wxString expOutputDirectory ;
		wxString expToLaunch = mIConfFileArray->Item(0) ;
		int i = 0 ;

		/* Create a new directory for the experiment */
		expOutputDirectory = expToLaunch + wxString::Format("%d", i ) ;
		while ( wxDir::Exists( expOutputDirectory ) )
  		{
   			i++ ;
			expOutputDirectory = expToLaunch + wxString::Format("%d", i ) ;
 		}

		if (!wxFileName::Mkdir( expOutputDirectory ) )
		{
			StatusLine("Unable to create directory : " + expOutputDirectory );		
			if(mIBatchLog->IsOpened()) {
				mIBatchLog->Write(wxDateTime::Now().Format() + "ER Unable to create directory : " + expOutputDirectory + "\n");
			}
			RunNextExperiment();
	   		return ;
		}

		mModelSettings->SetOutputDirectory( expOutputDirectory );
		StatusLine("Output directory set to " + expOutputDirectory);
		if(mIBatchLog->IsOpened()) {
			mIBatchLog->Write(wxDateTime::Now().Format() + " -- Output directory set to " + expOutputDirectory + "\n");
		}

		/* load the file and run the file */
		if(mIBatchLog->IsOpened()) {
			mIBatchLog->Write(wxDateTime::Now().Format() + " -- Loading Settings : " + expToLaunch + "\n");
		}
		LoadIConfFile( expToLaunch );

		wxSleep(5); // Wait for the reconnection of Webots
		if(mIBatchLog->IsOpened()) {
			mIBatchLog->Write(wxDateTime::Now().Format() + " -- Running evolution for " + expToLaunch + "\n");
		}
		EvolutionRun();
	} else { // if (!mIConfFileArray->IsEmpty() )
		StatusLine("BatchMode finished : No more experiment to launch.");
		if(mIBatchLog->IsOpened()) {
			mIBatchLog->Write(wxDateTime::Now().Format() + " -- BatchMode finished : No more experiment to launch." + "\n");
		}
		mMenuBar->Enable(MENU_EVOL_STOP_BATCH, false );

		delete mIConfFileArray ;
	  	mIConfFileArray = NULL ;
		delete mIBatchLog ;
	}
}

void Controller::OnMenuEvolutionPause(wxCommandEvent &event) {
	if(mEvoModel->IsRunning()) {
		mEvoModel->Pause();
		StatusLine("Paused evolution.");
	} else
		StatusLine("Couldn't pause evolution. (Resume with 'Start')");
}

void Controller::OnMenuEvolutionStop(wxCommandEvent &event) {
	EvolutionStop();
}

void Controller::OnMenuEvolutionStopBatch(wxCommandEvent &event) {
	if ( mIConfFileArray)
	{ // we are in batch mode
		delete mIConfFileArray ;
		mIConfFileArray = NULL ;
	}
	EvolutionStop();
	mMenuBar->Enable(MENU_EVOL_STOP_BATCH, false);
	StatusLine("BatchMode finished : Stopped by user.");
	if(mIBatchLog->IsOpened()) {
		mIBatchLog->Write(wxDateTime::Now().Format() + "BatchMode finished : Stopped by user" + "\n");
		delete mIBatchLog ;
	}
}

void Controller::EvolutionStop() {
	if(!mEvoModel->IsRunning() && !mEvoModel->IsPaused()) {
		StatusLine("Evolution isn't running or paused.");
		return;
	} 
	wxInt8 ret = 0;
	SetMenuMode(MENU_MODE_WAIT);
   	if((ret = mEvoModel->Stop()))
   		StatusLineErr("Couldn't stop evolution.", ret);
}

void Controller::OnMenuEvolutionTestPop(wxCommandEvent &event) {
	wxInt8 ret = 0;
	wxString mOutputDirectory; 
	wxFileDialog *fileDialog = new wxFileDialog(this, "Load population...", mOutputDirectory, "", "*.pop", wxOPEN);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString loadName = fileDialog->GetPath(); 
		if((ret = mEvoModel->TestPopulation(loadName)))
			StatusLineErr("Error while testing population from \"" + loadName + "\".", ret);
	}
	fileDialog->Destroy();
}

void Controller::OnMenuEvolutionTestInd(wxCommandEvent &event) {
	wxInt8 ret = 0;
	wxFileDialog *fileDialog = new wxFileDialog(this, "Load individual...", mOutputDirectory, "", "*.ind", wxOPEN);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString loadName = fileDialog->GetPath();
		Individual* individual = Individual::LoadFromDisk(loadName, mEvoModel->GetNNFactory()->GetGenomeLength());
		if(!individual)
			StatusLine("Couldn't load individual from \"" + loadName + "\".");
		else {
			StatusLine("Loaded individual from \"" + loadName + "\".");
			if((ret = mEvoModel->TestIndividual(individual)))
				StatusLineErr("Error while starting test.", ret);
			else
				SetMenuMode(MENU_MODE_EVOLUTION);
		}
	}
	fileDialog->Destroy();
}

void Controller::OnMenuEvolutionSettings(wxCommandEvent &event) {
	EvoSettingsDialog* dlg = new EvoSettingsDialog(this, mEvoModel);
	if(dlg->ShowModal() == wxID_OK)
		StatusLine("New evolution settings have been set.");
	else
		StatusLine("Evolution settings unchanged.");
	dlg->Destroy();
}

/* Control menu handlers */

void Controller::OnMenuControlManualStart(wxCommandEvent &event) {
	if(mManualModel) {
		StatusLine("Manual control already running.");
		return;
	}
	wxInt8 ret;
	mManualModel = new ManualModel(mModelSettings);
	if((ret = mManualModel->Start()) != wxTHREAD_NO_ERROR) {
		StatusLineErr("Couldn't start manual control!", ret);
		delete mManualModel;
		mManualModel = NULL;
		return;
	}
	SetMenuMode(MENU_MODE_MANUAL);
	if(mShowView) {
		mManualView = new ManualView(this, mManualModel);
		mManualView->SetConstraints(GetViewConstraints());
		mManualModel->AddObserver(mManualView);
		mEvoView->Enable(false);
		Layout();
	}
}

void Controller::OnMenuControlManualStop(wxCommandEvent &event) {
	if(mManualModel) {
		SetMenuMode(MENU_MODE_WAIT);
		mManualModel->Stop();
	} else
		StatusLine("Manual control isn't running.");
}


/* Observe menu handlers */

void Controller::OnMenuObserveView(wxCommandEvent &event) {
	mShowView = mMenuBar->FindItem(MENU_OBS_VIEW)->IsChecked();
	if(mShowView) {
		StatusLine("Enabled view.");
		mEvoView = new EvoView(this, mEvoModel);
		mEvoModel->SetView(mEvoView);
		mEvoView->SetConstraints(GetViewConstraints());
		if(mManualModel) {
			mManualView = new ManualView(this, mManualModel);
			mManualModel->AddObserver(mManualView);
			mManualView->SetConstraints(GetViewConstraints());
			mEvoView->Enable(false);
		}
	} else {
		StatusLine("Disabled view.");
		if(mEvoView) {
			mEvoModel->SetView(NULL);
			wxYield(); // process pending update events
			mEvoView->Destroy();
			mEvoView = NULL;
		}
		if(mManualView) {
			mManualModel->DeleteObserver(mManualView);
			wxYield(); // process pending update events
			mManualView->Destroy();
			mManualView = NULL;
		}
	}
	Layout();
}

void Controller::OnMenuObserveSensor(wxCommandEvent &event) {
	wxMenuItem* menuItem = mMenuBar->FindItem(event.GetId());
	if(menuItem->IsChecked()) {
		wxFileDialog *fileDialog = new wxFileDialog(this, "Log robot data to...", mOutputDirectory, "robot.log", "*.log", wxSAVE | wxOVERWRITE_PROMPT);
		if(fileDialog->ShowModal() == wxID_OK) {
			wxString saveName = fileDialog->GetPath();
			mModelSettings->SetObserveRobot(true, saveName);
			StatusLine("Logging robot data to " + saveName);
		} else
			menuItem->Check(false);
		delete fileDialog;
	} else
		mModelSettings->SetObserveRobot(false);
}

void Controller::OnMenuObserveExperiment(wxCommandEvent &event) {
	wxMenuItem* menuItem = mMenuBar->FindItem(event.GetId());
	if(menuItem->IsChecked()) {
		wxFileDialog *fileDialog = new wxFileDialog(this, "Log experiment data to...", mOutputDirectory, "experiment.log", "*.log", wxSAVE | wxOVERWRITE_PROMPT);
		if(fileDialog->ShowModal() == wxID_OK) {
			wxString saveName = fileDialog->GetPath();
			mModelSettings->SetObserveExperiment(true, saveName);
			StatusLine("Logging experiment data to " + saveName);
		} else
			menuItem->Check(false);
		delete fileDialog;
	} else
		mModelSettings->SetObserveExperiment(false);
}

void Controller::OnMenuObserveClearLog(wxCommandEvent &event) {
	mLogWindow->Clear();
}

void Controller::OnMenuObserveSaveLogAs(wxCommandEvent &event) {
	wxFileDialog *fileDialog = new wxFileDialog(this, "Save status log as...", mOutputDirectory, "status.log", "*.log", wxSAVE | wxOVERWRITE_PROMPT);
	if(fileDialog->ShowModal() == wxID_OK) {
		wxString saveName = fileDialog->GetPath();
		mLogWindow->SaveFile(saveName);
		StatusLine("Saved status log to " + saveName);
	}
	delete fileDialog;
}


/* Options menu handlers */
void Controller::OnMenuOptionsRobot(wxCommandEvent &event) {
	int oldID = mModelSettings->GetRobot()->GetID();
	wxUint16 id = event.GetId();
	int robotId = RobotFactory::GetIDByIndex(id - MENU_OPTIONS_ROBOT);
	StatusLine("Robot " + RobotFactory::GetDesc(robotId) + " selected.");
	mModelSettings->SetRobotByID(robotId);
	CheckMenu(MENU_OPTIONS_ROBOT, RobotFactory::GetNumTypes(), id);
	OnRobotChange(oldID);
}

void Controller::OnMenuOptionsCommLink(wxCommandEvent &event) {
	wxUint16 id = event.GetId();
	int commLinkID = CommLinkFactory::GetIDByIndex(id - MENU_OPTIONS_COMM);
	StatusLine("CommLink " + CommLinkFactory::GetDesc(commLinkID) + " selected.");
	mModelSettings->SetCommLinkByID(commLinkID);
	CheckMenu(MENU_OPTIONS_COMM, CommLinkFactory::GetNumTypes(), id);
}

void Controller::OnMenuOptionsExperiment(wxCommandEvent &event) {
	wxUint16 id = event.GetId();
	int expIndex = id - MENU_OPTIONS_EXP;
	mModelSettings->SetExperimentByIndex(expIndex);
	mEvoModel->OnGenomeLengthChanged();
	StatusLine("Experiment " + ExperimentFactory::GetIndexedDesc(expIndex, mModelSettings->GetRobotID()) + " selected.");
	CheckMenu(MENU_OPTIONS_EXP, ExperimentFactory::GetNumTypes(mModelSettings->GetRobotID()), id);
}

void Controller::OnMenuOptionsNeuralNetwork(wxCommandEvent &event) {
	wxUint16 id = event.GetId();
	NeuralNetworkType type = (NeuralNetworkType)(id - MENU_OPTIONS_NN + 1);
	wxString name;
	switch(type) {
		case NN_PERCEPTRON:		name = "perceptron";  break;
		case NN_SIGMOID:			name = "sigmoid";			break;
		case NN_SPIKING: 			name = "spiking";			break;
	}
	StatusLine("Selected " + name + " neural network factory.");
	mEvoModel->SetNNFactoryByType(type);
	CheckMenu(MENU_OPTIONS_NN, NeuralNetworkFactory::GetNumTypes(), id);
}

void Controller::OnMenuOptionsEncoding(wxCommandEvent &event) {
	mEvoModel->ConfigGenomeEncoding();
	StatusLine("Configured genome encoding.");
}



/* Help menu handlers */


void Controller::OnMenuHelpAbout(wxCommandEvent &event) {
	wxMessageDialog *msgDialog = new wxMessageDialog(this, "goevo 1.0\n\nEvolutionary robotics with neural controllers\n\nAutonomous Systems Lab (ASL2)\nSwiss Federal Institute of Technology in Lausanne (EPFL)\n\nAdaptive Intelligence Laboratory (AI Lab)\nUniversity Utrecht (UU)\n\nOriginal concept: Prof. Dario Floreano, LSA2\nReleased: October 2003\n\nThis software is freeware, please read docs/readme.txt\nand docs/gpl.txt.", "About goevo", wxOK | wxCENTRE | wxICON_INFORMATION);
	msgDialog->ShowModal();
	msgDialog->Destroy();
}
