/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _CONTROLLER_H
#define _CONTROLLER_H

#include <wx/file.h>

#include "robot/robotfactory.h"

/* Forward declarations */
class ModelSettings;
class ManualModel;
class ManualView;
class EvoModel;
class EvoView;

/**
 * The main window that controls the models and views. Implements
 * all interaction with the user.
 * @see Main
 * @see EvoView
 * @see EvoModel
 * @see ManualView
 * @see ManualModel
 */
class Controller : public wxFrame
{
public:
	/**
	 * Initialize the frame: setup controls, menubar, statusbar, etc.
	 */
	Controller();

	/**
	 * Delete any member variables that are not destroyed by wxWindows.
	 */
	~Controller();

	/** Returns the current output directory (used for populations, statistics, logs, etc.). */
	wxString& GetOutputDirectory() { return mOutputDirectory; }
	/** Sets the current output directory (full path). */
	void SetOutputDirectory(wxString& outputDirectory) { mOutputDirectory = outputDirectory; }

	void StatusLine(const wxString& message);
	void OnStatusLine(wxCommandEvent &event);
	void StatusLineErr(const wxString& message, wxInt8 errorCode);
	void OnStatusLineErr(wxCommandEvent &event);

	void OnSendMail(wxCommandEvent &event);
	void NotifyExperimentFinished();

protected:
	/** Enumeration of menu modes. */
	enum MenuMode {
		MENU_MODE_NORMAL, 		/**< Regular menu mode. */
		MENU_MODE_EVOLUTION, 	/**< Menu mode to use when evolving. */
		MENU_MODE_MANUAL, 		/**< Menu mode to use when in manual control mode. */
		MENU_MODE_WAIT				/**< Menu mode to use when waiting for somethings, thread termination for example. */
	};

	/** Initialize controls of this frame. */
	void InitControls();

	/** Returns layout constraints to be used for a view. */
	wxLayoutConstraints* GetViewConstraints();

	/**  Called when robot is changed, to handle experiment changes. */
	void OnRobotChange(int oldID);

	/**  Called when robot is loaded, to handle experiment changes. */
	void OnRobotLoad(int oldID);

	/* Menu handling methods */
	void InitMenuBar(); /**< Initialize and set the menu bar (in MENU_MODE_NORMAL). */
	void FillMenu(wxMenu *menu, wxUint16 startId, const wxString *descs, wxUint8 num); /**< Fill a menu with given items. */
	void CheckMenu(wxUint16 startId, wxUint8 num, wxUint16 checkId); /** Check menuitem checkId, uncheck all other in range [startId, startId + num]. */

	void SetMenuMode(MenuMode mode); /**< Set one of the MENU_MODE_* menu modes. */

	DECLARE_EVENT_TABLE()

	void OnClose(wxCloseEvent &event);

	/* Thread notifications methods */
	void OnManualStopped(wxCommandEvent &event);
	void OnEvolutionStopped(wxCommandEvent &event);

	/* File menu methods */
	void OnMenuFileNewPop(wxCommandEvent &event);
	void OnMenuFileLoadPop(wxCommandEvent &event);
	void OnMenuFileOutputDir(wxCommandEvent &event);
	void OnMenuFileLoadSettings(wxCommandEvent &event);
	void OnMenuFileSaveSettings(wxCommandEvent &event);
	void OnMenuFileLoadBatch(wxCommandEvent &event);
	void LoadIConfFile(wxString iconfFileName );
	void OnMenuFileExit(wxCommandEvent &event);

	/* Evolution menu methods */
	void OnMenuEvolutionRun(wxCommandEvent &event);
	void EvolutionRun();
	void RunNextExperiment();
	void OnMenuEvolutionPause(wxCommandEvent &event);
	void OnMenuEvolutionStop(wxCommandEvent &event);
	void OnMenuEvolutionStopBatch(wxCommandEvent &event);
	void EvolutionStop();
	void OnMenuEvolutionTestPop(wxCommandEvent &event);
	void OnMenuEvolutionTestInd(wxCommandEvent &event);
	void OnMenuEvolutionSettings(wxCommandEvent &event);

	/* Control menu methods */
	void OnMenuControlManualStart(wxCommandEvent &event);
	void OnMenuControlManualStop(wxCommandEvent &event);

	/* Observe menu methods */
	void OnMenuObserveView(wxCommandEvent &event);
	void OnMenuObserveSensor(wxCommandEvent &event);
	void OnMenuObserveExperiment(wxCommandEvent &event);
	void OnMenuObserveClearLog(wxCommandEvent &event);
	void OnMenuObserveSaveLogAs(wxCommandEvent &event);

	/* Options menu methods */
	void OnMenuOptionsRobot(wxCommandEvent &event);
	void OnMenuOptionsCommLink(wxCommandEvent &event);
	void OnMenuOptionsExperiment(wxCommandEvent &event);
	void OnMenuOptionsNeuralNetwork(wxCommandEvent &event);
	void OnMenuOptionsEncoding(wxCommandEvent &event);

	/* Help menu methods */
	void OnMenuHelpAbout(wxCommandEvent &event);

	/* Member variables */
	wxMenuBar 	*mMenuBar;
	wxMenu			*mExperimentMenu;
	MenuMode 		mMenuMode;
	wxTextCtrl 	*mLogWindow;
	bool				mShowView;
	
	ModelSettings *mModelSettings;

	ManualModel *mManualModel;
	ManualView 	*mManualView;
	EvoModel 		*mEvoModel;
	EvoView 		*mEvoView;

	wxString	mOutputDirectory ;
	wxArrayString *mIConfFileArray ;
	wxFile *mIBatchLog ;
};

/** Enumeration of all IDs used in the controller. */
enum {
	/* Status log window */
	LOG_WINDOW = 1000,

	/* Status line commands */
	CMD_STATUS_LINE,
	CMD_STATUS_LINE_ERR,

	/* Sendmail command */
	CMD_SENDMAIL,
	
	/* Update view command */
	UPDATE_VIEW,
	
	/* Thread notifications */
	MANUAL_STOPPED,
	EVOLUTION_STOPPED,

	/* File menu */
	MENU_FILE_NEW_POP,
	MENU_FILE_LOAD_POP,
	MENU_FILE_OUTPUT_DIR,
	MENU_FILE_LOAD_SETTINGS,
	MENU_FILE_SAVE_SETTINGS,
	MENU_FILE_LOAD_BATCH,
	MENU_FILE_EXIT,

	/* Evolution menu */
	MENU_EVOL_RUN,
	MENU_EVOL_PAUSE,
	MENU_EVOL_STOP,
	MENU_EVOL_STOP_BATCH,
	MENU_EVOL_TEST_POP,
	MENU_EVOL_TEST_IND,
	MENU_EVOL_SETTINGS,

	/* Control menu */
	MENU_CONTROL_MANUAL_SUB,
	MENU_CONTROL_MANUAL_START,
	MENU_CONTROL_MANUAL_STOP,

	/* Observe menu */
	MENU_OBS_VIEW,
	MENU_OBS_SENSOR,
	MENU_OBS_EXP,
	MENU_OBS_CLEAR_LOG,
	MENU_OBS_SAVE_LOG,

	/* Options menu */
	MENU_OPTIONS_ROBOT_SUB = 2500,
	MENU_OPTIONS_ROBOT, 	// range of 99 reserved for Robot menu item identifiers
	MENU_OPTIONS_COMM_SUB = 2600,
	MENU_OPTIONS_COMM,   	// range of 99 reserved for CommLink menu item identifiers
	MENU_OPTIONS_EXP_SUB = 2700,
	MENU_OPTIONS_EXP,			// range of 99 reserved for Exp menu item identifiers
	MENU_OPTIONS_NN_SUB = 2800,
	MENU_OPTIONS_NN,			// range of 98 reserved for NN menu item identifiers
	MENU_OPTIONS_ENCODING = 2899,

	/* Help menu */
	MENU_HELP_ABOUT = 4001,
	
	/* Sliders in regulated NN */
	SLIDER_THRESHOLDREG1 = 5000,
	SLIDER_THRESHOLDREG2 ,
	SLIDER_THRESHOLDREG3,

	/* Sliders in slow flyer */
	SLIDER_SLOWFLYER_PITCH = 5100,
	SLIDER_SLOWFLYER_DIR

};

#endif /* _CONTROLLER_H */
