/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/datetime.h>
#include <time.h>

#include "status.h"
#include "modelsettings.h"
#include "evoview.h"
#include "ga/crossover.h"
#include "ga/genome.h"
#include "ga/individual.h"
#include "ga/mutation.h"
#include "ga/population.h"
#include "ga/truncation.h"
#include "nn/perceptron/perceptronfactory.h"
#include "nn/sigmoid/sigmoidnnfactory.h"
#include "nn/spiking/spikingnnfactory.h"

#include "utils/randomtools.h"
#include "utils/ithread.h"
#include "utils/iutils.h"
#include "utils/xml/xml.h"

#include "evomodel.h"

EvoModel* EvoModel::mCurEvoModel = NULL;

EvoModel::EvoModel(ModelSettings *modelSettings) {
	mModelSettings = modelSettings;

	mNNFactory = new PerceptronFactory(mModelSettings);
	mPopulation = new Population(60, 0, true, mNNFactory->GetGenomeLength());
	mReproduce = new Truncation();
	mNumGenOperators = 2;
	mGenOperators = new IGenOperator *[2];
	mGenOperators[0] = new Crossover(0.1f);
	mGenOperators[1] = new Mutation(0.05f);

	mNumGenerations = 200;
	mNumEpochs = 2;
	mLifetime = 40;
	mCycleTime = 100;
	
	mTestPopulation = NULL;
	mTestIndividual = NULL;

	mEvoMode = NORMAL;
	mSeed = time(NULL);
	mIsPaused = false;
	mResume = false;
	
	mThread = NULL;
	mEvoView = NULL;
	mLogger = NULL;
	mPopLogger = NULL;
	mCurEvoModel = this;
}

EvoModel::~EvoModel() {
	delete mPopulation;
	delete mReproduce;
	delete mNNFactory;
	for(wxUint8 i=0; i<mNumGenOperators; i++)
		delete mGenOperators[i];
	delete [] mGenOperators;
	if(mLogger)
		delete mLogger;
	mCurEvoModel = NULL;
}

void EvoModel::Run() {
	
	if(mTestPopulation)
		TestPopulation();
	else if(mTestIndividual)
		TestIndividual();
	else
		Evolve();	
	
	
}

void EvoModel::Evolve() {
	Population *newPopulation = NULL;
	Individual *individualReplace=NULL;
	wxInt8 ret;
	wxUint16 curGeneration;

	/* Initialize everything */
	Experiment *experiment 	= mModelSettings->GetExperiment();
	Robot *robot 			= mModelSettings->GetRobot();
	experiment->SetRobot(robot);
	curGeneration = mPopulation->GetCurGeneration();
	

	// Additional initialization if not resuming
	if(!mResume) {
		RandomTools::Init(mSeed);
		mLogger = new Logger(mModelSettings->GetOutputDirectory() + "/evorun.log", this);
		AddObserver(mLogger);

		/* Save population if is the first one */
		if(curGeneration == 0)
			if((ret = mPopulation->SaveToDisk(mModelSettings->GetOutputDirectory() + "/000.pop"))) {
				StatusLineErr("Fatal error during saving of initial population.", ret);
				wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
				wxPostEvent(mModelSettings->GetController(), event);
				return;
			}
	}

	if((ret = InitRobot(experiment))) {
		StatusLineErr("Fatal error: couldn't initialize robot.", ret);
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
		wxPostEvent(mModelSettings->GetController(), event);
		
		return;
	}



	/* The main loop */
	for( ; curGeneration < mNumGenerations; curGeneration++) {
		StatusLine(wxString::Format("Starting evaluation of generation %d.", curGeneration));

		mPopulation->SetNumEpochs(mNumEpochs);
		/* Log statistics of each individual */
		mPopLogger = new Logger(mModelSettings->GetOutputDirectory() + wxString::Format("/generation%d.log", curGeneration), mPopulation);
		mPopulation->AddObserver(mPopLogger);

		/* Start with the most time-consuming: evaluating all individuals. */
		if((ret = mPopulation->Evaluate(this, experiment, mResume))) {
			if(ret > 0)
				StatusLineErr("Fatal error during evaluation of population.", ret);
			mPopulation->DeleteObserver(mPopLogger);
			delete mPopLogger;
			mPopLogger = NULL;
			break;
		}

		/* Remove log */
		mPopulation->DeleteObserver(mPopLogger);
		delete mPopLogger;
		mPopLogger = NULL;

		StatusLine(wxString::Format("Average fitness: %.3f, max fitness: %.3f.", mPopulation->GetAvgFitness(), mPopulation->GetMaxFitness()));
		mResume = false;
		SetChanged();
		NotifyObservers();

		/* Save best individual */
		if((ret = mPopulation->SaveBestIndividual(mModelSettings->GetOutputDirectory() + wxString::Format("/best%03d.ind", curGeneration))))
			StatusLineErr("Couldn't save best individual!", ret);

		/* Check if evolution interrupted. */
		if(!ContinueRun()) {
			StatusLine("Evolution interrupted.");
			break;
		}
		StatusLine("Evaluation done, creating next population.");

		/* Reproduction */
		newPopulation = mReproduce->Reproduce(mPopulation);
		if(!newPopulation) {
			StatusLine("Fatal error during reproduction.");
			break;
		}

		/* Apply genetic operators */
		if((ret = newPopulation->ApplyOperators(mGenOperators, mNumGenOperators))) {
			StatusLineErr("Fatal error during appliance of genetic operators.", ret);
			break;
		}

		/* Elitism */
		wxUint16 replace = RandomTools::RandInt16(newPopulation->GetSize());
		individualReplace = newPopulation->GetIndividual(replace);
		wxUint16	individualNum= individualReplace->GetGenome()->GetIndividualNum();
		wxUint16	populationNum=	individualReplace->GetGenome()->GetPopulationNum();
		delete newPopulation->GetIndividual(replace);
		//newPopulation->SetIndividual(replace, new Individual(new Genome(mPopulation->GetBestIndividual()->GetGenome(),newPopulation->GetCurGeneration(),replace)));
		newPopulation->SetIndividual(replace, new Individual(new Genome(mPopulation->GetBestIndividual()->GetGenome(),populationNum,individualNum)));

		/* Replace current population */
		delete mPopulation;
		mPopulation = newPopulation;
		newPopulation = NULL;

		/* Save new population to disk. */
		if((ret = mPopulation->SaveToDisk(mModelSettings->GetOutputDirectory() + wxString::Format("/%03d.pop", mPopulation->GetCurGeneration())))) {
			StatusLineErr("Fatal error during saving of population.", ret);
			break;
		}

		NotifyView();
		iYield();
	}

	if(ContinueRun()) { // evolution not interrupted but finished, clean up.
		experiment->GetRobot()->Disconnect();
		DeleteObservers();
		if(mLogger) {
			delete mLogger;
			mLogger = NULL;
		}
		StatusLine("Evolution finished.");
	}

	NotifyView();

	wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
	wxPostEvent(mModelSettings->GetController(), event);
}

void EvoModel::TestPopulation() {
	wxInt8 ret;
	float avgFitness, maxFitness, minFitness;

	minFitness = 10000000.0f;
	maxFitness = avgFitness = 0.0f;

	StatusLine("Starting population test.");

	/* Initialize everything */
	RandomTools::Init(mSeed);

	Experiment *experiment 	= mModelSettings->GetExperiment();
	Robot *robot 			= mModelSettings->GetRobot();
	experiment->SetRobot(robot);

	if((ret = InitRobot(experiment))) {
		StatusLineErr("Fatal error: couldn't initialize robot.", ret);
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
		wxPostEvent(mModelSettings->GetController(), event);		
		return;
	}

	if((ret = mTestPopulation->Test(this, experiment))) {
		if(ret > 0)
			StatusLineErr("Fatal error while testing population.", ret);
	}

	robot->Disconnect();

	delete mTestPopulation;
	mTestPopulation = NULL;
	wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
	wxPostEvent(mModelSettings->GetController(), event);
}

void EvoModel::TestIndividual() {
	wxInt8 ret;
	wxUint16 numTests = 0;
	float avgFitness, maxFitness, minFitness, fitness;

	minFitness = 10000000.0f;
	maxFitness = avgFitness = 0.0f;

	StatusLine("Starting individual test.");

	/* Initialize everything */
	RandomTools::Init(mSeed);

	Experiment *experiment 	= mModelSettings->GetExperiment();
	Robot *robot 			= mModelSettings->GetRobot();
	experiment->SetRobot(robot);

	if((ret = InitRobot(experiment))) {
		StatusLineErr("Fatal error: couldn't initialize robot.", ret);
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
		wxPostEvent(mModelSettings->GetController(), event);		
		return;
	}
	wxUint16 oldNumEpochs = mNumEpochs;
	mNumEpochs = 1;
	
	while(ContinueRun()) {
		StatusLine(wxString::Format("Test %d", numTests+1));
		wxString savename = mModelSettings->GetOutputDirectory() + wxString::Format("/neural%02d.log", numTests+1);
		if((ret = mTestIndividual->Evaluate(this, experiment, NULL, &savename))) {
			if(ret > 0)
      			StatusLineErr("Fatal error during test.", ret);
		} else {
			numTests++;
			fitness = mTestIndividual->GetFitness();
			if(maxFitness < fitness)
				maxFitness = fitness;
			if(minFitness > fitness)
				minFitness = fitness;
			avgFitness += fitness;
		}
	}
	avgFitness /= numTests;
	
	StatusLine(wxString::Format("Completed %d tests.", numTests));
	StatusLine(wxString::Format("Average fitness: %.3f, worst fitness: %.3f, best fitness: %.3f", avgFitness, minFitness, maxFitness));
	
	mNumEpochs = oldNumEpochs;
	delete mTestIndividual;
	mTestIndividual = NULL;
	wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
	wxPostEvent(mModelSettings->GetController(), event);		
}

/* ------------------------- CONTROL -------------------------- */

wxInt8 EvoModel::Start() {
	if(mThread)	return -1; // Already running...
	wxInt8 ret;

	mResume = false;
	mThread = new IThread(this);
	ret = mThread->Create();
	if(ret != wxTHREAD_NO_ERROR)
		return ret;
	ret = mThread->Run();
	return ret;
}

wxInt8 EvoModel::TestPopulation(wxString &loadname) {
	mTestPopulation = Population::LoadFromDisk(loadname, mNNFactory->GetGenomeLength());
	if(!mTestPopulation)
		return 1;
	StatusLine(wxString::Format("EvoModel - Loaded generation %d (%d individuals) for testing.", mTestPopulation->GetCurGeneration(), mTestPopulation->GetSize()));
	NotifyView();
	return Start();
}

wxInt8 EvoModel::TestIndividual(Individual *individual) {
	mTestIndividual = individual;
	return Start();
}

wxInt8 EvoModel::Pause() {
	if(!mThread) return 1;

	mThread->Delete();
	delete mThread;
	mThread = NULL;

	mModelSettings->GetExperiment()->GetRobot()->Disconnect();
	mIsPaused = true;

	NotifyView();

	return 0;
}

wxInt8 EvoModel::Resume() {
	if(mThread)	return -1; // Already running...
	wxInt8 ret;

	mResume = true;
	mThread = new IThread(this);
	ret = mThread->Create();
	if(ret != wxTHREAD_NO_ERROR)
		return ret;
	ret = mThread->Run();
	if(!ret) mIsPaused = false;
	return ret;
}

wxInt8 EvoModel::Stop() {
	if(mThread) {
		mThread->Delete();
		mModelSettings->GetExperiment()->GetRobot()->Disconnect();
	}
	
	return 0;
}

void EvoModel::CleanUp() {
	/* If we are in paused mode, don't delete all that */
	if (mIsPaused) return ;

	/* Create new population in case we want to launch another experiment after this one */
	NewPopulation( mPopulation->GetSize() );

	DeleteObservers();

	if(mThread) {
		mThread->Delete();
		mModelSettings->GetExperiment()->GetRobot()->Disconnect();
	}
	mThread = NULL;

	if(mLogger) {
		delete mLogger;
		mLogger = NULL;
	}

	if(mIsPaused) {
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, EVOLUTION_STOPPED);
		wxPostEvent(mModelSettings->GetController(), event);		
		mIsPaused = false;
	}

	NotifyView();

	StatusLine("Stopped evolution.");
}

bool EvoModel::IsRunning() {
	return mThread != NULL;
}

bool EvoModel::IsPaused() {
	return mIsPaused;
}

wxInt8 EvoModel::InitRobot(Experiment *experiment) {
	wxInt8 ret;
	if((ret = experiment->GetRobot()->AttemptConnect()))
		return ret;
	if((ret = experiment->InitRobot()))
		return ret;
	return 0;
}

/* ------------------------- ACCESSORS ------------------------- */

void EvoModel::NewPopulation(wxUint16 size) {
	delete mPopulation;
	mPopulation = new Population(size, 0, true, mNNFactory->GetGenomeLength());
	StatusLine(wxString::Format("Created new population of %d individuals.", size));
	NotifyView();
}

void EvoModel::SetGenOperators(wxUint8 num, IGenOperator **operators) {
	for(wxUint8 i=0; i<mNumGenOperators; i++)
		delete mGenOperators[i];
	delete [] mGenOperators;
	mNumGenOperators = num;
	mGenOperators = operators;	
	NotifyView();
}

void EvoModel::SetNNFactoryByType(NeuralNetworkType type) {
	SetNNFactoryByType(type, false);
}

void EvoModel::SetNNFactoryByType(NeuralNetworkType type, bool load) {
	if(mNNFactory->GetType() != type) {
		delete mNNFactory;
		switch(type) {
			case NN_PERCEPTRON:	mNNFactory = new PerceptronFactory(mModelSettings);	break;
			case NN_SIGMOID: 		mNNFactory = new SigmoidNNFactory(mModelSettings);	break;
			case NN_SPIKING:		mNNFactory = new SpikingNNFactory(mModelSettings);	break;
		}
	}
	if(!load) {
		mNNFactory->ConfigNeuralNetwork();
		OnGenomeLengthChanged();
	}
	NotifyView();
}

/* ------------------------- CONFIGURING ------------------------- */

void EvoModel::ConfigGenomeEncoding() {
	mNNFactory->ConfigGenomeEncoding();
	OnGenomeLengthChanged();
	NotifyView();
}

void EvoModel::OnGenomeLengthChanged() {
	wxUint16 newLength = mNNFactory->GetGenomeLength();
	StatusLine(wxString::Format("New genome length is %d.", newLength));
	if(newLength != mPopulation->GetIndividual(0)->GetGenome()->GetLength()) {
		wxUint16 size = mPopulation->GetSize();
		delete mPopulation;
		mPopulation = new Population(size, 0, true, newLength);
		StatusLine("Created new population to satify constraints of new genome encoding.");
	}
}

/* --------------------------- LOGGING ---------------------------- */

/** Called when a Log is started. */
void EvoModel::StartLog(Logger *logger) {
	logger->WriteString("Gen\tAvg\tMax\t%EndTime\n");
}

/** Called by logger when new data should be logged. */
void EvoModel::Log(Logger *logger) {
	logger->WriteString(wxString::Format("%d\t%.3f\t%.3f\t", mPopulation->GetCurGeneration(), mPopulation->GetAvgFitness(), mPopulation->GetMaxFitness()));
	logger->WriteString( "\t%\t" + wxDateTime::Now().Format("%X") + "\n");
}

/** Called when a Log is ended. */
void EvoModel::EndLog(Logger *logger) {

}

/* ------------------------- PAINTING ------------------------- */

void EvoModel::NotifyView() {
	if(mEvoView) {
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, UPDATE_VIEW);
		wxPostEvent(mEvoView, event);
	}
}

void EvoModel::Paint(wxDC& dc, wxPoint& position, wxSize& size) {
	wxUint16 offset = 0;
	wxString string;
	
	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	dc.DrawText("Evolution", position.x + 10, position.y + 15);
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);
	
	offset = 45;
	dc.DrawText("Status:", position.x + 10, position.y + offset);
	if(mIsPaused) string = "paused";
	else if(mThread) string = "running";
	else string = "stopped";
	dc.DrawText(string, position.x + 100, position.y + offset);
	offset += 20;
	dc.DrawText("Mode:", position.x + 10, position.y + offset);
	if(mTestIndividual) string = "testing individual";
	else if(mTestPopulation) string = "testing population";
	else string = "evolution";
	dc.DrawText(string, position.x + 100, position.y + offset);
	offset += 30;

	dc.DrawText("# generations:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d", mNumGenerations), position.x + 100, position.y + offset);
	offset += 20;
	dc.DrawText("# epochs:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d", mNumEpochs), position.x + 100, position.y + offset);
	offset += 20;
	dc.DrawText("Lifetime:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d", mLifetime), position.x + 100, position.y + offset);
	offset += 20;
	dc.DrawText("Cycle time:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d", mCycleTime), position.x + 100, position.y + offset);
	offset += 30;
	
	position.y += offset;
	size.SetHeight(size.GetHeight() - offset);
	if(mTestPopulation)
		mTestPopulation->Paint(dc, position, size);
	else
		mPopulation->Paint(dc, position, size);
}

/* ----------------------------- I/O --------------------------- */

wxInt8 EvoModel::LoadPopulation(wxString filename) {
	Population *population = Population::LoadFromDisk(filename, mNNFactory->GetGenomeLength());
	if(!population)
		return 1;
	delete mPopulation;
	mPopulation = population;
	StatusLine(wxString::Format("EvoModel - Loaded generation %d (%d individuals).", population->GetCurGeneration(), mPopulation->GetSize()));
	NotifyView();
	return 0;
}

/* ----------------------------- XML --------------------------- */
void EvoModel::Load (XMLElement *myself, XMLLoader *loader) {
	mSeed = myself->GetAttribute("seed")->GetIntValue();
	mNumGenerations = myself->GetAttribute("num_generations")->GetIntValue();
	mNumEpochs = myself->GetAttribute("num_epochs")->GetIntValue();
	mLifetime = myself->GetAttribute("life_time")->GetIntValue();
	mCycleTime = myself->GetAttribute("cycle_time")->GetIntValue();
	mEvoMode = (EvoMode) myself->GetAttribute("evomode")->GetIntValue();

	wxUint32 numGenOperators = myself->GetAttribute("num_gen_operators")->GetIntValue();
	IGenOperator** genOperators = new IGenOperator*[numGenOperators];
	bool crossover=(bool) myself->GetAttribute("crossover")->GetBoolValue();
	bool mutation=(bool) myself->GetAttribute("mutation")->GetBoolValue();

	for(wxUint32 i=0; i<numGenOperators; i++) {
		if(crossover)  {
			genOperators[i] = new Crossover();			
			crossover=false;
		} else if(mutation)  {
			genOperators[i] = new Mutation();
			mutation=false;
		}
	}
	SetGenOperators(numGenOperators, genOperators);

	NeuralNetworkType nnType = (NeuralNetworkType) myself->GetAttribute("factory_type")->GetIntValue();		
	SetNNFactoryByType(nnType, true);
	loader->Load(this, mNNFactory);
	OnGenomeLengthChanged();
}

XMLElement* EvoModel::SaveSelf () {
	XMLElement *result = new XMLElement("evomodel");
	result->AddAttribute(new XMLAttribute("seed",(wxInt32)mSeed));
	result->AddAttribute(new XMLAttribute("num_generations",mNumGenerations));
	result->AddAttribute(new XMLAttribute("num_epochs",mNumEpochs));
	result->AddAttribute(new XMLAttribute("life_time",(wxInt32) mLifetime));
	result->AddAttribute(new XMLAttribute("cycle_time",(wxInt32)mCycleTime));
	result->AddAttribute(new XMLAttribute("evomode", (wxInt32)mEvoMode));

	result->AddAttribute(new XMLAttribute("num_gen_operators",(wxInt32)mNumGenOperators));

	bool crossover=false, mutation=false;
	for(wxInt32 i=0; i<mNumGenOperators; i++) {
		if(mGenOperators[i]->GetType() == CROSSOVER)
			crossover=true;
		if(mGenOperators[i]->GetType() == MUTATION)
			mutation=true;
	}

	if(crossover)	result->AddAttribute(new XMLAttribute("crossover", "1"));
	else					result->AddAttribute(new XMLAttribute("crossover", "0"));
	if(mutation)	result->AddAttribute(new XMLAttribute("mutation", "1"));
	else					result->AddAttribute(new XMLAttribute("mutation", "0"));

	result->AddAttribute(new XMLAttribute("factory_type",(wxInt32)mNNFactory->GetType()));
	return result;
}

void EvoModel::SaveChilds (XMLSaver *saver) {
	saver->SaveObject(mNNFactory);
}

wxString EvoModel::GetXMLName () {
	return GetShortXMLName();
}

void EvoModel::SetParentName (const wxString &parentName) {

}
