/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _EVOMODEL_H
#define _EVOMODEL_H

#include <wx/datetime.h>

#include "utils/logger.h"
#include "utils/observable.h"

class EvoView;
class Experiment;
class Population;
class IGenOperator;
class INeuralNetworkFactory;
class ModelSettings;
class Individual;

#include "ga/ireproduce.h"
#include "nn/neuralnetworkfactory.h"
#include "utils/ithread.h"
#include "utils/xml/xmlserializable.h"

/** Enumeration of all possible evolution modes. */
enum EvoMode
{
	NORMAL = 0,
	SIMULATION,
	EMBEDDED_NN
};

/**
 * Model of the evolutionary process. Contains a Population of
 * Individuals. Can be observed, for example to display on the
 * screen (EvoView).
 * @see Population
 * @see Robot
 * @see Experiment
 * @see EvoView
 */
class EvoModel : public Observable, public ILoggable, public IRunnable, public XMLSerializable
{
public:
	/** Construct a default EvoModel. */
	EvoModel(ModelSettings *modelSettings);

	/** Destructor. */
	virtual ~EvoModel();

	/* ------------------------ EVOLUTION ------------------------- */

	/** Entry point for new thread, launches either Evolve() or TestIndividual(). */
	virtual void Run();
	
	static EvoModel* GetCurrentEvoModel() {return mCurEvoModel;}

private:
	/** Main loop implementing artificial evolution. */
	void Evolve();

	/** Test the set test population. */
	void TestPopulation();

	/** Test the set test individual. */
	void TestIndividual();

public:
	/* ------------------------- CONTROL -------------------------- */

	/** Start evolution. */
	wxInt8 Start();
	/** Start testing the given population. */
	wxInt8 TestPopulation(wxString &filename);
	/** Start testing the given individual. */
	wxInt8 TestIndividual(Individual *individual);

	/** Pause evolution/test. */
	wxInt8 Pause();
	/** Resume evolution/test. */
	wxInt8 Resume();
	/** Stop evolution/test. */
	wxInt8 Stop();
	
	/** Clean up thread object after stopping. */
	void CleanUp();

	/** Returns true if evolution is currently running. */
	bool IsRunning();
	/** Returns true if evolution is currently paused. */
	bool IsPaused();
	
	/** Returns false if evolution should stop right now. */
	bool ContinueRun() { return mThread->Proceed(); }
	
	/** Connect to robot and initialize it. */
	wxInt8 InitRobot(Experiment *experiment);

	/* ------------------------- ACCESSORS ------------------------- */

	/** @return The seed of the experiment. */
	wxUint32 GetSeed() { return mSeed; }
	/** @param seed The seed to initialize the random number generator. */
	void SetSeed(wxUint32 seed) { mSeed = seed; }

	EvoMode GetEvoMode() { return mEvoMode; }
	void SetEvoMode(EvoMode mode) { mEvoMode = mode; }

	wxUint16 GetNumGenerations() { return mNumGenerations; }
	void SetNumGenerations(wxUint16 numGenerations) { mNumGenerations = numGenerations; }
	wxUint16 GetNumEpochs() { return mNumEpochs; }
	void SetNumEpochs(wxUint16 numEpochs) { mNumEpochs = numEpochs; }
	wxUint32 GetLifetime() { return mLifetime; }
	void SetLifetime(wxUint32 lifetime) { mLifetime = lifetime; }
	wxUint32 GetCycleTime() { return mCycleTime; }
	void SetCycleTime(wxUint32 cycleTime) { mCycleTime = cycleTime; }
	
	/** Creates a new population with current genome length and specified size. */
	void NewPopulation(wxUint16 size);
	/** Returns the current population. */
	Population * GetPopulation() { return mPopulation; }
	/** Returns the current test population. */
	Population * GetTestPopulation() { return mTestPopulation; }

	IGenOperator ** GetGenOperators() { return mGenOperators; }
	wxUint8 GetNumGenOperators() { return mNumGenOperators; }
	void SetGenOperators(wxUint8 num, IGenOperator **operators);

	IReproduce * GetReproduce() { return mReproduce; }
	void SetReproduce(IReproduce *reproduce) { delete mReproduce; mReproduce = reproduce; }

	NeuralNetworkFactory * GetNNFactory() { return mNNFactory; }
	void SetNNFactoryByType(NeuralNetworkType type);
	void SetNNFactoryByType(NeuralNetworkType type, bool load);
	
	/** Returns the modelsettings. */
	ModelSettings * GetModelSettings() { return mModelSettings; }

	EvoView * GetView() { return mEvoView; }
	void SetView(EvoView *evoView) { mEvoView = evoView; }

	/* ------------------------- CONFIGURING ------------------------- */

	/** Configure the genome encoding. */
	void ConfigGenomeEncoding();

	/** Create new population if necessary, based on new genome length. */
	void OnGenomeLengthChanged();

	/* --------------------------- LOGGING ---------------------------- */

	/** Called when a Log is started. */
	void StartLog(Logger *logger);

	/** Called by logger when new data should be logged. */
	virtual void Log(Logger *logger);

	/** Called when a Log is ended. */
	void EndLog(Logger *logger);

	/* ------------------------- PAINTING ------------------------- */

	/** Notify view that it should repaint. */
	void NotifyView();

	/** Paint the model. */
	void Paint(wxDC& dc, wxPoint& position, wxSize& size);

	/* ----------------------------- I/O --------------------------- */

	/** Load a population from disk. */
	wxInt8 LoadPopulation(wxString filename);

	/* XML stuff */
	virtual void						Load (XMLElement *myself, XMLLoader *loader);
	virtual void						SaveChilds (XMLSaver *saver);
	virtual void						SetParentName (const wxString &parentName);
	virtual XMLElement*			SaveSelf ();
	virtual wxString				GetXMLName ();
	virtual const wxString	GetShortXMLName() { return "evomodel"; };

protected:
	EvoMode				mEvoMode;						/**< Normal, Simulation or Embedded-NN */
	wxUint32 			mSeed; 						/**< The seed to initialize the random number generator */
	bool					mIsPaused;				/**< True if evolution is currently paused. */
	bool					mResume;					/**< True if resuming evolution. */
	
	wxUint16 			mNumGenerations;	/**< Maximum number of generations */
	wxUint16			mNumEpochs;				/**< Number of epochs that each individual is tested. */
	wxUint32			mLifetime;				/**< Duration time (in s) of each epoch (discrete if in simulation). */
	wxUint32			mCycleTime;				/**< Duration time (in ms) of one cycle (discrete if in simulation). */
	
	Population 		*mPopulation; 		/**< Population that is evolved. */
	IGenOperator 	**mGenOperators;	/**< Pointer to a pointer to IgenOperator interface that is applied to the population.  */
	wxUint8				mNumGenOperators; /**< Size of mGenOperators array. */
	IReproduce 		*mReproduce; 			/**< Pointer to Ireproduce interface that corresponds to the model of selection */
	NeuralNetworkFactory *mNNFactory; /**< Pointer to NeuralNetworkFactory to be used to construct NeuralNetworks. */
	ModelSettings	*mModelSettings;	/**< Pointer to ModelSettings, containing robot, experiment and more. */
	
	Population		*mTestPopulation; /**< Population to be tested. */
	Individual		*mTestIndividual; /**< Individual to be tested. */

	EvoView				*mEvoView;				/**< View displaying the model. */
	IThread				*mThread;					/**< Thread running evolution or the testing process. */

	Logger				*mLogger;					/**< Used to log data of whole evolutionary run. */
	Logger				*mPopLogger;			/**< Used to log data of a single population. */	
	
	
	static EvoModel* mCurEvoModel;
};

#endif /* _EVOMODEL_H */
