/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "evomodel.h"
#include "status.h"
#include "ga/truncation.h"
#include "ga/crossover.h"
#include "ga/mutation.h"

#include "evosettingsdialog.h"

EvoSettingsDialog::EvoSettingsDialog(Controller *controller, EvoModel *evoModel) : wxDialog(controller, -1, "Evolution settings", wxDefaultPosition) {
	mEvoModel = evoModel;

	wxBoxSizer *topSizer = new wxBoxSizer(wxVERTICAL);
	AddRow(topSizer, "Random generator seed:", mSeedText = new wxTextCtrl(this, -1, ""));
	AddRow(topSizer, "Max. number of generations:", mGenerationsText = new wxTextCtrl(this, -1, ""));
	AddRow(topSizer, "Number of epochs:", mEpochsText = new wxTextCtrl(this, -1, ""));
	AddRow(topSizer, "Life time (in seconds):", mLifetimeText = new wxTextCtrl(this, -1, ""));
	AddRow(topSizer, "Cycle time (in ms):", mCycleTimeText = new wxTextCtrl(this, -1, ""));

	wxString optionsMode[3];
	wxUint8 optionsNumber = 2;
	optionsMode[0] = "Normal";
	optionsMode[1] = "Simulation";
	mEvoModeChoice = new wxRadioBox(this, -1, "", wxDefaultPosition, wxDefaultSize, optionsNumber, optionsMode, optionsNumber, wxRA_SPECIFY_ROWS);
	AddRow(topSizer, "Evolution mode:", mEvoModeChoice);

	wxString options[] = {"Truncation"};
	mReproductionChoice = new wxChoice(this, -1, wxDefaultPosition, wxDefaultSize, 1, options);
	mReproductionChoice->SetSelection(0);
	AddRow(topSizer, "Reproduction method:", mReproductionChoice);
	AddRow(topSizer, "Use crossover:", mCrossoverBox = new wxCheckBox(this, -1, ""));
	AddRow(topSizer, "Use mutation:", mMutationBox = new wxCheckBox(this, -1, ""));

	wxBoxSizer *horSizer = new wxBoxSizer(wxHORIZONTAL);
	horSizer->Add(new wxButton(this, wxID_OK, "Ok"));
	horSizer->Add(new wxButton(this, wxID_CANCEL, "Cancel"));
	topSizer->Add(horSizer, 0, wxALL | wxALIGN_CENTER, 5);
	
	SetAutoLayout(true);
	SetSizer(topSizer);
	topSizer->Fit(this);
	topSizer->SetSizeHints(this);
}

EvoSettingsDialog::~EvoSettingsDialog() {

}

bool EvoSettingsDialog::TransferDataToWindow() {
	mSeedText->SetValue(wxString::Format("%d", mEvoModel->GetSeed()));
	mGenerationsText->SetValue(wxString::Format("%d", mEvoModel->GetNumGenerations()));
	mEpochsText->SetValue(wxString::Format("%d", mEvoModel->GetNumEpochs()));
	mLifetimeText->SetValue(wxString::Format("%d", mEvoModel->GetLifetime()));

	mCycleTimeText->SetValue(wxString::Format("%d", mEvoModel->GetCycleTime()));
	mEvoModeChoice->SetSelection((wxInt32)mEvoModel->GetEvoMode());
	
	wxUint8 numOperators = mEvoModel->GetNumGenOperators();
	IGenOperator **operators = mEvoModel->GetGenOperators();
	mCrossoverBox->SetValue(false);
	mMutationBox->SetValue(false);
	for(wxUint8 i=0; i<numOperators; i++)
		switch(operators[i]->GetType()) {
			case CROSSOVER:
				mCrossoverBox->SetValue(true);
				break;
			case MUTATION:
				mMutationBox->SetValue(true);
				break;
		}
	return true;
}

bool EvoSettingsDialog::TransferDataFromWindow() {
	unsigned long number;
	
	if(mSeedText->GetValue().ToULong(&number))
		mEvoModel->SetSeed((wxUint32)number);
	if(mGenerationsText->GetValue().ToULong(&number))
		mEvoModel->SetNumGenerations((wxUint32)number);
	if(mEpochsText->GetValue().ToULong(&number))
		mEvoModel->SetNumEpochs((wxUint32)number);
	if(mLifetimeText->GetValue().ToULong(&number))
		mEvoModel->SetLifetime((wxUint32)number);
	if(mCycleTimeText->GetValue().ToULong(&number))
		mEvoModel->SetCycleTime((wxUint32)number);
	mEvoModel->SetEvoMode((EvoMode)mEvoModeChoice->GetSelection());
	
	wxUint8 numOperators = 0;
	if(mCrossoverBox->GetValue())
		numOperators++;
	if(mMutationBox->GetValue())
		numOperators++;
	IGenOperator **operators = new IGenOperator *[numOperators];
	numOperators = 0;
	if(mCrossoverBox->GetValue())
		operators[numOperators++] = new Crossover();
	if(mMutationBox->GetValue())
		operators[numOperators++] = new Mutation();
	mEvoModel->SetGenOperators(numOperators, operators);
	
	return true;
}

void EvoSettingsDialog::AddRow(wxBoxSizer *sizer, const wxString &desc, wxWindow *control) {
	wxBoxSizer *horSizer = new wxBoxSizer(wxHORIZONTAL);
	horSizer->Add(new wxStaticText(this, -1, desc), 0, wxALL, 5);
	horSizer->Add(control, 0, wxALL, 5);
	sizer->Add(horSizer, 0, wxALIGN_LEFT);
}
