/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/scrolwin.h>

#include "controller.h"
#include "evomodel.h"
#include "modelsettings.h"
#include "exp/experiment.h"
#include "ga/population.h"
#include "nn/neuralnetwork.h"
#include "robot/robot.h"

#include "evoview.h"

const wxUint8 EvoView::NumModes = 4;

EvoView::EvoView(Controller *controller, EvoModel *evoModel) : wxScrolledWindow(controller, -1, wxPoint(0,0) , wxDefaultSize  , wxVSCROLL | wxHSCROLL , "EvoView Window") {
	mEvoModel = evoModel;
	mModelSettings = mEvoModel->GetModelSettings();

	SetBackgroundColour(wxColour(255,255,255));
	SetForegroundColour(wxColour(0,0,255));

	mMode = 1;
	mViewString = wxString::Format("%d/%d", mMode, NumModes);
	mSize = GetSize();
	
    mEnabled = true;
	
}

EvoView::~EvoView() {

}

/* --------------------- EVENTS ---------------- */

BEGIN_EVENT_TABLE(EvoView, wxWindow)
	EVT_MENU(UPDATE_VIEW, EvoView::OnUpdateView)
	EVT_PAINT(EvoView::OnPaint)
	EVT_SIZE(EvoView::OnSize)
	EVT_LEFT_DOWN(EvoView::OnLeftClick)
	EVT_RIGHT_DOWN(EvoView::OnRightClick)
END_EVENT_TABLE()


void EvoView::OnPaint(wxPaintEvent& event) {
	wxPaintDC dc(this);
	Paint(dc);
}

void EvoView::OnSize(wxSizeEvent& event) {
	mSize = GetPaintSize();

	SetScrollbars( 8 , 8 , mSize.GetWidth()/8-1 , mSize.GetHeight()/8-1 /*, 0 , 0 , FALSE*/ ) ;

	wxClientDC dc(this);
	Paint(dc);
}

void EvoView::OnUpdateView(wxCommandEvent& event) {
	wxClientDC dc(this);
	Paint(dc);
}

void EvoView::OnLeftClick(wxMouseEvent& event) {
	if(++mMode > NumModes) mMode = 1;
	mViewString = wxString::Format("%d/%d", mMode, NumModes);

	mSize = GetPaintSize();
	SetScrollbars( 8 , 8 , mSize.GetWidth()/8-1 , mSize.GetHeight()/8-1 /*, 0 , 0 , FALSE*/ ) ;

	wxClientDC dc(this);
	Paint(dc);
}

void EvoView::OnRightClick(wxMouseEvent& event) {
	Population *population = mEvoModel->GetTestPopulation();
	if(population)
		population->SetInterrupt(event.GetX() < mSize.GetWidth() / 2 ? 1 : 2);
}

/* ------------------------ PAINTING --------------------- */

wxSize EvoView::GetPaintSize( void ) {
  switch(mMode) {

	case 2: // Neural network (architecture and activity)
	  {
	    wxSize neuralPaintSize ;
	
	  	NeuralNetwork *neuralNetwork = mModelSettings->GetExperiment()->GetNeuralNetwork();
	
	  	if(neuralNetwork){
			  neuralPaintSize = neuralNetwork->GetPaintSize();
			  if (neuralPaintSize.GetWidth() <= 0 || neuralPaintSize.GetHeight() <= 0) {
			  /* Old drawing solution for not defined paint sizes*/
					neuralPaintSize = GetSize();
	      }
				return neuralPaintSize ;
			} else {
				/* Nothing to draw. */
			  return GetSize();
			}
	  } /* case 2 */
		break;

  default: // Old system with no scroll bar available
    return GetSize();
		break;
	}

}

void EvoView::Paint(wxDC& dc) {
    PrepareDC(dc);

	if(!mEnabled) return; // don't draw at all.

	/* Initialize buffered image. */
	wxBitmap* bmpBlit = new wxBitmap(mSize.GetWidth(), mSize.GetHeight());
	wxMemoryDC* memDC= new wxMemoryDC();
	memDC->SelectObject(*bmpBlit);
	memDC->BeginDrawing();
	memDC->SetBackground(*wxWHITE_BRUSH);
	memDC->Clear();
	memDC->SetBackground(wxNullBrush);

	memDC->SetTextForeground(*wxBLACK);
	memDC->SetTextBackground(*wxWHITE);

	// Draw 'x/y' (view number x of y) text
	wxFont font = memDC->GetFont();
	font.SetPointSize(8);
	memDC->SetFont(font);
	memDC->DrawText(mViewString, mSize.GetWidth() - 25, 5);
	font.SetPointSize(10);
	memDC->SetFont(font);

	switch(mMode) {
		case 1: // Robot (sensor data)
			PaintRobot(*memDC);
			break;
		case 2: // Neural network (architecture and activity)
			PaintNeuralNetwork(*memDC);
			break;
		case 3: // Experiment
			PaintExperiment(*memDC);
			break;
		case 4: // EvoModel (statistics)
			PaintEvoModel(*memDC);
			break;

		default: // NA
			break;
	}

	/* Finish up, copy to screen. */
	memDC->SetBrush(wxNullBrush);
	memDC->SetPen(wxNullPen);

	memDC->EndDrawing();
	dc.Blit(0, 0, mSize.GetWidth(), mSize.GetHeight(), memDC, 0, 0, wxCOPY);
	delete bmpBlit;
	delete memDC;
}

void EvoView::PaintRobot(wxDC& dc) {
	wxPoint position;
	wxSize size;

	position.x = 0;
	position.y = 0;
	size.SetWidth(mSize.GetWidth());
	size.SetHeight(mSize.GetHeight());
	mModelSettings->GetRobot()->Paint(dc, position, size);
}

void EvoView::PaintNeuralNetwork(wxDC& dc) {
	wxPoint position;
	wxSize size;

	NeuralNetwork *neuralNetwork = mModelSettings->GetExperiment()->GetNeuralNetwork();

	if(neuralNetwork){
		position.x = 0;
		position.y = 0;
		size.SetWidth(mSize.GetWidth());
		size.SetHeight(mSize.GetHeight());
		neuralNetwork->Paint(dc, position, size);
	} else {
		/* Nothing to draw. */
		wxFont font = dc.GetFont();
		font.SetWeight(wxBOLD);
		dc.SetFont(font);
		dc.DrawText("There is currently no neural network to display", position.x + 10, position.y + 15);
		font.SetWeight(wxNORMAL);
		dc.SetFont(font);
	}
}

void EvoView::PaintExperiment(wxDC& dc) {
	wxPoint position;
	wxSize size;

	position.x = 0;
	position.y = 0;
	size.SetWidth(mSize.GetWidth());
	size.SetHeight(mSize.GetHeight());
	mModelSettings->GetExperiment()->Paint(dc, position, size);
}

void EvoView::PaintEvoModel(wxDC& dc) {
	wxPoint position;
	wxSize size;

	position.x = 0;
	position.y = 0;
	size.SetWidth(mSize.GetWidth());
	size.SetHeight(mSize.GetHeight());
	mEvoModel->Paint(dc, position, size);
}
