/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _EXPERIMENT_H
#define _EXPERIMENT_H

class NeuralNetwork;
class Robot;

#include "../robot/robot.h"
#include "../utils/xml/xml.h"
#include "../utils/observable.h"
#include "../utils/logger.h"
#include "experimentfactory.h"

/**
 * Experiments link the NeuralNetwork with the Robot and is also used
 * to determine the fitness of an Individual.
 * @see NeuralNetwork
 * @see Robot
 * @see Individual
 */
class Experiment : public Observable, public ILoggable, public XMLSerializable
{
public:
	/**
	 * Experiment constructor
	 */
	Experiment();

	/**
	 * Experiment destructor
	 */
	virtual ~Experiment();

	/* ------------------------------- BASIC ACCESSORS -------------------- */

	/**
	 * Accessor method
	 * @param robot is the robot for this experiment
	 */
	void SetRobot(Robot* robot) { mRobot = robot; }

	/**
	 * Accessor method
	 * @return the robot of this experiment
	 */
	Robot* GetRobot() { return mRobot; }

	virtual wxInt8 ReadSensors() { return mRobot->ReadSensors(); }

	/**
	 * Accessor method
	 * @param nn is the neural network for this experiment
	 */
	void SetNeuralNetwork(NeuralNetwork* neuralNetwork) { mNeuralNetwork = neuralNetwork; }

	/**
	 * Accessor method
	 * @return the neural network of this experiment
	 */
	NeuralNetwork* GetNeuralNetwork() { return mNeuralNetwork; }

	/**
	 * Accessor method
	 * @return the type of the experiment
	 */
	//wxUint8 GetType() { return mType; }

	/**
	 * Accessor method
	 * @return Description of this experiment.
	 */
	const wxString GetDesc() { return ExperimentFactory::GetDesc(mID.first, mID.second); }

	/* ------------------------------- VIRTUAL -------------------- */

	/**
	 * Method that returns the number of inputs for the neural network
	 * @return the number of sensors
	 */
	virtual wxUint8 GetNumSensorNeurons() =0;

	/**
	 * Method that returns the number of outputs for the neural network.
	 * @return The number of output neurons.
	 */
	virtual wxUint8 GetNumOutputNeurons() =0;

	/** Initialize the robot, called when new connection with robot has been established. */
	virtual wxInt8 InitRobot() =0;

	/** 
	 * Initialize an epoch.
	 * @param cycle Number of current init cycle.
	 */
	virtual wxInt8 InitEpoch(wxUint16 cycle) =0;
	
	/**
	 * Give to the experiement an oportunity to init the neural network before each epoch.
	 */
	virtual wxInt8 InitNeuralNetwork(NeuralNetwork* nn) {return 0;}

	/**
	 * Method that gets the sensors values of the robot and sets them to the neural network input.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 SetNNInput() =0;

	/**
	 * Method that gets the ouputs of the neural network and sets them to the robot
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ProcessNNOutput() =0;

	/**
	 * Comput the fitness value of the current individual at this very moment.
	 * @return The fitness value.
	 */
	virtual float ComputeFitness() =0;

	/** Called each time step to determine whether individual should be killed or not. */
	virtual bool KillIndividual() { return false; }

	/* ---------------- CONFIGURE ------------------ */

	/** Redefine this method to ask user for Experiment configuration. */
	virtual void Configure() {}
	
	/* ---------------- LOGGING -------------------- */

	/** Called when a Log is started. */
	void StartLog(Logger *logger);

	/** Has to be implemented by extensions to write 'legenda' of log file. */
	virtual void LogLegenda(Logger *logger) {};

	/** Called each time the Logger is notified by means of Update(). */
	virtual void Log(Logger *logger) {};
	
	/** Called when a Log is ended. */
	void EndLog(Logger *logger);

	/* ---------------- PAINTING -------------------- */

	/** Paint the experiment. */
	void Paint(wxDC& dc, wxPoint& position, wxSize& size);

	/** 
	 * Paint experiment data. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintExperiment(wxDC& dc, wxPoint& position, wxSize& size) {}

	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual void			SaveChilds (XMLSaver *saver);
	virtual void			SetParentName (const wxString &parentName);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "experiment"; }
	
	/* ---------------- REGISTER STUFF ------------------ */
	void SetID(DualID id) {mID = id;}
	DualID GetID() {return mID;}

protected:
	//ExperimentType mType; 			/**< The type of the experiment */
	Robot *mRobot; 					/**< The robot of the experiment */
	NeuralNetwork *mNeuralNetwork; 	/**< The neural Network of the experiment */
	float *mNeuronInput;			/**< Array of input values for neural network. */	
	DualID mID;						/**< ID of the experiment. */
};

#endif /* _EXPERIMENT_H */
