/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../status.h"
#include "../robot/khepera.h"
#include "../nn/neuralnetwork.h"
#include "../utils/iutils.h"
#include "../utils/xml/xml.h"

#include "kheperaexperiment.h"

const wxString KheperaExperiment::ItemDesc = "Infrared Braitenberg";

KheperaExperiment::KheperaExperiment() {

	mNeuronInput = new float[8];
	mMotorScale = 20;
}

KheperaExperiment::~KheperaExperiment() {

}

wxInt8 KheperaExperiment::InitRobot() {
	mKhepera = (Khepera *)mRobot;
	mKhepera->SetSensorMask(SENSOR_MOTOR_SPEED | SENSOR_DISTANCE);
	return 0;
}

wxInt8 KheperaExperiment::InitEpoch(wxUint16 cycle) {
	float motorSpeeds[3];
	motorSpeeds[2] = 1.0f;
	wxInt8 flagEpoch=1;
	
	if (cycle<10) {
		mKhepera->ReadDistanceSensors();
		mKhepera->CalcBraitenberg(motorSpeeds);
		mKhepera->SetMotors(motorSpeeds);
		flagEpoch=1;
	} else
		flagEpoch=0;
		
	return (flagEpoch);
}

wxInt8 KheperaExperiment::SetNNInput() {
	wxUint8 *distanceValues = mKhepera->GetDistanceData();
	for(wxUint8 i=0; i<8; i++)
		mNeuronInput[i] = distanceValues[i] / 255.0f;
		
	mNeuralNetwork->SetInput(mNeuronInput);
	return 0;
}

wxInt8 KheperaExperiment::ProcessNNOutput() {
	float *output = mNeuralNetwork->GetOutputValues();
	wxInt8 speedL = (wxInt8)(mMotorScale * (output[0] - output[1]));
	wxInt8 speedR = (wxInt8)(mMotorScale * (output[2] - output[3]));
	return mKhepera->SetMotors(speedL, speedR);
}

float KheperaExperiment::ComputeFitness() {
	/** Implicit version */
/*	wxInt8 speedL, speedR;
	speedL = mKhepera->GetSpeedLeft();
	speedR = mKhepera->GetSpeedRight();
	
	if(speedL < 0 || speedR < 0)
		return 0.0f;
	
	float fitness;
	fitness = (float)speedL / mMotorScale;
	fitness += (float)speedR / mMotorScale;
	return fitness; */


	/** More explicit version */
	float speedL, speedR;
	speedL = mKhepera->GetSpeedLeft() / (float)mMotorScale;
	speedR = mKhepera->GetSpeedRight() / (float)mMotorScale;

	float drive = (fabs(speedL) + fabs(speedR)) / 2;
	float rotate = 1.0f - fabs(speedL - speedR) / 2;
	float forward = ((speedL > 0 ? speedL : 0) + (speedR > 0 ? speedR : 0)) / 2;
	float fitness = 0.1f * drive + 0.5f * rotate + 0.4f * forward;
	return fitness;
}

/* --- XML --- */
void KheperaExperiment::Load (XMLElement *myself, XMLLoader *loader) {
	mMotorScale = myself->GetAttribute("motor_scale")->GetByteValue();
}

XMLElement* KheperaExperiment::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
	result->AddAttribute(new XMLAttribute("motor_scale",mMotorScale));
	return result;
}

wxString KheperaExperiment::GetShortXMLName() { return KheperaExperiment::GetStaticXMLType(); }
