/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../status.h"
#include "../robot/khepera.h"
#include "../nn/neuralnetwork.h"
#include "../utils/iutils.h"
#include "../utils/laplacefilter.h"
#include "../utils/xml/xml.h"

#include "kheperavisionexperiment.h"

const wxString KheperaVisionExperiment::ItemDesc = "Vision-based navigation with Khepera";

KheperaVisionExperiment::KheperaVisionExperiment() {

	// Default values that will be reset by Configure()
	mNumPixels = 32;
	//mNumPixels = 25;
	mMotorScale = 12;
	mPushPull = true;
	mMotion = false;
	mFitnessType = KHEPVIS_FITNESS_NO_BACK_MOTOR;
	mVisionSelection=NULL;
	mLaplaceFilter=NULL;
}

KheperaVisionExperiment::~KheperaVisionExperiment() {
	if (mVisionSelection) delete [] mVisionSelection;
	if (mLaplaceFilter) delete mLaplaceFilter;
}

wxInt8 KheperaVisionExperiment::InitRobot() {

	// Take the oportunity of robot initialization for instanciation of different arrays.
	// This cannot be done in the constructor since mNumPixels is not well defined before Configure() or XML Load() have been called.
	mNeuronInput = new float[mNumPixels];
	mVisionSelection = new wxUint8[mNumPixels];
	mLaplaceFilter = new LaplaceFilter(mNumPixels, 0, NULL); // default filter dim=3, coeff -0.5,1,-0.5
	for(wxUint8  i=0; i<mNumPixels; i++)
		mVisionSelection[i] = 0;
	mVisionInput = mLaplaceFilter->Apply(mVisionSelection);

	// Actual robot initialization
	mKhepera = (Khepera *)mRobot;
	if(mKhepera->GetSensorEnabled(SENSOR_DELTA_POSITION))
		mKhepera->SetSensorMask(SENSOR_MOTOR_SPEED | SENSOR_VISION | SENSOR_DISTANCE | SENSOR_DELTA_POSITION);
	else
		mKhepera->SetSensorMask(SENSOR_MOTOR_SPEED | SENSOR_VISION | SENSOR_DISTANCE);
	return 0;
}

wxInt8 KheperaVisionExperiment::InitEpoch(wxUint16 cycle) {
	float motorSpeeds[3];
	motorSpeeds[2] = 1.0f;
	wxInt8 flagEpoch=1;
	
	if (cycle<10) {	// 1 second
		if(cycle == 0)
			mKhepera->SetFlag(VIRTUAL_WALL, false);
		mKhepera->ReadDistanceSensors();
		mKhepera->CalcBraitenberg(motorSpeeds);
		mKhepera->SetMotors(motorSpeeds);
		flagEpoch=1;
	} else {
		mKhepera->SetFlag(VIRTUAL_WALL, true);
		flagEpoch=0;
	}
		
	return (flagEpoch);
}

wxInt8 KheperaVisionExperiment::SetNNInput() {
	wxUint8 i;

	if (!mMotion) { // Standard vision sensors => apply Laplace filter
		wxUint8 interval = mKhepera->GetNumVision() / mNumPixels;
		wxUint8 *visionValues = mKhepera->GetVisionData();
		for(i=0; i<mNumPixels; i++)
			mVisionSelection[i] = visionValues[i * interval];

		mVisionInput = mLaplaceFilter->Apply(mVisionSelection);
		for(i=0; i<mNumPixels; i++)
			mNeuronInput[i] = mVisionInput[i];
	} else { // Motion detector => interpret vision input as signed bytes
		wxUint8 interval = mKhepera->GetNumVision() / mNumPixels;
		wxUint8 *visionValues = mKhepera->GetVisionData();
		for(wxUint8  i=0; i<mNumPixels; i++)
//			mNeuronInput[i] = visionValues[i * interval] / 255.0f;
			mNeuronInput[i] = (wxInt8)visionValues[i * interval] / 128.0f;
	}
/***********Raw vision: inversed ************************/
/*
	wxUint8 interval = mKhepera->GetNumVision() / mNumPixels;
	wxUint8 *visionValues = mKhepera->GetVisionData();

	for(wxUint8  i=0; i<mNumPixels; i++){
		mVisionSelection[i] = visionValues[i * interval];
<<<<<<< kheperavisionexperiment.cpp
	mVisionInput = mLaplaceFilter->Apply(mVisionSelection);
	for(wxUint8 i=0; i<NumPixels; i++)
		mNeuronInput[i] = mVisionInput[i];

	// Vision sensors, raw vision
/*	wxUint8 interval = mKhepera->GetNumVision() / NumPixels;
	wxUint8 *visionValues = mKhepera->GetVisionData();
	for(wxUint8  i=0; i<NumPixels; i++)
		mNeuronInput[i] = visionValues[i * interval] / 255.0f;
=======
		mNeuronInput[i]=1-mVisionSelection[i]/255.0f;
		StatusLine(wxString::Format("%d %f",i,	mNeuronInput[i]));
	}
	mVisionInput=mNeuronInput;
>>>>>>> 1.44
*/
/*******************************************************/
	mNeuralNetwork->SetInput(mNeuronInput);
	SetChanged();
	return 0;
}

wxInt8 KheperaVisionExperiment::ProcessNNOutput() {
	mOutput = mNeuralNetwork->GetOutputValues();

	// Note: each mOutput is in [0..1], whatever the NN is.

	wxInt8 speedL, speedR;

	if(mPushPull)
	{	
		speedL = (wxInt8)(mMotorScale * (mOutput[3] - mOutput[2]));	
		speedR = (wxInt8)(mMotorScale * (mOutput[1] - mOutput[0]));
	}
	else
	{
		speedL = (wxInt8)(mMotorScale * (mOutput[0]-0.5)*2);
		speedR = (wxInt8)(mMotorScale * (mOutput[1]-0.5)*2);
	}

	SetChanged();
	return mKhepera->SetMotors(speedL, speedR);
}

float KheperaVisionExperiment::ComputeFitness() {
	/** Dario's version */
	float speedL, speedR;
	speedL = (float) mKhepera->GetSpeedLeft() / mMotorScale;
	speedR = (float) mKhepera->GetSpeedRight() / mMotorScale;
	
	float fitness;
	
	switch(mFitnessType)
	{
		case KHEPVIS_FITNESS_NO_BACK_MOTOR:
			{
				if (speedL < 0 || speedR < 0)
					fitness = 0.0f;
				else
					fitness = (speedL + speedR) * 0.5f;
			}
			break;	
	
		case KHEPVIS_FITNESS_NO_BACK_MOTION:
			{
				fitness = (speedL + speedR) * 0.5f;
				if(fitness < 0.0f)
					fitness = 0.0f;
			}
			break;
		
		case KHEPVIS_FITNESS_EXPLICIT:
			{
				float drive = (fabs(speedL) + fabs(speedR)) * 0.5f;
				float rotate = 1.0f - fabs(speedL - speedR) * 0.5f;
				float forward = ((speedL > 0 ? speedL : 0) + (speedR > 0 ? speedR : 0)) * 0.5f;
				fitness = 0.1f * drive + 0.5f * rotate + 0.4f * forward;
			}
			break;
		
		case KHEPVIS_FITNESS_STRAIGHT_MOVE_ONLY_SIMPLE:
			{
				fitness = (speedL + speedR) * 0.5f;
				float diff = speedL - speedR;
				
				// This mean that the forward fitness is taken into account only if
				// the radius of the trajectory is more than about 2 times the arena
				// i.e. about 1 m. This should force the robot to have a straight
				// trajectory.
				if(fitness < 0 || fabs(diff) > 0.005)
					fitness = 0;
			}
			break;
		
		default:
			fitness = 0.0f;
			break;
	}

	return fitness;
}

void KheperaVisionExperiment::LogLegenda(Logger *logger) {
	logger->WriteString(wxString::Format("NeuronInput#%d Output#%d\n", GetNumSensorNeurons(), GetNumOutputNeurons()));
}

void KheperaVisionExperiment::Log(Logger *logger) {
	for(wxUint16 i=0; i<GetNumSensorNeurons(); i++)
		logger->WriteString(wxString::Format("%.2f ", mNeuronInput[i]));
	for(wxUint16 i=0; i<GetNumOutputNeurons(); i++)
		logger->WriteString(wxString::Format("%.2f ", mOutput[i]));
	logger->WriteString(wxString::Format("\n"));
}

void KheperaVisionExperiment::PaintExperiment(wxDC& dc, wxPoint& position, wxSize& size) {
	Experiment::PaintExperiment(dc, position, size);
	
	wxUint8 width, height, value;
	float dTheta, dR;

	dc.SetPen(*wxBLACK_PEN);

	/* Selected vision */
	dc.DrawText("Selected vision", position.x + 150, position.y);
	width = 15;
	height = 20;
	for(int i=0; i<mNumPixels; i++) {
		value = (wxUint8)(mVisionSelection[i]*255);
		dc.SetBrush(wxBrush(wxColour(value, value, value), wxSOLID));
		dc.DrawRectangle(position.x + 150 + (width-1) * i, position.y + 20, width, height);
	}

	/* Prepocessed vision */
	dc.DrawText("Preprocessed vision", position.x + 150, position.y + 60);
	for(int i=0; i<mNumPixels; i++) {
		value = (wxUint8)(mVisionInput[i]*255);
		dc.SetBrush(wxBrush(wxColour(value, value, value), wxSOLID));
		dc.DrawRectangle(position.x + 150 + (width-1) * i, position.y + 80, width, height);
	}
	dc.DrawText("Black = 0%, white = 100%", position.x + 150, position.y + 120);

	/* Trajectory */
	if(mKhepera->GetSensorEnabled(SENSOR_DELTA_POSITION)) {
		dc.SetPen(*wxBLACK_PEN);
		dc.SetBrush(*wxWHITE_BRUSH);
		
		/* Compute trajectory */
		dR = (mKhepera->GetDeltaPositionLeft()+mKhepera->GetDeltaPositionRight())/16;	// the factor 8 is for display reasons...
		X += dR * cos(Theta);
		Y += dR * sin(Theta);
		dTheta = (mKhepera->GetDeltaPositionLeft()-mKhepera->GetDeltaPositionRight())*0.16/53;	// 0.16mm corresponds to one pulse (basic unit=0.08mm) * 2, 53 is the distance between the wheels of the Khepera
		Theta += dTheta;
		
		wxPoint *point = new wxPoint(X,Y);
		mPoints.Append((wxObject *)point);

		dc.DrawLines(&mPoints, position.x + 400, position.y + 400);
	}
}

void KheperaVisionExperiment::Configure()
{
	long number;
	
	wxString choices[] = {"Single output neuron", "Push-pull output"};
	wxString result = wxGetSingleChoice("Select output scheme:", "Configure Khepera Vision", 2, choices);
	if(result == "Single output neuron") {
		mPushPull = false;
	} else if(result == "Push-pull output")
		mPushPull = true;

	choices[0] = "Standard linear camera";
	choices[1] = "Motion detector";
	result = wxGetSingleChoice("Select visual sensor type:", "Configure Khepera Vision", 2, choices);
	if(result == "Standard linear camera") {
		mMotion = false;
		mNumPixels=25;
	} else if(result == "Motion detector"){
		mMotion = true;
		mNumPixels=32;
	}
	
	if((number = wxGetNumberFromUser("Enter the motor scale (max speed):", "motor scale", "Configure Khepera Vision", mMotorScale, 1, 128)) > 0)
    	mMotorScale = number;
    
    wxString choices2[] = {
				"Both motor must be forward (standard fitness)",
				"Motion must be forward",
				"Explicit fitness implementation",
				"Motion must be forward and trajectory straight"};
	int result2 = wxGetSingleChoiceIndex("Select fitness type:", "Configure Khepera Vision", 4, choices2);
	if(result2 != -1)
		mFitnessType = (KhepVisFitnessType) result2;
}

/* --- XML --- */
void KheperaVisionExperiment::Load (XMLElement *myself, XMLLoader *loader) {
	mMotorScale = myself->GetAttribute("motor_scale")->GetByteValue();
	mPushPull = myself->GetAttribute("push_pull")->GetBoolValue();
	mMotion = myself->GetAttribute("motion")->GetBoolValue();
	if (!mMotion)
			mNumPixels=25;
	else
		mNumPixels=32;
	mFitnessType = (KhepVisFitnessType) myself->GetAttribute("fitness_type")->GetIntValue();
}

XMLElement* KheperaVisionExperiment::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
	result->AddAttribute(new XMLAttribute("motor_scale",mMotorScale));
	result->AddAttribute(new XMLAttribute("push_pull", mPushPull));
	result->AddAttribute(new XMLAttribute("motion", mMotion));
	result->AddAttribute(new XMLAttribute("fitness_type", (int) mFitnessType));
	return result;
}

wxString KheperaVisionExperiment::GetShortXMLName() { return KheperaVisionExperiment::GetStaticXMLType(); }
