/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _KHEPERAVISIONEXPERIMENT_H
#define _KHEPERAVISIONEXPERIMENT_H

class Khepera;
class LaplaceFilter;

#include "../utils/xml/xml.h"
#include "experiment.h"

enum KhepVisFitnessType
{
	KHEPVIS_FITNESS_NO_BACK_MOTOR = 0,
	KHEPVIS_FITNESS_NO_BACK_MOTION,
	KHEPVIS_FITNESS_EXPLICIT,
	KHEPVIS_FITNESS_STRAIGHT_MOVE_ONLY_SIMPLE
};

/**
 * Default Experiment for Khepera.
 * Uses IR sensor data for network input, forward speed means fitness.
 * @see Experiment
 * @see NeuralNetwork
 * @see Khepera
 */
class KheperaVisionExperiment : public Experiment
{
public:

	// Registeration stuff.
	enum { ItemID = 'KVEx' };
	const static wxString ItemDesc;

	/** Default constructor. */
	KheperaVisionExperiment();

	/** Destructor. */
	virtual ~KheperaVisionExperiment();

	/**
	 * Method that returns the number of inputs for the neural network
	 * @return the number of sensors
	 */
	virtual wxUint8 GetNumSensorNeurons() { return mNumPixels; }

	/**
	 * Method that returns the number of outputs for the neural network.
	 * @return The number of output neurons.
	 */
	virtual wxUint8 GetNumOutputNeurons() { return mPushPull ? 4 : 2; }

	/** Initialize the robot, called when new connection with robot has been established. */
	virtual wxInt8 InitRobot();

	/** Initialize an epoch. */
	virtual wxInt8 InitEpoch(wxUint16 cycle);
	
	/**
	 * Method that gets the sensors values of the robot and sets them to the neural network input.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 SetNNInput();

	/**
	 * Method that gets the ouputs of the neural network and sets them to the robot
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ProcessNNOutput();

	/**
	 * Compute the fitness value of the current individual at this very moment.
	 * @return The fitness value.
	 */
	virtual float ComputeFitness();
	
	/** Redefine this method to ask user for Experiment configuration. */
	virtual void Configure();
	
	/* ---------------- LOGGING -------------------- */

	/** Write 'legenda' to log file. */
	virtual void LogLegenda(Logger *logger);

	/** Called by logger when new data should be logged. */
	virtual void Log(Logger *logger);

	/* ---------------- PAINTING -------------------- */
		
	/** 
	 * Paint experiment data. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintExperiment(wxDC& dc, wxPoint& position, wxSize& size);


	/* ---------------- XML -------------------- */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual XMLElement*		SaveSelf ();	
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "experiment-kheperavision"; }

protected:
	wxUint8 mNumPixels;

	wxUint8 mMotorScale;
	bool mPushPull;
	bool mMotion;
	KhepVisFitnessType mFitnessType;
	Khepera *mKhepera;
	LaplaceFilter *mLaplaceFilter;
	wxUint8 *mVisionSelection;
	const float *mVisionInput;
	const float *mOutput;

	float X,Y,Theta;		/**< For trajectory display. */
	wxList mPoints;			/**< List of the trajectory points. */
};

#endif /* _KHEPERAVISIONEXPERIMENT_H */
