/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

#include <wx/wx.h>
#include <wx/textfile.h>
#include <math.h>
#ifdef WIN32
	#include <wx/msw/winundef.h>
#endif

#include "../status.h"
#include "../utils/randomtools.h"

#include "timeseriesexperiment.h"
#include "../nn/neuralnetwork.h"
#include "../evomodel.h"
#include "../nn/spiking/spikingnn.h"

#include <wx/listimpl.cpp>
WX_DEFINE_LIST(PointListT);

/*********************** DATA FILE FORMAT ************************

<NumInput:wxUint8> <NumOutput:wxUint8>
<SetLength:wxUint16>
<Input:float>[SetLength][NumInput]
<Target:float>[SetLength][NumOutput]

Data can be created and saved into the right format using the "create_time_series.m" matlab script

See "exemple.dat" in folder "data/timeseries" to see an exemple of time series format

See "docs/experiments/" for more documentation
*****************************************************************/

const wxString TimeSeriesExperiment::ItemDesc = "Time Series Analysis";

TimeSeriesExperiment::TimeSeriesExperiment() {
	mNumInput = 0;
	mNumOutput = 0;
	mFileName="";
	mInput = mOutput = mTarget = NULL;
	mSetLength = mCurIndex = 0;
//	fitnessTemp=0;
	mEvoModel= EvoModel::GetCurrentEvoModel();
	mOutputPoints=NULL;
	mTargetPoints=NULL;
	SpikingNN::SetBiasActivity(1.0f);
}

TimeSeriesExperiment::~TimeSeriesExperiment() {
	if(mInput) {
		for(wxUint16 i=0; i<mSetLength; i++)
			delete [] mInput[i];
		delete mInput;
	}
	if(mOutput) {
		for(wxUint16 i=0; i<mSetLength; i++)
			delete [] mOutput[i];
		delete mOutput;
	}
	if(mTarget) {
		for(wxUint16 i=0; i<mSetLength; i++)
			delete [] mTarget[i];
		delete mTarget;
	}
	if (mOutputPoints){
		delete[] mOutputPoints;
	}
	if (mTargetPoints){
		delete[] mTargetPoints;
	}
}

wxInt8 TimeSeriesExperiment::SaveEpoch(wxString &filename) {
	wxFile *file = new wxFile(filename.c_str(), wxFile::write);
	if(!file->IsOpened()) {
		StatusLine("TimeSeriesExperiment - Couldn't open file \"" + filename + "\" for writing.");
		delete file;
		return 1;
	}

	file->Write("Input, target output and actual output\n");
	
	for(wxUint16 j=0; j<mNumInput; j++) {
		file->Write("Input:  ");
		for(wxUint16 i=0; i<mSetLength; i++)
			file->Write(wxString::Format("%.03f ", mInput[i][j]));
		file->Write("\n");
	}
	for(wxUint16 j=0; j<mNumOutput; j++) {
		file->Write("Target: ");
		for(wxUint16 i=0; i<mSetLength; i++)
			file->Write(wxString::Format("%.03f ", mTarget[i][j]));
		file->Write("\nActual: ");
		for(wxUint16 i=0; i<mSetLength; i++)
			file->Write(wxString::Format("%.03f ", mOutput[i][j]));
		file->Write("\n");
	}
			
	file->Write("End of log\n\n");
	file->Close();
	delete file;

	return 0;
}

wxInt8 TimeSeriesExperiment::SetNNInput() { 
//	if(++mCurIndex == mSetLength) mCurIndex = 0;
	mNeuralNetwork->SetInput(mInput[mCurIndex]);
		SetChanged();
	return 0;
}

wxInt8 TimeSeriesExperiment::ProcessNNOutput() {
	float *output = mNeuralNetwork->GetOutputValues();

	if(++mCurIndex == mSetLength) mCurIndex = 0;

	for(wxUint8 i=0; i<mNumOutput; i++) {
		mOutput[mCurIndex][i] = output[i];
	}
	SetChanged();
	return 0;
}




float TimeSeriesExperiment::ComputeFitness(){
	wxUint8 num = 0;
	double error = 0;
	double outputTmp = 0;
	float fitness = 0;
	
	float *target = mTarget[mCurIndex];

	for(wxUint8 i=0; i<mNumOutput; i++) {
		error+=sqrt(fabs(mOutput[mCurIndex][i]-target[i])); 
		num++;
	}
	
	if(num > 0) {
		error/=num; 
		fitness = 1.0f - error;
	} else
		fitness = -1.0f;
	return fitness;
}





void TimeSeriesExperiment::Configure(){
	mFileName =wxFileSelector("Load data set", "data/timeseries","","","*.dat",wxOPEN);
	if (mFileName!="")
		LoadFromDisk(mFileName);
	else
		StatusLine("You should give a name for the file");
}

void TimeSeriesExperiment::LoadFromDisk(const wxString &filename, bool xmlTag) {
	long unsigned temp;
	double tempDouble;
	wxInt32 pos, idx;

	mEvoModel= EvoModel::GetCurrentEvoModel();

	wxTextFile *file = new wxTextFile(filename);
	if(!file->Open()) {
		StatusLine("TimeSeriesExperiment - Couldn't open file \"" + filename + "\" for reading.");
		delete file;
		return;
	}
	
	// Parameters
	wxString line = file->GetFirstLine();
	if(!line.Mid(0, line.Index(' ')).ToULong(&temp)) {
		StatusLine("TimeSeriesExperiment - Couldn't load, no NumInput.");
		delete file;
		return;
	}
	mNumInput = (wxUint8)temp;




	if(!line.Mid(line.Index(' ')).ToULong(&temp)) {
		StatusLine("TimeSeriesExperiment - Couldn't load, no NumOutput.");
		delete file;
		return;
	}
	mNumOutput = (wxUint8)temp;

	line = file->GetNextLine();
	if(!line.ToULong(&temp)) {
		StatusLine("TimeSeriesExperiment - Couldn't load, no SetLength.");
		delete file;
		return;
	}
	mSetLength = (wxUint16)temp;
	mCurIndex = mSetLength - 1;

	// Input sets
	mInput = new float *[mSetLength];
	for(wxUint16 i=0; i<mSetLength; i++)
		mInput[i] = new float[mNumInput];
	
	for(wxUint8 num=0; num<mNumInput; num++) {
		line = file->GetNextLine();
		pos = 0;
		
	//debug
 //size_t const
 size_t bb=line.Length();

		for(wxUint16 i=0; i<mSetLength; i++) {
			idx = line.Mid(pos).Index(' ');
			if(!line.Mid(pos, idx == wxNOT_FOUND ? wxSTRING_MAXLEN : idx).ToDouble(&tempDouble)) {
				StatusLine("TimeSeriesExperiment - Couldn't load input data, not enough values available.");
				return;
			}
			pos += idx + 1;
			mInput[i][num] = (float)tempDouble;
		}
	}

	// Output sets
	mOutput = new float *[mSetLength];
	for(wxUint16 i=0; i<mSetLength; i++)
		mOutput[i] = new float[mNumOutput];

	// Target sets
	mTarget = new float *[mSetLength];
	for(wxUint16 i=0; i<mSetLength; i++)
		mTarget[i] = new float[mNumOutput];

	for(wxUint8 num=0; num<mNumOutput; num++) {
		line = file->GetNextLine();
		pos = 0;
		
		for(wxUint16 i=0; i<mSetLength; i++) {
			idx = line.Mid(pos).Index(' ');
			if(!line.Mid(pos, idx == wxNOT_FOUND ? wxSTRING_MAXLEN : idx).ToDouble(&tempDouble)) {
				StatusLine("TimeSeriesExperiment - Couldn't load input data, not enough values available.");
				return;
			}
			pos += idx + 1;
			mTarget[i][num] = (float)tempDouble;
		}
	}
	
	//set for drawing experiment

	mOutputPoints=new PointListT[mNumOutput];
	mTargetPoints=new PointListT[mNumOutput];

	//adjusting evomodel parameters
	if (xmlTag!=true && mEvoModel==NULL)
		StatusLine("WARNING: The evomodel parameters couldn't be set correctly according to the datas");

	if (xmlTag!=true && mEvoModel!=NULL){
		wxUint32 cycleTime=1000; // [milli seconds] //long enoug to be sure, that the network has got enough time to compute all..
		if (mSetLength > 4.2e+006)
			StatusLine("warning, getting close to overflow, too many datas...");
		wxUint32 lifeTime=(wxUint32) mSetLength*1; // [seconds] //=mSetlength * mCycleTime
		mEvoModel->SetCycleTime(cycleTime);
		mEvoModel->SetLifetime(lifeTime);
		mEvoModel->SetNumEpochs(1); //only one epoch, because all datas are proceeded in one epoch
		mEvoModel->SetEvoMode(SIMULATION);
		mEvoModel->SetNumGenerations(1000);
		StatusLine("New evolution settings has been set according to time series analysis settings");
	}
		file->Close();
		delete file;
	
}

void TimeSeriesExperiment::PaintExperiment(wxDC& dc, wxPoint& position, wxSize& size) {

	/* Graphics */
	dc.SetBrush(*wxWHITE_BRUSH);
	
	wxUint16 temp;
	if (mSetLength!=0)
		temp = (size_t) size.GetWidth()/mSetLength;
	else
	 temp = 2;

	if (temp==0)
		temp=1;

	const wxUint16 pixPerStep = temp;
	const wxUint16 graphHeight = 120;
	wxPoint *point;
	

	// clear lists.
	if((mOutputPoints[0].GetCount() > (size_t) size.GetWidth()/pixPerStep) || mCurIndex == 0)
	{
		for (wxUint16 i=0; i<mNumOutput; i++){
		mOutputPoints[i].Clear();
		mTargetPoints[i].Clear();
		}
	}
	
	position.y -= 130;

	for (wxUint16 i = 0; i < mNumOutput; i++){
		position.y += 130;
	
		// Draw 
		dc.SetPen(*wxBLACK_PEN);
		dc.DrawText("Output", position.x + 120, position.y + 70);
		dc.SetTextForeground(*wxGREEN);
		dc.DrawText("Target", position.x + 175, position.y + 70);
		dc.SetTextForeground(*wxRED);


		dc.SetPen(*wxBLACK_PEN);
		int x = mOutputPoints[0].GetCount()* pixPerStep;
		dc.DrawRectangle(position.x, position.y + 90, position.x+size.GetWidth(), graphHeight+1);
		

		point = new wxPoint(x, graphHeight-5 - (int) ((float)(graphHeight-10)*fabs(mOutput[mCurIndex][i])));
		mOutputPoints[i].Append( point);			point = new wxPoint(x, graphHeight-5 - (int) ((float)(graphHeight-10)*fabs(mTarget[mCurIndex][i])));
		mTargetPoints[i].Append(point);
	
		dc.DrawLines((wxList*) &mOutputPoints[i], position.x, position.y + 90);
		dc.SetPen(*wxGREEN_PEN);
		dc.DrawLines((wxList*) &mTargetPoints[i], position.x, position.y + 90);
		dc.SetPen(*wxRED_PEN);
	}
}


wxInt8 TimeSeriesExperiment::InitRobot() {
	mDummy = (Dummy *)mRobot;
	return 0;
}

wxInt8 TimeSeriesExperiment::InitEpoch(wxUint16 cycle) {
	if (cycle==0)
		mCurIndex=mSetLength-1;
	return 0;
}

void TimeSeriesExperiment::LogLegenda(Logger *logger) {
	logger->WriteString(wxString::Format("NeuronInput#%d Output#%d Target#%d\n", GetNumSensorNeurons(), GetNumOutputNeurons(), GetNumOutputNeurons()));
}

void TimeSeriesExperiment::Log(Logger *logger) {
	for(wxUint16 i=0; i<GetNumSensorNeurons(); i++)
		logger->WriteString(wxString::Format("%.2f ", mInput[mCurIndex][i]));
	for(wxUint16 i=0; i<GetNumOutputNeurons(); i++){
		logger->WriteString(wxString::Format("%.2f ",mOutput[mCurIndex][i]));
		logger->WriteString(wxString::Format("%.2f ", mTarget[mCurIndex][i]));
	}
	
	logger->WriteString(wxString::Format("\n"));
}

/* --- XML --- */
void TimeSeriesExperiment::Load (XMLElement *myself, XMLLoader *loader) {
			mFileName = myself->GetAttribute("fileName")->GetValue();
			bool xmlTag=true;
			LoadFromDisk(mFileName,xmlTag );
}

XMLElement* TimeSeriesExperiment::SaveSelf () {
		XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
		result->AddAttribute(new XMLAttribute("fileName",mFileName));
return result; 
}

wxString TimeSeriesExperiment::GetShortXMLName() { return TimeSeriesExperiment::GetStaticXMLType(); }
