/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "genomelayout.h"
#include "../utils/randomtools.h"
#include "../nn/neuralnetworktopology.h"
#include "../status.h"


void GenomeLayout::AddDigitalParam(ParamScope scope, ParamID id, const char* name, wxUint8 encodingSize, wxInt32 defaultValue, wxInt32 lowerBound)
{
	ParamDesc desc = {id, "", encodingSize, false, lowerBound, 0.0f, 0.0f, (float)defaultValue};
	strcpy(desc.name, name);
	AddParam(scope, desc);
}


void GenomeLayout::AddAnalogParam(ParamScope scope, ParamID id, const char* name, wxUint8 encodingSize, float defaultValue, float lowBound, float highBound)
{
	ParamDesc desc = {id, "", encodingSize, true, 0, lowBound, highBound, defaultValue};
	strcpy(desc.name, name);
	AddParam(scope, desc);
}


void GenomeLayout::AddParam(ParamScope scope, const ParamDesc &paramDesc)
{
	switch(scope)
	{
		case PARAM_NEURALNETWORK:
			mNetworkParams.push_back(paramDesc);
			mNetworkEncodingSize += paramDesc.encodingSize;
			break;
			
		case PARAM_NEURON:
			mNeuronParams.push_back(paramDesc);
			mNeuronEncodingSize += paramDesc.encodingSize;
			break;		
			
		case PARAM_SYNAPSE:
			mSynapseParams.push_back(paramDesc);
			mSynapseEncodingSize += paramDesc.encodingSize;
			break;
	};
}


void GenomeLayout::ClearParameters()
{
	mNetworkParams.clear();
	mNeuronParams.clear();
	mSynapseParams.clear();
	mNetworkEncodingSize = mNeuronEncodingSize = mSynapseEncodingSize = 0;
}


wxUint32 GenomeLayout::GetGenomeLength()
{
	return mNetworkEncodingSize + mNumNeurons * mNeuronEncodingSize + mNumSynapses * mSynapseEncodingSize;	
}



// NOT AWARE OF NON EVOLVED PARAMETER // TO BE VERIFIED AND CORRECTED
	// ANYWAY: USEFULNESS TO BE CONFIRMED


/*void GenomeLayout::RandomizeGenome(Genome* genome)
{
	if(genome->GetLength() != GetGenomeLength())
		return;
	
	// Randomize global parameters.
	wxUint16 pos = 0;
	for(int i = 0; i < mNetworkParams.size(); i++)
	{
		wxUint16 r = rand();
		for(int j = 0; j < mNetworkParams[i].encodingSize; j++, pos++)
		{
			genome->SetValue(pos, r & 0x1);
			r >>= 1;
		}
	}
	
	// Randomize neuron parameters.
	RandomizeParameter(genome, mNeuronParams, pos, mNeuronEncodingSize, mNumNeurons);
	pos += mNumNeurons * mNeuronEncodingSize;
	
	// Randomize synapse parameters.
	RandomizeParameter(genome, mSynapseParams, pos, mSynapseEncodingSize, mNumSynapses);
}

void GenomeLayout::RandomizeParameter(Genome *genome, std::vector<ParamDesc> &params, wxUint16 startingPos, wxUint16 encodingSize, wxUint32 numElement)
{
	wxUint16 posInEnc = 0;
	// Recurse on each parameters
	for(int i = 0; i < params.size(); i++)
	{
		wxUint16 enc = params[i].encodingSize;
		
		// Choose a random mean [0.1, 0.9]
		float mean = 0.1f + 0.8f * (float) rand() / (float) RAND_MAX;
		
		// For each element (neuron/synapse).
		for(int j = 0; j < numElement; j++)
		{
			// Generate a random value.
			wxUint16 r;
			if(enc == 1)
			{
				r = RandomTools::Flip(mean);
			}
			else
			{
				float unitary = rand() / (float) RAND_MAX;
				
				// We apply a piece-wise linear function to shift the mean to m.
				if(unitary < 0.5f)
					unitary = 2.0f * mean * unitary;
				else
					unitary = 2.0f * unitary * (1 - mean) - 1.0f + 2.0f * mean;
				
				// Map this unitary random value to the parameter size. We'll use only
				// the "enc" most significant bits.
				r = unitary * 0xFFFF;
				r >>= 16 - enc;
			}
			
			// Copy to genome.
			for(int k = 0; k < enc; k++)
			{
				genome->SetValue(startingPos + j*encodingSize + posInEnc, r & 0x1);
				r >>= 1;
			}
		}
			
		posInEnc += enc;
	}
}


wxUint32 GenomeLayout::CountCrossoverLocations()
{
	return mNetworkParams.size() + mNeuronParams.size() * mNumNeurons + mSynapseParams.size() * mNumSynapses - 1;
}


wxUint32 GenomeLayout::GetNthCrossoverLocation(wxUint32 index)
{
	wxUint32 position = 0;
	
	if(index < mNetworkParams.size())
	{
		// We are in the global parameter zone.
		for(int i = 0; i <= index; i++)
			position += mNetworkParams[i].encodingSize;
	}
	else if(index < mNetworkParams.size() + mNeuronParams.size() * mNumNeurons)
	{
		// We are in the neuron parameter zone.
		index -= mNetworkParams.size();		
		int neuronNum = index / mNeuronParams.size();
		int paramNum = index % mNeuronParams.size();
		
		position = neuronNum * mNeuronEncodingSize;
		for(int i = 0; i <= paramNum; i++)
			position += mNeuronParams[i].encodingSize;
	}
	else
	{
		// We are in the synapse parameter zone.
		index -= mNetworkParams.size() + mNeuronParams.size() * mNumNeurons;
		int synapseNum = index / mSynapseParams.size();
		int paramNum = index % mSynapseParams.size();
		
		position = synapseNum * mSynapseEncodingSize;
		for(int i = 0; i <= paramNum; i++)
			position += mSynapseParams[i].encodingSize;
	}
	
	return position;
}

*/

void GenomeLayout::DecodeNetworkParameters(Genome* genome, std::vector<ParamRecord> &params)
{
#if !OldLayout

	// Network parameters are at the beginning of the genome.
	DecodeParameters(genome, 0, 0, mNetworkParams, params);
#endif
}

void GenomeLayout::DecodeNeuronParameters(Genome* genome, wxUint32 index, std::vector<ParamRecord> &params)
{
	if(index >= mNumNeurons)
		return;
#if OldLayout
	DecodeParameters(genome, index, mNetworkEncodingSize, mNeuronParams, params);
#else		
	DecodeParameters(genome, index, mNetworkEncodingSize + index * mNeuronEncodingSize, mNeuronParams, params);
#endif
}

void GenomeLayout::DecodeSynapseParameters(Genome* genome, wxUint32 index, std::vector<ParamRecord> &params)
{
	if(index >= mNumSynapses)
		return;
	
#if OldLayout
	DecodeParameters(genome, index, mNetworkEncodingSize + mNumNeurons * mNeuronEncodingSize, mSynapseParams, params);
#else
	DecodeParameters(genome, index, mNetworkEncodingSize + mNumNeurons * mNeuronEncodingSize + index * mSynapseEncodingSize,
		mSynapseParams, params);
#endif
}

void GenomeLayout::DecodeParameters(Genome *genome, wxUint32 index, wxUint32 pos, std::vector<ParamDesc> &paramList, std::vector<ParamRecord> &params)
{
#if OldLayout
	int cumulEncSize = 0;
#endif

	// Iterate on descriptor list.
	for(std::vector<ParamDesc>::iterator iter = paramList.begin(); iter < paramList.end(); iter++)
	{
		ParamRecord pRec;
		pRec.id = iter->id;
		
		if(iter->encodingSize == 0)
			pRec.value = iter->defaultValue;
		else
		{
			wxUint32 value;
			
#if OldLayout
			genome->DecodeInt(value, pos + mNumNeurons * cumulEncSize + index * iter->encodingSize, iter->encodingSize);
			cumulEncSize += iter->encodingSize;
#else
			genome->DecodeInt(value, pos, iter->encodingSize);
			pos += iter->encodingSize;
#endif
			
			// Convert into final parameter value.
			if(iter->analog)
			{
				float f = (float) value / (float) ((1 << iter->encodingSize) - 1);
				pRec.value = f * (iter->higherAnalogBound - iter->lowerAnalogBound) + iter->lowerAnalogBound;
			}
			else
				pRec.value = iter->lowerDigitalBound + value;
			
			
		}
		
		params.push_back(pRec);
	}
}

wxUint32 GenomeLayout::GetParameterNumber(ParamScope scope)
{
	switch(scope)
	{
		case PARAM_NEURALNETWORK: return mNetworkParams.size();
		case PARAM_NEURON: return mNeuronParams.size();
		case PARAM_SYNAPSE: return mSynapseParams.size();
		default: return 0;
	};
}

GenomeLayout::ParamDesc& GenomeLayout::GetParameterDesc(ParamScope scope, wxUint32 index)
{
	switch(scope)
	{
		case PARAM_NEURALNETWORK: return mNetworkParams[index];
		case PARAM_NEURON: return mNeuronParams[index];
		case PARAM_SYNAPSE: return mSynapseParams[index];
		default:
		{
			// This is for the compiler not to complain. Should never execute.
			static ParamDesc pd;
			return pd;
		}
	};
}

void GenomeLayout::SetParameterEncodingSize(ParamScope scope, wxUint32 index, wxUint8 encodingSize)
{
	switch(scope)
	{
		case PARAM_NEURALNETWORK:
			mNetworkEncodingSize -= mNetworkParams[index].encodingSize;
			mNetworkEncodingSize += (mNetworkParams[index].encodingSize = encodingSize);
			break;
		
		case PARAM_NEURON:
			mNeuronEncodingSize -= mNeuronParams[index].encodingSize;
			mNeuronEncodingSize += (mNeuronParams[index].encodingSize = encodingSize);
			break;
		
		case PARAM_SYNAPSE:
			mSynapseEncodingSize -= mSynapseParams[index].encodingSize;
			mSynapseEncodingSize += (mSynapseParams[index].encodingSize = encodingSize);
			break;
	};
}

void GenomeLayout::SetParameterName(ParamScope scope, wxUint32 index, const char* name)
{
	::memcpy(GetParameterDesc(scope, index).name, name, (::strlen(name)+1 > ParamNameLength ? ParamNameLength : ::strlen(name)+1)); 
}
