/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _GENOMELAYOUT_H
#define _GENOMELAYOUT_H

#include <vector>
#ifdef WIN32
	#include <wx/msw/winundef.h>
#endif

#include "../nn/neuralnetwork.h"
#include "../nn/neuron.h"
#include "../nn/synapse.h"
#include "genome.h"


class NeuralNetworkTopology;

/** Old genome layout, for compatibility with old individuals. */
#define OldLayout	0


/** Max size of parameter name */
#define ParamNameLength 20

/** Scope of a parameter. */
enum ParamScope
{
	PARAM_NEURALNETWORK = 0,	///< Parameter for the whole neural network.
	PARAM_NEURON,	///< Parameter for a neuron.
	PARAM_SYNAPSE	///< Parameter for a synapse.
};



/**
 * Class describing the layout of a genome. This class is responsible of 
 * decoding a genome into a set of parameters. This responsibility imply
 * a knowledge of the genome layout. So we take advantage of this, and
 * build the following feature in the class;
 * 		- logging of the params
 * 		- parameter-aware genetic operation helpers
 *
 * The "knowledge of the layout" is modelised in the following way: the genome
 * is a list of parameter. Parameters fall in thee categories:
 *		- parameters that are global to the nn
 *		- parameters that refer to a single neuron
 *		- parameters that refer to a single synapse
 * This distinction is called the "scope" of a parameter.
 *
 * A "parameter" is modelised in the following way: it can be "analog" or "digital"
 * that is floating point or integer. In the case of analog parameter, the range
 * of the parameter must be described. It will be encoded in the genome as an 
 * integer value that uniformly samples this range. In the case of a digital parameter
 * only the low bound of the range is described, since the dynamic of the parameter
 * is given by the encoding size. For example, if the encoding size is 2 and 
 * the low bound is 10, the parameter will take values between 10 and 10+2^2-1
 * (i.e. 10-13).
 *
 * In both case, the folowing attribute are defined for a parameter:
 *		- a name (for logging, saving in xml and user interaction)
 *		- the encoding size (e.g. the # of bits used for encoding)
 *		- the default value (e.g. the value to be used when encoding size is 0)
 *		- the ID, a unique identifier used for parameter attribution (see IParameterizable)
 *
 * Note that the default value, as well as the actual values during decoding,
 * are always typed as float, even in the case of digital parameter, because it
 * would be too complicated otherwise. Of course, in the case of digital parameter,
 * the float value will be actually integer.
 */
class GenomeLayout
{
public:

	/* ------------------------- CONSTRUCTION ------------------------------ */

	/**
	 * GenomeLayout constructor.
	 * @param numEvolvNeurons Number of neurons to be evolved.
	 * @param numEvolvSynapses Number of synapses to be evolved.
	 */
	GenomeLayout(wxUint32 numEvolvNeurons, wxUint32 numEvolvSynapses) :
		mNetworkEncodingSize(0), mNeuronEncodingSize(0), mSynapseEncodingSize(0),
		mNumNeurons(numEvolvNeurons), mNumSynapses(numEvolvSynapses) {}
						
	/**
	 * GenomeLayout desctructor.
	 */
	virtual ~GenomeLayout() {}
	
	
	/* ------------------------- ADD PARAMETER ------------------------------ */
	
	/**
	 * Add a new digital parameter to be encoded in genome.
	 * @param scope Global, neuron or synapse param.
	 * @param id Unique ID of the parameter.
	 * @param name Name of the parameter, to be used in logging.
	 * @param encodingSize Number of bit to be used for encoding.
	 * @param defaultValue Default value when the parameter is not evolved.
	 * @param lowerBound Lower bound to be used during decoding (for example 0b10 with a bound of ten is 10 + 2 = 12)
	 */
	virtual void AddDigitalParam(ParamScope scope, ParamID id, const char* name, wxUint8 encodingSize, wxInt32 defaultValue = 0, wxInt32 lowerBound = 0);
	
	/**
	 * Add a new analog (floating point) parameter to be encoded in genome.
	 * @param scope Global, neuron or synapse param.
	 * @param id Unique ID of the parameter.
	 * @param name Name of the parameter, to be used in logging.
	 * @param encodingSize Number of bit to be used for encoding.
	 * @param defaultValue Default value when the parameter is not evolved.
	 * @param lowBound Low bound of the param.
	 * @param highBound High bound of the param.
	 */
	virtual void AddAnalogParam(ParamScope scope, ParamID id, const char* name, wxUint8 encodingSize, float defaultValue = 0.0f, float lowBound = 0.0f, float highBound = 1.0f);
	
	
	/** Clears all parameters from the genome layout. */
	virtual void ClearParameters();

	/**
	 * Calculate the lenght of the genome to encode the currently added parameters.
	 * @return Length of the genome, in bits.
	 */
	virtual wxUint32 GetGenomeLength();
	
	
	/* ------------------------- ACCESSORS ------------------------------ */
	
	/** Set number of evolvable neurons. */
	virtual void SetEvolvableNeuronNumber(wxUint32 num) {mNumNeurons = num;}
	
	/** Set number of evolvable synapse. */
	virtual void SetEvolvableSynapseNumber(wxUint32 num) {mNumSynapses = num;}
	
	/** Return the number of parameters for a particular scope (nn, neuron, synapse). */
	virtual wxUint32 GetParameterNumber(ParamScope scope);
	
	/** Return whether a particular parameter is analog (floating point) or digital. */
	virtual bool GetParameterAnalog(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).analog;}

	/** Return the low bound of a parameter. */
	virtual float GetAnalogParameterLowBound(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).lowerAnalogBound;}
	
	/** Set the low bound of a parameter. */
	virtual void SetAnalogParameterLowBound(ParamScope scope, wxUint32 index, float lowBound) {GetParameterDesc(scope, index).lowerAnalogBound = lowBound;}
	
	/** Return the high bound of a parameter. */
	virtual float GetAnalogParameterHighBound(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).higherAnalogBound;}
	
	/** Set the high bound of a parameter. */
	virtual void SetAnalogParameterHighBound(ParamScope scope, wxUint32 index, float highBound) {GetParameterDesc(scope, index).higherAnalogBound = highBound;}
	
	/** Return the low bound of a parameter. */
	virtual wxInt32 GetDigitalParameterLowBound(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).lowerDigitalBound;}
	
	/** Set the low bound of a parameter. */
	virtual void SetDigitalParameterLowBound(ParamScope scope, wxUint32 index, wxInt32 lowBound) {GetParameterDesc(scope, index).lowerDigitalBound = lowBound;}
	
	/** Return the default value of a parameter. */
	virtual float GetParameterDefaultValue(ParamScope scope, wxUint32 index)  {return GetParameterDesc(scope, index).defaultValue;}

	/** Set the default value of a parameter. */
	virtual void SetParameterDefaultValue(ParamScope scope, wxUint32 index, float defaultValue) {GetParameterDesc(scope, index).defaultValue = defaultValue;}
	
	/** Return the encoding size (in bits) of a parameter. */
	virtual wxUint8 GetParameterEncodingSize(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).encodingSize;}
	
	/** Set the encoding size (in bits) of a parameter. */
	virtual void SetParameterEncodingSize(ParamScope scope, wxUint32 index, wxUint8 encodingSize);
	
	/** Return the name of a parameter. */
	virtual wxString GetParameterName(ParamScope scope, wxUint32 index) {return wxString(GetParameterDesc(scope, index).name);}
	
	/** Set the name of a parameter. Size in character of the name must be less than ParamNameLength-1. */
	virtual void SetParameterName(ParamScope scope, wxUint32 index, const char* name);
	
	/** Return the ID of a parameter. */
	virtual ParamID GetParameterID(ParamScope scope, wxUint32 index) {return GetParameterDesc(scope, index).id;}
	
	/** Set the ID of a parameter. */
	virtual void SetParameterID(ParamScope scope, wxUint32 index, ParamID id) {GetParameterDesc(scope, index).id = id;}
	
	
	/* ------------------------- DECODING ------------------------------ */
	
	/**
	 * Decode the network parameters.
	 * @param genome Source genome.
	 * @param network Target neural network.
	 */
	virtual void DecodeNetworkParameters(Genome* genome, NeuralNetwork* network)
	{
		std::vector<ParamRecord> params;
		DecodeNetworkParameters(genome, params);
		network->SetParameters(params);
	}
	
	/**
	 * Decode network parameters.
	 * @param genome Source genome.
	 * @param params On output: a list of parameter (ID, Value) pairs.
	 */
	virtual void DecodeNetworkParameters(Genome* genome, std::vector<ParamRecord> &params);
	
	/**
	 * Decode neuron parameters.
	 * @param genome Source genome.
	 * @param index Neuron index.
	 * @param neuron Destination neuron.
	 */
	virtual void DecodeNeuronParameters(Genome* genome, wxUint32 index, Neuron* neuron)
	{
		std::vector<ParamRecord> params;
		DecodeNeuronParameters(genome, index, params);
		neuron->SetParameters(params);
	}
	
	/**
	 * Decode neuron parameters.
	 * @param genome Source genome.
	 * @param index Neuron index.
	 * @param params On output: a list of parameter (ID, Value) pairs.
	 */
	virtual void DecodeNeuronParameters(Genome* genome, wxUint32 index, std::vector<ParamRecord> &params);
	
	/**
	 * Decode synapse parameters.
	 * @param genome Source genome.
	 * @param index Synapse index.
	 * @param synapse Destination synapse.
	 */
	virtual void DecodeSynapseParameters(Genome* genome, wxUint32 index, Synapse* synapse)
	{
		std::vector<ParamRecord> params;
		DecodeSynapseParameters(genome, index, params);
		synapse->SetParameters(params);				
	}
	
	/**
	 * Decode synapse parameters.
	 * @param genome Source genome.
	 * @param index Synapse index.
	 * @param params On output: a list of parameter (ID, Value) pairs.
	 */
	virtual void DecodeSynapseParameters(Genome* genome, wxUint32 index, std::vector<ParamRecord> &params);
	
	
	/* ------------------------- GENETIC OPERATIONS ------------------------------ */
	
						
	/**
	 * Randomize the content of the genome.
	 * @param genome Genome to be randomized. Its size MUST be of the same value than GetGenomeLength().
	 */
//	virtual void RandomizeGenome(Genome* genome);
	
	/**
	 * Count the number of crossover locations location between parameter.
	 * @retrun Number of crossover locations.
	 */
//	virtual	wxUint32 CountCrossoverLocations();
	
	/**
	 * Return the bit index of the ith crossover location.
	 * @param i Index of the crossover location to be determined.
	 * @return Position of the first bit after the crossover location.
	 */
//	virtual wxUint32 GetNthCrossoverLocation(wxUint32 i);
	
protected:

	struct ParamDesc;
	
	/* ------------------------- IMPLEMENTATION FUNCTION ------------------------------ */
	
	/** Helper function to add a parameter descriptor in the appropriate list. */
	virtual void AddParam(ParamScope scope, const ParamDesc &paramDesc);
	
	/** Helper function to decode a list of parameter into an IParametrizable. */
	virtual void DecodeParameters(Genome *genome, wxUint32 index, wxUint32 pos, std::vector<ParamDesc> &paramList, std::vector<ParamRecord> &params);

	/** Helper function to randomize the genome for a list of parameters. */
	//virtual void RandomizeParameter(Genome *genome, std::vector<ParamDesc> &params, wxUint16 startingPos, wxUint16 encodingSize, wxUint32 numElement);

	virtual ParamDesc& GetParameterDesc(ParamScope scope, wxUint32 index);


	/* ------------------------- INTERNAL DATA TYPES ------------------------------ */

	/** Description of a parameter to be encoded in a genome. */
	struct ParamDesc
	{
		ParamID id;					///< Unique ID of the parameter.
		char name[ParamNameLength];	///< Name of the parameter (for logging).
		wxUint8 encodingSize;		///< Number of bits to be used for encoding.
		bool analog;				///< Is the parameter floating point or digital?
		wxInt32 lowerDigitalBound;	///< Lower bound (digital parameter).
		float lowerAnalogBound;		///< Lower bound (analog parameter).
		float higherAnalogBound;	///< Higher bound (analog parameter).
		float defaultValue;			///< Default value (when not evolved).
	};


	/* ------------------------- DATA MEMBERS ------------------------------ */

	std::vector<ParamDesc> mNetworkParams;	///< Array of network parameters.
	std::vector<ParamDesc> mNeuronParams;	///< Array of neuron parameters.
	std::vector<ParamDesc> mSynapseParams;	///< Array of synapse parameters.
	
	wxUint16 mNetworkEncodingSize;			///< Total size in bit of the network parameters.
	wxUint16 mNeuronEncodingSize;			///< Total size in bit of the parameters for one neuron.
	wxUint16 mSynapseEncodingSize;			///< Total size in bit of the parameters for one synapse.
	
	wxUint32 mNumNeurons;					///< Number of neurons to be evolved.
	wxUint32 mNumSynapses;					///< Number of synapses to be evolved.
	
	
};

#endif
