/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/file.h>
#include <wx/textfile.h>
#include <math.h>

#include "../status.h"
#include "../evomodel.h"
#include "../exp/experiment.h"
#include "../nn/neuralnetwork.h"
#include "../nn/neuralnetworkfactory.h"
#include "../utils/iutils.h"
#include "genome.h"
#include "population.h"

#include "individual.h"

Individual::Individual(Genome *genome) {
	mGenome = genome;
	mFitness = 0.0f;
	mEpochFitness = NULL;
	mEpochCycles = NULL;
	mNumEpochs=2;
}

Individual::~Individual() {
	delete mGenome;
	if(mEpochFitness)
		delete [] mEpochFitness;
	if(mEpochCycles)
		delete [] mEpochCycles;
}

wxInt8 Individual::Evaluate(EvoModel *evoModel, Experiment *experiment, Population *population, wxString *nnSavename) {
		return NormalEvaluate(evoModel, experiment, population, nnSavename);
}

wxInt8 Individual::NormalEvaluate(EvoModel *evoModel, Experiment *experiment, Population *population, wxString *nnSavename) {
	wxInt8 ret;
	wxUint16 currentCycle;
	wxUint32 lifetime = evoModel->GetLifetime() * 1000, cycleTime = evoModel->GetCycleTime();
	wxUint32 numCycles = lifetime / cycleTime;
	StatusLine(wxString::Format("-- NumCycles: %d ", numCycles));
	float epochFitness;
	wxUint32 lastCycleStart, now=0;
	mFitness = 0.0f;
	wxInt8 flagInitEpoch;

	mNumEpochs = evoModel->GetNumEpochs();
	mEpochFitness = new float[mNumEpochs];
	mEpochCycles = new wxUint16[mNumEpochs];

	NeuralNetwork *neuralNetwork = evoModel->GetNNFactory()->Create(mGenome);
	if(!neuralNetwork)
		return 10;
	experiment->SetNeuralNetwork(neuralNetwork);

	neuralNetwork->AllocActivityStorage((wxUint16)(ceil((float)lifetime / cycleTime)));
	for(wxUint16 epoch=0; epoch<mNumEpochs; epoch++) { // epochs
		StatusLine(wxString::Format("-- Epoch %d", epoch));

		StatusLine("--- InitEpoch");
		wxStopWatch stopwatch;	// This starts the stopwatch. The Time() method returns elapsed milliseconds.

		currentCycle = 0; lastCycleStart = 0;
		flagInitEpoch=1;
		while( flagInitEpoch !=0 && evoModel->ContinueRun() && 
				(!population || !population->GetInterrupted()) ) { // init epoch
			now = stopwatch.Time();
			if(evoModel->GetEvoMode()==SIMULATION || (wxUint32)(now - lastCycleStart) > cycleTime) { // cycle
				lastCycleStart = now;
				flagInitEpoch=experiment->InitEpoch(currentCycle);
				experiment->GetRobot()->NotifyObservers();
				evoModel->NotifyView();
				iYield();
				currentCycle++;
			} // end of cycle			
		} // end of init epoch

		StatusLine("--- Live");
		stopwatch.Start();	// This starts the stop watch. The Time() method returns elapsed milliseconds.

		epochFitness = 0.0f;
		currentCycle = 0;
		lastCycleStart = 0;
		
		// Here we must initialize the synapse randomly.
		if((ret = experiment->InitNeuralNetwork(neuralNetwork)))
			StatusLineErr("Experiement - Error while initializing neural network.", ret);

		// In real-time mode (physical robots), the lifetime is used and the actually reached numCycles is always less than the theoretical numCycles.
		// In simulation mode, we just don't wait after each cycle and the numCycles will be completed (hopefully) in less than the real lifetime.		
		
		while(currentCycle < numCycles && (now = stopwatch.Time()) < lifetime && 
						evoModel->ContinueRun() && (!population || !population->GetInterrupted())) { // life
			if(evoModel->GetEvoMode()==SIMULATION || (wxUint32)(now - lastCycleStart) > cycleTime) { // cycle
				lastCycleStart = now;
				neuralNetwork->SetCurrentCycle(currentCycle);
				
				// 1) Send actuator commands resulting from previous NN cycle
				if((ret = experiment->ProcessNNOutput()))
					StatusLineErr("Individual - Error while processing neural network output.", ret);

				// 2) Get sensor values from robot
				if((ret = experiment->ReadSensors()))
					StatusLineErr("Individual - Error while reading sensor data.", ret);

				// 3) Apply them to NN inputs
				if((ret = experiment->SetNNInput()))
					StatusLineErr("Individual - Error while setting neural network input.", ret);

				// 4) Run a NN cycle
				neuralNetwork->Compute();

				epochFitness += experiment->ComputeFitness();

				if(experiment->KillIndividual()) {
					StatusLine("--- KILLED!");
					break;
				}

				experiment->GetRobot()->NotifyObservers();
				experiment->NotifyObservers(); // log experiment data if enable
				evoModel->NotifyView();
				iYield();
				currentCycle++;
			} // end of cycle
		} // end of life time

		// check for interruption
		if(!evoModel->ContinueRun()) {
			delete neuralNetwork;
			experiment->SetNeuralNetwork(NULL);
			return -1;
		}

		// log neural activity
		if(nnSavename)
			neuralNetwork->SaveActivity(*nnSavename);

		mEpochCycles[epoch] = currentCycle;
		// In simulation mode, the epochFitness is simply normalized with respect to the theoretical numCycles since there is no variation 
		//   possible in the number of cycles except when the individual is killed.
		// In real-time mode, the actual number of cycles is often less than the theoretical numCycles. Therefore, the epochFitness has to be
		//   normalized with respect to the actual number of cycles (currentCycle). In order to take into account the case when the individual
		//   is killed, the denominator is multiplied by the ratio liftime/now.
		if(evoModel->GetEvoMode()==SIMULATION) {
			epochFitness /= (double)numCycles; // numCycles is used instead of currentCycle to decrease the fitness of killed individual
		} else {
			if(currentCycle != 0) epochFitness /= ((double)currentCycle * ((double)lifetime / now));
			else epochFitness = 0;
		}
		mEpochFitness[epoch] = epochFitness;
		mFitness += epochFitness;
		StatusLine(wxString::Format("-- %d cycles, fitness %.3f", currentCycle, epochFitness));
	} // end of epochs loop

	mFitness /= mNumEpochs;
	StatusLine(wxString::Format("- Fitness: %.3f", mFitness));

	experiment->SetNeuralNetwork(NULL);
	delete neuralNetwork;
	return 0;
}


void Individual::LogLegenda(Logger *logger) {
	wxFAIL_MSG ("this function is not anymore in use, beware");
}

void Individual::Log(Logger *logger) {
	for(wxUint16 i=0; i<mNumEpochs; i++)
		logger->WriteString(wxString::Format("%d ", mEpochCycles[i]));
	for(wxUint16 i=0; i<mNumEpochs; i++)
		logger->WriteString(wxString::Format("%.3f ", mEpochFitness[i]));
	logger->WriteString(wxString::Format("%.3f\n", mFitness));
}

wxInt8 Individual::SaveToDisk(wxString filename) {
	wxFile *file = new wxFile(filename.c_str(), wxFile::write);
	if(!file->IsOpened()) {
		StatusLine("Individual - Couldn't open file \"" + filename + "\" for writing.");
		delete file;
		return 1;
	}
	mGenome->WriteToFile(file);
	file->Close();
	delete file;
	return 0;
}

wxInt8 Individual::WriteToFile(wxFile *file) {
	return mGenome->WriteToFile(file);
}

Individual *Individual::ReadFromFile(wxTextFile *file, wxUint16 curGenomeLength) {
	Genome *genome = Genome::ReadFromFile(file, curGenomeLength);
	if(!genome)
		return NULL;
	return new Individual(genome);
}

Individual* Individual::LoadFromDisk(wxString filename, wxUint16 curGenomeLength) {
	wxTextFile *file = new wxTextFile(filename);
	if(!file->Open()) {
		StatusLine("Individual - Couldn't open file \"" + filename + "\" for reading.");
		delete file;
		return NULL;
	}
	file->GoToLine((wxUint32)-1);
	Genome *genome = Genome::ReadFromFile(file, curGenomeLength);
	if(!genome)
		return NULL;
	file->Close();
	delete file;
	return new Individual(genome);
}
