/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _INDIVIDUAL_H
#define _INDIVIDUAL_H

class wxTextFile;

class Experiment;
class EvoModel;
class Population;
class Genome;
class Logger;

/**
 * One Individual, part of a Population, consisting of a Genome encoding
 * a NeuralNetwork. This NeuralNetwork is used to control a Robot by an
 * Experiment.
 * @see Population
 * @see Genome
 * @see NeuralNetwork
 * @see Robot
 * @see Experiment
 */
class Individual
{
public:
	/**
	 * Individual constructor
	 * @param genome Genome for this individual (will be deleted by Individual upon destruction).
	 */
	Individual(Genome *genome);

	/** Individual destructor. */
	~Individual();

	/* ----------------------- EVALUATION --------------------- */

	/**
	 * Evaluate the individual.
	 * @param experiment Experiment to use in evaluation.
	 * @param nnSavename Filename to save neural activity to.
	 * @return a error code
	 */
	wxInt8 Evaluate(EvoModel *evoModel, Experiment *experiment, Population *population, wxString *nnSavename);

	/**
	 * Test the individual.
	 * @param experiment Experiment to use in evaluation.
	 * @return a error code
	 */
	wxInt8 Test(EvoModel *evoModel, Experiment *experiment);

	/* ------------------------ ACCESSORS --------------------- */

	/**
	 * Accessor method
	 * @return the genome of this individual
	 */
	Genome* GetGenome() { return mGenome; }

	/**
	 * Accessor method
	 * @param genome is the new genome of this individual
	 */
	void SetGenome(Genome* genome) { mGenome = genome; }

	/**
	 * Accessor method
	 * @return The fitness of this individual.
	 */
	float GetFitness() { return mFitness; }

	/** Returns total number of epochs. */
	wxUint16 GetNumEpochs() { return mNumEpochs; }

	/** Returns array with fitness value for each separate epoch. */
	float *GetEpochFitness() { return mEpochFitness; }
	
	/** Returns array with number of cycles for each separate epoch. */
	wxUint16 *GetEpochCycles() { return mEpochCycles; }

	/* -------------------------- LOGGING -------------------------- */
	
	/** Called by population when legenda should be logged. */
	void LogLegenda(Logger *logger);

	/** Called by population when new data should be logged. */
	void Log(Logger *logger);	
	
	/* ----------------------------- I/O ----------------------------- */
	
	/** Save individual to the specified file. */
	wxInt8 SaveToDisk(wxString filename);
	/** Load an individual from the specified file. */
	static Individual *LoadFromDisk(wxString filename, wxUint16 curGenomeLength);

	/** Writes individual to file. */
	wxInt8 WriteToFile(wxFile *file);
	/**load an individual from a population stored on the disk*/
	static Individual *ReadFromFile(wxTextFile *file, wxUint16 curGenomeLength);

protected:
	Genome* mGenome; 				/**< The genome of this individual. */
	float mFitness;					/**< The fitness of this individual. */

	wxUint16 mNumEpochs;		/**< Total number of epochs. */
	float *mEpochFitness;		/**< Array containing fitness value for each epoch. */
	wxUint16 *mEpochCycles;	/**< Array containing number of cycles for each epoch. */

	wxInt8 NormalEvaluate(EvoModel *evoModel, Experiment *experiment, Population *population, wxString *nnSavename);

};

#endif /* _INDIVIDUAL_H */
