/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/file.h>
#include <wx/textfile.h>

#include "../main.h"
#include "../status.h"
#include "../evomodel.h"
#include "../robot/robot.h"
#include "../utils/randomtools.h"
#include "../exp/experiment.h"
#include "igenoperator.h"
#include "individual.h"
#include "genome.h"
#include "population.h"

Population::Population(wxUint16 size, wxUint16 curGeneration, bool newIndividuals, wxUint16 genomeLength) {
	mSize = size;
	mIndividuals = new Individual *[mSize];
	if(newIndividuals)
		for(wxUint16 i=0; i<mSize; i++)
			mIndividuals[i] = new Individual(new Genome(genomeLength,curGeneration,i));
	else		
		for(wxUint16 i=0; i<mSize; i++)
			mIndividuals[i] = NULL;
	mBestIndividual = NULL;
	mMaxFitness = 0.0f;
	mAvgFitness = 0.0f;
	mWorstFitness = 10000000.0f;	
	mCurIndividual = 0;
	mCurGeneration = curGeneration;
	mInterrupt = 0;
	mNumEpochs=2;
}

Population::Population(Population *population) {
	mSize = population->GetSize();
	mIndividuals = new Individual *[mSize];
	for(int i=0; i<mSize; i++)
		mIndividuals[i] = new Individual(new Genome(population->GetIndividual(i)->GetGenome()));
	mBestIndividual = population->GetBestIndividual();
	mMaxFitness = population->GetMaxFitness();
	mAvgFitness = population->GetAvgFitness();
	mWorstFitness = population->GetWorstFitness();
	mCurGeneration = population->GetCurGeneration();
}

Population::~Population() {
	for(wxUint16 i=0; i<mSize; i++)
		delete mIndividuals[i];
	delete [] mIndividuals;
}

wxInt8 Population::ApplyOperators(IGenOperator **genOperators, wxUint8 numGenOperators) {
	wxInt8 ret;
	for(wxUint8 i=0; i<numGenOperators; i++)
		for(wxUint16 j=0; j<mSize/2; j++)
			if((ret = genOperators[i]->Apply(mIndividuals[j*2]->GetGenome(), mIndividuals[j*2+1]->GetGenome())))
				return ret;
	return 0;
}

wxInt8 Population::Evaluate(EvoModel *evoModel, Experiment *experiment, bool resume) {
	wxInt8 ret;
	float fitness;
	Robot *robot = experiment->GetRobot();
	wxString saveName;

	if(!resume) {
		mMaxFitness = 0;
		mWorstFitness = 10000000.0f;
		mAvgFitness = 0;
		mCurIndividual = 0;
		mBestIndividual = mIndividuals[0];
	} else
		StatusLine(wxString::Format("Resuming evaluation from individual %d", mCurIndividual));
	
	// For each individual
	for(; mCurIndividual<mSize; mCurIndividual++) {
		// Evaluate
		StatusLine(wxString::Format("Individual %d", mCurIndividual));
		if((ret = mIndividuals[mCurIndividual]->Evaluate(evoModel, experiment, this, NULL)))
			return ret;

		// Maintain statistical values
		fitness = mIndividuals[mCurIndividual]->GetFitness();
		
		if(mMaxFitness < fitness) {
			mMaxFitness = fitness;
			mBestIndividual = mIndividuals[mCurIndividual];
		} else if(mWorstFitness > fitness)
			mWorstFitness = fitness;
		mAvgFitness += fitness;
		
		// Make sure that everyone is aware
		SetChanged();
		NotifyObservers();
		evoModel->NotifyView();		
		
		// Change batteries if necessary
		if(!robot->BatteriesOK())
			robot->ChangeBatteries();
	}
	mAvgFitness /= mSize;
	return 0;
}

wxInt8 Population::Test(EvoModel *evoModel, Experiment *experiment) {
	wxInt8 ret;
	float fitness;
	Robot *robot = experiment->GetRobot();

	mMaxFitness = 0;
	mWorstFitness = 10000000.0f;
	mCurIndividual = 0;
	mInterrupt = 0;

	while(evoModel->ContinueRun()) {
		// Evaluate
		StatusLine(wxString::Format("Individual %d", mCurIndividual));
		if((ret = mIndividuals[mCurIndividual]->Evaluate(evoModel, experiment, this, NULL))) {
			if(ret > 0) {
				robot->Disconnect();
				return ret;
			}
		}

		switch(mInterrupt) {
			case 0: // not interrupted
				fitness = mIndividuals[mCurIndividual]->GetFitness();
				if(mMaxFitness < fitness)
					mMaxFitness = fitness;
				else if(mWorstFitness > fitness)
					mWorstFitness = fitness;
				break;
			case 1: // previous
				if(mCurIndividual > 0) mCurIndividual--;
				break;
			case 2: // next
				if(mCurIndividual < mSize-1) mCurIndividual++;
				break;
			default:
				break;
		}
		mInterrupt = 0;

		// Make sure that everyone is aware
		SetChanged();
		NotifyObservers();
		evoModel->NotifyView();		
		
		// Change batteries if necessary
		if(!robot->BatteriesOK())
			robot->ChangeBatteries();
	}

	return 0;
}

void Population::StartLog(Logger *logger) {
	LogLegenda(logger);
}

void Population::LogLegenda(Logger *logger) {
	wxString name=wxString::Format("numIndividual\tEpochCycles#%d\tEpochFitness#%d\tFitness\n",mNumEpochs,mNumEpochs);
	logger->WriteString(name);
}





void Population::Log(Logger *logger) {
	logger->WriteString(wxString::Format("%d ", mCurIndividual));
	mIndividuals[mCurIndividual]->Log(logger);
}

void Population::EndLog(Logger *logger) {
	
}

void Population::Paint(wxDC& dc, wxPoint& position, wxSize& size) {
	wxUint16 offset = 0;
	wxString string;

	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	dc.DrawText("Population", position.x + 10, position.y + 15);
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);
	
	offset = 45;
	dc.DrawText("Generation:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d", mCurGeneration), position.x + 100, position.y + offset);
	offset += 20;
	dc.DrawText("Individual:", position.x + 10, position.y + offset);
	dc.DrawText(wxString::Format("%d of %d", mCurIndividual + 1, mSize), position.x + 100, position.y + offset);
	offset += 20;
	if(mWorstFitness < 10000000.0f) {
		dc.DrawText("Worst fitness:", position.x + 10, position.y + offset);
		dc.DrawText(wxString::Format("%.03f", mWorstFitness), position.x + 100, position.y + offset);
		offset += 20;
	}
	if(mMaxFitness > 0.0f) {
		dc.DrawText("Best fitness:", position.x + 10, position.y + offset);
		dc.DrawText(wxString::Format("%.03f", mMaxFitness), position.x + 100, position.y + offset);
		offset += 20;
	}
	if(mCurIndividual > 0 && mAvgFitness > 0.0f) {
		dc.DrawText("Average:", position.x + 10, position.y + offset);
		dc.DrawText(wxString::Format("%.03f", mAvgFitness / mCurIndividual), position.x + 100, position.y + offset);
	}
}

wxInt8 Population::SaveBestIndividual(wxString filename) {
	return mBestIndividual->SaveToDisk(filename);
}

wxInt8 Population::SaveToDisk(wxString filename) {
	wxFile *file = new wxFile(filename.c_str(), wxFile::write);
	if(!file->IsOpened()) {
		StatusLine("Population - Couldn't open file \"" + filename + "\" for writing.");
		delete file;
		return 1;
	}
	file->Write(wxString::Format("%d %d\n", mSize, mCurGeneration));
	for(wxUint16 i=0; i<mSize; i++)
		mIndividuals[i]->WriteToFile(file);
	file->Close();
	delete file;
	return 0;
}

Population *Population::LoadFromDisk(wxString filename, wxUint16 curGenomeLength) {
	wxTextFile *file = new wxTextFile(filename);
	if(!file->Open()) {
		StatusLine("Population - Couldn't open file \"" + filename + "\" for reading.");
		delete file;
		return NULL;
	}
	wxString line = file->GetFirstLine();
	long unsigned temp;
	wxUint16 size;
	if(!line.Mid(0, line.Index(' ')).ToULong(&temp)) {
		StatusLine("Population - Couldn't load, no size available.");
		delete file;
		return NULL;
	}
	size = (wxUint16)temp;
	if(!line.Mid(line.Index(' ')).ToULong(&temp)) {
		StatusLine("Population - Couldn't load, no generation number available.");
		delete file;
		return NULL;
	}
	Population *newPop = new Population(size, (wxUint16)temp);
	Individual *ind;
	for(wxUint32 i=0; i<size; i++) {
		ind = Individual::ReadFromFile(file, curGenomeLength);
		if(!ind) {
			StatusLine("Population - Couldn't load population, are you sure this is a valid i population?");
			delete newPop;
			newPop = NULL;
			break;
		} else
			newPop->SetIndividual(i, ind);
	}
	file->Close();
	delete file;
	return newPop;
}
