/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _POPULATION_H
#define _POPULATION_H

#include "../utils/logger.h"
#include "../utils/observable.h"

class wxFile;

class EvoModel;
class Experiment;
class NeuralNetworkFactory;
class Individual;
class IGenOperator;

/**
 * Population of Individuals, used in EvoModel.
 * @see EvoModel
 * @see Individual
 * @see Experiment
 */
class Population : public Observable, public ILoggable
{
public:
	/** 
	 * Create a new population. 
	 * @pararm size Size of the new population.
	 * @param newIndividuals Create new (random) individuals if true.
	 */
	Population(wxUint16 size, wxUint16 curGeneration, bool newIndividuals = false, wxUint16 genomeLength = 0);

	/**
	 * Population copy constructor
	 * @param population is the population that is copied
	 */
	Population(Population* population);

	/** Population destructor */
	virtual ~Population();

	/* --------------------------- EVOLUTION ---------------------------- */

	/** Apply specified genetic operators to the population. */
	wxInt8 ApplyOperators(IGenOperator **genOperators, wxUint8 numGenOperators);
	
	/**
	 * Evaluate the population
	 * @param resume Resumes from the current individual if true, starts from first otherwise.
	 * @return a error code
	 */
	wxInt8 Evaluate(EvoModel *evoModel, Experiment *experiment, bool resume);

	/**
	 * Test the population
	 * @return a error code
	 */
	wxInt8 Test(EvoModel *evoModel, Experiment *experiment);

	/* -------------------------- ACCESSORS ------------------------- */

	wxInt8 GetInterrupted() { return mInterrupt; }
	void SetInterrupt(wxInt8 interrupt) { mInterrupt = interrupt; }

	/**
	 * Accessor method
	 * @return the size of the population
	 */
	wxUint16 GetSize() { return mSize; }

	/**
	 * Accessor method
	 * @param size is the size of the population
	 */
	void SetSize(wxUint16 size);

	wxUint16 GetCurGeneration() { return mCurGeneration; }
	void SetCurGeneration(wxUint16 curGeneration) { mCurGeneration = curGeneration; }

	/**
	 * Accessor method
	 * @return the array of individuals
	 */
	Individual** GetIndividuals() { return mIndividuals; }


	/**
	 * Accessor method
	 * @param num is the position of the  wanted individual
	 * @return the wanted individual
	 */
	Individual* GetIndividual(wxUint16 num) { return mIndividuals[num]; }

	/**
	 * Accessor method
	 * @param num is the position of the wanted individual.
	 */
	void SetIndividual(wxUint16 num, Individual *individual) { mIndividuals[num] = individual; }

	/**
	 * Accessor method
	 * @return the best individual of the population
	 */
	Individual* GetBestIndividual() { return mBestIndividual; }

	/**
	 * Accessor method
	 * @return the fitness of the best individual of the population
	 */
	float GetMaxFitness() { return mMaxFitness; }

	/**
	 * Accessor method
	 * @return the average fitness of the population
	 */
	float GetAvgFitness() { return mAvgFitness; }

	/**
	 * Accessor method
	 * @return the fitness of the worst individual of the population
	 */
	float GetWorstFitness() { return mWorstFitness; }

	void SetNumEpochs(wxUint16 numEpochs){mNumEpochs = numEpochs;}
	/* --------------------------- LOGGING ---------------------------- */

	/** Called when a Log is started. */
	void StartLog(Logger *logger);

	void LogLegenda(Logger *logger);

	/** Called by logger when new data should be logged. */
	void Log(Logger *logger);

	/** Called when a Log is ended. */
	void EndLog(Logger *logger);

	/* ------------------------- PAINTING ------------------------- */

	/** Paint the population. */
	void Paint(wxDC& dc, wxPoint& position, wxSize& size);

	/* --------------------------- I/O ---------------------------- */
	
	/** Saves the best individual to the specified file. */
	wxInt8 SaveBestIndividual(wxString filename);
	
	/** Save population to the specified file. */
	wxInt8 SaveToDisk(wxString filename);
	/** Load a population from the specified file. */
	static Population *LoadFromDisk(wxString filename, wxUint16 curGenomeLength);

protected:
	wxInt8 mInterrupt; /**< Makes it possible to interrupt individuals. */

	/** the number of individuals there are in the population */
	wxUint16 mSize;

	wxUint16 mCurGeneration; /**< Current generation. */
	wxUint16 mCurIndividual; /**< Number of the last individual evaluated. */

	wxUint16 mNumEpochs;

	/** Array of individuals  */
	Individual **mIndividuals;

	/** The best individual */
	Individual *mBestIndividual;

	/** The fitness of the best individual */
	float mMaxFitness;

	/** The average fitness of the population */
	float mAvgFitness;

	/** The fitness of the worst individual */
	float  mWorstFitness;
};

#endif /* _POPULATION_H */
