/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>
#define _USE_MATH_DEFINES 
#include "utils/js.h"
#ifdef WIN32
	#include <wx/msw/winundef.h>
#endif

#include "status.h"
#include "controller.h"
#include "modelsettings.h"
#include "robot/robot.h"
#include "utils/iutils.h"

#include "manualmodel.h"

ManualModel::ManualModel(ModelSettings *modelSettings) : wxThread(wxTHREAD_JOINABLE) {
	mModelSettings = modelSettings;
	mRobot = mModelSettings->GetRobot();
	for(int i=0; i<3; i++)
		mMotorSpeeds[i] = 0.0f;
	mLedState = false;
	mLoopDelay = 0;
}

ManualModel::~ManualModel() {

}

wxInt8 ManualModel::Start() {
	if(IsRunning()) return -1; // Already running...
	wxInt8 ret;

	mLoopDelay = wxGetNumberFromUser("Enter the loop delay (in ms):", "Loop delay", "Start manual control", 0, 0, 1000);

	ret = Create();
	if(ret != wxTHREAD_NO_ERROR)
		return ret;
	ret = Run();
	return ret;
}

wxInt8 ManualModel::Stop() {
	DeleteObservers();
	Delete();
	return 0;
}

/* Return type should be ExitCode, but this typedef seems to be undefined here. */
void *ManualModel::Entry() {
	wxInt8 ret;
	float temp;

	mJoystick = new jsJoystick(0);
	if(mJoystick->notWorking()) {
		StatusLine("Joystick doesn't work properly!");
//		delete mJoystick;
//		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, MANUAL_STOPPED);
//		wxPostEvent(mModelSettings->GetController(), event);
//		return NULL;
	}
	StatusLine("Joystick found!");

	float jsAxes[6];
	int jsButtons[6];

	mMotorSpeeds[0] = mMotorSpeeds[1] = mMotorSpeeds[2] = 0.0f;
	mRollAxis = mPitchAxis = mYawAxis = mThrottleAxis = 0.0f;

	if((ret = mRobot->AttemptConnect())) {
		StatusLineErr("Fatal error: couldn't connect to robot!", ret);
		delete mJoystick;
		wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, MANUAL_STOPPED);
		wxPostEvent(mModelSettings->GetController(), event);
		return NULL;
	}

	wxInt8 ledCounter = 0;
	wxUint32 loopCounter = 0;
	wxStopWatch stopwatch;

	while(!TestDestroy()) {
		if(ledCounter == 0) {
			mLedState = !mLedState;
			if((ret = mRobot->SetLED(0, mLedState)))
				StatusLineErr("Couldn't set LED state.", ret);
		}
		if(++ledCounter == 10) ledCounter = 0;

		if(mJoystick->notWorking()) {
			jsAxes[0]=0;
			jsAxes[1]=0;
			jsAxes[2]=0;
			jsAxes[3]=0;
		} else {
			mJoystick->read(jsButtons, jsAxes);
		}

		// Handling of platform dependencies. Linux is the reference.
		//
		// Roll:	jsAxes[0]
		// Pitch:	jsAxes[1]
		// Yaw:		jsAxes[2] under Linux ([3] under Windows)
		// Power:	jsAxes[3] under Linux ([2] under Windows)
#ifdef WIN32
		temp = jsAxes[2];
		jsAxes[2] = jsAxes[3];
		jsAxes[3] = - temp;
#else
		jsAxes[3] = - jsAxes[3];
#endif
		mRollAxis = jsAxes[0];
		mPitchAxis = jsAxes[1];
		mYawAxis = jsAxes[2];
		mThrottleAxis = jsAxes[3];

		if((ret = mRobot->ReadSensors()))
			StatusLineErr("Couldn't read sensor data!", ret);
		if((ret = mRobot->ManualControl(jsButtons, jsAxes)))
			StatusLineErr("Couldn't do control of robot actuators!", ret);

		SetChanged();
		NotifyObservers();
		mRobot->NotifyObservers();

		if(!mRobot->BatteriesOK())
			mRobot->ChangeBatteries();

		loopCounter++;

		if(mLoopDelay)
			iSleep(mLoopDelay);
		iYield();
	}

	StatusLine(wxString::Format("Average loop time: %d ms", (int) (stopwatch.Time() / loopCounter)));
	if((ret = mRobot->Disconnect()))
		StatusLineErr("Couldn't disconnect from robot.", ret);
	delete mJoystick;
	wxCommandEvent event(wxEVT_COMMAND_MENU_SELECTED, MANUAL_STOPPED);
	wxPostEvent(mModelSettings->GetController(), event);
	return NULL;
}

void ManualModel::Paint(wxDC& dc, wxPoint& position, wxSize& size) {
	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	wxUint16 height, width, x, y;

	/* LED */
	dc.SetPen(*wxBLACK_PEN);
	if(mLedState)
		dc.SetBrush(*wxRED_BRUSH);
	else
		dc.SetBrush(*wxWHITE_BRUSH);
	dc.DrawEllipse(position.x + 115, position.y + 15, 10, 10);

	/* Joystick */
	dc.DrawText("Joystick", position.x + 20, position.y + 10);
	
	// Throttle axis
	dc.SetBrush(*wxWHITE_BRUSH);
	dc.DrawRectangle(position.x + 25, position.y + 30, 5, 50);
	dc.SetBrush(*wxBLUE_BRUSH);
	height = (wxUint16)(fabs(mThrottleAxis) * 25);
	if(mThrottleAxis > 0.0f)
		dc.DrawRectangle(position.x + 25, position.y + 54 - height, 5, height);
	else
		dc.DrawRectangle(position.x + 25, position.y + 55, 5, height);

	// Pitch and roll axes
	dc.SetPen(*wxBLACK_PEN);
	dc.SetBrush(*wxWHITE_BRUSH);
	dc.DrawRectangle(position.x + 35, position.y + 30, 50, 50);
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.SetBrush(*wxBLUE_BRUSH);
	x = (wxUint16)((mRollAxis + 1.0f) * 25);
	y = (wxUint16)((mPitchAxis + 1.0f) * 25);
	dc.DrawEllipse(position.x + 32 + x, position.y + 27 + y, 6, 6);

	// Yaw axis
	dc.SetPen(*wxBLACK_PEN);
	dc.SetBrush(*wxWHITE_BRUSH);
	dc.DrawRectangle(position.x + 35, position.y + 85, 50, 5);
	dc.SetBrush(*wxBLUE_BRUSH);
	width = (wxUint16)(fabs(mYawAxis) * 25);
	if(mYawAxis < 0.0f)
		dc.DrawRectangle(position.x + 59 - width, position.y + 85, width, 5);
	else
		dc.DrawRectangle(position.x + 60, position.y + 85, width, 5);

	/* Motor speeds (should be done in robot implementation, if necessary) */
/*	dc.DrawText("Motor speeds", position.x + 20, position.y + 100);
	dc.SetPen(*wxBLACK_PEN);
	dc.SetBrush(*wxWHITE_BRUSH);
	dc.DrawRectangle(position.x + 25, position.y + 120, 5, 50);
	dc.DrawRectangle(position.x + 40, position.y + 120, 5, 50);
	dc.DrawRectangle(position.x + 55, position.y + 120, 5, 50);

	dc.SetBrush(*wxBLUE_BRUSH);
	height = (wxUint16)(fabs(mMotorSpeeds[0]) * 25);
	if(mMotorSpeeds[0] > 0.0f)
		dc.DrawRectangle(position.x + 25, position.y + 144 - height, 5, height);
	else
		dc.DrawRectangle(position.x + 25, position.y + 145, 5, height);
	height = (wxUint16)(fabs(mMotorSpeeds[1]) * 25);
	if(mMotorSpeeds[1] > 0.0f)
		dc.DrawRectangle(position.x + 40, position.y + 144 - height, 5, height);
	else
		dc.DrawRectangle(position.x + 40, position.y + 145, 5, height);
	height = (wxUint16)(fabs(mMotorSpeeds[2]) * 25);
	if(mMotorSpeeds[2] > 0.0f)
		dc.DrawRectangle(position.x + 55, position.y + 144 - height, 5, height);
	else
		dc.DrawRectangle(position.x + 55, position.y + 145, 5, height);
*/
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);
}
