% FUNCTION [act, spikes] = sim_srm_neuron(input, delay, tau_s, tau_m, thresh, windowLen, updates, cycleTime)
%
%	Simulate a SRM neuron.
%
%	INPUT is the input spikes, where row are presynaptic neuron and columns
%	are time step. DELAY is the delay of the synapse. TAU_S is the synapse
%	time constant. TAU_M is the membrane time constant. THRESH is the
%	neuron threshold and WINDOWLEN is the time window for the kernels.
%
%	DELAY and TAU_S can be a single number to be used for all synapses, or
%	a column vector containing differente value for each synapse.
%
%	UPDATES: #updates per cycle
%	CYCLETIME: cycle time in ms
%
%	ACT is the activation of the neuron for each timestep. SPIKES is set to
%	one for each spike emitted by the neuron. T is the corresponding time
%	value in ms.
%
%	ex:
%	[act, spikes, t] = sim_srm_neuron([1 1 1 0 0 1], 2, 10, 4, 0.2, 20, 40, 100); plot(t, act, t, spikes, 'x');

function [act, spikes, t] = sim_srm_neuron(input, delay, tau_s, tau_m, thresh, windowLen, updates, cycleTime)

% check size of delay
if length(delay) == 1
	delay = delay .* ones(size(input, 1), 1);
elseif length(delay) ~= size(input, 1)
	error('Wrong matrix size for delay');
end

% check size of tau_s
if length(tau_s) == 1
	tau_s = tau_s .* ones(size(input, 1), 1);
elseif length(tau_s) ~= size(input, 1)
	error('Wrong matrix size for tau_s');
end

% compute spike kernels
[kern, tk] = calc_spike_kernel(delay, tau_s, tau_m, windowLen, updates, cycleTime);

% compute refrac kernel
% should be in a custom script...
rkern(1:windowLen) = -exp(-tk / tau_m);
lastSpike = -1;

% for all times
for i=1:size(input, 2)+windowLen-1
	activity = 0;
	spikes(i) = 0;
	t(i) = (i-1) * cycleTime / updates;
	
	% for all times in the time window
	for j=max(1, i-windowLen+1):min(i, size(input,2))
		% for all neurons
		for n = 1:size(input,1)
			if input(n, j) == 1
				% we've spike
				activity = activity + kern(n, i-j+1);
			end
		end
	end
	
	if lastSpike ~= -1
		if lastSpike < i - windowLen
			lastSpike = -1;
		else
			activity = activity + rkern(i - lastSpike);
		end
	end
	
	if activity > thresh
		lastSpike = i;
		spikes(i) = 1;
	end
	
	act(i) = activity;
end