/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "status.h"
#include "comm/commlinkfactory.h"
#include "comm/serialport.h"
#include "robot/robotfactory.h"
#include "exp/experimentfactory.h"
#include "utils/logger.h"
#include "utils/xml/xml.h"

#include "modelsettings.h"

ModelSettings::ModelSettings(Controller *controller, wxString outputDirectory) {
	mController = controller;
	mOutputDirectory = outputDirectory;
	
	mCommLink = CommLinkFactory::CreateByIndex(0);
	mRobot = RobotFactory::CreateByIndex(0, mCommLink);
	mRobot->Initialize();
	mExperiment = ExperimentFactory::CreateByIndex(0, GetRobotID());
	
	mObserveRobot = false;
	mObserveExperiment = false;
}

ModelSettings::~ModelSettings() {
	delete mCommLink;
	delete mRobot;	
	delete mExperiment;
	if(mObserveRobot)
		delete mRobotLogger;
	if(mObserveExperiment)
		delete mExperimentLogger;
}

void ModelSettings::SetRobot(Robot *robot) {
	delete mRobot; 
	mRobot = robot;
	mRobot->SetCommLink(mCommLink); 
	if(mObserveRobot)
		mRobot->AddObserver(mRobotLogger);
}

void ModelSettings::SetRobotByID(int id) { 
	Robot *robot;
	robot = RobotFactory::Create(id, mCommLink);
	if(!robot) {
		StatusLine("ModelSettings - Invalid RobotType, RobotFactory returned null pointer!");
		return;
	}
	robot->Configure();
	robot->Initialize();
	SetRobot(robot);		
}

void ModelSettings::SetCommLinkByID(int id) { 
	CommLink *commLink;
	commLink = CommLinkFactory::Create(id); 
	if(!commLink) {
		StatusLine("ModelSettings - Invalid CommLink ID, CommLinkFactory returned null pointer!");
		return;
	}
	delete mCommLink; 
	mCommLink = commLink;
	mCommLink->Configure();
	mRobot->SetCommLink(mCommLink);
}

void ModelSettings::SetExperimentByIndex(wxUint32 subIndex) { 
	Experiment *experiment = ExperimentFactory::CreateByIndex(subIndex, GetRobotID());
	if(!experiment) {
		StatusLine("ModelSettings - Invalid ExperimentType, ExperimentFactory returned null pointer!");
		return;
	}
	delete mExperiment; 
	mExperiment = experiment;
	mExperiment->Configure();
}

void ModelSettings::SetObserveRobot(bool enable, const wxString& saveName) {
	if(!mObserveRobot && enable) {
		mRobotLogger = new Logger(saveName, mRobot);
		mRobot->AddObserver(mRobotLogger);
		mObserveRobot = true;
	} else if(mObserveRobot && !enable) {
		mRobot->DeleteObserver(mRobotLogger);
		delete mRobotLogger;
		mObserveRobot = false;
	}
}

void ModelSettings::SetObserveExperiment(bool enable, const wxString& saveName) {
	if(!mObserveExperiment && enable) {
		mExperimentLogger = new Logger(saveName, mExperiment);
		mExperiment->AddObserver(mExperimentLogger);
		mObserveExperiment = true;
	} else if(mObserveExperiment && !enable) {
		mExperiment->DeleteObserver(mExperimentLogger);		delete mExperimentLogger;
		mObserveExperiment = false;
	}
}



/* ----------------------------- XML --------------------------- */
void ModelSettings::Load (XMLElement *myself, XMLLoader *loader) {
	CommLink *commLink = CommLinkFactory::Create(loader->GetNextElementName());
	loader->Load(this, commLink);
	SetCommLink(commLink);

	Robot *robot = RobotFactory::Create(loader->GetNextElementName(), commLink);
	loader->Load(NULL,robot);
	robot->Initialize();
	SetRobot(robot);

	Experiment *experiment = ExperimentFactory::Create(loader->GetNextElementName(), GetRobotID());
	loader->Load(NULL,experiment);
	SetExperiment(experiment);
}

XMLElement* ModelSettings::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());		
	return result;
}

void ModelSettings::SaveChilds (XMLSaver *saver) {
	saver->SaveObject(GetCommLink());
	saver->SaveObject(GetRobot());
	saver->SaveObject(GetExperiment());
}

wxString ModelSettings::GetXMLName () {
	return GetShortXMLName();
}

void ModelSettings::SetParentName (const wxString &parentName) {

}

