/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _MODELSETTINGS_H
#define _MODELSETTINGS_H

#include "comm/commlink.h"
#include "robot/robot.h"
#include "exp/experiment.h"
#include "utils/xml/xml.h"

class Controller;
class Logger;

/** 
 * Maintains main settings used by most Models.
 * @see EvoModel
 * @see ManualModel
 */
class ModelSettings : public XMLSerializable
{
public:
	/** Construct a default ModelSettings. */
	ModelSettings(Controller *controller, wxString outputDirectory);

	/** Destroy this ManualModel. */
	virtual ~ModelSettings();
	
	/** Returns directory to be used for writing files to. */
	const wxString& GetOutputDirectory() { return mOutputDirectory; }
	void SetOutputDirectory(wxString outputDirectory) { mOutputDirectory=outputDirectory; }
	
	/** Returns the controller. */
	Controller *GetController() { return mController; }
	/** Returns the currently used Robot. */
	Robot *GetRobot() { return mRobot; }
	/** Returns the current robot's ID. */
	int GetRobotID() { return mRobot->GetID(); }
	/** Set the currently used Robot. */
	void SetRobot(Robot *robot);
	/** Set the current Robot by specifying its type. */
	void SetRobotByID(int id);
	/** Adds Robot observers if argument is true, remove otherwise. */
	void SetObserveRobot(bool enable, const wxString& saveName = wxString());
	/** Adds Experiment observers if argument is true, remove otherwise. */
	void SetObserveExperiment(bool enable, const wxString& saveName = wxString());	
	
	/** Returns the currently used CommLink. */
	CommLink *GetCommLink() { return mCommLink; }
	/** Set the CommLink to be used. */
	void SetCommLink(CommLink *commLink) { delete mCommLink; mCommLink = commLink; mRobot->SetCommLink(mCommLink); }
	/** Set the current CommLink by specifying its type. */
	void SetCommLinkByID(int id);

	/** Returns the currently used Experiment. */	
	Experiment *GetExperiment() { return mExperiment; }
	/** Set the Experiment to be used. */
	void SetExperiment(Experiment *experiment) { delete mExperiment; mExperiment = experiment; }
	/** Set the current Experiment by specifying its type. */
	void SetExperimentByIndex(wxUint32 subIndex);
	
	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual void			SaveChilds (XMLSaver *saver);
	virtual void			SetParentName (const wxString &parentName);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual const wxString	GetShortXMLName() { return "modelsettings"; };

protected:
	wxString mOutputDirectory; /**< Same as mOutputDirectory in Controller. */

	Controller *mController; /**< Pointer to Controller object. */
	Robot *mRobot; /**< Pointer to Robot object. */
	CommLink *mCommLink;  /**< Pointer to CommLink object. */
	Experiment *mExperiment; /**< Pointer to Experiment object. */
	
	bool mObserveRobot; /**< True if Robot currently observed. */
	Logger *mRobotLogger; /**< Pointer to logger assigned to robot. */
	
	bool mObserveExperiment; /**< True if Experiment currently observed. */
	Logger *mExperimentLogger; /**< Pointer to logger assigned to experiment. */
};

#endif
