/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _NEURALNETWORK_H
#define _NEURALNETWORK_H

class Neuron;

#include <wx/file.h>
#include "../utils/observable.h"
#include "../utils/iparameterizable.h"

/** 
 * The base artificial neural network class.
 * @see Synapse
 * @see Neuron
 */

class NeuralNetwork : public Observable, public IParameterizable
{
public:
	/**
	 * NeuralNetwork constructor
	 * @param size the number of neurons
	 * @param numOutput the number of output values
	 */
	NeuralNetwork(wxUint8 size, wxUint8 numOutput);

	/** Destructor. */
	virtual ~NeuralNetwork();

	/* ------------------------- USAGE ------------------------------ */

	/** 
	 * Allocate storage for network activity (required before using network!).
	 * @param numCycles Number of sensorimotor cycles to allocate storage for.
	 */
	virtual void AllocActivityStorage(wxUint16 numCycles) =0;
	
	/** Sets the current cycle, used to store neural activity. */
	void SetCurrentCycle(wxUint16 currentCycle) { mCurrentCycle = currentCycle; }

	/**  Returns the current cycle. */
	wxUint16 GetCurrentCycle() { return mCurrentCycle; }

	/**
	 * Abstract method to feed input into the neural network.
	 * @param input The array of input values.
	 */
	virtual void SetInput(float* input) =0;

	/** Abstract method to compute the output values of the neural network. */
	virtual void Compute() =0;

	/* ------------------------- ACCESSORS ---------------------------- */

	/**
	 * Returns the output value asked for.
	 * @param position is the position of the output value wanted
	 * @return output value (between 0.0 and 1.0)
	 */
	float GetOutputValue(wxUint8 num) { return mOutput[num]; }
	/** Returns an array (of size GetNumOutput()) containing all output values. */
	float * GetOutputValues() { return mOutput; }
	/** Returns the number of output values. */
	wxUint8 GetNumOutput() { return mNumOutput; }

	/** Returns the size (number of neurons) of the network. */
	wxUint8 GetSize() { return mSize; }
  
	/** Returns the neuron asked for. */
	Neuron *GetNeuron(wxUint8 num) { return mNeurons[num]; }
	
	/** Set a particular neuron. */
	void SetNeuron(wxUint8 num, Neuron *neuron) { mNeurons[num] = neuron; }

	/* ---------------- LOGGING -------------------- */

	/** Saves the stored neural activity to the specified file. */
	wxInt8 SaveActivity(wxString &filename);

protected:
	/** Does the actual saving, different for each subclass. */
	virtual void SaveActivity(wxFile *file) =0;
	
public:
	/* ---------------- PAINTING -------------------- */

	/** Return the size needed to paint the neuralNetwork stuffs */
	wxSize NeuralNetwork::GetPaintSize() ;

	/** Return the size needed to paint the neuralNetwork architecture */
	virtual wxSize GetPaintArchitectureSize() ;

	/** Return the size needed to paint the neuralNetwork activity */
	virtual wxSize GetPaintActivitySize() ;
	
	/** Paint the neural network. */
	void Paint(wxDC& dc, wxPoint& position, wxSize& size);

	/** 
	 * Paint the neural network architecture, which is neural network dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintArchitecture(wxDC& dc, wxPoint& position, wxSize& size) =0;

	/** 
	 * Paint the neural network activity, which is neural network dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintActivity(wxDC& dc, wxPoint& position, wxSize& size) = 0;

protected:
	wxUint8 mSize;					/**< Size of the network. i.e. number of neurons (internal + output + input + bias) */
	Neuron **mNeurons;			/**< The neurons of this network. */
	wxUint8 mNumOutput;			/**< Number of output values. */
	float *mOutput;					/** Output values. */
	wxUint16 mCurrentCycle;	/**< Current sensorimotor cycle. */	
};

#endif /* _NEURALNETWORK_H */
