/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _NEURALNETWORKFACTORY_H
#define _NEURALNETWORKFACTORY_H

class Genome;
class NeuralNetwork;
class ModelSettings;

#include "../utils/xml/xml.h"

/**
 * Enum of the different types of available neural networks.
 * Please note: GetNumTypes() below should also be changed for both public and private versions!
 */
enum NeuralNetworkType {
	NN_PERCEPTRON=1,	/**< Perceptron (2-layered, feed forward) */
	NN_SIGMOID,				/**< Sigmoid NN. */
	NN_SPIKING,				/**< Spiking NN. */
	
};

/**
 * The interface to be implemented by concrete factories, able to create concrete NeuralNetworks.
 * @see Genome
 * @see Experiment
 * @see NeuralNetwork
 */
class NeuralNetworkFactory : public XMLSerializable
{
public:
	NeuralNetworkFactory(ModelSettings *modelSettings) { mModelSettings = modelSettings; }
	virtual ~NeuralNetworkFactory() {}

	/**
	 * Method to create a neural network from a genome and an experiment
	 * @param genome genome encoding parameters evolved
	 * @return a neural network of the wanted type
	 */
	virtual NeuralNetwork* Create(Genome* genome) =0;

	/** Returns the genome length required with current genome encoding. */
	virtual wxUint16 GetGenomeLength() =0;
	
	/** 
	 * Configure the genome encoding settings. 
	 * @return The new required genome length (zero on error).
	 */
	virtual wxUint16 ConfigGenomeEncoding() =0;

	/** Configure neural network specific settings. */
	virtual void ConfigNeuralNetwork() =0;
	
	/** Returns the type of neural network the factory creates. */
	virtual NeuralNetworkType GetType() =0;

	/** Returns the number of neural network factories that are available. */
	static wxUint8 GetNumTypes() { return 3; };

	/* XML stuff */
	virtual void			SaveChilds (XMLSaver *saver) {};
	virtual void			SetParentName (const wxString &parentName) {};	


protected:
	ModelSettings *mModelSettings;
};

#endif /* _NEURALNETWORKFACTORY_H */
