/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <map>

#include "neuralnetworktopology.h"


NeuralNetworkTopology::NeuralNetworkTopology(wxUint32 numNeuron, wxUint32 numSensor)
	: mNumNeuron(numNeuron), mNumSensor(numSensor), mTotNeurons(numNeuron + numSensor), mTotSynapses(numNeuron * mTotNeurons)
{
	mSynapseDescs = new ItemDesc[mTotSynapses];
	mNeuronDescs = new ItemDesc[mTotNeurons];
	
	// init to default value
	for(wxUint32 i = 0; i < mTotSynapses; i++)
		((long*) mSynapseDescs)[i] = 0;
	for(wxUint32 i = 0; i < mTotNeurons; i++)
		((long*) mNeuronDescs)[i] = 0;
	
	mNumClasslessNeuron = mTotNeurons;
	mNumClasslessSynapse = mTotSynapses;
}


NeuralNetworkTopology::~NeuralNetworkTopology()
{
	delete[] mSynapseDescs;
	delete[] mNeuronDescs;
}


wxUint16 NeuralNetworkTopology::GetNthNeuronClass(wxUint16 classIndex)
{
	std::set<wxUint16>::iterator iter = mNeuronClasses.begin();
	while(classIndex--) ++iter;
	return (*iter);
}


bool NeuralNetworkTopology::GetNextEvolvableNeuronOfClass(wxInt32 &neuron, wxUint16 neuronClass)
{
	while((wxUint32)++neuron < mTotNeurons)
	{
		if(mNeuronDescs[neuron].itemClass == neuronClass && mNeuronDescs[neuron].evolv == EVOLVABILITY_EVOLVABLE)
			return true;
	}
	
	return false;
}

	
wxUint16 NeuralNetworkTopology::GetNthSynapseClass(wxUint16 classIndex)
{
	std::set<wxUint16>::iterator iter = mSynapseClasses.begin();
	while(classIndex--) ++iter;
	return (*iter);
}


bool NeuralNetworkTopology::GetNextEvolvableSynapseOfClass(wxInt32 &synapse, wxUint16 synapseClass)
{
	while((wxUint32) ++synapse < mTotSynapses)
	{
		if(mSynapseDescs[synapse].itemClass == synapseClass && mSynapseDescs[synapse].evolv == EVOLVABILITY_EVOLVABLE)
			return true;
	}
	
	return false;
}


wxUint32 NeuralNetworkTopology::GetNumberOfEvolvableItems(ItemDesc* descs, wxUint32 num)
{
		//###### map useless... could be implemented with a set
	std::map<wxUint16, bool> classes; // set to true if the coresponding class has been accounted for
	wxUint32 evolvable = 0;
	
	for(wxUint32 i = 0; i < num; i++)
	{
		if(descs[i].evolv == EVOLVABILITY_EVOLVABLE)
		{
			wxUint16 nClass = descs[i].itemClass;
			if(nClass == 0)
				++evolvable;
			else
			{
				if(!classes[nClass])
				{
					++evolvable;
					classes[nClass] = true;
				}
			}
		}
	}
	
	return evolvable;
}

void NeuralNetworkTopology::SetSynapsesInfo(wxUint32 neuronFromLow, wxUint32 numNeuronFrom, wxUint32 neuronToLow, wxUint32 numNeuronTo,
					SynapseType type, EvolvabilityType prop, wxUint16 sClass)
{
	for(wxUint32 from = neuronFromLow; from < neuronFromLow+numNeuronFrom; from++)
	{
		for(wxUint32 to = neuronToLow; to < neuronToLow+numNeuronTo; to++)
		{
			SetSynapseInfo(from, to, type, prop, sClass);
		}
	}
}

void NeuralNetworkTopology::SetNeuronDesc(wxUint32 neuron, ItemDesc desc)
{
	if(neuron < mTotNeurons)
	{
		/*if(desc.itemClass != 0)
		{
			wxUint16 i = desc.itemClass;
			mNeuronClasses[i]--;
		}
		else
			mNumClasslessNeuron--;*/
			
		mNeuronDescs[neuron] = desc;
		
		if(desc.itemClass != 0)
		{
			wxUint16 i = desc.itemClass;
			mNeuronClasses.insert(i);
		}
		/*else
			mNumClasslessNeuron++;*/
	}
}

NeuralNetworkTopology::ItemDesc NeuralNetworkTopology::GetNeuronDesc(wxUint32 neuron)
{
	if(neuron < mTotNeurons)
		return mNeuronDescs[neuron];
	else
		return ItemDesc();
}

void NeuralNetworkTopology::SetSynapseDesc(wxUint32 neuronFrom, wxUint32 neuronTo, ItemDesc desc)
{
	if(neuronFrom < mTotNeurons && neuronTo < mNumNeuron)
		SetSynapseDesc(NeuronIndiceToSynapseIndex(neuronFrom, neuronTo), desc);
}

void NeuralNetworkTopology::SetSynapseDesc(wxUint32 synapse, ItemDesc desc)
{
	if(synapse < mTotSynapses)
	{
		/*if(desc.itemClass != 0)
		{
			wxUint16 i = desc.itemClass;
			mSynapseClasses[i]--;
		}
		else
			mNumClasslessSynapse--;*/
	
		mSynapseDescs[synapse] = desc;
		
		if(desc.itemClass != 0)
		{
			wxUint16 i = desc.itemClass;
			mSynapseClasses.insert(i);
		}
		/*else
			mNumClasslessSynapse++;*/
	}
}

NeuralNetworkTopology::ItemDesc NeuralNetworkTopology::GetSynapseDesc(wxUint32 neuronFrom, wxUint32 neuronTo)
{
	if(neuronFrom < mTotNeurons && neuronTo < mNumNeuron)
		return GetSynapseDesc(NeuronIndiceToSynapseIndex(neuronFrom, neuronTo));
	else
		return ItemDesc();
}

NeuralNetworkTopology::ItemDesc NeuralNetworkTopology::GetSynapseDesc(wxUint32 synapse)
{
	if(synapse < mTotSynapses)
		return mSynapseDescs[synapse];
	else
		return ItemDesc();
}

