/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _NEURALNETWORKTOPOLOGY_H
#define _NEURALNETWORKTOPOLOGY_H

#include <set>
#ifdef WIN32
	#include <wx/msw/winundef.h>
#endif

/**
 * Type of neurons.
 */
enum NeuronType
{
	// Spiking neurons.
	NTYPE_SPIKING = 0,
	NTYPE_SPIKING_RATE_RECEPTOR = 10,
	NTYPE_SPIKING_RATE_RECEPTOR_NOREFRAC,
	NTYPE_SPIKING_RATE_AC = 30,
	
	// Sigmoidal neurons.
	NTYPE_SIGMOID = 100
};


/**
 * Type of synapses.
 */
enum SynapseType
{
	// Spiking synapses.
	STYPE_SPIKING,
	STYPE_SPIKING_OUTPUT,
	
	// Sigmoidal synapses.
	STYPE_SIGMOID
};


/**
 * Describe the evolvability of an entity (neuron or synapse).
 */
enum EvolvabilityType
{
	EVOLVABILITY_UNCONNECTED = 0x00,	///< (For synapse only) The synapse will not be connected.
	EVOLVABILITY_DEFAULT = 0x01,		///< The item will take default values (synapse are connected).
	EVOLVABILITY_EVOLVABLE = 0x02		///< The parameter for this item will be evolved.
};

/**
 * This class describe the topology of a network, and in particular which neurons and synapse will
 * be evolved. Evolved neurons and synapse can be assigned a class. Neurons of the same class will
 * share their parameters, and thus those are represented only once in the genome. The same holds
 * true for synapses. This general class handle plain neurons and sensors. For handling of internal
 * neurons, output neurons and AC neurons, see the specific subclasses.
 *
 * NOTE: a neuron layout is implicitely defined by this class. Standard neurons have indice ranging
 * from 0 to numNeuron-1 (see constructor) while sensor neurons have indice ranging from numNeuron
 * to numNeuron + numSensor - 1. The layout within non-sensor neurons (internal, output, etc) is
 * defined by the subclasses.
 *
 * @see SpikingNNTopology
 * @see SigmoidNNTopology
 */
class NeuralNetworkTopology
{
public:

	/* ------------------------- CONSTRUCTION ------------------------------ */

	/**
	 * Constructor.
	 * @param numNeuron Number of neurons (not sensors).
	 * @param numSensor Number of sensors.
	 */
	NeuralNetworkTopology(wxUint32 numNeuron, wxUint32 numSensor);
	
	/**
	 * Destructor.
	 */
	virtual ~NeuralNetworkTopology();
	
	
	/* ------------------------- NEURON INFO ACCESSORS ------------------------------ */
	
	/**
	 * Set neuron information.
	 * @param neuron Neuron index.
	 * @param type Neuron type.
	 * @param prop Neuron evolvability type.
	 * @param neuronClass Neuron class (0 for no class).
	 */
	virtual void SetNeuronInfo(wxUint32 neuron, NeuronType type, EvolvabilityType prop, wxUint16 neuronClass)
	{
		ItemDesc desc = {type, prop, neuronClass & 0x1FFF};
		SetNeuronDesc(neuron, desc);
	}
	
	/**
	 * Get the type of a neuron.
	 * @param neuron Neuron index.
	 * @return Neuron type.
	 */
	virtual NeuronType GetNeuronType(wxUint32 neuron) {return (NeuronType) GetNeuronDesc(neuron).type;}
	
	/**
	 * Get the evolvability of a neuron.
	 * @param neuron Neuron index.
	 * @return Neuron evolvability.
	 */
	virtual EvolvabilityType GetNeuronEvolvability(wxUint32 neuron) {return (EvolvabilityType) GetNeuronDesc(neuron).evolv;}
	
	/**
	 * Get the class of a neuron.
	 * @param neuron Neuron index.
	 * @return Neuron class.
	 */
	virtual wxUint16 GetNeuronClass(wxUint32 neuron) {return GetNeuronDesc(neuron).itemClass;}
	
	
	
	/* ------------------------- SYNAPSE INFO ACCESSORS ------------------------------ */
	
	/**
	 * Set synapse information.
	 * @param neuronFrom Pre-synaptic neuron.
	 * @param neuronTo Post-synaptic neuron (cannot be a sensor neuron).
	 * @param type Synapse type.
	 * @param prop Synapse evolvability type.
	 * @param neuronClass Synapse class (0 for no class).
	 */
	virtual	void SetSynapseInfo(wxUint32 neuronFrom, wxUint32 neuronTo, SynapseType type, EvolvabilityType prop, wxUint16 sClass)
	{
		ItemDesc desc = {type, prop, sClass & 0x1FFF};
		SetSynapseDesc(neuronFrom, neuronTo, desc);
	}
	
	/**
	 * Set synapse information.
	 * @param synapse Synapse index.
	 * @param type Synapse type.
	 * @param prop Synapse evolvability type.
	 * @param neuronClass Synapse class (0 for no class).
	 */
	virtual void SetSynapseInfo(wxUint32 synapse, SynapseType type, EvolvabilityType prop, wxUint16 sClass)
	{
		ItemDesc desc = {type, prop, sClass & 0x1FFF};
		SetSynapseDesc(synapse, desc);
	}
	
	/** Set synapse information for a range of synapses. */
	virtual void SetSynapsesInfo(wxUint32 neuronFromLow, wxUint32 numNeuronFrom, wxUint32 neuronToLow, wxUint32 numNeuronTo,
					SynapseType type, EvolvabilityType prop, wxUint16 sClass);
	
	/**
	 * Get the type of a synapse.
	 * @param neuronFrom Pre-synaptic neuron.
	 * @param neuronTo Post-synaptic neuron (cannot be a sensor neuron).
	 * @return Synapse type.
	 */
	virtual SynapseType GetSynapseType(wxUint32 neuronFrom, wxUint32 neuronTo) {return (SynapseType) GetSynapseDesc(neuronFrom, neuronTo).type;}
	
	/**
	 * Get the type of a synapse.
	 * @param synapse Synapse index.
	 * @return Synapse type.
	 */
	virtual SynapseType GetSynapseType(wxUint32 synapse) {return (SynapseType) GetSynapseDesc(synapse).type;}
	
	/**
	 * Get the evolvability of a synapse.
	 * @param neuronFrom Pre-synaptic neuron.
	 * @param neuronTo Post-synaptic neuron (cannot be a sensor neuron).
	 * @return Synapse evolvability.
	 */
	virtual EvolvabilityType GetSynapseEvolvability(wxUint32 neuronFrom, wxUint32 neuronTo) {return (EvolvabilityType) GetSynapseDesc(neuronFrom, neuronTo).evolv;}
	
	/**
	 * Get the evolvability of a synapse.
	 * @param synapse Synapse index.
	 * @return Synapse evolvability.
	 */
	virtual EvolvabilityType GetSynapseEvolvability(wxUint32 synapse) {return (EvolvabilityType) GetSynapseDesc(synapse).evolv;}
	
	/**
	 * Get the class of a synapse.
	 * @param neuronFrom Pre-synaptic neuron.
	 * @param neuronTo Post-synaptic neuron (cannot be a sensor neuron).
	 * @return Synapse class.
	 */
	virtual wxUint16 GetSynapseClass(wxUint32 neuronFrom, wxUint32 neuronTo) {return GetSynapseDesc(neuronFrom, neuronTo).itemClass;}
	
	/**
	 * Get the class of a synapse.
	 * @param synapse Synapse index.
	 * @return Synapse class.
	 */
	virtual wxUint16 GetSynapseClass(wxUint32 synapse) {return GetSynapseDesc(synapse).itemClass;}



	/* ------------------------- ITERATIVE ACCESS ------------------------------ */


	/**
	 * Returns the number of neuron in the network.
	 */
	virtual wxUint32 GetNeuronNumber() {return mTotNeurons;}

	
	/**
	 * Returns the number of (potential) synapse in the network. (Synapse that are UNCONNECTED
	 * are also counted here.)
	 */
	virtual wxUint32 GetSynapseNumber() {return mTotSynapses;}
	
	/**
	 * Get the number of neurons to be evolved, taking into account neuron classes. (Neurons
	 * of the same class are counted only once).
	 */
	virtual wxUint32 GetEvolvableNeuronNumber() {return GetNumberOfEvolvableItems(mNeuronDescs, mTotNeurons);}
	
	/**
	 * Get the number of synapses to be evolved, taking into account synapse classes. (Synapses
	 * of the same class are counted only once).
	 */
	virtual wxUint32 GetEvolvableSynapseNumber() {return GetNumberOfEvolvableItems(mSynapseDescs, mTotSynapses);}
	
	
	virtual wxUint32 GetOutputNumber() = 0;
	virtual wxUint32 GetOutputNeuronIndex(wxUint32 outputNum) = 0;
	virtual wxUint32 GetInputNumber() {return mNumSensor;}
	virtual wxUint32 GetInputNeuronIndex(wxUint32 inputNum) {return mNumNeuron + inputNum;}
	virtual wxUint32 GetNonInputNumber() {return mNumNeuron;}
	virtual wxUint32 GetNonInputNeuronIndex(wxUint32 nonInputNum) {return nonInputNum;}

	
	
	
	/* ------------------------- CLASS FUNCTION ------------------------------ */
	
	/** Get number of neuron classes. */
	virtual wxUint16 GetNeuronClassNumber() {return mNeuronClasses.size();}
	
	/** Get an indexed neuron class. */
	virtual wxUint16 GetNthNeuronClass(wxUint16 classIndex);
	
	/**
	 * Iterate on neurons of a particular class.
	 * @param neuron The index of the neuron (initialize to -1).
	 * @param neuronClass Class to consider (pass 0 to iterate on class-less neurons).
	 * @return False if no more neuron of this class.
	 */
	virtual bool GetNextEvolvableNeuronOfClass(wxInt32 &neuron, wxUint16 neuronClass);
	
	/** Get number of synapse classes. */
	virtual wxUint16 GetSynapseClassNumber() {return mSynapseClasses.size();}
	
	/** Get an indexed synapse class. */
	virtual wxUint16 GetNthSynapseClass(wxUint16 classIndex);
	
	/**
	 * Iterate on synapses of a particular class.
	 * @param synapse The index of the synapse (initialize to -1).
	 * @param synapseClass Class to consider (pass 0 to iterate on class-less synapses).
	 * @return False if no more synapse of this class.
	 */
	virtual bool GetNextEvolvableSynapseOfClass(wxInt32 &synapse, wxUint16 synapseClass);
	
	
	
	/* ------------------------- UTILS ------------------------------ */
	
	/** Convert from synapse index to pre- and post-synaptic neuron indice. */
	virtual	void SynapseIndexToNeuronIndice(wxUint32 sIndex, wxUint32 &neuronFrom, wxUint32 &neuronTo) {neuronFrom = sIndex % mTotNeurons; neuronTo = sIndex / mTotNeurons;}

	/** Convert from pre- and post-synaptic neuron indice to synapse index. */
	virtual wxUint32 NeuronIndiceToSynapseIndex(wxUint32 neuronFrom, wxUint32 neuronTo) {return neuronTo * mTotNeurons + neuronFrom;}

protected:

	/**
	 * Description of an item (neuron / synapse).
	 */
	struct ItemDesc
	{
		unsigned type: 16;			///< Item type.
		unsigned evolv: 3;			///< Item evolvability.
		unsigned itemClass: 13;		///< Item class.
	};
	
	/* ------------------------- BASIC ACCESSORS ------------------------------ */
	
	/**  Set neuron description. */
	virtual void SetNeuronDesc(wxUint32 neuron, ItemDesc desc);
	
	/** Get neuron description. */
	virtual ItemDesc GetNeuronDesc(wxUint32 neuron);
	
	/** Set synapse description. */
	virtual void SetSynapseDesc(wxUint32 neuronFrom, wxUint32 neuronTo, ItemDesc desc);
	
	/** Set synapse description. */
	virtual void SetSynapseDesc(wxUint32 synapse, ItemDesc desc);
	
	/** Get synapse description. */
	virtual ItemDesc GetSynapseDesc(wxUint32 neuronFrom, wxUint32 neuronTo);
	
	/** Get synapse description. */
	virtual ItemDesc GetSynapseDesc(wxUint32 synapse);
	
	
	/* ------------------------- DATA MEMBERS ------------------------------ */
	
	const wxUint32 mNumNeuron;		///< Number of (non-sensor) neurons.
	const wxUint32 mNumSensor;		///< Number of sensor neurons.
	const wxUint32 mTotNeurons;		///< Total number of neurons.
	const wxUint32 mTotSynapses;	///< Total number of synapses.
	ItemDesc* mNeuronDescs;			///< Neuron descriptors.
	ItemDesc* mSynapseDescs;		///< Synapse descriptors.
	std::set<wxUint16> mNeuronClasses;	///< Set of neuron classes.
	std::set<wxUint16> mSynapseClasses;	///< Set of synapse classes.
	wxUint32 mNumClasslessNeuron;
	wxUint32 mNumClasslessSynapse;

private:

	/** Helper function to count the number of evolvable items. */
	wxUint32 GetNumberOfEvolvableItems(ItemDesc* descs, wxUint32 num);
	
};


#endif

