/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _NEURON_H
#define _NEURON_H

class Synapse;

#include <wx/dynarray.h>
#include "../utils/iparameterizable.h"

WX_DEFINE_ARRAY(Synapse *, wxArraySynapse);

/** 
 * Base Neuron class.
 * @see Synapse
 * @see NeuralNetwork
 */
class Neuron : public IParameterizable
{
public:
	/** Neuron constructor. */
	Neuron();

	/** Neuron destructor. */ 
	virtual ~Neuron();
	

	/* ----------------------- BUILDING ---------------------------- */
	
	/** 
	 * Adds an incoming synapse to this neuron.
	 * @param syn A synapse which arrives to this neuron.
	 */
	void AddIncoming(Synapse *synapse) { mIncoming->Add(synapse); }

	/** 
	 * Adds an outgoing synapse to this neuron.
	 * @param syn A synapse departs from this neuron.
	 */
	void AddOutgoing(Synapse *synapse) { mOutgoing->Add(synapse); }

	/* -------------------------- ACCESSORS ----------------------- */

	/** 
	 * Accessor method.
	 * @return Membrane potential.
	 */
	virtual float GetMembranePotential() { return mMembranePotential; }
  
	/** 
	 * Accessor method.
	 * @param  membranePotential Membrane potential.
	 */
	virtual void SetMembranePotential(float membranePotential) { mMembranePotential = membranePotential; }
  
	/** 
	 * Accessor method
	 * @return the array of synapses which arrive to this neuron.
	 */
	wxArraySynapse *GetIncoming() { return mIncoming; }

	/** 
	 * Accessor method
	 * @param num
	 * @return the synapse at this position 
	 */
	Synapse *GetIncoming(size_t num) { return mIncoming->Item(num); }

	wxArraySynapse *GetOutgoing() { return mOutgoing; }
	Synapse *GetOutgoing(size_t num) { return mOutgoing->Item(num); }


protected:

	float mMembranePotential; /**< The membrane potential of this neuron. */
	wxArraySynapse *mIncoming; /**< List of incoming synapses for this neuron. */
	wxArraySynapse *mOutgoing; /**< List of outgoing synapses for this neuron. */

};

#endif /* _NEURON_H */
