/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../../status.h"
#include "../../modelsettings.h"
#include "../../ga/genome.h"
#include "../../exp/experiment.h"
#include "../../utils/xml/xml.h"
#include "perceptron.h"

#include "perceptronfactory.h"

PerceptronFactory::PerceptronFactory(ModelSettings *modelSettings) : NeuralNetworkFactory(modelSettings) {
	mWeightNumBits = 6;
}

PerceptronFactory::~PerceptronFactory() {

}

NeuralNetwork *PerceptronFactory::Create(Genome *genome) {
	Experiment *experiment = mModelSettings->GetExperiment();
	wxUint8 input = experiment->GetNumSensorNeurons();
	wxUint8 output = experiment->GetNumOutputNeurons();
	Perceptron *neuralNetwork = new Perceptron(input, output);

	wxUint32 decodeValue;
	double maxValue = pow(2.0, mWeightNumBits) - 1.0;
	wxUint16 numWeights = (input + 1) * output;
	float *weights = new float[numWeights];

	for(wxUint16 i=0; i<numWeights; i++) {
		if(genome->DecodeInt(decodeValue, i * mWeightNumBits, mWeightNumBits))
			return NULL;
		weights[i] = 5.0f * ((decodeValue / maxValue) - 0.5);
	}
	neuralNetwork->SetWeights(weights);
	
	return neuralNetwork;
}

wxUint16 PerceptronFactory::GetGenomeLength() {
	Experiment *experiment = mModelSettings->GetExperiment();
	wxUint8 input = experiment->GetNumSensorNeurons();
	wxUint8 output = experiment->GetNumOutputNeurons();
	return (input + 1) * output * mWeightNumBits; // add 1 for bias
}

wxUint16 PerceptronFactory::ConfigGenomeEncoding() {
	long result = wxGetNumberFromUser("Enter the number of bits to be used for the encoding of one weight. (An integer between 1 and 16.)", "# of bits / weight:", "Configure genome encoding Perceptron", mWeightNumBits, 1, 16);
	if(result > 0)
		mWeightNumBits = (wxUint8)result;
	return GetGenomeLength();
}

void PerceptronFactory::ConfigNeuralNetwork() {

}

/* --- XML --- */
void PerceptronFactory::Load (XMLElement *myself, XMLLoader *loader) {
	mWeightNumBits = myself->GetAttribute("weight_num_bits")->GetByteValue();
}

XMLElement* PerceptronFactory::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());	
	result->AddAttribute(new XMLAttribute("weight_num_bits",(wxInt8)mWeightNumBits));
	return result;
}

wxString PerceptronFactory::GetXMLName () { return GetShortXMLName(); }
wxString PerceptronFactory::GetShortXMLName() { return PerceptronFactory::GetStaticXMLType(); }
