/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../../status.h"
#include "sigmoidneuron.h"
#include "../synapse.h"

const wxUint8 SigmoidNeuron::MembranePotentialDrawHeight = 10; // the height of the potential drawings


SigmoidNeuron::SigmoidNeuron() {
	mAlpha = -1.0f;
	mType = InputNeuron;
	mActivity = NULL;
	mInhibitory = false;
}

SigmoidNeuron::~SigmoidNeuron() {
	if(mActivity) delete [] mActivity;
}

void SigmoidNeuron::AllocActivityStorage(wxUint16 numCycles) {
	mMaxNumCycles = numCycles;
	mActivity = new float[numCycles]; 
}

void SigmoidNeuron::ComputeMembranePotential(wxUint16 currentCycle) {
	size_t i, count = mIncoming->GetCount();
	Synapse *synapse;
	Neuron *preNeuron;
	SigmoidNeuronType type;
	float tmp = 0.0f;
	
	for(i=0; i<count; i++) {
		synapse = mIncoming->Item(i);
		preNeuron = synapse->GetPreNeuron();

/* 
		Here, for the non-output neurons, the problem is: do we want to make sure 
	 	to take the value of the incoming neurons one step before just the membrane potential
		which may have been updated in between, by the process itself ?? 
		For the recurrent connections it will always be the old value (cannot be updated
		before itself => no problem) but for the lateral connections, it depends on which 
		is updated first (processed from right to left). 
		Solution: using GetActivity(currentCycle -1) instead of GetMembranePotential(), and 
		this is exactly what we do here.
*/

		type = ((SigmoidNeuron *)preNeuron)->GetType();
		if(type == InternalNeuron || type == OutputNeuron) {
			if(currentCycle != 0) {
				if(((SigmoidNeuron *)preNeuron)->GetInhibitory())
					tmp += synapse->GetStrength() * ((SigmoidNeuron *)preNeuron)->GetActivity(currentCycle - 1);
				else
					tmp -= synapse->GetStrength() * ((SigmoidNeuron *)preNeuron)->GetActivity(currentCycle - 1);
			}
		} else // input neurons
			tmp += synapse->GetStrength() * preNeuron->GetMembranePotential();

	}
	
	if (mType != InputNeuron && mType != BiasNeuron) 
		mMembranePotential = 1 / (1 + exp(mAlpha * tmp));	// sigmoidal filter (not tanh)

	mActivity[currentCycle] = mMembranePotential;	// also done for Bias and Input neurons
}

void SigmoidNeuron::PaintActivity(wxDC& dc, wxPoint& position, wxUint16 cyclesVisible, wxUint16 currentCycle) {
	wxUint16 x, start = 0;
	if(currentCycle > cyclesVisible)
		start = currentCycle - cyclesVisible;

	x = position.x + 20;
	for(wxUint16 cycle = start; cycle<currentCycle; cycle++)
		dc.DrawLine(x, (int)(position.y + 18 - mActivity[cycle] * MembranePotentialDrawHeight), x++, position.y + 18);

	position.y += MembranePotentialDrawHeight + 5;
}

void SigmoidNeuron::SaveActivity(wxFile *file, wxUint16 numCycles) {
	for(wxUint16 i=0; i<numCycles; i++)
		file->Write(wxString::Format("%.2f ", mActivity[i]));
	file->Write("\n"); 
}
