/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _SIGMOIDNEURON_H
#define _SIGMOIDNEURON_H

#include <wx/file.h>

#include "../neuron.h"

enum SigmoidNeuronType {
	BiasNeuron,
	InputNeuron,
	InternalNeuron,
	OutputNeuron
};

/** 
 * The class to construct an artificial sigmoid neuron, inherited of Neuron
 * @see Neuron
 */
class SigmoidNeuron : public Neuron
{
public:
	/** SigmoidNeuron constructor */
	SigmoidNeuron();

	/** SigmoidNeuron destructor */ 
	virtual ~SigmoidNeuron();

	/* ------------------------ USAGE ----------------------------- */

	/** 
	 * Allocates storage for the activity of a certain number of network cycles. 
	 * @param numCycles Number of cycles to reserve space for.
	 */
	void AllocActivityStorage(wxUint16 numCycles);
	
	/** Method that computes the membrane potential. */
	void ComputeMembranePotential(wxUint16 currentCycle);

	/* ------------------------- ACCESSORS --------------------------- */
	
	/** 
	 * Accessor method
	 * @return alpha value
	 */
	float GetAlpha() { return mAlpha; }
	/** 
	 * Accessor method
	 * @param alpha New alpha value.
	 */
	void SetAlpha(float alpha) { mAlpha = alpha; }
	
	/** 
	 * Accessor method
	 * @return true if the neuron is inhibitory
	 */
	bool GetInhibitory() { return mInhibitory; }
	/** 
	 * Accessor method
	 * define the neuron as inhibitory
	 */
	void SetInhibitory() { mInhibitory = true; }

	/** Returns activity for all cycles. */
	float *GetActivity() { return mActivity; }
	/** Returns activity for a given cycle. */
	float GetActivity(wxUint16 cycle) { return mActivity[cycle]; }

	/** Set neuron type. */
	void SetType(SigmoidNeuronType type) { mType = type; }
	/** Returns neuron type. */
	SigmoidNeuronType GetType() { return mType; }	

	/* ----------------------- PAINTING -------------------------------- */

	/** 
	 * Paint the activity of this neuron.
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 */
	void PaintActivity(wxDC& dc, wxPoint& position, wxUint16 cyclesVisible, wxUint16 currentCycle);

	/* ----------------------------- I/O ---------------------------------- */

	/** Saves the past activity of this neuron. */
	void SaveActivity(wxFile *file, wxUint16 numCycles);
	
protected:
	float mAlpha;						/**< alpha parameter of this sigmoid neuron	*/
	wxUint16 mMaxNumCycles; /**< Maximum number of cycles that can be stored. */
	float *mActivity;				/**< Neural activity (membrane potential along time) storage. */
	SigmoidNeuronType mType;
	bool mInhibitory;
	
	const static wxUint8 MembranePotentialDrawHeight; /**< Used when painting neural activity. */
};

#endif /* _SIGMOIDNEURON_H */
