/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/list.h>

#include "../../status.h"
#include "../../utils/randomtools.h"
#include "sigmoidneuron.h"
#include "../synapse.h"

#include "sigmoidnn.h"

SigmoidNN::SigmoidNN(wxUint8 numInput, wxUint8 numInternal, wxUint8 numOutput, wxUint8 strengthRange) : NeuralNetwork(numInternal + numOutput + numInput + 1, numOutput) {
	mNumInput = numInput;
	mNumBias = 1;
	mNumInternal = numInternal;
	mNumNeurons = numInternal + numOutput;
	
	mStrengthRange = strengthRange;

	// The neurons are ordered as follows: internal [if any] .. output .. input .. bias [only one]
	for(wxUint8 i=0; i<mSize; i++) {
		mNeurons[i] = new SigmoidNeuron();	// deletion is done in neuralnetwork.cpp
		((SigmoidNeuron *)mNeurons[i])->SetType(
			i < mNumInternal ? InternalNeuron :
			i < mNumNeurons  ? OutputNeuron   :
			i < mNumNeurons + mNumInput ? InputNeuron :
			BiasNeuron		
		);
	}
}

SigmoidNN::~SigmoidNN() {
	/* Note: deletion of neurons is handled by neuralnetwork.cpp */
}

void SigmoidNN::SetInput(float *input) {
	for(wxUint8 i=mNumNeurons; i<mNumNeurons+mNumInput; i++) 
		mNeurons[i]->SetMembranePotential(input[i-mNumNeurons]);
	
	/* Bias neuron */
	mNeurons[mNumNeurons+mNumInput]->SetMembranePotential(-1);
}

void SigmoidNN::Compute() {
	wxInt16 i;
	
	// for each neuron
	for(i=mSize-1; i>=0; i--) 
		((SigmoidNeuron *)mNeurons[i])->ComputeMembranePotential(mCurrentCycle);
	
	// for each output 
	for(i=0; i<mNumOutput; i++)
		mOutput[i] = ((SigmoidNeuron *)mNeurons[i+mNumInternal])->GetMembranePotential();
}

void SigmoidNN::AllocActivityStorage(wxUint16 numCycles) {
	for(wxUint8 i=0; i<mSize; i++)
		((SigmoidNeuron *)mNeurons[i])->AllocActivityStorage(numCycles);
}

void SigmoidNN::SaveActivity(wxFile *file) {
	file->Write("Neural activity of SigmoidNN\n");
	for(wxUint8 i=0; i<mSize; i++)
		((SigmoidNeuron *)mNeurons[i])->SaveActivity(file, mCurrentCycle);
	file->Write("End of log\n\n");
}

wxSize SigmoidNN::GetPaintArchitectureSize() {
  wxSize architectureSize ;
  wxUint8 width = 12, height = 12 ;

	// in the width, 250 is the width of the largest string to draw
  architectureSize.SetWidth(
  	(10 + width * mSize + 40) > 440 ?
    (10 + width * mSize + 40) : 440 ) ;
  architectureSize.SetHeight( 50 + height * (mSize - mNumInput) + 20 );

  return architectureSize ;
}

wxSize SigmoidNN::GetPaintActivitySize() {
  wxSize activitySize ;

  /* The width of Activity graph is always 3/2*architecture size (see neuralNetwork.cpp)*/
  activitySize.SetWidth( (GetPaintArchitectureSize().GetWidth() * 3) / 2 ) ;
	// 15 = MembranePotentialDrawHeight + 5 (see sigmoidneuron.cpp)
  activitySize.SetHeight( mSize*15 + 130 );

  return activitySize ;
}

void SigmoidNN::PaintArchitecture(wxDC& dc, wxPoint& position, wxSize& size) {
	position.x += 10;
	position.y += 50;
 
 	wxUint16 i=0, n=0, j, k, count;
 	Neuron *neuron;
 	wxArraySynapse *synapses;
	wxUint8 width = 12, height = 12;
	float strength;
	
	// Prepare for writing
	wxFont font = dc.GetFont();
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);
	
	// draw horizontal line
	dc.DrawLine(position.x, position.y + width * i + 18, position.x + height * (mSize + 1) + 10, position.y + width * i + 18);

	// draw vertical line
	dc.DrawLine(position.x + width * i + 18, position.y, position.x + width * i + 18, position.y + height * (mNumNeurons + 1) + 10);

	// draw each internal neuron twice (horizontal and vertical)
	for( ; i<mNumInternal; i++) {	
		if(((SigmoidNeuron *)mNeurons[n])->GetInhibitory()) dc.SetTextForeground(*wxGREEN);
		else dc.SetTextForeground(*wxBLACK);	
		dc.DrawText("I", position.x, position.y + height * i + 20 - 3);
		dc.DrawText("I", position.x + width * i + 20, position.y - 3);
		n++;
	}
	dc.SetTextForeground(*wxBLACK);
	
	// draw horizontal line
	dc.DrawLine(position.x, position.y + width * i + 18, position.x + height * (mSize + 1) + 10, position.y + width * i + 18);
	
	// draw vertical line
	dc.DrawLine(position.x + width * i + 18, position.y, position.x + width * i + 18, position.y + height * (mNumNeurons + 1) + 10);
	
	// draw each output neuron twice (horizontal and vertical)
	for( ; i<mNumNeurons; i++) {
		if(((SigmoidNeuron *)mNeurons[n])->GetInhibitory()) dc.SetTextForeground(*wxGREEN);
		else dc.SetTextForeground(*wxBLACK);	
		dc.DrawText("O", position.x, position.y + height * i + 20 - 3);
		dc.DrawText("O", position.x + width * i + 20, position.y - 3);
		n++;
	}
	dc.SetTextForeground(*wxBLACK);

	// draw horizontal line
	dc.DrawLine(position.x, position.y + width * i + 18, position.x + height * (mSize + 1) + 10, position.y + width * i + 18);
	
	// legend
	dc.DrawText("Synaptic strength: red = positive, blue = negative", position.x, position.y + width * i + 30);
	dc.DrawText("I: Internal neuron | O: Output neuron | S: Sensor input | B: Bias neuron", position.x, position.y + width * i + 45);
	dc.SetTextForeground(*wxGREEN);	
	dc.DrawText("When inhibitory, neurons are painted in green", position.x, position.y + width * i + 60);
	dc.SetTextForeground(*wxBLACK);

	// draw vertical line
	dc.DrawLine(position.x + width * i + 18, position.y, position.x + width * i + 18, position.y + height * (mNumNeurons + 1) + 10);
	
	// draw each sensory receptor
	dc.SetBrush(wxBrush(wxColour(0,0,0),wxTRANSPARENT));
	for( ; i<mNumNeurons+mNumInput; i++) {
		dc.DrawText("S", position.x + width * i + 20, position.y - 3);
	}
	
	// draw vertical line
	dc.DrawLine(position.x + width * i + 18, position.y, position.x + width * i + 18, position.y + height * (mNumNeurons + 1) + 10);

	// draw bias
	dc.SetBrush(wxBrush(wxColour(0,0,0),wxSOLID));
	for( ; i<mNumNeurons+mNumInput+mNumBias; i++) {
		dc.DrawText("B", position.x + width * i + 20, position.y - 3);
	}

	// draw vertical line
	dc.DrawLine(position.x + width * i + 18, position.y, position.x + width * i + 18, position.y + height * (mNumNeurons + 1) + 10);
	

	// draw each synapse
	for(i=0; i<mNumNeurons; i++) {  // each line
		synapses = mNeurons[i]->GetIncoming();
		count = synapses->GetCount();
		for(j=0; j<count; j++) {			// each column
			neuron = synapses->Item(j)->GetPreNeuron();
			strength = ((synapses->Item(j)->GetStrength())/mStrengthRange)*255;
			if(strength>0) dc.SetBrush(wxBrush(wxColour(255,255-(wxUint8)strength,255-(wxUint8)strength),wxSOLID));
			else {
   			strength=-strength;
   			dc.SetBrush(wxBrush(wxColour(255-(wxUint8)strength,255-(wxUint8)strength,255),wxSOLID));
			}
			for(k=0; mNeurons[k] != neuron; k++)
				;
			dc.DrawRectangle(position.x + width * k + 20, position.y + height * i + 20, width - 3, height - 3);
		}
	}
}

void SigmoidNN::PaintActivity(wxDC& dc, wxPoint& position, wxSize& size) {
	SigmoidNeuronType type;
	SigmoidNeuron *neuron;
	wxString s;
		
	/* Draw neural network name */
	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	dc.DrawText("Sigmoid Neural Network", position.x + 10, position.y + 15);
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);

//	dc.SetPen(*wxBLACK_PEN);              
//	dc.SetBrush(*wxBLACK_BRUSH);

	wxUint16 cyclesVisible = (size.GetWidth() - 40);
	if(cyclesVisible < 1) cyclesVisible = 1;

	position.y += 40;
	for(wxUint8 i=0; i<mSize; i++) {
		neuron = (SigmoidNeuron *)mNeurons[i];
		if(neuron->GetInhibitory()) { dc.SetTextForeground(*wxGREEN); dc.SetPen(*wxGREEN_PEN); }
		else { dc.SetTextForeground(*wxBLACK); dc.SetPen(*wxBLACK_PEN); }
		type = neuron->GetType();
		if(type == InternalNeuron) s = "I";
		else if(type == OutputNeuron) s = "O";
		else if(type == InputNeuron) s = "S";
		else if(type == BiasNeuron) s = "B";
		dc.DrawText(s, position.x + 10, position.y + 5);
		neuron->PaintActivity(dc, position, cyclesVisible, mCurrentCycle);
	}
}
