/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _SIGMOIDNN_H
#define _SIGMOIDNN_H

class SigmoidNeuron;

#include "../neuralnetwork.h"

/** 
 * Class to construct artificial neural networks made of sigmoid neurons and simple weighted synapses, inherited from NeuralNetwork
 * The neurons are ordered as follows: internal (if existing) .. output .. input .. bias (only one)
 * The network is fully connected or has one hidden layer (no direct connections between input and output). In this case numInternal must be != 0.
 * @see SigmoidNeuron
 * @see SigmoidSynapse
 * @see NeuralNetwork
 */

class SigmoidNN : public NeuralNetwork
{
public:

	/**
	 * SigmoidNN constructor
	 * @param inputSize the number of input neurons
	 * @param internalSize the number of internal neurons
	 * @param outputSize the number of output neurons
	 */
	SigmoidNN(wxUint8 numInput, wxUint8 numInternal, wxUint8 numOutput, wxUint8 stengthRange);

	/**
	 * SigmoidNN destructor
	 */
	~SigmoidNN();  

	/**
	 * Method to set the input values into the input neurons of the neural network
	 * @param input is an array of inputs 
	 */
	virtual void SetInput(float *input);

	/**
	 * Method to compute the output values of the neural network 
	 */
	virtual void Compute();
  
	/** set storage of activity*/
	virtual void AllocActivityStorage(wxUint16 size);

	/* ---------------- LOGGING -------------------- */

protected:
	/** Does the actual saving of the neural activity. */
	virtual void SaveActivity(wxFile *file);

public:
	/* ---------------- PAINTING -------------------- */

	/** Return the size needed to paint the neuralNetwork architecture */
	virtual wxSize GetPaintArchitectureSize();

	/** Return the size needed to paint the neuralNetwork activity */
	virtual wxSize GetPaintActivitySize();

	/** 
	 * Paint the neural network data, which is neural network dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintArchitecture(wxDC& dc, wxPoint& position, wxSize& size) ;

	/** 
	 * Paint the neural network activity, which is neural network dependent.
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintActivity(wxDC& dc, wxPoint& position, wxSize& size);
	
	/* ---------------- ACCESSORS -------------------- */
	
	/** 
	 * Accessor method
	 * @return synaptic strength range (only the positive part)
	 */
	wxUint8 GetStrengthRange() { return mStrengthRange; }
 
	/** 
	 * Accessor method
	 * @param strengthRange New synaptic strength range value.
	 */
	void SetStrengthRange(wxUint8 strengthRange) { mStrengthRange = strengthRange; }

protected:
	wxUint8 mNumInput, mNumBias, mNumInternal, mNumNeurons;
	wxUint8 mStrengthRange;	/* The real range is +- mStrengthRange */
};

#endif /* _SIGMOIDNN_H */
