/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../../status.h"
#include "../../modelsettings.h"
#include "../../ga/genome.h"
#include "../../exp/experiment.h"
#include "../../utils/xml/xml.h"
#include "sigmoidnn.h"
#include "sigmoidneuron.h"
#include "../synapse.h"

#include "sigmoidnnfactory.h"

SigmoidNNFactory::SigmoidNNFactory(ModelSettings *modelSettings) : NeuralNetworkFactory(modelSettings) {
  // Number of bits encoding neuron parameters:
	// 0 -> alpha [0..8]
	// 1 -> sign [0..1]
	mNeuronBits[0] = 0;
	mNeuronBits[1] = 0;

	// Number of bits encoding synapse parameters:
	// 0 -> existence [0..1]
	// 1 -> strength [0..16]
	mSynapseBits[0] = 0;
	mSynapseBits[1] = 4;
  
	// The NN is fully connected with recurrent connections by default (no hidden layer)
	mIsFullyConnected = true;
	mIsRecurrent = true;
	mIsRecurrentOnlyOutput= false;
	
 // Default synaptic range (only positive part => final range will be [-4,4])
	mStrengthRange = 4;
  
  // Defaut number of input neurons which are connecting directly to output neurons
	mNumInputDirectlyConnectedOutput = 0;
  
	// Default number of internal neurons, additional to neurons used for output
	mNumInternal = 0;
  
	// Connecting a bias neuron to each neuron of the NN is equivalent to a threshold in the output function
	mNumBias = 1;
}

SigmoidNNFactory::~SigmoidNNFactory() {

}

NeuralNetwork *SigmoidNNFactory::Create(Genome *genome) {
	Experiment *experiment = mModelSettings->GetExperiment();
	SigmoidNN *neuralNetwork;
	SigmoidNeuron *neuron;
	Synapse *synapse;
  
	float decodedFloat;
	wxUint16 i, j, genomeIndex = 0;
	wxUint8 numInput, numInputDirectlyConnectedOutput, numBias, numInternal, numOutput, numNeurons;

	// Instantiate the neural network
	numInput = experiment->GetNumSensorNeurons();
	numInputDirectlyConnectedOutput =mNumInputDirectlyConnectedOutput;
	numBias = mNumBias;
	numInternal = mNumInternal;
	numOutput = experiment->GetNumOutputNeurons();
	numNeurons = mNumInternal + numOutput;
	neuralNetwork = new SigmoidNN(numInput, numInternal, numOutput, mStrengthRange);

	// For each neuron (input and bias are not included, because they have no parameters genetically coded)
	for(i=0; i<numNeurons; i++) {	// For each "neuron"
		neuron = (SigmoidNeuron *)neuralNetwork->GetNeuron(i);
		
		// 0 -> alpha [0..8]
		if(mNeuronBits[0]) {
			if(genome->DecodeFloat(decodedFloat, genomeIndex, mNeuronBits[0])) {
				StatusLine("SigmoidNNFactory - Error while trying to decode the alpha parameter of a neuron.");
				delete neuralNetwork;
				return NULL;
			}
			genomeIndex += mNeuronBits[0];
			neuron->SetAlpha(decodedFloat);
		}
		
		// 1 -> sign [0..1]
		if(mNeuronBits[1]) {
			if(genome->GetValue(genomeIndex++)==1) neuron->SetInhibitory();
			genomeIndex += mNeuronBits[1];
		}

		// For each synapse (j : Internal..Output..Input..Bias)
		for(j=0; j<numNeurons+numInput+numBias; j++) { 	
  
			// Check for non connectivity	if NN has no direct connection between input and output (one hidden layer)
			if (!mIsFullyConnected) {
				if (i<numInternal) { // this one is an internal neuron
					if (j>=numInternal && j<numNeurons) continue;  // do not connect to output neurons
					if (numInputDirectlyConnectedOutput!=0 && j>= numNeurons && j<(numNeurons+numInputDirectlyConnectedOutput)) continue; // do not connect to  input neurons which are directly connecting to output neurons 
				} else {  					// this one is an output neuron
					if (j>=(numNeurons+numInputDirectlyConnectedOutput) && j<numNeurons+numInput) continue;  // do not connect to input neurons which aren't directly to them
				}
			}
			if (!mIsRecurrent) {
				if (i<numInternal && j<numInternal) continue; // do not connect between internal neurons 
				else if ( !mIsRecurrentOnlyOutput && i>=numInternal && i<numNeurons && j>=numInternal && j<numNeurons) continue; // do not connect between output neurons
			}
		
			// 0 -> existence [0..1]
			if(mSynapseBits[0] && !genome->GetValue(genomeIndex++)) {
				genomeIndex += mSynapseBits[1];
				continue;
			}
			
			// Instanciate a synapse, synaptic strength is mStrengthRange by default
			synapse = new Synapse(neuralNetwork->GetNeuron(j), neuron, mStrengthRange);

			// 0 -> strength [0..16]
			if(mSynapseBits[1]) {
				if(genome->DecodeFloat(decodedFloat, genomeIndex, mSynapseBits[1])) {
					StatusLine("SigmoidNNFactory - Error while trying to decode synapse strength.");
					delete neuralNetwork;
					return NULL;
				}
				genomeIndex += mSynapseBits[1];
				synapse->SetStrength((decodedFloat - 0.5) * mStrengthRange * 2); // Synaptic strenghs are in [-mStrengthRange,mStrengthRange]
			}
		}
	}
	return neuralNetwork;
}

wxUint16 SigmoidNNFactory::GetGenomeLength() {
	wxUint8 numInput, numOutput, numNeurons, neuronBits, synapseBits;
	wxUint16 length;
	Experiment *experiment = mModelSettings->GetExperiment();
	numInput = experiment->GetNumSensorNeurons();	// receptors
	numOutput = experiment->GetNumOutputNeurons(); // outputs
	numNeurons = experiment->GetNumOutputNeurons() + mNumInternal;	// neurons
	neuronBits = mNeuronBits[0] + mNeuronBits[1];
	synapseBits = mSynapseBits[0] + mSynapseBits[1];
				
	if (mIsFullyConnected)
		length = numNeurons * neuronBits + numNeurons * (numNeurons + numInput + mNumBias) * synapseBits;
	else {
		if (mNumInputDirectlyConnectedOutput !=0)
			length = numNeurons * neuronBits
							+ mNumInternal * (mNumInternal + numInput - mNumInputDirectlyConnectedOutput + mNumBias) * synapseBits
							+ numOutput * (mNumInputDirectlyConnectedOutput + numNeurons + mNumBias) * synapseBits;
		else
			length = numNeurons * neuronBits
							+ mNumInternal * (mNumInternal + numInput + mNumBias) * synapseBits
							+ numOutput * (numNeurons + mNumBias) * synapseBits;
	}					
	if (!mIsRecurrent) {
		if (mIsRecurrentOnlyOutput)
			length -= (mNumInternal*mNumInternal) * synapseBits;
		else
			length -= (mNumInternal*mNumInternal + numOutput*numOutput) * synapseBits;
	}

	StatusLine(wxString::Format("Genome length is %d.", length));

	return length;
}

wxUint16 SigmoidNNFactory::ConfigGenomeEncoding() {
	wxInt8 ret;
	
	ret=wxGetNumberFromUser("Choose the number of bits to code the sigmoid neuron parameter", "Alpha :","Configure genome encoding", mNeuronBits[0], 0, 255);
	if(ret!=-1)
		mNeuronBits[0]=ret;
		
	ret=wxGetNumberFromUser("Choose the number of bits (0 or 1) to code the sign of the neuron", "Sign :","Configure genome encoding", mNeuronBits[1], 0, 1);
	if(ret!=-1)
		mNeuronBits[1]=ret;
		
	ret=wxGetNumberFromUser("Choose the number of bits to code the existence of a synapse", "Existence :","Configure genome encoding", mSynapseBits[0], 0, 1);
	if(ret!=-1)
		mSynapseBits[0]=ret;
		
	ret=wxGetNumberFromUser("Choose the number of bits to code the strength of a synapse", "Strength :","Configure genome encoding", mSynapseBits[1], 0, 255);
	if(ret!=-1)
		mSynapseBits[1]=ret;
		
  return GetGenomeLength();
}

void SigmoidNNFactory::ConfigNeuralNetwork() {
	long number;
	wxString result;
	wxUint8 minNumInternal=0;
	wxUint8 minNumInputDirectlyConnectedOutput=0;
	wxString choices[3];
	bool isOutputSomeInputConnecetd = false;
	
	if (mIsFullyConnected) {
		choices[0] = "Fully connected";
		choices[1] = "No direct connection between input and output (one hidden layer)";
  	choices[2] = "No direct connection between input and output except for certain input neurons (one hidden layer)"	;
 	} else { 	
		choices[0] = "No direct connection between input and output (one hidden layer)";
		choices[1] = "No direct connection between input and output except for certain input neurons (one hidden layer)"	;
		choices[2] = "Fully connected";
 	}
	result = wxGetSingleChoice("Please select the network architecture:                                                            ", "Network architecture", 3, choices);
	if(!result.IsEmpty()) {
		if(result == "Fully connected") {
			mIsFullyConnected = true;	
   		isOutputSomeInputConnecetd = false;	
   	}	else if(result == "No direct connection between input and output (one hidden layer)") {
			mIsFullyConnected = false;
			isOutputSomeInputConnecetd = false;
   	} else if(result == "No direct connection between input and output except for certain input neurons (one hidden layer)"){
			mIsFullyConnected = false;
			isOutputSomeInputConnecetd=true;
		}
	}
	
	if(mIsRecurrent) {
		choices[0] = "With recurent and lateral connections for all output and internal neurons";
		choices[1] = "With recurent and lateral connections for only output neurons";
  	choices[2] = "Feedforward (perceptron)";
 	} else { 	
		choices[0] = "Feedforward (perceptron)";
		choices[1] = "With recurent and lateral connections for only output neurons";
		choices[2] = "With recurent and lateral connections for all output and internal neurons";
 	}
	result = wxGetSingleChoice("Please select the network architecture:                                                            ", "Network architecture", 3, choices);
	if(!result.IsEmpty()) {
		if(result == "With recurent and lateral connections for all output and internal neurons") {
			mIsRecurrent = true;
			mIsRecurrentOnlyOutput = false ;
		}
		else if (result == "With recurent and lateral connections for only output neurons") {
			mIsRecurrent = false;
			mIsRecurrentOnlyOutput = true;
  	} else if(result == "Feedforward (perceptron)") {
			mIsRecurrent = false;
			mIsRecurrentOnlyOutput = false;
		}
	}
	
	if (!mIsFullyConnected) {
		minNumInternal=1;
		if (mNumInternal==0) mNumInternal++;
		if (isOutputSomeInputConnecetd) {
  		minNumInputDirectlyConnectedOutput =1;
  		mNumInputDirectlyConnectedOutput =1;
		}
	} else
		minNumInternal=0;
	
		
	if((number = wxGetNumberFromUser("Enter the number of internal neurons, in addition to the neurons used for output (defined in experiment).", "# neurons", "Configure neural network", mNumInternal, 0, 50)) >= minNumInternal)
		mNumInternal = (wxUint16)number;
		
	if (!mIsFullyConnected && mNumInternal!=0 && isOutputSomeInputConnecetd){
		if((number = wxGetNumberFromUser("Enter the number of input neurons which are directly connecting to output neurons, among the neurons used for input (defined in experiment).", "# neurons", "Configure neural network",mNumInputDirectlyConnectedOutput , 0, 50)) >= minNumInputDirectlyConnectedOutput)
		mNumInputDirectlyConnectedOutput = (wxUint16)number;
 }
	if((number = wxGetNumberFromUser("Enter the positive range of synaptic strength.", "Strength range", "Configure neural network", mStrengthRange, 0, 50)) > 0)
		mStrengthRange = (wxUint8)number;
	
	// Dirty trick for handling correctly the non-recurrent version... ...
	if (!mIsRecurrent) {
 		if (mNumInternal==0) mIsFullyConnected = true;
 		else mIsFullyConnected = false;
	}
}


/* --- XML --- */
void SigmoidNNFactory::Load (XMLElement *myself, XMLLoader *loader) {
	mNumInternal = myself->GetAttribute("num_internal")->GetByteValue();
	mNumBias = myself->GetAttribute("num_bias")->GetByteValue();
	mNumInputDirectlyConnectedOutput = myself->GetAttribute("num_input_directly_connected_output")->GetByteValue();
	mStrengthRange = myself->GetAttribute("strength_range")->GetByteValue();
	
	mIsFullyConnected = myself->GetAttribute("fully_connected")->GetBoolValue();
	mIsRecurrent = myself->GetAttribute("recurrent")->GetBoolValue();
	mIsRecurrentOnlyOutput = myself->GetAttribute("recurrent_only_output")->GetBoolValue();
	
	mNeuronBits[0] = (wxUint8) myself->GetAttribute("neuron_bits_num")->GetByteValue();
	mNeuronBits[1] = (wxUint8) myself->GetAttribute("neuron_bits_sign")->GetByteValue();
	mSynapseBits[0] = (wxUint8) myself->GetAttribute("synapse_bits_num")->GetByteValue();
	mSynapseBits[1] = (wxUint8) myself->GetAttribute("synapse_bits_sign")->GetByteValue();
}

XMLElement* SigmoidNNFactory::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());
	result->AddAttribute(new XMLAttribute("num_internal",(wxInt8)mNumInternal));
	result->AddAttribute(new XMLAttribute("num_bias",(wxInt8)mNumBias));
	result->AddAttribute(new XMLAttribute("num_input_directly_connected_output",(wxInt8)mNumInputDirectlyConnectedOutput));
	result->AddAttribute(new XMLAttribute("strength_range",(wxInt8)mStrengthRange));
	
	result->AddAttribute(new XMLAttribute("fully_connected",(bool)mIsFullyConnected));
	result->AddAttribute(new XMLAttribute("recurrent",(bool)mIsRecurrent));
	result->AddAttribute(new XMLAttribute("recurrent_only_output",(bool)mIsRecurrentOnlyOutput));

	result->AddAttribute(new XMLAttribute("neuron_bits_num",(wxUint8)mNeuronBits[0]));
	result->AddAttribute(new XMLAttribute("neuron_bits_sign",(wxUint8)mNeuronBits[1]));
	result->AddAttribute(new XMLAttribute("synapse_bits_num",(wxUint8)mSynapseBits[0]));
	result->AddAttribute(new XMLAttribute("synapse_bits_sign",(wxUint8)mSynapseBits[1]));

	return result;
}

wxString SigmoidNNFactory::GetXMLName () { return GetShortXMLName(); }
wxString SigmoidNNFactory::GetShortXMLName() { return SigmoidNNFactory::GetStaticXMLType(); }
