/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../../status.h"
#include "../../utils/randomtools.h"
#include "spikingnn.h"
#include "spikingbase.h"

const wxUint8 SpikingBase::SpikeDrawHeight = 8;
const wxUint8 SpikingBase::NumUpdatesToDrawRatio = 1;
const wxUint8 SpikingBase::DisplaySpaceSpikes = 2;

SpikingBase::SpikingBase() {
	// Init parameters to default
	mSign = 1;
	mThreshold = 0.5f;
	mMembraneTimeC = 4.0f;
	
	// Init variables
	mTimeFromLastSpike = 255; 
	mActivity = NULL;
	wxUint16 spikeWinLen = SpikingNN::GetSpikeWindowLen();
	mSpikeWindow = new bool[spikeWinLen];
	for(wxUint16 i=0; i<spikeWinLen; i++)
		mSpikeWindow[i] = false;
	mSpikeWindowIdx = 0;
}

SpikingBase::~SpikingBase() {
	if(mActivity) {
		for(wxUint16 i=0; i<mMaxNumCycles; i++)
			delete [] mActivity[i];
		delete [] mActivity;
	}
	delete [] mSpikeWindow;
}

void SpikingBase::SetParameterSelf(ParamID id, float param)
{
	switch(id)
	{
		case PARAMID_SPIKE_SIGN:
			SetSign(param != 0.0f ? 1 : -1);
			break;
		
		case PARAMID_SPIKE_THRESHOLD:
			SetThreshold(param);
			break;
		
		case PARAMID_SPIKE_MEMBRANE_TIME_CONST:
			SetMembraneTimeC(param);
			break;
		
		default:
			Neuron::SetParameterSelf(id, param);
			break;
	};
}

bool SpikingBase::GetSpikeHistory(wxUint16 index)
{
	return mSpikeWindow[mSpikeWindowIdx - index + (index > mSpikeWindowIdx ? SpikingNN::GetSpikeWindowLen(): 0)];
}

void SpikingBase::AllocActivityStorage(wxUint16 numCycles, wxUint8 numUpdates) {
	mMaxNumCycles = numCycles;
	mActivity = new bool *[numCycles]; 
	for(wxUint16 i=0; i<numCycles; i++) {
		mActivity[i] = new bool[numUpdates];
		for(wxUint8 j=0; j<numUpdates; j++)
			mActivity[i][j] = false;
	}
}

void SpikingBase::ComputeMembranePotential(wxUint16 currentCycle, wxUint16 currentUpdate) {
	mMembranePotential = ComputeExternalPotential(currentUpdate) + ComputeRefractoryKernel();

	// check for spike
	if(mMembranePotential > mThreshold) { // spike!
		mMembranePotential = 5.0f;
		mActivity[currentCycle][currentUpdate] = true;
		mSpikeWindow[mSpikeWindowIdx++] = true;
		mTimeFromLastSpike = 0;
	} else { // no spike
		mActivity[currentCycle][currentUpdate] = false;
		mSpikeWindow[mSpikeWindowIdx++] = false;
		mTimeFromLastSpike++;
	}
	if(mSpikeWindowIdx == SpikingNN::GetSpikeWindowLen()) mSpikeWindowIdx = 0;
}

float SpikingBase::ComputeRefractoryKernel() {
	if(mTimeFromLastSpike > 0)
			return RandomTools::RandUnityDouble() * -exp(-mTimeFromLastSpike / (double)mMembraneTimeC);
	//		return  -exp(-mTimeFromLastSpike / (double)mMembraneTimeC);
	return -999.0f;
}

void SpikingBase::PaintActivity(wxDC& dc, wxPoint& position, wxUint16 cyclesVisible, wxUint16 currentCycle, wxUint8 updatesPerCycle) {
	if(!mActivity || !currentCycle)
		return;
	wxUint16 x, start = 0, index = 0;
	float aveActi, sumActi;
	wxUint8 update;

	if(currentCycle > cyclesVisible)
		start = currentCycle - cyclesVisible;
	x = position.x + 20;
	for(wxUint16 cycle = start; cycle<currentCycle; cycle++){
		for(update = 0; update < (updatesPerCycle/NumUpdatesToDrawRatio); update++) {
			sumActi = 0.0f;

			for(index = 0; index<NumUpdatesToDrawRatio; index++)
				sumActi += 1.0f * mActivity[cycle][(update * NumUpdatesToDrawRatio) + index];
			aveActi = sumActi / index;
			if(aveActi >= 0.5f) {
				dc.DrawLine(x, position.y + SpikeDrawHeight + 5, x, position.y + 5);
				dc.DrawLine(x, position.y + SpikeDrawHeight + 5, x, position.y + 5);
			} else
				dc.DrawLine(x, position.y + SpikeDrawHeight + 5, x, position.y + SpikeDrawHeight + 6);
			x += DisplaySpaceSpikes;
		}
		/* Draw a vertical line to separate sensory-motor cycle */
		dc.DrawLine(x,position.y + SpikeDrawHeight + 5,x,position.y - SpikeDrawHeight + 5);
		x+=DisplaySpaceSpikes;
	}
	position.y += SpikeDrawHeight + 5;
}

void SpikingBase::SaveActivity(wxFile *file, wxUint16 numCycles, wxUint16 numUpdates, wxUint16 numNeuron) {	

#if 1 /* 0->Version 2 , 1->Version 1*/
	// Version 1 (default)
	for(wxUint16 i=0; i<numCycles; i++){
		for(wxUint8 j=0; j<numUpdates; j++){
			file->Write(mActivity[i][j] ? "1 " : "0 ");
		}
	}
	file->Write( "\n" ) ;

#else
	// Version 2
	for(wxUint16 i=0; i<numCycles; i++){
		file->Write(wxString::Format("%d %d ",numNeuron,i));
		for(wxUint8 j=0; j<numUpdates; j++){
			file->Write(mActivity[i][j] ? "1 " : "0 ");
		}
		file->Write("\n"); //newline=newcycle
	}
#endif
}
