/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _SPIKINGBASE_H
#define _SPIKINGBASE_H

#include <wx/file.h>
#include <wx/dc.h>
#include "../neuron.h"


enum
{
	PARAMID_SPIKE_SIGN = 100,
	PARAMID_SPIKE_THRESHOLD,
	PARAMID_SPIKE_MEMBRANE_TIME_CONST
};

/** 
 * Base class for spiking neurons and receptors, subclass of Neuron.
 */
class SpikingBase : public Neuron
{
public:
	/** SpikingBase constructor */
	SpikingBase();

	/** SpikingBase destructor */ 
	virtual ~SpikingBase();


	

	/* ------------------------ USAGE ----------------------------- */

	/** 
	 * Allocates storage for the activity of a certain number of network updates. 
	 * @param numCycles Number of cycles to reserve space for.
	 * @param numUpdates Number of network updates per cycle to reserve space for.
	 */
	virtual void AllocActivityStorage(wxUint16 numCycles, wxUint8 numUpdates);

protected:
	/** Computes external contributions to the membrane potential. */
	virtual float ComputeExternalPotential(wxUint16 currentUpdate) =0;

	/** Computes the refractory kernel to add to the membrane potential. */
	virtual float ComputeRefractoryKernel();
	
	
	virtual void SetParameterSelf(ParamID id, float param);

public:
	/** 
	 * Method that computes the new membrane potential.
	 * @param currentCycle the cycle that is performed
	 * @param currentUpdate The update of the current cycle.
	 */
	virtual void ComputeMembranePotential(wxUint16 currentCycle, wxUint16 currentUpdate);
	/** 
	 * Method that computes an analogue neuron's output.

	 * @param currentCycle the cycle that is performed
	 */
	//virtual float ComputeOutput(wxUint16 currentCycle,wxUint16 updatesPerCycle,wxUint16 outputWindowLen);
	
	
	/**
	 * Perform any learning or adaptation, for the neuron or the synapses.
	 */
	virtual void PerformLearning(){};
	
	
	/* ------------------------- ACCESSORS --------------------------- */

	wxInt8 GetSign() { return mSign; }
	void SetSign(wxInt8 sign) { mSign = sign; }
	float GetThreshold() { return mThreshold; }
	void SetThreshold(float threshold) { mThreshold = threshold; }
	float GetMembraneTimeC() { return mMembraneTimeC; }
	void SetMembraneTimeC(float membraneTimeC) { mMembraneTimeC = membraneTimeC; }

	/** Returns activity for all cycles and updates. */
	bool **GetActivity() { return mActivity; }
	/** Returns activity for a given cycle. */
	bool *GetActivity(wxUint16 cycle) { return mActivity[cycle]; }
	/** Returns activity for a given cycle and update. */
	bool GetActivity(wxUint16 cycle, wxUint16 update) { return mActivity[cycle][update]; }
	/** Returns the spike window for this neuron. */
	bool *GetSpikeWindow() { return mSpikeWindow; }
	/** Returns true if neuron spiked at time step now minus index. */
	bool GetSpikeHistory(wxUint16 index);

	/** returns the ratio (=updatesPerCycle/NumNumUpdatesToDraw) of updates to draw. */
	static const wxUint8 GetNumUpdatesToDrawRatio() { return NumUpdatesToDrawRatio ; }
	
	/** return the space between the display of two spikes*/
	static const wxUint8 GetDisplaySpaceSpikes() { return DisplaySpaceSpikes ; }

	/* ----------------------- PAINTING -------------------------------- */

	/** 
	 * Paint the activity of this neuron.
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 */
	virtual void PaintActivity(wxDC& dc, wxPoint& position, wxUint16 cyclesVisible, wxUint16 currentCycle, wxUint8 updatesPerCycle);

	/* ----------------------------- I/O ---------------------------------- */

	/** Saves the past activity of this neuron. */
	virtual void SaveActivity(wxFile *file, wxUint16 numCycles, wxUint16 numUpdates, wxUint16 numNeuron);

protected:
	wxInt8 mSign; /**< sign parameter of this spiking neuron */
	float mThreshold; /**< threshold parameter of this spiking neuron */
	float mMembraneTimeC; /**< membrane time constant parameter of this spiking neuron */
	
	wxUint16 mTimeFromLastSpike; /**< store the elapsted time for the last spike emmited */
	wxUint16 mMaxNumCycles; /**< Maximum number of cycles that can be stored. */
	bool **mActivity; /**< Neural activity storage. */
	bool *mSpikeWindow; /**< Store spikes for a certain time (SpikeWindowLen) to use for potential computation. */
	wxUint16 mSpikeWindowIdx; /**< Index for spike window. */

	/*-----------------------------display settings-----------------------------*/
	/* How to use display settings:
	A dot "." shows the absence of spike and a bar "|" shows the presence of a spike.
	The vertical line is splitting all inputs into sensory-motor cycles.
	
	NumUpdatesToDrawRatio sets the number of spike to be displayed. A spike is displayed, if the number of spike averaged is bigger than 0.5.
 	DisplaySpaceSpikes sets the "screen definition".
	Increasing NumUpdatesToDrawRatio or decreasing DisplaySpaceSpikes allow more cycle to be displayed 
 	on the same screen. The vertical line can easily be removed, if necessary. */
 
	/**< Used when painting neural activity. */
	const static wxUint8 SpikeDrawHeight; 
	
	/**< Ratio Number (=updatesPerCycle/NumUpdatesToDraw) of updates to draw for each cycle. */
	const static wxUint8 NumUpdatesToDrawRatio; 
	
	/** space between the display of two spikes. */
	const static wxUint8 DisplaySpaceSpikes; 
};

#endif /* _SPIKINGBASE_H */

