/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/list.h>
#include "math.h"

#include "../../status.h"
#include "../../utils/randomtools.h"
#include "spikingreceptor.h"
#include "spikingneuron.h"
#include "spikingsynapse.h"
#include "spikingoutputac.h"
#include "../neuralnetworktopology.h"

#include "spikingnn.h"

wxUint16 SpikingNN::UpdatesPerCycle = 20;
wxUint16 SpikingNN::OutputWindowLen = 20;
const wxUint16 SpikingNN::SpikeWindowLen  = 20;
//const wxUint16 SpikingNN::SpikeWindowLen  = 100;

float SpikingNN::BiasActivity = 0.5f;

/*
SpikingNN::SpikingNN(wxUint8 numInput, wxUint8 internalSize, wxUint8 numOutput) : NeuralNetwork(numInput + internalSize + numOutput, numOutput) {
	mNumInput = numInput;
	mNumInternal= internalSize;
//	mRegulator = NULL ;
}
*/

SpikingNN::SpikingNN(NeuralNetworkTopology* topo) : NeuralNetwork(topo->GetNeuronNumber(), topo->GetOutputNumber())
{
	mNNTopology = topo;
	
	//mNumInput = topo->GetInputNumber();
	//mNumInternal= internalSize;

	mNumAC=mNumOutput;
	//mNumInput=mSize-mNumOutput-mNumAC-mNumInternal;
}

SpikingNN::~SpikingNN() {

}

void SpikingNN::AllocActivityStorage(wxUint16 numCycles) {
	for(wxUint8 i=0; i<mSize; i++)
		((SpikingBase *)mNeurons[i])->AllocActivityStorage(numCycles, UpdatesPerCycle);
}

void SpikingNN::SetInput(float *input) {
	((SpikingReceptor *)mNeurons[mNNTopology->GetInputNeuronIndex(0)])->SetExternalCurrent(BiasActivity);
	for(wxUint8 i=1; i<mNNTopology->GetInputNumber(); i++)
		((SpikingReceptor*)mNeurons[mNNTopology->GetInputNeuronIndex(i)])->SetExternalCurrent(input[i-1]);
}


void SpikingNN::Compute() {
	wxUint16 j;
	wxInt32 i;
	
	// update network UpdatesPerCycle times
	for(i=0; i<UpdatesPerCycle; i++)
	{
		// update nn with current synaptic weights
		for(j=0; j<mSize; j++) {
			((SpikingBase*)mNeurons[j])->ComputeMembranePotential(mCurrentCycle, i);
		}
		
		// update synaptic weights
		for(j=0; j<mSize; j++)
			((SpikingBase*)mNeurons[j])->PerformLearning();
	}
	
	// for each output, compute the ouput value 

	for(j=0; j<mNNTopology->GetOutputNumber(); j++) {		
		((SpikingOutputAC *)mNeurons[mNNTopology->GetOutputNeuronIndex(j)])->ComputeOutput(mCurrentCycle,UpdatesPerCycle, OutputWindowLen);
		mOutput[j]=((SpikingOutputAC *)mNeurons[mNNTopology->GetOutputNeuronIndex(j)])->GetOutput();
	}



// THE CODE BELOW IS NO LONGER VALID

/***********************************************************
	//scaling (optional)

	//The output are scaled, such that:
	//E.G. the firing rate of one of the output neuron always equal its maximum value
	//Why: because maximum fitness are related to the maximum firing rate
	//	
	//		M.S.												

	float maxOutput=0.0f;
	for(j=mSize-mNumOutput; j<mSize; j++) 
		if (maxOutput<mOutput[j-(mSize-mNumOutput)])
			maxOutput=mOutput[j-(mSize-mNumOutput)];

	if (maxOutput!=0){
		float scale=1/maxOutput;
		for(j=mSize-mNumOutput; j<mSize; j++) {//normalisation
			mOutput[j-(mSize-mNumOutput)]*=scale;
			//StatusLine(wxString::Format("neuronSCALE %d output %f",j,mOutput[j-(mSize-mNumOutput)]));//Debug
		}
	}	
	
***********************************************************/

/***********************************************************
	//Debug:  old firing rate implementation (used for comparison with new one)
	
	bool *activity;
	float mOutputOld[4];
	for(j=0; j<mNumOutput; j++) {
		// outputWindowLen may be any length 	
		outCount = 0;
		mOutputOld[j] = 0.0f;
		for(curCycle=mCurrentCycle; curCycle>=0; curCycle--) {
			//activity = ((SpikingNeuron*)mNeurons[mSize - mNumOutput + j])->GetActivity(curCycle);
			activity = ((SpikingNeuron*)mNeurons[mSize - 2*mNumOutput + j])->GetActivity(curCycle);
			
			for(i=UpdatesPerCycle-1; i>=0; i--) {
				mOutputOld[j] += activity[i] ? 1.0f : 0.0f;
				outCount++;
				if(outCount==OutputWindowLen) break;
			} if(outCount==OutputWindowLen) break;
		}
		mOutputOld[j] /= outCount;
		mOutputOld[j] *= 2.0f;
		StatusLine(wxString::Format("mOutputOLD %d  %f",j , mOutputOld[j])); //debug
	}
***********************************************************/

}

void SpikingNN::RandomizeWeights()
{
	for(wxUint32 i = 0; i < mNNTopology->GetSynapseNumber(); i++)
	{
		if(mNNTopology->GetSynapseEvolvability(i) == EVOLVABILITY_EVOLVABLE)
		{
			wxUint32 from, to;
			mNNTopology->SynapseIndexToNeuronIndice(i, from, to);
			
			for(size_t j = 0; j < mNeurons[to]->GetIncoming()->GetCount(); j++)
			{
				if(mNeurons[to]->GetIncoming()->Item(j)->GetPreNeuron() == mNeurons[from])
					mNeurons[to]->GetIncoming()->Item(j)->SetStrength(RandomTools::RandUnityFloat() * 0.1f + 0.1f);
			}	
		}
	}
}

void SpikingNN::SaveActivity(wxFile *file) {
#if 1
	//version A
	file->Write("neuron's_number number_of_cycle Neural_activity_of_SpikingNN\n");
	for(wxUint8 i=0; i<mSize; i++){
		((SpikingBase *)mNeurons[i])->SaveActivity(file, mCurrentCycle, UpdatesPerCycle, i);
 	}
	file->Write("End of log\n\n");
#else 
	//	version B 
	//	you can use Dario's mathematika correlation script, but then use SpikingBase::SaveActivity with version 1 (default)
	
	//	only dynamics neural are logged
	/*file->Write(wxString::Format("%d %d\n", mNumInternal,mCurrentCycle*UpdatesPerCycle));
		for(wxUint8 i=mNumInput; i<mNumInternal+mNumInput; i++)
			((SpikingBase *)mNeurons[i])->SaveActivity(file, mCurrentCycle, UpdatesPerCycle, i);*/
	
	//or all neurons's are logged
	
	// [AB] i am not sure of what's going on here. Version A seems fine to me. Beside, I don't see
	// any significative difference between the commented version and version B (exept the legenda).

		file->Write(wxString::Format("%d %d\n",mNumInternal+mNumInput,mCurrentCycle*UpdatesPerCycle));
		for(wxUint8 i=0; i<mNumInternal+mNumInput; i++){
			((SpikingBase *)mNeurons[i])->SaveActivity(file, mCurrentCycle, UpdatesPerCycle, i);
 	}

#endif
}	

wxSize SpikingNN::GetPaintArchitectureSize() {
  wxSize architectureSize ;
  wxUint8 width = 12, height = 12 ;
  architectureSize.SetWidth( 10 + width * mSize + 40) ;
  architectureSize.SetHeight( 50 + height * (mSize - mNNTopology->GetInputNumber()) + 20 );

  return architectureSize ;
}

wxSize SpikingNN::GetPaintActivitySize() {
  wxSize activitySize ;
  wxUint8 width = 12;
  /* The width of Activity graph is always 3/2*architecture size (see neuralNetwork.cpp)*/
  activitySize.SetWidth( ((10 + width * mSize + 40) * 3) / 2 ) ;
	// 13 = SpikeDrawHeight + 5 (see sigmoidneuron.cpp)
  activitySize.SetHeight( mSize*13 + 130 );

  return activitySize ;
}

void SpikingNN::PaintArchitecture(wxDC& dc, wxPoint& position, wxSize& size) {
	//position.x += 10;
	position.x += 30;
	position.y += 50;
 
 	wxUint16 i, j, k, count;
 	Neuron *neuron;
 	wxArraySynapse *synapses;
	wxUint8 width = 12, height = 12, numNeurons = mSize - mNNTopology->GetInputNumber();


	// Draw each neuron
	for(i = 0; i < mNNTopology->GetNonInputNumber(); i++)
	{
		if(((SpikingBase*)mNeurons[i])->GetSign() == -1)
			dc.SetBrush(wxBrush(wxColour(0,0,0),wxSOLID));
		else
			dc.SetBrush(wxBrush(wxColour(0,0,0),wxTRANSPARENT));
		dc.DrawEllipse(position.x, position.y + height * mNNTopology->GetNonInputNeuronIndex(i) + 20, width - 2, height - 2);
		dc.DrawEllipse(position.x + width * mNNTopology->GetNonInputNeuronIndex(i) + 20, position.y, width - 2, height - 2);
	}
	
	
	//sperate internal neurons and outputs neurons
	dc.DrawLine((int)(position.x-width*1.5),position.y+height *(numNeurons-mNumOutput)+20,(int)(position.x+width*(mSize+1.5)),position.y+height *(numNeurons-mNumOutput)+20);
	dc.DrawText("N", (int)(position.x-width*1.5),position.y+height *(numNeurons-mNumOutput-mNumAC-1)+20);
	//separate InternalNeuron and AC converter
	dc.DrawLine((int)(position.x-width*1.5),position.y+height *(numNeurons - mNumOutput-mNumAC)+20,(int)(position.x+width*(mSize+1.5)),position.y+height *(numNeurons - mNumOutput-mNumAC)+20);
	dc.DrawText("O", (int)(position.x-width*1.5),position.y+height *(numNeurons - mNumOutput-1)+20);
	dc.DrawText("AC", (int)(position.x-width*1.5),position.y+height *(numNeurons - mNumOutput)+20);
	

	// draw vertical lines
	
	dc.DrawLine(position.x + width * (numNeurons - mNumOutput-mNumAC) + 18, position.y, position.x + width * (numNeurons - mNumOutput-mNumAC) + 18, position.y + height * numNeurons + 20);
	dc.DrawText("N", (int)(position.x + width * (numNeurons - mNumOutput-mNumAC-1)+18),position.y-20);

	dc.DrawLine(position.x + width * (numNeurons - mNumOutput) + 18, position.y, position.x + width * (numNeurons -mNumOutput) + 18, position.y + height * numNeurons + 20);
	dc.DrawText("O", (int)(position.x + width * (numNeurons - mNumOutput-1)+18),position.y-20);

	dc.DrawLine(position.x + width * numNeurons + 18, position.y, position.x + width * numNeurons + 18, position.y + height * numNeurons + 20);
	dc.DrawText("A", (int)(position.x + width * (numNeurons-1) + 18),position.y-20);

	dc.DrawLine(position.x + width * (numNeurons + 1) + 18, position.y, position.x + width * (numNeurons + 1) + 18, position.y + height * numNeurons + 20);
	dc.DrawText("B", (int)(position.x + width * (numNeurons + 1-1)+18),position.y-20);

	dc.DrawText("S", (int)(position.x + width * (mSize-1)+18),position.y-20);

	// draw each sensory receptor
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	for(i=0; i<mNNTopology->GetInputNumber(); i++)
		dc.DrawEllipse(position.x + width * mNNTopology->GetInputNeuronIndex(i) + 20, position.y, width - 2, height - 2);

	// draw each synapse
	for(i=0; i<mNNTopology->GetNonInputNumber(); i++)
	{
		wxUint32 neuronIndex = mNNTopology->GetNonInputNeuronIndex(i);
		
		
		synapses = mNeurons[neuronIndex]->GetIncoming();
		count = synapses->GetCount();
		
		for(j=0; j<count; j++)
		{
			if(!synapses->Item(j)->GetExistence())
				continue;
			
			neuron = synapses->Item(j)->GetPreNeuron();
			
			for(k=0; mNeurons[k] != neuron; k++)
				;
			
			//if(k < mNumInput) k += numNeurons;
			//else k -= mNumInput;
			
			wxPoint pt(position.x + width * k + 20, position.y + height * neuronIndex + 20);
			wxSize sz(width-2, height-2);
			synapses->Item(j)->PaintArchitecture(dc, pt, sz);
		}
	}
}

void SpikingNN::PaintActivity(wxDC& dc, wxPoint& position, wxSize& size) {
	// Draw neural network name
	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	dc.DrawText("Spiking Neural Network", position.x + 10, position.y + 15);
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);

	dc.SetPen(*wxBLACK_PEN);              
	dc.SetBrush(*wxBLACK_BRUSH);

	wxUint16 cyclesVisible = (size.GetWidth() - 40) / ((UpdatesPerCycle * SpikingBase::GetDisplaySpaceSpikes()) / SpikingBase::GetNumUpdatesToDrawRatio());
	if(cyclesVisible < 1) cyclesVisible = 1;

	position.y += 40;
	
	for(wxUint32 i=0; i<mSize; i++) {
		
		// This is hacky...
		wxString s;
		if(i < mNNTopology->GetNonInputNumber() - 2*mNNTopology->GetOutputNumber())
			s = "N";
		else if(i < mNNTopology->GetNonInputNumber() - mNNTopology->GetOutputNumber())
			s = "O";
		else if(i < mNNTopology->GetNonInputNumber())
			s = "AC";
		else if(i == mNNTopology->GetNonInputNumber())
			s = "B";
		else
			s = "S";
		
		dc.DrawText(s, position.x + 10, position.y + 5);
		
		((SpikingBase*)mNeurons[i])->PaintActivity(dc, position, cyclesVisible, mCurrentCycle, UpdatesPerCycle);
	}

}
