/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _SPIKINGNN_H
#define _SPIKINGNN_H

#include "spikingbase.h"
#include "../neuralnetwork.h"


class NeuralNetworkTopology;

/** 
 * Class to construct artificial neural networks composed of spiking neurons and spiking synapses, inherited from NeuralNetwork.
 * @see SpikingNeuron
 * @see SpikingSynapse
 */
class SpikingNN : public NeuralNetwork
{
public:
	/**
	 * SpikingNN constructor
	 * @param inputSize the number of input neurons
	 * @param internalSize the number of internal neurons
	 * @param outputSize the number of output neurons
	 */
	//SpikingNN(wxUint8 inputSize, wxUint8 internalSize, wxUint8 outputSize);
	SpikingNN(NeuralNetworkTopology* topo);

	/** SpikingNN destructor */
	~SpikingNN();  

	/* ---------------------- USING THE NN --------------------------- */

	/** 
	 * Allocate storage for network activity (required before using network!).
	 * @param numCycles Number of sensorimotor cycles to allocate storage for.
	 */
	virtual void AllocActivityStorage(wxUint16 numCycles);

	/**
	 * Feeds input values into the neural network.
	 * @param input is an array of inputs 
	 */
	virtual void SetInput(float *input);


	/** Method to compute the output values of the neural network */
	virtual void Compute();
	
	virtual void RandomizeWeights();
	
	virtual NeuralNetworkTopology* GetNNTopology() {return mNNTopology;}
	
  
	/* ----------------------- ACCESSORS --------------------------- */

	/** Returns the number of input neurons. */
	wxUint8 GetNumInput() { return mNumInput ; }

	

	static void SetUpdatesPerCycle(wxUint16 updatesPerCycle) { SpikingNN::UpdatesPerCycle = updatesPerCycle; }
	static wxUint16 GetUpdatesPerCycle() { return UpdatesPerCycle; }
	static void SetOutputWindowLen(wxUint16 outputWindowLen) { SpikingNN::OutputWindowLen = outputWindowLen; }
	static wxUint16 GetOutputWindowLen() { return OutputWindowLen; }
	static const wxUint16 GetSpikeWindowLen() { return SpikeWindowLen; }
	static void SetBiasActivity(float biasActivity){ SpikingNN::BiasActivity = biasActivity ;}
	

	/* ------------------------ LOGGING ------------------------------ */

protected:
	/** Does the actual saving of the neural activity. */
	virtual void SaveActivity(wxFile *file);

public:
	/* ----------------------- PAINTING -------------------------------- */

	/** Return the size needed to paint the neuralNetwork architecture */
	virtual wxSize GetPaintArchitectureSize();

	/** Return the size needed to paint the neuralNetwork activity */
	virtual wxSize GetPaintActivitySize();

	/** 
	 * Paint the neural network data, which is neural network dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintArchitecture(wxDC& dc, wxPoint& position, wxSize& size);

	/** 
	 * Paint the neural network activity, which is neural network dependent.
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintActivity(wxDC& dc, wxPoint& position, wxSize& size);

protected:

	// Parameters
	static wxUint16 UpdatesPerCycle;	/**< Number of network updates to be done each sensorimotor cycle. */
	static wxUint16 OutputWindowLen;	/**< Length of output window used to measure fire rate (# of updates). */

	// Fixed parameter
	const static wxUint16 SpikeWindowLen;	/**< Length of spike window used to determine synaptic contributions
													to the membrane potential. */
	static float BiasActivity;	/**< Activity value for bias receptors. */
	
	// Member variables
	wxUint8 mNumInput;					/**< Number of receptors. */
	wxUint8 mNumInternal;			/**<num of internal neurons, i.e. dynamics neurons "new style">**/
	wxUint8 mNumAC;


	NeuralNetworkTopology* mNNTopology;
};

#endif /* _SPIKINGNN_H */
