/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>
#include <wx/dynarray.h>
#include <wx/file.h>
#include <wx/filename.h>


#include "../../status.h"
#include "../../modelsettings.h"
#include "../../ga/genome.h"
#include "../../exp/experiment.h"
#include "../../utils/xml/xml.h"
#include "../../utils/iutils.h"

#include "spikingnn.h"
#include "spikingreceptornewnorefrac.h"
#include "spikingreceptornewrefrac.h"
#include "spikingoutputacrate.h"
#include "spikingneuron.h"
#include "spikingsynapse.h"
#include "spikingoutputsynapse.h"
#include "spikingnntopology.h"
#include "../../ga/genomelayout.h"

#include "spikingnnfactory.h"


SpikingNNFactory::SpikingNNFactory(ModelSettings *modelSettings) : NeuralNetworkFactory(modelSettings)
{
	// Default number of internal neurons, additional to neurons used for output
	mInternalSize = 0; 


	mNNTopology = NULL;

	// The NN is fully connected with recurrent connections by default (no hidden layer)
	mTopologyType = SPTOPO_FULLY_CONNECTED;
	mReceptorType = NTYPE_SPIKING_RATE_RECEPTOR;
	mOutputACType = NTYPE_SPIKING_RATE_AC;
	
	mParameterChanged = false;
	
	// Create genome layout.
	mGenomeLayout = new GenomeLayout(GetNNTopology()->GetEvolvableNeuronNumber(), GetNNTopology()->GetEvolvableSynapseNumber());
	
	
	// Setup parameter set.
		// Neuron params
	mGenomeLayout->AddDigitalParam(PARAM_NEURON, PARAMID_SPIKE_SIGN, "Sign", 1, 1);
	mGenomeLayout->AddAnalogParam(PARAM_NEURON, PARAMID_SPIKE_THRESHOLD, "Threshold", 0, 0.5f, 0.0f, 2.0f);
	mGenomeLayout->AddAnalogParam(PARAM_NEURON, PARAMID_SPIKE_MEMBRANE_TIME_CONST, "MembraneTimeC", 0, 4.0f, 1.0f, 10.f);
		// Synapse params.
	mGenomeLayout->AddDigitalParam(PARAM_SYNAPSE, PARAMID_EXISTENCE, "Existence", 1, 1);
	mGenomeLayout->AddAnalogParam(PARAM_SYNAPSE, PARAMID_STRENGTH, "Strength", 0, 1.0f, 0.0f, 1.0f);
	mGenomeLayout->AddDigitalParam(PARAM_SYNAPSE, PARAMID_SPIKE_DELAY, "Delay", 0, 2);
	mGenomeLayout->AddAnalogParam(PARAM_SYNAPSE, PARAMID_SPIKE_SYNAPSE_TIME_cONST, "TimeC", 0, 10.0f, 1.0f, 20.0f);	
}

SpikingNNFactory::~SpikingNNFactory() {

	delete mGenomeLayout;
	if(mNNTopology) delete mNNTopology;

}
	
void SpikingNNFactory::LogNN(Genome *genome, NeuralNetwork* nn)
{
	// THIS IS QUITE GENERALIZED AND MAY BE MOVABLE TO NN FACT
	// WITH FEW MODIFICATION (TO BE USED BACK IN SIGM NN FACT)




	wxUint32 i,j;

		// 1) HANDLE FILE HIERARCHY.

	//create directory (genome)
	wxString dirName= mModelSettings->GetOutputDirectory() +"/genome";
		if (!wxFileName::DirExists(dirName))
			if (!wxFileName::Mkdir( dirName ) ){
				StatusLine("Unable to create directory : " + dirName );	
				return;
			}
	//create directory (population)
	wxString dirNamePop= dirName + wxString::Format("/population_%d", genome->GetPopulationNum());
		if (!wxFileName::DirExists(dirNamePop))
			if (!wxFileName::Mkdir( dirNamePop ) ){
				StatusLine("Unable to create directory : " + dirNamePop );	
				return;
			}

	
	wxString filename = dirNamePop + "/"+ wxString::Format("genome_%d_%d.log", genome->GetPopulationNum(), genome->GetIndividualNum());
	wxFile *file = new wxFile(filename.c_str(), wxFile::write);
		if(!file->IsOpened()) {
			StatusLine("SNN Genome Log - Couldn't open file \"" + filename + "\" for writing.");
			delete file;
			wxASSERT_MSG(0,"error couldn't open the SNN Genome Log file");
		}

	file->Write("% genome_M_N.log M is the population number, N is the individual number\n\n");
	
	
	
	
		// 2) WRITE LABEL
	
	GenomeLayout* layout = GetGenomeLayout();
	NeuralNetworkTopology* topo = GetNNTopology();
	
	// neuron number
	file->Write("NeuronNumber\t");
	
	// neuron parameters
	for(i = 0; i < layout->GetParameterNumber(PARAM_NEURON); i++)
	{
		if(layout->GetParameterEncodingSize(PARAM_NEURON, i) > 0)
			file->Write(wxString(layout->GetParameterName(PARAM_NEURON, i)) + '\t');
	}
	
	// synapse evolvability
	file->Write("SynapseEvolved#" + wxString::Format("%d\t", topo->GetNeuronNumber()));
	
	// synapse parameters
	for(i = 0; i < layout->GetParameterNumber(PARAM_SYNAPSE); i++)
	{
		if(layout->GetParameterEncodingSize(PARAM_SYNAPSE, i) > 0)
			file->Write(wxString::Format("%s#%d\t", layout->GetParameterName(PARAM_SYNAPSE, i).c_str(), topo->GetNeuronNumber()));
	}
	
	file->Write("\n");
	
	
	
	
		// 3) WRITE PARAMETERS
		
	for(i = 0; i < topo->GetNonInputNumber(); i++)
	{
		wxUint32 index = topo->GetNonInputNeuronIndex(i);
		
		// neuron index
		file->Write(wxString::Format("%d\t", index));
		
		// neuron parameters
		for(j = 0; j < layout->GetParameterNumber(PARAM_NEURON); j++)
		{
			if(layout->GetParameterEncodingSize(PARAM_NEURON, j) == 0)
				continue;
				
			float param;
			if(topo->GetNeuronEvolvability(i) == EVOLVABILITY_EVOLVABLE)
				param = nn->GetNeuron(i)->GetParameter(layout->GetParameterID(PARAM_NEURON, j));
			else
				param = layout->GetParameterDefaultValue(PARAM_NEURON, j);
				
			if(layout->GetParameterAnalog(PARAM_NEURON, j))
				file->Write(wxString::Format("%f\t", param));
			else
				file->Write(wxString::Format("%d\t", (int) param));
		}
		
		// synapse evolvability
		for(j = 0; j < topo->GetNeuronNumber(); j++)
		{
			if(topo->GetSynapseEvolvability(j, i) == EVOLVABILITY_EVOLVABLE)
				file->Write("1\t");
			else
				file->Write("0\t");
		}
		
		// synapse parameter
		for(j = 0; j < layout->GetParameterNumber(PARAM_SYNAPSE); j++)
		{
			if(layout->GetParameterEncodingSize(PARAM_SYNAPSE, j) == 0)
				continue;
				
			for(wxUint32 index = 0; index < topo->GetNeuronNumber(); index++)
			{
				float param = 0.0f;
				
				if(topo->GetSynapseEvolvability(index, i) == EVOLVABILITY_EVOLVABLE)
				{
					// we must get the synapse
					Neuron* from = nn->GetNeuron(index), *to = nn->GetNeuron(i);
					Synapse* synapse = NULL;
					
					for(wxUint32 synIndex = 0; synIndex < from->GetOutgoing()->GetCount(); synIndex++)
					{
						if(from->GetOutgoing()->Item(synIndex)->GetPostNeuron() == to)
							synapse = from->GetOutgoing()->Item(synIndex);
					}
					
					if(synapse != NULL)
						param = synapse->GetParameter(layout->GetParameterID(PARAM_SYNAPSE, j));
				}
				else if(topo->GetSynapseEvolvability(index, i) == EVOLVABILITY_DEFAULT)
				{
					param = layout->GetParameterDefaultValue(PARAM_SYNAPSE, j);
				}
				
				if(layout->GetParameterAnalog(PARAM_SYNAPSE, j))
					file->Write(wxString::Format("%f\t", param));
				else
					file->Write(wxString::Format("%d\t", (int) param));
			}
		}
		
		file->Write('\n');
	}
	


	file->Close();
	delete file;
}



NeuralNetworkTopology* SpikingNNFactory::GetNNTopology()
{
	// Check if parameters were changed.
	if(mParameterChanged && mNNTopology)
	{
		delete mNNTopology;
		mNNTopology = NULL;
	}
	
	// Create nn topology if necessary.
	if(!mNNTopology)
	{
		Experiment* experiment = mModelSettings->GetExperiment();
		mNNTopology = new SpikingNNTopology(
						mInternalSize,
						experiment->GetNumOutputNeurons(),
						experiment->GetNumSensorNeurons()+1,			// with bias neuron
						mReceptorType, mOutputACType, mTopologyType);
	}
	
	mParameterChanged = false;
	
	return mNNTopology;
}

GenomeLayout* SpikingNNFactory::GetGenomeLayout()
{
	return mGenomeLayout;
}

NeuralNetwork* SpikingNNFactory::CreateNeuralNetwork()
{
	return new SpikingNN(GetNNTopology());
}

Neuron* SpikingNNFactory::CreateNeuron(NeuronType type)
{
	switch(type)
	{
		case NTYPE_SPIKING: return new SpikingNeuron;
		case NTYPE_SPIKING_RATE_RECEPTOR: return new SpikingReceptorNewRefrac;
		case NTYPE_SPIKING_RATE_RECEPTOR_NOREFRAC: return new SpikingReceptorNewNoRefrac;
		case NTYPE_SPIKING_RATE_AC: return new SpikingOutputACRate;
		default: return NULL;
	}
}

Synapse* SpikingNNFactory::CreateSynapse(Neuron* preNeuron, Neuron* postNeuron, SynapseType type)
{
	switch(type)
	{
		case STYPE_SPIKING: return new SpikingSynapse(preNeuron, postNeuron);
		case STYPE_SPIKING_OUTPUT: return new SpikingOutputSynapse(preNeuron, postNeuron);
		default: return NULL;
	}
}

NeuralNetwork *SpikingNNFactory::Create(Genome *genome)
{
	
	/////////// NN FACTORY GENERAL ALGORITHM
	
	/////////// This part of the code should be included in NeuralNetworkFactory
	/////////// and SigmoidNN should be simplified accordingly. It will be done
	/////////// later.
	
	
	// Get helper objects.
	NeuralNetworkTopology* topo = GetNNTopology();
	GenomeLayout* genLay = GetGenomeLayout();
	
	
		// Neural network.
	
	// Create network.
	NeuralNetwork* nn = CreateNeuralNetwork();
	
	// Decode network parameters.
	genLay->DecodeNetworkParameters(genome, nn);
	
	
	
	
		// Neurons.
	
	// Create neurons.
	for(wxUint32 n = 0; n < topo->GetNeuronNumber(); n++)
	{
		Neuron* neuron = CreateNeuron(topo->GetNeuronType(n));
		nn->SetNeuron(n, neuron);
	}
	
	// Decode neuron parameters for non-class neurons.
	wxUint32 genIndex = 0;
	wxInt32 index = -1;
	while(topo->GetNextEvolvableNeuronOfClass(index, 0))
		genLay->DecodeNeuronParameters(genome, genIndex++, nn->GetNeuron(index));
	
	// Decode neuron parameters for class neurons.
	for(wxUint16 classIdx = 0; classIdx < topo->GetNeuronClassNumber(); classIdx++)
	{
		wxUint32 neuronClass = topo->GetNthNeuronClass(classIdx);
		std::vector<ParamRecord> params;
		
		genLay->DecodeNeuronParameters(genome, genIndex++, params);
		index = -1;
		while(topo->GetNextEvolvableNeuronOfClass(index, neuronClass))
			nn->GetNeuron(index)->SetParameters(params);
	}
	
	
	// NOTE: here we should apply default parameter set to EVOLVED_DEFAULT neuron
	
	
	
		// Synapses.
		//
		// Note: when the existence of synapses is evolved, it is difficult with
		// this design to selectivly create (or not) synapse depending on evolved
		// setting. So the synapse is always created, and the existence parameter
		// is set in the same way than otherparameters (see IParameterizable).
		// If a synapse is told not to exist, GetStrength() will return 0.of whatever
		// its real strength is. Beside, SpikingSynpase take also into account the
		// existence to optimize its computation.
	
	// Create synapses.
	Synapse** synArray = new Synapse*[topo->GetSynapseNumber()];
	for(wxUint32 s = 0; s < topo->GetSynapseNumber(); s++)
	{
		if(topo->GetSynapseEvolvability(s) != EVOLVABILITY_UNCONNECTED)
		{
			wxUint32 neuronFrom, neuronTo;
		
			topo->SynapseIndexToNeuronIndice(s, neuronFrom, neuronTo);
			synArray[s] = CreateSynapse(nn->GetNeuron(neuronFrom), nn->GetNeuron(neuronTo), topo->GetSynapseType(s));
		}
		else
			synArray[s] =  NULL;
	}
	
	// Decode synapse parameters for non-class synapses.
	genIndex = 0;
	index = -1;
	while(topo->GetNextEvolvableSynapseOfClass(index, 0))
		genLay->DecodeSynapseParameters(genome, genIndex++, synArray[index]);
	
	// Decode synapse parameters for class synapses.
	for(wxUint16 classIdx = 0; classIdx < topo->GetSynapseClassNumber(); classIdx++)
	{
		wxUint32 synapseClass = topo->GetNthSynapseClass(classIdx);
		std::vector<ParamRecord> params;
		
		genLay->DecodeSynapseParameters(genome, genIndex++, params);
		index = -1;
		while(topo->GetNextEvolvableSynapseOfClass(index, synapseClass))
			synArray[index]->SetParameters(params);
	}
	
	
	// NOTE: here we should apply default parameter set to EVOLVED_DEFAULT synapse
	
	
	
	
	LogNN(genome, nn);
	
	
	// Cleanup.
	delete[] synArray;
	
	
	
	return nn;
}

wxUint16 SpikingNNFactory::GetGenomeLength()
{
	// Update helper objects.
	NeuralNetworkTopology* topo = GetNNTopology();
	mGenomeLayout->SetEvolvableNeuronNumber(topo->GetEvolvableNeuronNumber());
	mGenomeLayout->SetEvolvableSynapseNumber(topo->GetEvolvableSynapseNumber());
	


	// THIS IS GOING TO NNFACTORY
	GenomeLayout* layout = GetGenomeLayout();
	if(layout)
		return layout->GetGenomeLength();
	else
		return 0;
}

wxUint16 SpikingNNFactory::ConfigGenomeEncoding()
{
	// This is an attempt to do a generalized ConfigGenomeEncoding
	// THIS GOES IN NN FACT

	
	wxString scopeStr[] = {"nn", "neuron", "synapse"};

	// For nn, neuron and synapse.
	for(ParamScope scope = PARAM_NEURALNETWORK; scope <= PARAM_SYNAPSE; scope = (ParamScope) (scope+1))
	{
		// For each params.
		for(wxUint32 index = 0; index < mGenomeLayout->GetParameterNumber(scope); index++)
		{
			// encoding size first.
			long num = wxGetNumberFromUser(
						"Enter the number of bits that should be used to encode " + mGenomeLayout->GetParameterName(scope, index) + "(" + scopeStr[scope] + ").",
						"encoding size", "Configure genome encoding", mGenomeLayout->GetParameterEncodingSize(scope, index), 0, 16);
			if(num != -1)
				mGenomeLayout->SetParameterEncodingSize(scope, index, num);
			
			if(mGenomeLayout->GetParameterEncodingSize(scope, index) > 0)
			{
				// Ask range.
				if(mGenomeLayout->GetParameterAnalog(scope, index))
				{
					float value = mGenomeLayout->GetAnalogParameterLowBound(scope, index);
					if(iGetFloatFromUser("Enter the low bound of the parameter "  + mGenomeLayout->GetParameterName(scope, index) + "(" + scopeStr[scope] + ").", "Configure genome encoding", value))
						mGenomeLayout->SetAnalogParameterLowBound(scope, index, value);
					
					value = mGenomeLayout->GetAnalogParameterHighBound(scope, index);
					if(iGetFloatFromUser("Enter the high bound of the parameter "  + mGenomeLayout->GetParameterName(scope, index) + "(" + scopeStr[scope] + ").", "Configure genome encoding", value))
						mGenomeLayout->SetAnalogParameterHighBound(scope, index, value);
				}
				else
					; // we dont ask lowerBound -- always 0 -- change if needed
			}
			else
			{
				// Ask default value.
				if(mGenomeLayout->GetParameterAnalog(scope, index))
				{
					float value = mGenomeLayout->GetParameterDefaultValue(scope, index);
					if(iGetFloatFromUser("Enter the value to use for the parameter "  + mGenomeLayout->GetParameterName(scope, index) + "(" + scopeStr[scope] + ").", "Configure genome encoding", value))
						mGenomeLayout->SetParameterDefaultValue(scope, index, value);
				}
				else
				{
					long num = (long) mGenomeLayout->GetParameterDefaultValue(scope, index);
					if((num = wxGetNumberFromUser("Enter the value to use for the parameter "  + mGenomeLayout->GetParameterName(scope, index) + "(" + scopeStr[scope] + ").", "default value", "Configure genome encoding", num, -10000, 10000)) != -1)
						mGenomeLayout->SetParameterDefaultValue(scope, index, num);
				}
			}
		}
	}	
	
	
	
	// THIS IS THE SP NN FACT OVERRIDE PART
	
	mParameterChanged = true;

	return GetGenomeLength();
}

void SpikingNNFactory::ConfigNeuralNetwork()
{
	long number;
	wxString result;

	// receptor type


		wxString choices[] = { "RATE (with refractory)", "RATE ( without refractory)" };
		number = wxGetSingleChoiceIndex("Please select a spiking receptor.", "Configure neural network", 2, choices);

	if(number != -1)
		mReceptorType = (NeuronType) (NTYPE_SPIKING_RATE_RECEPTOR + number);
	

	// network update
	if((number = wxGetNumberFromUser("Enter the number of network updates that the network should do in one time step.", "# updates", "Configure neural network", SpikingNN::GetUpdatesPerCycle(), 1, 5000)) > 0)
    	SpikingNN::SetUpdatesPerCycle((wxUint16)number);
    	
    // network update to measure firing rate
	if((number = wxGetNumberFromUser("Enter the number of network updates that should be used to measure fire rates (used as output values).", "# updates", "Configure neural network", SpikingNN::GetOutputWindowLen() < SpikingNN::GetUpdatesPerCycle() ? SpikingNN::GetOutputWindowLen() : SpikingNN::GetUpdatesPerCycle(), 1, 2500)) > 0)
		SpikingNN::SetOutputWindowLen((wxUint16)number);
		
	// number of internal neuron
	if((number = wxGetNumberFromUser("Enter the number of neurons that should be added additional to the neurons used for output.", "# neurons", "Configure neural network", mInternalSize, 0, 50)) >= 0)
    	mInternalSize = (wxUint16)number;

	// network topology
	wxString choices3[] = {"Fully connected", "Fully connected (no recurent output)", "Feed forward (recurent ouput)", "Feed forward (no recurent)","No direct connection between sensors and outputs (NO Recurrents outputs)", "No direct connection between sensors and outputs  (Recurrents outputs)"};
	number = wxGetSingleChoiceIndex("Choose a network topology", "Configure neural network", 6, choices3);
	if(number != -1)
		mTopologyType = (SpikingNNTopologyType) (SPTOPO_FULLY_CONNECTED + number);
	

	
	
	
	
	
	mParameterChanged = true;
}


/* --- XML --- */
void SpikingNNFactory::Load (XMLElement *myself, XMLLoader *loader)
{
	// spiking nn factory attribute.
	mInternalSize = myself->GetAttribute("internal_size")->GetByteValue();	
	mReceptorType = (NeuronType)myself->GetAttribute("receptor_type")->GetIntValue();
	mOutputACType = (NeuronType)myself->GetAttribute("output_ac_type")->GetIntValue();
	SpikingNN::SetUpdatesPerCycle((wxUint16)myself->GetAttribute("update_cycle")->GetIntValue());
	SpikingNN::SetOutputWindowLen((wxUint16)myself->GetAttribute("window_length")->GetIntValue());
	mTopologyType = (SpikingNNTopologyType) myself->GetAttribute("topology_type")->GetIntValue();



	mParameterChanged = true;



	// generalized genome layout saving. TO BE MOVED IN NN FACT.
	for(ParamScope scope = PARAM_NEURALNETWORK; scope <= PARAM_SYNAPSE; scope = (ParamScope) (scope+1))
	{
		for(wxUint32 index = 0; index < mGenomeLayout->GetParameterNumber(scope); index++)
		{
			wxString prefix = mGenomeLayout->GetParameterName(scope, index) + "_";
		
			mGenomeLayout->SetParameterEncodingSize(scope, index, myself->GetAttribute(prefix+"bits")->GetByteValue());
			mGenomeLayout->SetParameterDefaultValue(scope, index, myself->GetAttribute(prefix+"default_value")->GetFloatValue());
			if(mGenomeLayout->GetParameterAnalog(scope, index))
			{
				mGenomeLayout->SetAnalogParameterLowBound(scope, index, myself->GetAttribute(prefix+"min")->GetFloatValue());
				mGenomeLayout->SetAnalogParameterHighBound(scope, index, myself->GetAttribute(prefix+"max")->GetFloatValue());
			}
			else
				mGenomeLayout->SetDigitalParameterLowBound(scope, index, myself->GetAttribute(prefix+"min")->GetIntValue());
		}
	}
}

XMLElement* SpikingNNFactory::SaveSelf () {
	XMLElement *result = new XMLElement(GetShortXMLName().c_str());
	result->AddAttribute(new XMLAttribute("internal_size",(wxInt8)mInternalSize));
	result->AddAttribute(new XMLAttribute("receptor_type",(wxInt32)mReceptorType));
	result->AddAttribute(new XMLAttribute("output_ac_type",(wxInt32)mOutputACType));
	result->AddAttribute(new XMLAttribute("update_cycle",SpikingNN::GetUpdatesPerCycle()));
	result->AddAttribute(new XMLAttribute("window_length",SpikingNN::GetOutputWindowLen()));
	result->AddAttribute(new XMLAttribute("topology_type", (wxInt32) mTopologyType));

	
	
	// GENERALIZED ==> NN TOPO
	for(ParamScope scope = PARAM_NEURALNETWORK; scope <= PARAM_SYNAPSE; scope = (ParamScope)(scope+1))
	{
		for(wxUint32 index = 0; index < mGenomeLayout->GetParameterNumber(scope); index++)
		{
			wxString prefix = mGenomeLayout->GetParameterName(scope, index) + "_";
			
			result->AddAttribute(new XMLAttribute(prefix+"bits", mGenomeLayout->GetParameterEncodingSize(scope, index)));
			result->AddAttribute(new XMLAttribute(prefix+"default_value", mGenomeLayout->GetParameterDefaultValue(scope, index)));
			if(mGenomeLayout->GetParameterAnalog(scope, index))
			{
				result->AddAttribute(new XMLAttribute(prefix+"min", mGenomeLayout->GetAnalogParameterLowBound(scope, index)));
				result->AddAttribute(new XMLAttribute(prefix+"max", mGenomeLayout->GetAnalogParameterHighBound(scope, index)));
			}
			else
				result->AddAttribute(new XMLAttribute(prefix+"min", mGenomeLayout->GetDigitalParameterLowBound(scope, index)));
		}
	}
	
	return result;
}

wxString SpikingNNFactory::GetXMLName () { return GetShortXMLName(); }
wxString SpikingNNFactory::GetShortXMLName() { return SpikingNNFactory::GetStaticXMLType(); }
