/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

 
 
/*

This class concentrate the somewhat criptic code that plays with neuron
type (internal, AC, etc), indexes, etc. It is not necessarily easy to read
and understand, but it is quite concentrated, so it minimize the risk of
error. Some schema are provided here to better understand the code. To had
new layout of neural network, refer the the /docs.

This class defines the layout of neuron index. Remember that NNTopology defined
the following layout:

0                                   n-1
+----------------------+-------------+
|   neurons            | sensors     |
+----------------------+-------------+


For spiking nn, we refine the layout in this way:

0                                   n-1
+-------------+---+----+-------------+
|     i       | o | ac |      s      |
+-------------+---+----+-------------+

	i: internal (mNumInternal)
	o: output (mNumOutput)
	ac: analog-converter (mNumAC)
	s: sensor (mNumSensor)

So we have:

	mNumNeuron = mNumInternal + mNumOutput + mNumAC


This give us the following connectivity matrix


 				pre-neuron
 			
        i    o    ac         s
    +------+----+----+----------------+
    |      |    |    |                |
 i  |  X   | X  | U  |       X        |
    |      |    |    |                |
    +------+----+----+----------------+
    |      |    |    |                |    post-neuron
 o  |  X   | X  | U  |       X        |
    +------+----+----+----------------+
    |      |    |    |                |
ac  |  U   | O  | U  |       U        |
    +------+----+----+----------------+


U: always unconnected
O: connectivity depends on output coding (typically non evolved)
X: user configurable (layout dependent)
E: evolved


Now, the following layout are implemented (we use simplified schema
where AC are not represented.

	SPTOPO_FULLY_CONNECTED
	
        i       o    b      s
   +---------+-----+---------------+
   |         |     | |             |
   |         |     | |             |
i  |    E    |  E  |E|    E        |
   |         |     | |            |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  E  |E|    E        |
   |         |     | |             |
   +---------+-----+---------------+
   
   
	SPTOPO_FULLY_CONNECTED_NO_RECURENT_OUTPUT
	
        i       o   b     s
   +---------+-----+---------------+
   |         |     | |             |
   |         |     | |             |
i  |    E    |  E  |E|    E        |
   |         |     | |             |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  U  |?|     E       |
   |         |     | |             |
   +---------+-----+---------------+
   
   
	SPTOPO_FEEDFORWARD_RECURENT
	
        i       o   b     s
   +---------+-----+---------------+
   |         |     | |             |
   |         |     | |             |
i  |    U    |  U  |E|    E        |
   |         |     | |             |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  E  |?|    U        |
   |         |     | |             |
   +---------+-----+---------------+
   
   
	SPTOPO_FEEDFORWARD
	
        i       o         s
   +---------+-----+---------------+
   |         |     | |            |
   |         |     | |             |
i  |    U    |  U  |E|    E        |
   |         |     | |             |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  U  |?|    U        |
   |         |     | |             |
   +---------+-----+---------------+
   
	 SPTOPO_NO_DIRECT_CONNECT_NO_RECURENT

	       i       o  b      s
   +---------+-----+---------------+
   |         |     | |             |
   |         |     | |             |
i  |    E    |  U  |E|    E        |
   |         |     | |             |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  U  |E|    U        |
   |         |     | |             |
   +---------+-----+---------------+

	SPTOPO_NO_DIRECT_CONNECT

		   i       o    b    s
   +---------+-----+---------------+
   |         |     | |             |
   |         |     | |             |
i  |    E    |  U  |E|    E        |
   |         |     | |             |
   |         |     | |             |
   +---------+-----+---------------+
   |         |     | |             |
o  |    E    |  E  |E|    U        |
   |         |     | |             |
   +---------+-----+---------------+

    
*/


#include <wx/wx.h>

#include "spikingnntopology.h"


SpikingNNTopology::SpikingNNTopology(wxUint32 numInternal, wxUint32 numOutput, wxUint32 numSensor,
	NeuronType inputCoding, NeuronType outputCoding, SpikingNNTopologyType topoType)
	: NeuralNetworkTopology(numInternal+numOutput+numOutput, numSensor),
	mNumInternal(numInternal), mNumOutput(numOutput), mNumAC(mNumOutput)
																	// here we count 2 times numOutput, to take
																	// into account the output neuron and the
																	// coresponding AC. This is fine for rate
																	// and delay decoding, but not for other,
																	// many-neuron-per-output decoding scheme!
{	
	mInputCoding = inputCoding;
	mOutputCoding = outputCoding;
	mTopologyType = topoType;
	
	SetupNeurons();
	SetupSynapses();
	
}

void SpikingNNTopology::SetupNeurons()
{
	wxUint32 i = 0;
	
	for(; i < mNumInternal+mNumOutput; i++)
		SetNeuronInfo(i, NTYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
	
	for(; i < mNumInternal+mNumOutput+mNumAC; i++)
		SetNeuronInfo(i, mOutputCoding, EVOLVABILITY_DEFAULT, 0);
	
	for(; i < mNumNeuron+mNumSensor; i++)
		SetNeuronInfo(i, mInputCoding, EVOLVABILITY_DEFAULT, 0);
}

void SpikingNNTopology::SetupSynapses()
{
		// THIS IS THE ALL IMPORTANT PLACE TO
		// GO IF YOU WANT TO CUSTOMIZE YOUR NN
		// LAYOUT.
		//
		// TO BETTER UNDERSTAND THIS CODE, REFER
		// TO THE COMMENT AT THE TOP OF THIS
		// FILE.


	// Synapses that do not depends on network topology type. These
	// are represented by U on the connectivity schema.
	
	SetSynapsesInfo(
		0, mNumInternal,						// from: internal neuron
		mNumInternal+mNumOutput, mNumAC,		// to: ac
		STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
	
	SetSynapsesInfo(
		mNumNeuron, mNumSensor,				// from: sensors
		mNumInternal+mNumOutput, mNumAC,		// to: ac
		STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
		
	SetSynapsesInfo(
		mNumInternal+mNumOutput, mNumAC+mNumSensor,	// from: ac
		0, mNumNeuron,								// to: all
		STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
	
	
	// Output to ac connection. These are represented by O on the schema.
	// Both in the case of rate and delay coding, there is a one to one
	// connectivity here. This would be different in the case of population
	// coding or similar, many-neuron-per-output coding.
	SetSynapsesInfo(
		mNumInternal, mNumOutput, 			// from: output
		mNumInternal+mNumOutput, mNumAC,	// to: ac
		STYPE_SPIKING_OUTPUT, EVOLVABILITY_UNCONNECTED, 0);
	
	for(wxUint32 i = 0; i < mNumOutput; i++)
	{
		SetSynapseInfo(
			mNumInternal+i, 			// from: output i
			mNumInternal+mNumOutput+i,	// to: ac i
			STYPE_SPIKING_OUTPUT, EVOLVABILITY_DEFAULT, 0);
	}
	
	
	// The remaining synapses depends on the particular layout scheme, thus
	// the switch statement.
	
	/// some help
	///
	///SetSynapsesInfo
	///		(wxUint32 neuronFromLow, wxUint32 numNeuronFrom, wxUint32 neuronToLow, wxUint32 numNeuronTo,
	///		SynapseType type, EvolvabilityType prop, wxUint16 sClass)
	///
	///		THEREFORE
	///		FROM neuronFromLow >>> (neuronFromLow+ numNeuronFrom) TO neuronToLow >>> (neuronToLow+numNeuronTo)

	
	switch(mTopologyType)
	{
		case SPTOPO_FULLY_CONNECTED_NO_RECURENT_OUTPUT:
		case SPTOPO_FULLY_CONNECTED:
			SetSynapsesInfo(
				0, mNumInternal+mNumOutput, 0, mNumInternal+mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
			SetSynapsesInfo(mNumNeuron, mNumSensor, 0, mNumInternal+mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
			
			if(mTopologyType == SPTOPO_FULLY_CONNECTED_NO_RECURENT_OUTPUT)
				SetSynapsesInfo(mNumInternal, mNumOutput, mNumInternal, mNumOutput,
					STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
			break;
		
		case SPTOPO_FEEDFORWARD_RECURENT:
		case SPTOPO_FEEDFORWARD:
			// From sensor to internal => evolvable.
			SetSynapsesInfo(
				mNumNeuron, mNumSensor, 0, mNumInternal,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
			
			// From sensor to output => unconnected.
			SetSynapsesInfo(
				mNumNeuron, mNumSensor, mNumInternal, mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
			
			SetSynapsesInfo(
				mNumNeuron, 1 , mNumInternal, mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);

			// From internal and output to internal => unconnected.
			SetSynapsesInfo(
				0, mNumInternal+mNumOutput, 0, mNumInternal,
				STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);
			
			// From internal to output => evolvable.
			SetSynapsesInfo(
				0, mNumInternal, mNumInternal, mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
			
			// From output to output => evolvable if recurent.
			SetSynapsesInfo(
				mNumInternal, mNumOutput, mNumInternal, mNumOutput,
				STYPE_SPIKING, mTopologyType == SPTOPO_FEEDFORWARD ?
				EVOLVABILITY_UNCONNECTED : EVOLVABILITY_EVOLVABLE, 0);
			break;
		case SPTOPO_NO_DIRECT_CONNECT:
		case SPTOPO_NO_DIRECT_CONNECT_NO_RECURENT:
			SetSynapsesInfo(
				0, mNumInternal, 0, mNumInternal+mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);

			SetSynapsesInfo(mNumNeuron, mNumSensor, 0, mNumInternal,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);

			SetSynapsesInfo (mNumNeuron, mNumSensor,mNumInternal,mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_UNCONNECTED, 0);


		
			if(1) // bias connected
			SetSynapsesInfo(
				mNumNeuron, 1 , mNumInternal, mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);

			if(mTopologyType == SPTOPO_NO_DIRECT_CONNECT){
				SetSynapsesInfo(mNumInternal, mNumOutput , mNumInternal, mNumOutput,
				STYPE_SPIKING, EVOLVABILITY_EVOLVABLE, 0);
			}
	};
}
