/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "spikingnn.h"
#include "spikingbase.h"

#include "spikingsynapse.h"

SpikingSynapse::SpikingSynapse(Neuron *preNeuron, Neuron *postNeuron) : Synapse(preNeuron, postNeuron, 1.0f) {
	// Init parameters to defaults
	mDelay = 2;
	mTimeC = 10.0f;
	
	
	// Init variables
	mMustBuildKernel = true;
	mKernel = NULL;
}

SpikingSynapse::~SpikingSynapse() {
	if(mKernel) delete [] mKernel;
}


void SpikingSynapse::SetParameterSelf(ParamID id, float param)
{
	switch(id)
	{	
		case PARAMID_SPIKE_DELAY:
			SetDelay((wxUint8) param);
			break;
		
		case PARAMID_SPIKE_SYNAPSE_TIME_cONST:
			SetTimeC(param);
			break;
		
		default:
			Synapse::SetParameterSelf(id, param);
			break;
	};
}

void SpikingSynapse::InitKernel() {
	if(!mMustBuildKernel)
		return;
	
	wxUint16 i, spikeWinLen = SpikingNN::GetSpikeWindowLen();
	
	if(mKernel)
		delete[] mKernel;
	mKernel = new float[spikeWinLen];

	for(i=0; i<mDelay; i++)
		mKernel[i] = 0.0f;
	for(i=mDelay; i<spikeWinLen; i++)
		mKernel[i] = exp((- i + mDelay) / ((SpikingBase *)mPostNeuron)->GetMembraneTimeC()) * (1 - exp((- i + mDelay) / mTimeC));
}

float SpikingSynapse::ComputePotential(wxUint16 spikeWindowIdx) {
	// optimization: we return zero if we do not exist
	if(!mExists)
		return 0.0f;
		
	InitKernel();

	float potential = 0.0f;
	wxUint16 spikeWinLen = SpikingNN::GetSpikeWindowLen();
	SpikingBase *neuron = (SpikingBase *)mPreNeuron;
	bool *spikeWindow = neuron->GetSpikeWindow();

	for(wxUint16 i=0; i<spikeWinLen; i++)
		if(spikeWindow[i]) // check for spikes in the spike window
			potential += mKernel[(spikeWinLen + spikeWindowIdx - i) % spikeWinLen];
	return potential * GetStrength() * neuron->GetSign();
}

void SpikingSynapse::PaintArchitecture(wxDC &dc, wxPoint& position, wxSize& size)
{
	dc.SetBrush(*wxTRANSPARENT_BRUSH);
	dc.SetPen(*wxBLACK_PEN);
	dc.DrawRectangle(position.x, position.y, size.GetWidth(), size.GetHeight());
	
	// draw strength
	dc.SetBrush(wxBrush(wxColor(255, (1.0f - GetStrength()) * 255, (1.0f - GetStrength()) * 255), wxSOLID));
	dc.SetPen(*wxTRANSPARENT_PEN);
	dc.DrawRectangle(position.x + 2, position.y + 2, size.GetWidth() - 4, size.GetHeight() - 4);
}
