/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../utils/randomtools.h"

#include "../status.h"
#include "../utils/xml/xml.h"
#include "khepera/basickhep.h"
#include "khepera/evopickhep.h"

#include "khepera.h"

const wxString Khepera::ItemDesc = "Khepera";

Khepera::Khepera() {
	mImpl = new BasicKhep(&mSensorMask);

	mDistanceData = NULL;
	mFloorData = NULL;
	mAmbientData = NULL; 
	mVisionData = NULL;
	mAnalogData = NULL;
	
	DEL_SENSOR(SENSOR_VISION)
	DEL_SENSOR(SENSOR_FLOOR)
	DEL_SENSOR(SENSOR_ANALOG)
	DEL_SENSOR(SENSOR_DELTA_POSITION)

	mDPLeft=mDPRight=0;	// initialize delta positions
}

Khepera::~Khepera() {
	if(mDistanceData) delete [] mDistanceData;
	if(mFloorData) delete [] mFloorData;
	if(mAmbientData) delete [] mAmbientData;
	if(mVisionData) delete [] mVisionData;
	if(mAnalogData) delete [] mAnalogData;
	delete mImpl;
}

void Khepera::Initialize() {
	wxUint8 num = GetNumDistance();
	mDistanceData = num ? new wxUint8[num] : NULL;
	mAmbientData = num ? new wxUint8[num] : NULL;
	for(int i=0; i<num; i++) {
		mDistanceData[i] = 0;
		mAmbientData[i] = 0;
	}
	mSpeedLeft = 0;
	mSpeedRight = 0;
	if(GetSensorEnabled(SENSOR_VISION)) {
		wxUint8 num = GetNumVision();
		mVisionData = new wxUint8[num];
		for(int i=0; i<num; i++)
			mVisionData[i] = 0;
	}
	if(GetSensorEnabled(SENSOR_FLOOR)) {
		wxUint8 num = GetNumFloorSensors();
		mFloorData = new wxUint8[num];
		for(int i=0; i<num; i++)
			mFloorData[i] = 0;
	}
	if(GetSensorEnabled(SENSOR_ANALOG)) {
		wxUint8 num = GetNumAnalog();
		mAnalogData = new wxUint8[num];
		for(wxUint8 i=0; i<num; i++)
			mAnalogData[i] = 0;	
	}
}

wxInt8 Khepera::Connect() {
	wxInt8 ret;
	if(mCommLink->GetConnected()) return 0;
	mImpl->SetCommLink(mCommLink);

	StatusLine("Connecting to " + mImpl->GetDesc() + "...");	
	
	if((ret = mImpl->Connect())) {
		StatusLineErr("Connection failure.", ret);
		return ret;
	}

	StatusLine("Connection established.");
	SetChanged();
	return 0;   
}

wxInt8 Khepera::Disconnect() {
	if(!mCommLink->GetConnected()) return 0;
	float speeds[3];
	speeds[0] = speeds[1] = speeds[2] = 0.0f;
	SetMotors(speeds);
	return Robot::Disconnect();
}

wxInt8 Khepera::ResetPos() {
	wxByte buffer[2];
	buffer[0]=0xEE;
	buffer[1]=0xEE;
	mCommLink->SendBuffer(buffer,2);
	return 1;
}

wxInt8 Khepera::ReadSensors() {
	wxInt8 ret;
	if(GetSensorEnabled(SENSOR_MOTOR_SPEED) && (ret = ReadEncoders()))
		return ret;	
	if(GetSensorEnabled(SENSOR_DISTANCE) && (ret = ReadDistanceSensors())) 
		return ret;
	if(GetSensorEnabled(SENSOR_FLOOR) && (ret = ReadFloorSensor())) 
		return ret;
	if(GetSensorEnabled(SENSOR_AMBIENT) && (ret = ReadAmbientLightSensors())) 
		return ret;
	if(GetSensorEnabled(SENSOR_VISION) && (ret = ReadVisionSensors())) 
		return ret;
	if(GetSensorEnabled(SENSOR_ANALOG) && (ret = ReadAnalogSensors())) 
		return ret;
	if(GetSensorEnabled(SENSOR_DELTA_POSITION) && (ret = ReadDeltaPosition())) 
		return ret;
	return 0;
}

wxInt8 Khepera::ReadEncoders() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadEncoders(mSpeedLeft, mSpeedRight)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadDeltaPosition() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadDeltaPosition(mDPLeft, mDPRight)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadDistanceSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadDistanceSensors(mDistanceData)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadFloorSensor() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadFloorSensor(mFloorData)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadAmbientLightSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadAmbientLightSensors(mAmbientData)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadVisionSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadVisionSensors(mVisionData)))
		return ret;

	SetChanged();
	return 0;
}

wxInt8 Khepera::ReadAnalogSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;

	if((ret = mImpl->ReadAnalogSensors(mAnalogData)))
		return ret;

	SetChanged();
	return 0;
}

void Khepera::CalcBraitenberg(float *motorSpeeds) {
	wxUint16 i,j;
	const float motscale = 20;
	const float interconn[16] = {4,4,6,-24,-15,-8,6,3, -8,-15,-24,6,4,4,3,6};
 	const float offset[2] = {(motscale/3+RandomTools::RandFloat(4.0)),(motscale/4+RandomTools::RandFloat(4.0))};

	for(j=0; j<2; j++)
		for(i=0, motorSpeeds[j]=0; i<8; i++)
			motorSpeeds[j] += mDistanceData[i]*4*interconn[(j*8)+i];
	for(j=0; j<2; j++)
		motorSpeeds[j] = motorSpeeds[j]/800;
	for(j=0; j<2; j++)
		motorSpeeds[j] += offset[j];

	for(j=0; j<2; j++)	// scale the values in the range ~ -1.0f to 1.0f
		motorSpeeds[j] /= 100;

	if((motorSpeeds[0]+motorSpeeds[1])<0){
		if (fabs(motorSpeeds[0])>fabs(motorSpeeds[1]))
			motorSpeeds[1]=fabs(motorSpeeds[0]);
		else
			motorSpeeds[0]=fabs(motorSpeeds[1]);
	}
}


wxInt8 Khepera::SetMotors(float *speeds) {
	const wxUint8 scale = 100;

	speeds[2] = (speeds[2] + 1.0f) / 2;
	for(int i=0; i<2; i++) {
		if(speeds[i] < -1.0f) speeds[i] = -1.0f;
		if(speeds[i] >  1.0f) speeds[i] =  1.0f;
		speeds[i] *= speeds[2]; // use z-axis to scale speed!
	}
	return SetMotors((wxInt8)(speeds[0] * scale), (wxInt8)(speeds[1] * scale));
}

wxInt8 Khepera::SetMotors(wxInt8 speedL, wxInt8 speedR) {
	if(!mCommLink->GetConnected()) return 1;
	return mImpl->SetMotors(speedL, speedR);
}

wxInt8 Khepera::SetLED(wxUint8 num, bool enable) {
	return mImpl->SetLED(num, enable);
}

wxInt8 Khepera::SetFlag(wxUint8 type, bool enable){
	return mImpl->SetFlag(type,enable);
}

wxInt8 Khepera::ManualControl(int *jsButtons, float *jsAxes) {
	wxInt8 ret;
	float speeds[] = { 0.0f, 0.0f, 0.0f };

	// Those are the joystick axes:
	// Roll:	jsAxes[0]
	// Pitch:	jsAxes[1]
	// Yaw:		jsAxes[2]
	// Power:	jsAxes[3]

	speeds[0] = + jsAxes[2] - jsAxes[1];
	speeds[1] = - jsAxes[2] - jsAxes[1];
	speeds[2] = jsAxes[3];	// speed factor
		
	for (int i=0;i<3;i++) { // limit the motor values between 1 and -1
		if (speeds[i]>1) speeds[i] = 1;
		else if (speeds[i]<-1) speeds[i] = -1;
	}

	speeds[0] = 2*asin(speeds[0])/3.14f;
	speeds[1] = 2*asin(speeds[1])/3.14f;

	if((ret = SetMotors(speeds)))
		StatusLineErr("Couldn't set motor speeds!", ret);

	return mImpl->ManualControl(jsButtons, jsAxes);	// further implementation of button handling in some cases...
}


void Khepera::Configure() {
	wxString result;
	DEL_SENSOR(SENSOR_DELTA_POSITION)

	wxString choices[] = { "Basic Khepera", "kevopic or kevobot"};
	result = wxGetSingleChoice("Please select your Khepera of choice:", "Configure Khepera", 2, choices);
	if(result.IsEmpty() || result == "Basic Khepera") {
		result = "Basic Khepera";
		SetImpl(KHEPERA_BASIC);
	} else if(result == "kevopic or kevobot")
		SetImpl(KHEPERA_EVOPIC);
	StatusLine("Selected " + result + ".");

	if(wxMessageBox("Enable delta positions reading?", "Configure Khepera", wxYES_NO) == wxYES)
		ADD_SENSOR(SENSOR_DELTA_POSITION)

	mImpl->Configure();
}

void Khepera::LogLegenda(Logger *logger) {
	logger->WriteString(wxString("speed_left speed_right\n"));
}

void Khepera::Log(Logger *logger) {
	logger->WriteString(wxString::Format("%d %d\n", mSpeedLeft, mSpeedRight));
}

void Khepera::PaintSensorData(wxDC& dc, wxPoint& position, wxSize& size) {
	wxUint8 num = GetNumDistance();
	wxUint8 width, height, value;

	dc.SetPen(*wxBLACK_PEN);
	
	dc.DrawText(mImpl->GetDesc(), position.x, position.y - 18);

	/* Ambient light */
	if(GetSensorEnabled(SENSOR_AMBIENT)) {
		dc.DrawText("Ambient light", position.x, position.y + 10);
		width = 15;
		height = 30;
		for(int i=0; i<num; i++) {
			value = mAmbientData[i];
			dc.SetBrush(wxBrush(wxColour(value, value, value), wxSOLID));
			dc.DrawRectangle(position.x + (width-1) * i, position.y + 30, width, height);
		}
	}
	
	/* Distance */
	if(GetSensorEnabled(SENSOR_DISTANCE)) {
		dc.DrawText("Distance", position.x, position.y + 70);
		dc.SetBrush(*wxWHITE_BRUSH);
		width = 5;
		for(int i=0; i<num; i++) {
			height = (wxUint8)((0xFF - mDistanceData[i]) / 5);
			dc.DrawRectangle(position.x + 2 * width * i, position.y + 90, width, height);
		}
	}

	/* Floor sensor */
	if(GetSensorEnabled(SENSOR_FLOOR)) {
		num = GetNumFloorSensors();
		dc.DrawText("Floor sensor", position.x + 100, position.y + 70);
		dc.SetBrush(*wxWHITE_BRUSH);
		width = 5;
		for(int i=0; i<num; i++) {
			height = (wxUint8)((0xFF - mFloorData[i]) / 5);
			dc.DrawRectangle(position.x + 100 + 2 * width * i, position.y + 90, width, height);
		}
	}
	
	/* Motor speeds */
	if(GetSensorEnabled(SENSOR_MOTOR_SPEED)) {
		dc.DrawText("Motor speeds", position.x, position.y + 150);
		dc.SetPen(*wxBLACK_PEN);
		dc.SetBrush(*wxWHITE_BRUSH);
		dc.DrawRectangle(position.x, position.y + 170, 5, 50);
		dc.DrawRectangle(position.x + 15, position.y + 170, 5, 50);

		dc.SetBrush(*wxBLUE_BRUSH);
		height = (wxUint16)(fabs(mSpeedLeft / 127.0f) * 25);
		if(mSpeedLeft > 0)
			dc.DrawRectangle(position.x, position.y + 194 - height, 5, height);
		else
			dc.DrawRectangle(position.x, position.y + 195, 5, height);
		height = (wxUint16)(fabs(mSpeedRight / 127.0f) * 25);
		if(mSpeedRight > 0)
			dc.DrawRectangle(position.x + 15, position.y + 194 - height, 5, height);
		else
			dc.DrawRectangle(position.x + 15, position.y + 195, 5, height);
	}

	/* Delta positions */
	if(GetSensorEnabled(SENSOR_DELTA_POSITION)) {
		dc.DrawText("Delta positions", position.x + 100, position.y + 150);
		dc.SetPen(*wxBLACK_PEN);
		dc.SetBrush(*wxWHITE_BRUSH);
		dc.DrawRectangle(position.x + 100, position.y + 170, 5, 50);
		dc.DrawRectangle(position.x + 100 + 15, position.y + 170, 5, 50);

		dc.SetBrush(*wxBLUE_BRUSH);
		height = (wxUint16)(fabs(mDPLeft / 127.0f) * 25);
		if(mDPLeft > 0)
			dc.DrawRectangle(position.x + 100, position.y + 194 - height, 5, height);
		else
			dc.DrawRectangle(position.x + 100, position.y + 195, 5, height);
		height = (wxUint16)(fabs(mDPRight / 127.0f) * 25);
		if(mDPRight > 0)
			dc.DrawRectangle(position.x + 100 + 15, position.y + 194 - height, 5, height);
		else
			dc.DrawRectangle(position.x + 100 + 15, position.y + 195, 5, height);
	}

	/* Vision */
	if(GetSensorEnabled(SENSOR_VISION)) {
		const wxUint8 num = GetNumVision();

		dc.SetPen(*wxBLACK_PEN);
	
		/* Vision */
		dc.DrawText("Vision", position.x + 150, position.y + 10);
		width = 15; //360 / num;	
		height = 30;
		for(int i=0; i<num; i++) {
			value = mVisionData[i];
			dc.SetBrush(wxBrush(wxColour(value, value, value), wxSOLID));
			dc.DrawRectangle(position.x + 150 + (width-1) * i, position.y + 30, width, height);
		} 

		/* Display Vision as optic flow with color codes: red and blue for left and right  */
		wxInt8 temp;
		dc.DrawText("Optic flow", position.x + 150, position.y + 70);
		width = 15;
		height = 30;
		for(int i=0; i<num; i++) {
			temp = (wxInt8)mVisionData[i];
			if (temp<0) {
				value = (wxUint8)(-temp)*2;
				dc.SetBrush(wxBrush(wxColour(value, 0, 0), wxSOLID));
			} else {
				value = (wxUint8)temp*2;
				dc.SetBrush(wxBrush(wxColour(0, 0, value), wxSOLID));
			}
			dc.DrawRectangle(position.x + 150 + (width-1) * i, position.y + 90, width, height);
		}  
	}

}

void Khepera::SetImpl(KheperaType kheperaType) {
	switch(kheperaType) {
		case KHEPERA_BASIC:		delete mImpl; mImpl = new BasicKhep(&mSensorMask); 		break;
		case KHEPERA_EVOPIC: 	delete mImpl; mImpl = new EvopicKhep(&mSensorMask); 	break;
		default:				break;
	}
}

/* --- XML --- */
void Khepera::Load (XMLElement *myself, XMLLoader *loader) {	
	Robot::Load(myself, loader);
	KheperaType kheperaType = KheperaImpl::GetKheperaTypeByXMLTagName(loader->GetNextElementName());
	SetImpl(kheperaType);
	loader->Load(this, mImpl);
}

XMLElement* Khepera::SaveSelf () {	
	XMLElement *result = Robot::SaveSelf();
	result->SetName(GetShortXMLName());
	return result;
}

void Khepera::SaveChilds (XMLSaver *saver) {
	Robot::SaveChilds(saver);
	saver->SaveObject(mImpl);
}

wxString Khepera::GetXMLName () { return GetShortXMLName(); }
wxString Khepera::GetShortXMLName() { return Khepera::GetStaticXMLType(); }
