/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _KHEPERA_H
#define _KHEPERA_H

#include "robot.h"
#include "../utils/xml/xml.h"
#include "khepera/kheperaimpl.h"

#define VIRTUAL_WALL 0 

/** 
 * Implements control of actuators and sensors of the Khepera.
 * @see RobotFactory
 */
class Khepera : public Robot
{
public:
	// register stuff
	enum { ItemID = 'Khep' };
	const static wxString ItemDesc;
	

	Khepera();
	virtual ~Khepera();

	/** 
	 * Initialize buffers and related variables. Called by RobotFactory
	 * immediately after creating the instance.
	 */
	virtual void Initialize();

	/* ---------------- ROBOT CONTROL -------------------- */

	/**
	 * Connect to the robot using the current CommLink.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect();

	/**
	 * Disconnect from the robot.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Disconnect();

	/** 
	 * Reset position of the robot to its initial position
	 */
	virtual wxInt8 ResetPos();

	/** 
	 * Read and store sensor data of all enabled sensors.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadSensors();

	/**
	 * Read and store encoder data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadEncoders();

	/**
	 * Read and store delta position data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadDeltaPosition();

	/**
	 * Read and store distance sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadDistanceSensors();
	
	/**
	 * Read and store floor sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadFloorSensor();
	
	/**
	 * Read and store ambient light sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadAmbientLightSensors();

	/** 
	 * Read and store vision data. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadVisionSensors();

	/** 
	 * Read and store analog inputs. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadAnalogSensors();

	/** Calculate new motorspeeds necessary for Braitenberg behaviour. */
	void CalcBraitenberg(float *motorSpeeds);

	/**
	 * Set the motor speeds to the values given.
	 * @param speeds Three floats between -1.0f and 1.0f. (Third is used as speed scale.)
	 */
	virtual wxInt8 SetMotors(float *speeds);

	/**
	 * Set the motor speeds to the values given.
	 * @param speedL Speed for left motor.
	 * @param speedR Speed for right motor.
	 */
	virtual wxInt8 SetMotors(wxInt8 speedL, wxInt8 speedR);

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable);

	/**
		*Enable or disable the component of environment 
		* @param type the type of conponent
		* @param enable  True to enable, false to disable.
		*/
	virtual wxInt8 SetFlag(wxUint8 type, bool enable);
	
	/**
	 * Use axis and button states of the joystick to control the robot.
	 * @param jsButtons Joystick button states.
	 * @param jsAxes Joystick axis states.
	 */
	virtual wxInt8 ManualControl(int *jsButtons, float *jsAxes);


	/* ---------------- ACCESSORS -------------------- */

	/** Returns current Khepera type. */
	KheperaType GetKheperaType() { return mImpl->GetType(); }

	/** Returns a pointer to an array of 10-bit distance values. */
	wxUint8* GetDistanceData() { return mDistanceData; }
	/** Returns a pointer to an array of 10-bit ambient light values. */
	wxUint8* GetAmbientData() { return mAmbientData; }
	/** Return the speed of left motor. */
	wxInt8 GetSpeedLeft() { return mSpeedLeft; }
	/** Return the speed of right motor. */
	wxInt8 GetSpeedRight() { return mSpeedRight; }
	/** Returns a pointer to an array of 8-bit greylevel values. */
	wxUint8* GetVisionData() { return mVisionData; }
	/** Returns a pointer to an array of 8-bit analog input values. */
	wxUint8* GetAnalogData() { return mAnalogData; }
	/** Returns a pointer to an array of 8-bit floor input values. */
	wxUint8* GetFloorData() { return mFloorData; }
	/** Return the delta position of left motor. */
	wxInt8 GetDeltaPositionLeft() { return mDPLeft; }
	/** Return the delta position of right motor. */
	wxInt8 GetDeltaPositionRight() { return mDPRight; }
	 
	/** Number of vision values returned by robot. */
	 const wxUint8 GetNumVision() { return GetSensorEnabled(SENSOR_VISION) ? mImpl->GetNumVision() : 0; }
	/** Number of distance values returned by robot. */
	 const wxUint8 GetNumDistance() { return GetSensorEnabled(SENSOR_DISTANCE) ? mImpl->GetNumDistance() : 0; }
	/** Number of floor sensors. */
	 const wxUint8 GetNumFloorSensors() { return GetSensorEnabled(SENSOR_FLOOR) ? mImpl->GetNumFloorSensors() : 0; }
	/** Number of analog values returned by robot. */
	 const wxUint8 GetNumAnalog() { return GetSensorEnabled(SENSOR_ANALOG) ? mImpl->GetNumAnalog() : 0; }

	/* ---------------- CONFIGURE ------------------ */

	/** Ask user for robot configuration. */
	virtual void Configure();

	/** Set a given robot configuration. */
	void SetImpl(KheperaType kheperaType);

	/* ---------------- LOGGING -------------------- */

	/** Write 'legenda' to log file. */
	virtual void LogLegenda(Logger *logger);

	/** Called by logger when new data should be logged. */
	virtual void Log(Logger *logger);

	/* ---------------- PAINTING -------------------- */
	
	/** 
	 * Paint the sensor data, which is robot dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintSensorData(wxDC& dc, wxPoint& position, wxSize& size);


	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual XMLElement*		SaveSelf ();
	virtual void			SaveChilds (XMLSaver *saver);
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "robot-khepera"; }

protected:
	KheperaImpl *mImpl;		/** Khepera implementation to be used. */

	wxUint8 *mDistanceData; /**< Array for maintaining distance data. */
	wxUint8 *mFloorData;	/**< Array for maintaining floor data. */
	wxUint8 *mAmbientData;	/**< Array for maintaining ambient light data. */
	wxInt8 mSpeedLeft;		/**< Speed of left motor. */
	wxInt8 mSpeedRight;		/**< Speed of right motor. */
	wxUint8 *mVisionData;	/**< Array for maintaining vision data. */
	wxUint8 mAnalogNum;		/**< Number of analog value to be received. */
	wxUint8 *mAnalogData;	/**< Array for maintaining analog data. */
	wxInt8 mDPLeft;			/**< Delta position of left motor. */
	wxInt8 mDPRight;		/**< Delta position of right motor. */

};

#endif /* _KHEPERA_H */
