/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "../../utils/xml/xml.h"
#include "../../status.h"
#include "../khepera.h"

#include "basickhep.h"

BasicKhep::BasicKhep(wxUint32 *sensorMask) : KheperaImpl(sensorMask) {
	mType = KHEPERA_BASIC;
}

BasicKhep::~BasicKhep() {

}

wxInt8 BasicKhep::Connect() {
	wxInt8 ret;
	wxString command, reply;
//	StatusLine("Please turn on the robot or press the 'Reset' button.");
	if((ret = mCommLink->Connect()))
		return ret;
/*	for(wxUint8 counter = 0; (ret = mCommLink->ReceiveString(string)) && counter < 15; counter++)
		iSleep(50);
	if(ret) {
		StatusLine("Couldn't connect, time-out.");
		mCommLink->Disconnect();
		return 4;
	}
	total.Append(string);
	while(!mCommLink->ReceiveString(string))
		total.Append(string);
	total.Trim(true);
	total.Trim(false);
	StatusLine(GetDesc() + " said \"" + total + "\"");
*/
	mCommLink->Synchronize();

    /* Set the PID controller settings */
    command = "A,3800,800,100\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
    /* Set the scan period of the K213 turret */
	if(SENSOR_ENABLED(*mpSensorMask, SENSOR_VISION)) {  
    	/********************
    	Value  Period
          1  ->  50 ms
          2  -> 100 ms
          3  -> 150 ms
          0  -> 200 ms (default)
          4  -> 250 ms
          5  -> 500 ms
          6  ->   1 s
          7  ->   5 s
        ********************/
		command = "T,2,U,1\n";    // 1 -> 50ms
		if((ret = ExecCommand(command, reply)))
			return ret;
	} 
	if(SENSOR_ENABLED(*mpSensorMask, SENSOR_FLOOR)) {
		command = "W,19,48\n";   // selecting the right channel on the IF turret (0x30 = 48) & switching on the sensor
		if((ret = ExecCommand(command, reply)))
			return ret;
	}
	return 0;
}

wxInt8 BasicKhep::ReadEncoders(wxInt8 &speedLeft, wxInt8 &speedRight) {
	wxInt8 ret;
	wxString command, reply;
	int temp1, temp2;

	/* Speed */
	command = "E\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	const char *buff = reply.c_str();
	if(sscanf(buff, "e,%d,%d", &temp1, &temp2) != 2) {
		StatusLine("BasicKhep - Couldn't parse motor speeds!");
		StatusLine("BasicKhep - Received \"" + reply + "\"");
		return 4;
	}
	speedLeft = (wxInt8)temp1;
	speedRight = (wxInt8)temp2;
	return 0;
}

wxInt8 BasicKhep::ReadDeltaPosition(wxInt8 &dPLeft, wxInt8 &dPRight) {
	wxInt8 ret;
	wxString command, reply;
	wxInt32 temp1, temp2;

	/* Positions*/
	command = "H\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	const char *buff = reply.c_str();
	if(sscanf(buff, "h,%d,%d", &temp1, &temp2) != 2) {
		StatusLine("BasicKhep - Couldn't parse wheel positions!");
		StatusLine("BasicKhep - Received \"" + reply + "\"");
		return 4;
	}
	dPLeft = (wxInt8)(temp1/2);
	dPRight = (wxInt8)(temp2/2);

	command = "G,0,0\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	return 0;
}

wxInt8 BasicKhep::ReadDistanceSensors(wxUint8 *distanceData) {
	wxInt8 ret, index;
	wxString command, reply;
	int temp1;
	const char *buff;
	
	command = "N\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	buff = reply.c_str();
	buff++;
	index = 0;
	while(buff[0] == ',' && index < GetNumDistance()) {
		sscanf(buff, ",%d", &temp1);
		distanceData[index++] = temp1 >> 2;
		buff++;
		while(buff[0] != ',' && buff[1] != '\r')
			buff++;
	}
	if(index != GetNumDistance()) {
		StatusLine(wxString::Format("BasicKhep - Expected %d distance values, got only %d!", GetNumDistance(), index));
		return 4;
	}
	return 0;
}

wxInt8 BasicKhep::ReadFloorSensor(wxUint8 *floorData) {
	wxInt8 ret;
	wxString command, reply;
	int temp1;
	const char *buff;
	
	command = "I,3\n";
	if((ret = ExecCommand(command, reply))) {
		StatusLine("BasicKhep - Error while reading floor sensor value!");
		return ret;
	}
	buff = reply.c_str();
	sscanf(buff, "i,%d", &temp1);
	floorData[0] = temp1 >> 2;

	return 0;
}

wxInt8 BasicKhep::ReadAmbientLightSensors(wxUint8 *ambientData) {
	wxInt8 ret, index;
	wxString command, reply;
	int temp1;
	const char *buff;
	
	command = "O\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	buff = reply.c_str();
	buff++;
	index = 0;
	while(buff[0] == ',' && index < GetNumDistance()) {
		sscanf(buff, ",%d", &temp1);
		ambientData[index++] = temp1 >> 2;
		buff++;
		while(buff[0] != ',' && buff[1] != '\r')
			buff++;
	}
	if(index != GetNumDistance()) {
		StatusLine(wxString::Format("BasicKhep - Expected %d ambient light values, got %d!", GetNumDistance(), index));
		return 4;
	}
	return 0;
}

wxInt8 BasicKhep::ReadVisionSensors(wxUint8 *visionData) {
	if(!SENSOR_ENABLED(*mpSensorMask, SENSOR_VISION)) return 10;
	wxInt8 ret, index;
	wxUint16 low;
	wxString command, reply;
	
	command = "T,2,T\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	const char *buff = reply.c_str() + 5;
	index = 0;
	while(buff[0] == ',' && index < GetNumVision()) {
		sscanf(buff, ",%hd", &low);
		visionData[index++] = (wxUint8)low;
		buff++;
		while(buff[0] != ',' && buff[1] != '\r')
			buff++;
	}
	if(index != GetNumVision()) {
		StatusLine(wxString::Format("BasicKhep - Expected %d vision values, got only %d!", GetNumVision(), index));
		StatusLine("BasicKhep - Received \"" + reply + "\"");
		return 4;
	}
	return 0;
}

wxInt8 BasicKhep::SetMotors(wxInt8 speedL, wxInt8 speedR) {
	wxInt8 ret;
	wxString reply, command = wxString::Format("D,%d,%d\n", speedL, speedR);
	if((ret = ExecCommand(command, reply)))
		return ret;
	return 0;
}

wxInt8 BasicKhep::SetLED(wxUint8 num, bool enable) {
	wxInt8 ret;
	wxString reply, command = wxString::Format("L,%d,%d\n", num, (wxUint8)enable);
	if((ret = ExecCommand(command, reply)))
		return ret;
	return 0;
}

void BasicKhep::Configure() {
	if(wxMessageBox("Is the K213 turret available?", "Configure Basic Khepera", wxYES_NO) == wxYES)
		*mpSensorMask |= SENSOR_VISION;
	else
		*mpSensorMask &= ~SENSOR_VISION;
	if(wxMessageBox("Is the floor sensor available?", "Configure Basic Khepera", wxYES_NO) == wxYES) {
		*mpSensorMask |= SENSOR_FLOOR;
		*mpSensorMask &= ~SENSOR_AMBIENT;	// ambient light sensors produce strange results when floor sensor is enabled
	} else {
		*mpSensorMask &= ~SENSOR_FLOOR;
		*mpSensorMask |= SENSOR_AMBIENT;
	}	
}

wxInt8 BasicKhep::ExecCommand(const wxString &command, wxString &reply) {
	wxInt8 ret = -1;
	wxString firstChar = command.SubString(0,0).Lower();
	
	for(wxUint8 tries = 0; ret != 0 && tries < 3; tries++) {
		ret = 0;
		if(mCommLink->SendString(command)) {
			mCommLink->Synchronize();
			ret = 1;
			break;
		}
		if(mCommLink->ReceiveString(reply)) {
			mCommLink->Synchronize();
			ret = 2;
			break;
		}
		if(!reply.StartsWith(firstChar, NULL)) {
			mCommLink->Synchronize();
			ret = 3;
			break;
		}
	}
	return ret;
}

/* --- XML --- */
void BasicKhep::Load (XMLElement *myself, XMLLoader *loader) {

}

XMLElement* BasicKhep::SaveSelf () {
	XMLElement *result = KheperaImpl::SaveSelf();
	result->SetName(GetShortXMLName());
	return result;
}

wxString BasicKhep::GetXMLName () { return GetShortXMLName(); }
wxString BasicKhep::GetShortXMLName() { return BasicKhep::GetStaticXMLType(); }
