/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _BASICKHEP_H
#define _BASICKHEP_H

#include "../../utils/xml/xml.h"
#include "kheperaimpl.h"

/** 
 * Khepera implementation: basic Khepera (I or II).
 * @see Khepera
 */
class BasicKhep : public KheperaImpl
{
public:
	BasicKhep(wxUint32 *sensorMask);
	virtual ~BasicKhep();

	/* ---------------- ROBOT CONTROL -------------------- */

	/**
	 * Connect to the robot using the specified CommLink.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect();

	/**
	 * Read and store encoder data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadEncoders(wxInt8 &speedLeft, wxInt8 &speedRight);

	/**
	 * Read and store delta position data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadDeltaPosition(wxInt8 &dPLeft, wxInt8 &dPRight);

	/**
	 * Read and store distance sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadDistanceSensors(wxUint8 *distanceData);
	
	/**
	 * Read and store floor sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadFloorSensor(wxUint8 *floorData);
	
	/**
	 * Read and store ambient light sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadAmbientLightSensors(wxUint8 *ambientData);

	/** 
	 * Read and store vision data. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadVisionSensors(wxUint8 *visionData);

	/**
	 * Set the motor speeds to the values given.
	 * @param speedL Speed for left motor.
	 * @param speedR Speed for right motor.
	 */
	virtual wxInt8 SetMotors(wxInt8 speedL, wxInt8 speedR);

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable);

	/* --------------------------- ACCESSORS ------------------------ */

	/** Returns description of implementation. */
	virtual const wxString GetDesc() { return "Basic Khepera"; }

	/** Number of vision values returned by implementation. */
	virtual const wxUint8 GetNumVision() { return 16; }
	
	/** Number of floor sensors. */
	virtual const wxUint8 GetNumFloorSensors() { return 1; }

	/* --------------------------- CONFIGURE ------------------------ */

	/** Ask user for robot configuration. */
	virtual void Configure();

	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "khepimpl-basic"; }

protected:
	/** 
	 * Executes a command.
	 * @param command The command to be executed.
	 * @param reply Contains the reply of the robot after successfull execution.
	 * @return Error code, zero if ok.
	 */
	wxInt8 ExecCommand(const wxString &command, wxString &reply);
};

#endif /* _BASICKHEP_H */
