/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>

#include "../../status.h"
#include "../../utils/iutils.h"
#include "../../utils/xml/xml.h"
#include "../khepera.h"

#include "evopickhep.h"

EvopicKhep::EvopicKhep(wxUint32 *sensorMask) : KheperaImpl(sensorMask) {
	mType = KHEPERA_EVOPIC;
	mBuffer = new wxByte[EVOPICKHEP_BUFFERLEN];

	mNumAnalog = 50;
	mVisionMode = KEvoPic_NoVision;
}

EvopicKhep::~EvopicKhep() {
	delete [] mBuffer;
}

wxInt8 EvopicKhep::Connect() {
	wxInt8 ret;
	wxString total, string;

	if((ret = mCommLink->Connect()))
		return ret;

	StatusLine("kevopic - Testing connection.");
	mBuffer[0] = 0xFF; // NOP
	mBuffer[1] = 0x00; // data length
	if((ret = mCommLink->SendBuffer(mBuffer, 2)))
		return ret;
	if((ret = ReceivePacket()))
		return ret;
	if(mBuffer[0] != 0xFF) {
		mCommLink->Disconnect();
		return 4;
	}

	if(mVisionMode == KEvoPic_SMDM) {
		StatusLine("kevopic - Setting up SMDM camera.");
		/* Put here some initialization if needed. */ 
	} else if(mVisionMode == KEvoPic_TSL3301) {
		StatusLine("kevopic - Setting up TLS3301 camera.");
		mBuffer[0] = 0x11; // set exposition time
		mBuffer[1] = 0x01; // data length
		mBuffer[2] = mTSL3301ExpTime;   // exp time (times 100 us)
		if((ret = mCommLink->SendBuffer(mBuffer, 3)))
			return ret;
		if((ret = ReceivePacket()))
			return ret;
		if(mBuffer[0] != 0x11) {
			mCommLink->Disconnect();
			return 6;
		}
	}

	return 0;
}

wxInt8 EvopicKhep::ReadEncoders(wxInt8 &speedLeft, wxInt8 &speedRight) {
	wxInt8 ret;

	mBuffer[0] = 0x3A;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request speed data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+2))) { 
		StatusLine("kevopic - Couldn't receive speed data.");
		return ret;
	}
	if(mBuffer[0] != 0x3A || mBuffer[1] != 2) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive speed data.", mBuffer[0], mBuffer[1]));
		return 7;
	}
	speedLeft = (wxInt8)mBuffer[2];
	speedRight = (wxInt8)mBuffer[3];
	return 0;
}

wxInt8 EvopicKhep::ReadDeltaPosition(wxInt8 &dPLeft, wxInt8 &dPRight) {
	wxInt8 ret;

	mBuffer[0] = 0x3B;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request delta position data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+2))) { 
		StatusLine("kevopic - Couldn't receive delta position data.");
		return ret;
	}
	if(mBuffer[0] != 0x3B || mBuffer[1] != 2) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive delta position data.", mBuffer[0], mBuffer[1]));
		return 7;
	}
	dPLeft = (wxInt8)mBuffer[2];
	dPRight = (wxInt8)mBuffer[3];
	return 0;
}

wxInt8 EvopicKhep::ReadDistanceSensors(wxUint8 *distanceData) {
	wxInt8 ret;

	mBuffer[0] = 0x20;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request distance data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+8))) { 
		StatusLine("kevopic - Couldn't receive distance data.");
		return ret;
	}
	if(mBuffer[0] != 0x20 || mBuffer[1] != 8) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive distance data.", mBuffer[0], mBuffer[1]));
		return 8;
	}
	for(wxUint8 i=0; i<8; i++)
		distanceData[i] = mBuffer[i+2];
	return 0;
}

wxInt8 EvopicKhep::ReadFloorSensor(wxUint8 *floorData) {
	wxInt8 ret, num = GetNumFloorSensors();

	mBuffer[0] = 0x25;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request floor sensor data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+num))) { 
		StatusLine("kevopic - Couldn't receive floor sensor data.");
		return ret;
	}
	if(mBuffer[0] != 0x25 || mBuffer[1] != num) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive floor sensor data.", mBuffer[0], mBuffer[1]));
		return 8;
	}
	floorData[0] = mBuffer[2];
	return 0;
}

wxInt8 EvopicKhep::ReadAmbientLightSensors(wxUint8 *ambientData) {
	wxInt8 ret;

	mBuffer[0] = 0x22;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request ambient data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+8))) { 
		StatusLine("kevopic - Couldn't receive ambient data.");
		return ret;
	}
	if(mBuffer[0] != 0x22 || mBuffer[1] != 8) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive ambient light data.", mBuffer[0], mBuffer[1]));
		return 8;
	}
	for(wxUint8 i=0; i<8; i++)
		ambientData[i] = mBuffer[i+2];
	return 0;
}

wxInt8 EvopicKhep::ReadVisionSensors(wxUint8 *visionData) {
	if(mVisionMode == KEvoPic_NoVision) 
		return 10;
	wxInt8 ret;
	wxUint8 num = GetNumVision();
	wxByte cmd = mVisionMode == KEvoPic_K213 ? 0x1A : (mVisionMode == KEvoPic_SMDM ? 0x14 : ( mVisionMode == KEvoPic_TSL3301_Laplace ? 0x12 : 0x10));	// 0x10 is for TSL3301

	mBuffer[0] = cmd;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request vision data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+num))) { 
		StatusLine("kevopic - Couldn't receive vision data.");
		return ret;
	}
	if(mBuffer[0] != cmd || mBuffer[1] != num) {
		StatusLine(wxString::Format("kevopic - Got %d, expected %d vision values.", mBuffer[1], num));
		return 9;
	}
	for(wxUint8 i=0; i<num; i++)
		visionData[i] = mBuffer[i+2];
	return 0;
}

wxInt8 EvopicKhep::ReadAnalogSensors(wxUint8 *analogData) {
	wxInt8 ret;

	mBuffer[0] = 0x21;
	mBuffer[1] = 0x00;
	if((ret = mCommLink->SendBuffer(mBuffer, 2))) { 
		StatusLine("kevopic - Couldn't request analog data.");
		return ret;
	} else if((ret = mCommLink->ReceiveBuffer(mBuffer, 2+mNumAnalog))) { 
		StatusLine("kevopic - Couldn't receive analog data.");
		return ret;
	}
	if(mBuffer[0] != 0x21)
		return 9;
	mNumAnalog = mBuffer[1];
	for(wxUint8 i=0; i<mNumAnalog; i++)
		analogData[i] = mBuffer[2 + i];
	return 0;
}

wxInt8 EvopicKhep::SetMotors(wxInt8 speedL, wxInt8 speedR) {
	wxInt8 ret;
	mBuffer[0] = 0x30;
	mBuffer[1] = 0x02;
	mBuffer[2] = speedL;
	mBuffer[3] = speedR;

	if((ret = mCommLink->SendBuffer(mBuffer, 4)))
		return ret;
	if((ret = ReceivePacket())) {
		StatusLine("kevopic - Couldn't receive acknowledge on set motor speeds.");
		return ret;
	}
	if(mBuffer[0] != 0x30 || mBuffer[1] != 0x00) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on receive acknowledge for set motor speeds.", mBuffer[0], mBuffer[1]));
		return 7;
	}
	return 0;
}

wxInt8 EvopicKhep::SetLED(wxUint8 num, bool enable) {
// Disabled because of use of LEDs for debugging purposes.
/*	wxInt8 ret;
	
	mBuffer[0] = 0x10;
	mBuffer[1] = 0x03;
	mBuffer[2] = 0x00;
	mBuffer[3] = 0x10;
	mBuffer[4] = 0x01;
	
	if((ret = mCommLink->SendBuffer(mBuffer, 5)))
		return ret;
	if((ret = ReceivePacket()))
		return ret;
	if(mBuffer[0] != 0x10)
		return 7;
*/
	return 0;
}

wxInt8 EvopicKhep::SetFlag(wxUint8 type, bool enable){
	wxInt8 ret;

	if (type == VIRTUAL_WALL) {
		mBuffer[0] = 0x01;
		mBuffer[1]=0x01;
		if (enable)
			mBuffer[2] =0x01;
		else 
			mBuffer[2] =0x00;
	}
	
 if((ret = mCommLink->SendBuffer(mBuffer, 3)))
		return ret;
	if((ret = ReceivePacket())) {
		StatusLine("kevopic - Couldn't determine the activation of virtual_wall.");
		return ret;
	}
	if(mBuffer[0] != 0x01) {
		StatusLine(wxString::Format("kevopic - Incorrect packet header (0x%2X,%d) on determination to the activation of virtual_wall.", mBuffer[0], mBuffer[1]));
		return 1;
	}
	return 0;
}

wxInt8 EvopicKhep::ManualControl(int *jsButtons, float *jsAxes) {
	wxInt8 ret;
	switch(jsButtons[0]) {
		case 8:
			if(mVisionMode != KEvoPic_TSL3301) break;
			mTSL3301ExpTime++;
			StatusLine(wxString::Format("kevopic - Increasing TSL3301 exposition time to %d us.", mTSL3301ExpTime * 100));
			mBuffer[0] = 0x11; // set exposition time
			mBuffer[1] = 0x01; // data length
			mBuffer[2] = mTSL3301ExpTime;   // exp time (times 100 us)
			if((ret = mCommLink->SendBuffer(mBuffer, 3)))
				return ret;
			if((ret = ReceivePacket()))
				return ret;
			if(mBuffer[0] != 0x11)
				return 7;
			break;
		case 16:
			if(mVisionMode != KEvoPic_TSL3301) break;
			if(mTSL3301ExpTime == 0) {
				StatusLine("kevopic - TSL3301 exposition time is 0, cannot decrease.");
				break;
			}
			mTSL3301ExpTime--;
			StatusLine(wxString::Format("kevopic - Decreasing TSL3301 exposition time to %d us.", mTSL3301ExpTime * 100));
			mBuffer[0] = 0x11; // set exposition time
			mBuffer[1] = 0x01; // data length
			mBuffer[2] = mTSL3301ExpTime;   // exp time (times 100 us)
			if((ret = mCommLink->SendBuffer(mBuffer, 3)))
				return ret;
			if((ret = ReceivePacket()))
				return ret;
			if(mBuffer[0] != 0x11)
				return 7;
			break;
		default:
			break;
	}
	return 0;
}


void EvopicKhep::Configure() {
	*mpSensorMask &= ~SENSOR_AMBIENT;
	*mpSensorMask &= ~SENSOR_VISION;
	mVisionMode = KEvoPic_NoVision;
	*mpSensorMask &= ~SENSOR_FLOOR;
	mTSL3301ExpTime = 4;


	if(wxMessageBox("Is the TSL3301 camera available?", "Configure kevopic", wxYES_NO) == wxYES) {
		if(wxMessageBox("Use embedded Laplace filter?", "Configure kevopic", wxYES_NO) == wxYES)
			mVisionMode = KEvoPic_TSL3301_Laplace;
		else 
			mVisionMode = KEvoPic_TSL3301;
		*mpSensorMask |= SENSOR_VISION;
		if(wxMessageBox("Use EL-20 lens? (Marshall otherwise.)", "Configure kevopic", wxYES_NO) == wxYES)
			mTSL3301ExpTime = 40;
		else
			mTSL3301ExpTime = 4;
	} else if (wxMessageBox("Is the SMDM available?", "Configure kevopic", wxYES_NO) == wxYES) {
		mVisionMode = KEvoPic_SMDM;
		*mpSensorMask |= SENSOR_VISION;
	} else if(wxMessageBox("Is the K213 camera available (only kevobot)", "Configure kevopic", wxYES_NO) == wxYES) {
		mVisionMode = KEvoPic_K213;
		*mpSensorMask |= SENSOR_VISION;
	} else if(wxMessageBox("Is analog input available?", "Configure kevopic", wxYES_NO) == wxYES) {
		*mpSensorMask |= SENSOR_ANALOG;
	}
	if(wxMessageBox("Is the floor sensor available?", "Configure kevopic", wxYES_NO) == wxYES)
		*mpSensorMask |= SENSOR_FLOOR;
}

wxInt8 EvopicKhep::ReceivePacket() {
	wxInt8 ret;
	if((ret = mCommLink->ReceiveBuffer(mBuffer, 2))) {
		StatusLine("kevopic - ReceivePacket error 1");
		return ret;
	}
	if(mBuffer[1] != 0)
		if((ret = mCommLink->ReceiveBuffer(mBuffer + 2, mBuffer[1]))) {
			StatusLine("EvoPic - ReceivePacket error 2");
			return ret + 3;
		}
	return 0;
}



/* --- XML --- */
void EvopicKhep::Load (XMLElement *myself, XMLLoader *loader) {
	mTSL3301ExpTime = (wxUint8)myself->GetAttribute("tsl3301exptime")->GetIntValue();
	mVisionMode = (KEvoPicVision)myself->GetAttribute("visionmode")->GetIntValue();
}

XMLElement* EvopicKhep::SaveSelf () {
	XMLElement *result = KheperaImpl::SaveSelf();
	result->SetName(GetShortXMLName());
	result->AddAttribute(new XMLAttribute("tsl3301exptime", mTSL3301ExpTime));
	result->AddAttribute(new XMLAttribute("visionmode",(wxUint8)mVisionMode));
	return result;
}

wxString EvopicKhep::GetXMLName () { return GetShortXMLName(); }
wxString EvopicKhep::GetShortXMLName() { return EvopicKhep::GetStaticXMLType(); }
