/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _EVOPICKHEP_H
#define _EVOPICKHEP_H

#include "../../utils/xml/xml.h"
#include "kheperaimpl.h"

#define EVOPICKHEP_BUFFERLEN 700

enum KEvoPicVision {
	KEvoPic_NoVision,
	KEvoPic_TSL3301,
	KEvoPic_SMDM,
	KEvoPic_K213, // only in Webots
	KEvoPic_TSL3301_Laplace
};

/** 
 * Khepera implementation: Khepera with evopic turret.
 * @see Khepera
 */
class EvopicKhep : public KheperaImpl
{
public:
	EvopicKhep(wxUint32 *sensorMask);
	virtual ~EvopicKhep();

	/* ---------------- ROBOT CONTROL -------------------- */

	/**
	 * Connect to the robot using the specified CommLink.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect();

	/**
	 * Read and store encoder data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadEncoders(wxInt8 &speedLeft, wxInt8 &speedRight);

	/**
	 * Read and store delta position data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadDeltaPosition(wxInt8 &dPLeft, wxInt8 &dPRight);

	/**
	 * Read and store distance sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadDistanceSensors(wxUint8 *distanceData);
	
	/**
	 * Read and store floor sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadFloorSensor(wxUint8 *floorData);

	/**
	 * Read and store ambient light sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadAmbientLightSensors(wxUint8 *ambientData);

	/** 
	 * Read and store vision data. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadVisionSensors(wxUint8 *visionData);

	/** 
	 * Read and store analog inputs. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadAnalogSensors(wxUint8 *analogData);
	
	/**
	 * Set the motor speeds to the values given.
	 * @param speedL Speed for left motor.
	 * @param speedR Speed for right motor.
	 */
	virtual wxInt8 SetMotors(wxInt8 speedL, wxInt8 speedR);

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable);
	
	/**
		*Enable or disable the component of environment 
		* @param type the type of conponent
		* @param enable  True to enable, false to disable.
		*/
	virtual wxInt8 SetFlag(wxUint8 type, bool enable);

	/**
	 * Use axis and button states of the joystick to control the robot.
	 * @param jsButtons Joystick button states.
	 * @param jsAxes Joystick axis states.
	 */
	virtual wxInt8 ManualControl(int *jsButtons, float *jsAxes);


	/* --------------------------- ACCESSORS ------------------------ */

	/** Returns description of implementation. */
	virtual const wxString GetDesc() { return "kevopic"; }
	
	/* Note: TSL3301 can generate 50 or 25 pixels, depending on the TSL_NUM_PIXELS value in the EvoPic software */
	virtual const wxUint8 GetNumVision() { return mVisionMode == KEvoPic_TSL3301 ? 25 :
																								(mVisionMode == KEvoPic_SMDM ? 32 :
																								(mVisionMode == KEvoPic_K213 ? 16 : 
																								(mVisionMode == KEvoPic_TSL3301_Laplace ? 32 :
																									0))); }

	
	/** Number of floor sensors. */
	virtual const wxUint8 GetNumFloorSensors() { return 1; }

	/** Number of analog values returned by implementation. */
	virtual const wxUint8 GetNumAnalog() { return mNumAnalog; }

	/* --------------------------- CONFIGURE ------------------------ */

	/** Ask user for robot configuration. */
	virtual void Configure();

	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "khepimpl-evopic"; }

protected:
	/**
	 * Receives one packet from Khepera turret and stores it in mBuffer.
	 * @return Error code (zero if ok).
	 */
	wxInt8 ReceivePacket();

	wxByte *mBuffer; /**< Buffer used for transceiving data. */
	wxUint8 mNumAnalog; /**< Number of analog values returned. */
	KEvoPicVision mVisionMode; /**< Determines vision mode. */
	wxUint8 mTSL3301ExpTime; /**< TSL3301 exposition time (x 100 us). */
};

#endif /* _EVOPICKHEP_H */
