/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _KHEPERAIMPL_H
#define _KHEPERAIMPL_H

#include "../../utils/xml/xml.h"

class CommLink;

enum KheperaType {
	KHEPERA_UNDEFINED,
	KHEPERA_BASIC,
	KHEPERA_EVOPIC
};

/** 
 * Abstract Khepera implementation.
 * @see Khepera
 */
class KheperaImpl : public XMLSerializable
{
public:
	KheperaImpl(wxUint32 *sensorMask);
	virtual ~KheperaImpl();

	/* ---------------- ROBOT CONTROL -------------------- */

	/**
	 * Connect to the robot using the specified CommLink.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect() =0;

	/**
	 * Read and store encoder data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadEncoders(wxInt8 &speedLeft, wxInt8 &speedRight) =0;

	/**
	 * Read and store delta position data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadDeltaPosition(wxInt8 &dPLeft, wxInt8 &dPRight) =0;

	/**
	 * Read and store distance sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadDistanceSensors(wxUint8 *distanceData) =0;
	
	/**
	 * Read and store floor sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadFloorSensor(wxUint8 *floorData) =0;
	
	/**
	 * Read and store ambient light sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadAmbientLightSensors(wxUint8 *ambientData) { return 123; }

	/** 
	 * Read and store vision data. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadVisionSensors(wxUint8 *visionData) { return 123; }

	/** 
	 * Read and store analog inputs. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadAnalogSensors(wxUint8 *analogData) { return 123; }

	/**
	 * Set the motor speeds to the values given.
	 * @param speedL Speed for left motor.
	 * @param speedR Speed for right motor.
	 */
	virtual wxInt8 SetMotors(wxInt8 speedL, wxInt8 speedR) =0;

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable) { return 0; }

	/**
		*Enable or disable the component of environment 
		* @param type the type of conponent
		* @param enable  True to enable, false to disable.
		*/
	virtual wxInt8 SetFlag(wxUint8 type, bool enable){ return 0; }
	
	/**
	 * Use axis and button states of the joystick to control the robot.
	 * @param jsButtons Joystick button states.
	 * @param jsAxes Joystick axis states.
	 */
	virtual wxInt8 ManualControl(int *jsButtons, float *jsAxes) { return 0; }


	/* --------------------------- ACCESSORS ------------------------ */

	/** Sets CommLink. */
	void SetCommLink(CommLink *commLink) { mCommLink = commLink; }

	/** Returns implementation type. */	
	const KheperaType GetType() { return mType; }
	/** Returns description of implementation. */
	virtual const wxString GetDesc() { return "Undefined"; }

	/** Number of distance values returned by implementation. */
	virtual const wxUint8 GetNumDistance() { return 8; }
	/** Number of floor sensors. */
	virtual const wxUint8 GetNumFloorSensors() { return 0; }
	/** Number of vision values returned by implementation. */
	virtual const wxUint8 GetNumVision() { return 0; }
	/** Number of analog values returned by implementation. */
	virtual const wxUint8 GetNumAnalog() { return 0; }

	/* --------------------------- CONFIGURE ------------------------ */

	/** Ask user for robot configuration. */
	virtual void Configure() {}

	static KheperaType GetKheperaTypeByXMLTagName(wxString tagName);

	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader)=0;
	virtual XMLElement*		SaveSelf ()=0;
	virtual void			SaveChilds (XMLSaver *saver);
	virtual void			SetParentName (const wxString &parentName);
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static const wxString	GetStaticXMLType() { return "khepimpl";}
	static const wxUint8	mNumKhepTypes;

protected:
	wxUint32 *mpSensorMask; /**< Pointer to sensormask of Khepera. */
	KheperaType mType; /**< Type of Khepera implementation. */
	CommLink *mCommLink; /**< CommLink to be used for communication. */
	static const wxString mXMLKheperaNames[]; /**< Array of xmlnames belonging to types. */
};

#endif /* _KHEPERAIMPL_H */
