/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "../status.h"
#include "../utils/xml/xml.h"

#include "koala.h"

const wxString Koala::ItemDesc = "Koala";

Koala::Koala() {
}

Koala::~Koala() {
	if(mDistanceData) delete [] mDistanceData;
	if(mAmbientData) delete [] mAmbientData;
}

void Koala::Initialize() {
	wxUint8 num = GetNumDistance();
	mDistanceData = num ? new wxUint8[num] : 0;
	mAmbientData = num ? new wxUint8[num] : 0;
	for(int i=0; i<num; i++) {
		mDistanceData[i] = 0;
		mAmbientData[i] = 0;
	}
	mSpeedLeft = 0;
	mSpeedRight = 0;
}

wxInt8 Koala::Connect() {
	if(mCommLink->GetConnected()) return 0;
	wxInt8 ret;
	wxString total, string;

	StatusLine("Connecting to " + RobotFactory::GetDesc(GetID()) + "...");	
	
//	StatusLine("Please turn on the robot or press the 'Reset' button.");
	if((ret = mCommLink->Connect()))
		return ret;
/*	for(wxUint8 counter = 0; (ret = mCommLink->ReceiveString(string)) && counter < 15; counter++)
		iSleep(50);
	if(ret) {
		StatusLine("Couldn't connect, time-out.");
		mCommLink->Disconnect();
		return 4;
	}
	total.Append(string);
	while(!mCommLink->ReceiveString(string))
		total.Append(string);
	total.Trim(true);
	total.Trim(false);
	StatusLine(RobotFactory::GetDescByType(mType) + " said \"" + total + "\"");
*/
	mCommLink->Synchronize();

	StatusLine("Connection established.");
	SetChanged();
	return 0;
}

wxInt8 Koala::Disconnect() {
	if(!mCommLink->GetConnected()) return 0;
	float speeds[3];
	speeds[0] = speeds[1] = speeds[2] = 0.0f;
	SetMotors(speeds);
	return Robot::Disconnect();
}

wxInt8 Koala::ReadSensors() {
	wxInt8 ret;
	if(GetSensorEnabled(SENSOR_MOTOR_SPEED) && (ret = ReadEncoders()))
		return ret;	
	if(GetSensorEnabled(SENSOR_DISTANCE) && (ret = ReadDistanceSensors())) 
		return ret;
	if(GetSensorEnabled(SENSOR_AMBIENT) && (ret = ReadAmbientLightSensors())) 
		return ret;
	return 0;
}

wxInt8 Koala::ReadEncoders() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret;
	wxString command, reply;
	int temp1, temp2;

	/* Speed */
	command = "E\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	const char *buff = reply.c_str();
	if(sscanf(buff, "e,%d,%d", &temp1, &temp2) != 2) {
		StatusLine("Koala - Couldn't parse motor speeds!");
		StatusLine("Koala - Received \"" + reply + "\"");
	}
	mSpeedLeft = (wxInt8)temp1;
	mSpeedRight = (wxInt8)temp2;
	
	SetChanged();
	return 0;
}

wxInt8 Koala::ReadDistanceSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret, index;
	wxString command, reply;
	int temp1;
	const char *buff;
	
	/* Distance */
	command = "N\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	buff = reply.c_str();
	buff++;
	index = 0;
	while(buff[0] == ',' && index < GetNumDistance()) {
		sscanf(buff, ",%d", &temp1);
		mDistanceData[index++] = temp1 >> 2;
		buff++;
		while(buff[0] != ',' && buff[1] != '\r')
			buff++;
	}
	if(index != GetNumDistance())
		StatusLine(wxString::Format("Koala - Expected %d distance values, got %d!", GetNumDistance(), index));

	SetChanged();
	return 0;
}

wxInt8 Koala::ReadAmbientLightSensors() {
	if(!mCommLink->GetConnected()) return 1;
	wxInt8 ret, index;
	wxString command, reply;
	int temp1;
	const char *buff;
	
	command = "O\n";
	if((ret = ExecCommand(command, reply)))
		return ret;
	buff = reply.c_str();
	buff++;
	index = 0;
	while(buff[0] == ',' && index < GetNumDistance()) {
		sscanf(buff, ",%d", &temp1);
		mAmbientData[index++] = temp1 >> 2;
		buff++;
		while(buff[0] != ',' && buff[1] != '\r')
			buff++;
	}
	if(index != GetNumDistance())
		StatusLine(wxString::Format("Koala - Expected %d ambient light values, got %d!", GetNumDistance(), index));

	SetChanged();
	return 0;
}

wxInt8 Koala::SetMotors(float *speeds) {
	const wxUint8 scale = 100;

	speeds[2] = (speeds[2] + 1.0f) / 2;
	for(int i=0; i<2; i++) {
		if(speeds[i] < -1.0f) speeds[i] = -1.0f;
		if(speeds[i] >  1.0f) speeds[i] =  1.0f;
		speeds[i] *= speeds[2]; // use z-axis to scale speed!
	}
	return SetMotors((wxInt8)(speeds[0] * scale), (wxInt8)(speeds[1] * scale));
}

wxInt8 Koala::SetMotors(wxInt8 speedL, wxInt8 speedR) {
	wxInt8 ret;
	wxString reply, command = wxString::Format("D,%d,%d\n", speedL, speedR);
	if((ret = ExecCommand(command, reply)))
		return ret;
	SetChanged();
	return 0;
}

wxInt8 Koala::SetLED(wxUint8 num, bool enable) {
	wxInt8 ret;
	wxString reply, command = wxString::Format("L,%d,%d\n", num, (wxUint8)enable);
	if((ret = ExecCommand(command, reply)))
		return ret;
	SetChanged();
	return 0;
}

wxInt8 Koala::ManualControl(int *jsButtons, float *jsAxes) {
	wxInt8 ret;
	float speeds[] = { 0.0f, 0.0f, 0.0f };

	// Those are the joystick axes:
	// Roll:	jsAxes[0]
	// Pitch:	jsAxes[1]
	// Yaw:		jsAxes[2]
	// Power:	jsAxes[3]

	speeds[0] = + jsAxes[2] - jsAxes[1];
	speeds[1] = - jsAxes[2] - jsAxes[1];
	speeds[2] = jsAxes[3];	// speed factor
		
	for (int i=0;i<3;i++) { // limit the motor values between 1 and -1
		if (speeds[i]>1) speeds[i] = 1;
		else if (speeds[i]<-1) speeds[i] = -1;
	}

	speeds[0] = 2*asin(speeds[0])/3.14f;
	speeds[1] = 2*asin(speeds[1])/3.14f;

	if((ret = SetMotors(speeds)))
		StatusLineErr("Couldn't set motor speeds!", ret);

	return ret;
}

void Koala::LogLegenda(Logger *logger) {
	logger->WriteString(wxString("speed_left speed_right distance[0..15]\n"));
}

void Koala::Log(Logger *logger) {
	logger->WriteString(wxString::Format("%d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d\n", mSpeedLeft, mSpeedRight, mDistanceData[0], mDistanceData[1], mDistanceData[2], mDistanceData[3], mDistanceData[4], mDistanceData[5], mDistanceData[6], mDistanceData[7], mDistanceData[8], mDistanceData[9], mDistanceData[10], mDistanceData[11], mDistanceData[12], mDistanceData[13], mDistanceData[14], mDistanceData[15]));
}

void Koala::PaintSensorData(wxDC& dc, wxPoint& position, wxSize& size) {
	const wxUint8 num = GetNumDistance();
	wxUint8 width, height, value;

	dc.SetPen(*wxBLACK_PEN);
	
	/* Ambient light */
	if(GetSensorEnabled(SENSOR_AMBIENT)) {
		dc.DrawText("Ambient light", position.x, position.y);
		width = 15;
		height = 30;
		for(int i=0; i<num; i++) {
			value = (1023 - mAmbientData[i]) >> 2;
			dc.SetBrush(wxBrush(wxColour(value, value, value), wxSOLID));
			dc.DrawRectangle(position.x + (width-1) * i, position.y + 20, width, height);
		}
	}
	
	/* Distance */
	if(GetSensorEnabled(SENSOR_DISTANCE)) {
		dc.DrawText("Distance", position.x, position.y + 60);
		dc.SetBrush(*wxWHITE_BRUSH);
		width = 5;
		for(int i=0; i<num; i++) {
			height = (wxUint8)((1023 - mDistanceData[i]) / 20);
			dc.DrawRectangle(position.x + 2 * width * i, position.y + 80, width, height);
		}
	}
	
	/* Motor speeds */
	if(GetSensorEnabled(SENSOR_MOTOR_SPEED)) {
		dc.DrawText("Motor speeds", position.x, position.y + 140);
		dc.SetPen(*wxBLACK_PEN);
		dc.SetBrush(*wxWHITE_BRUSH);
		dc.DrawRectangle(position.x, position.y + 160, 5, 50);
		dc.DrawRectangle(position.x + 15, position.y + 160, 5, 50);

		dc.SetBrush(*wxBLUE_BRUSH);
		height = (wxUint16)(fabs(mSpeedLeft / 127.0f) * 25);
		if(mSpeedLeft > 0)
			dc.DrawRectangle(position.x, position.y + 184 - height, 5, height);
		else
			dc.DrawRectangle(position.x, position.y + 185, 5, height);
		height = (wxUint16)(fabs(mSpeedRight / 127.0f) * 25);
		if(mSpeedRight > 0)
			dc.DrawRectangle(position.x + 15, position.y + 184 - height, 5, height);
		else
			dc.DrawRectangle(position.x + 15, position.y + 185, 5, height);
	}
}

wxInt8 Koala::ExecCommand(const wxString &command, wxString &reply) {
	wxInt8 ret = -1;
	wxString firstChar = command.SubString(0,0).Lower();
	
	for(wxUint8 tries = 0; ret != 0 && tries < 3; tries++) {
		ret = 0;
		if(mCommLink->SendString(command)) {
			mCommLink->Synchronize();
			ret = 1;
			break;
		}
		if(mCommLink->ReceiveString(reply)) {
			mCommLink->Synchronize();
			ret = 2;
			break;
		}
		if(!reply.StartsWith(firstChar, NULL)) {
			mCommLink->Synchronize();
			ret = 3;
			break;
		}
	}
	return ret;
}

/* --- XML --- */
void Koala::Load (XMLElement *myself, XMLLoader *loader) {
	Robot::Load(myself,loader);
}

XMLElement* Koala::SaveSelf () {	
	XMLElement *result = Robot::SaveSelf();
	result->SetName(GetShortXMLName());
	return result;
}

wxString Koala::GetXMLName () { return GetShortXMLName(); }
wxString Koala::GetShortXMLName() { return Koala::GetStaticXMLType(); }

