/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _KOALA_H
#define _KOALA_H

#include "robot.h"

/** 
 * Implements control of actuators and sensors of the Koala.
 * @see RobotFactory
 */
class Koala : public Robot
{
public:
	// register stuff
	enum { ItemID = 'Koal' };
	const static wxString ItemDesc;
	

	Koala();
	virtual ~Koala();

	/** 
	 * Initialize buffers and related variables. Called by RobotFactory
	 * immediately after creating the instance.
	 */
	virtual void Initialize();

	/* ---------------- ROBOT CONTROL -------------------- */

	/**
	 * Connect to the robot using the current CommLink.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect();

	/**
	 * Disconnect from the robot.
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Disconnect();

	/**
	 * Read and store the sensor data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadSensors();
	

	/**
	 * Read and store only the encoder data.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadEncoders();

	/**
	 * Read and store only the distance sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadDistanceSensors();
	
	/**
	 * Read and store only the ambient light sensor data.
	 * @return Error code, zero if ok.
	 */	
	virtual wxInt8 ReadAmbientLightSensors();

	/**
	 * Set the motor speeds to the values given.
	 * @param speeds Three floats between -1.0f and 1.0f. (Third is used as speed scale.)
	 */
	virtual wxInt8 SetMotors(float *speeds);

	/**
	 * Set the motor speeds to the values given.
	 * @param speedL Speed for left motor.
	 * @param speedR Speed for right motor.
	 */
	virtual wxInt8 SetMotors(wxInt8 speedL, wxInt8 speedR);

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable);

	/**
	 * Use axis and button states of the joystick to control the robot.
	 * @param jsButtons Joystick button states.
	 * @param jsAxes Joystick axis states.
	 */
	virtual wxInt8 ManualControl(int *jsButtons, float *jsAxes);


	/* ---------------- ACCESSORS -------------------- */

	/** Returns a pointer to an array of 10-bit distance values. */
	wxUint8* GetDistanceData() { return mDistanceData; }
	/** Returns a pointer to an array of 10-bit ambient light values. */
	wxUint8* GetAmbientData() { return mAmbientData; }
	/** Return the speed of left motor. */
	wxInt8 GetSpeedLeft() { return mSpeedLeft; }
	/** Return the speed of right motor. */
	wxInt8 GetSpeedRight() { return mSpeedRight; }


	/* ---------------- LOGGING -------------------- */

	/** Write 'legenda' to log file. */
	virtual void LogLegenda(Logger *logger);

	/** Called by logger when new data should be logged. */
	virtual void Log(Logger *logger);


	/* ---------------- PAINTING -------------------- */

	/** 
	 * Paint the sensor data, which is robot dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintSensorData(wxDC& dc, wxPoint& position, wxSize& size);


	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "robot-koala"; }


protected:
	/** 
	 * Executes a command.
	 * @param command The command to be executed.
	 * @param reply Contains the reply of the robot after successfull execution.
	 * @return Error code, zero if ok.
	 */
	wxInt8 ExecCommand(const wxString &command, wxString &reply);

	wxUint8 *mDistanceData; /**< Array for maintaining distance data. */
	wxUint8 *mAmbientData; /**< Array for maintaining ambient light data. */
	wxInt8 mSpeedLeft; /**< Speed of left motor. */
	wxInt8 mSpeedRight; /**< Speed of right motor. */

	/** Number of distance values returned by robot. Default sixteen. */
	virtual const wxUint8 GetNumDistance() { return 16; }
};

#endif /* _KOALA_H */
