/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/



#include <wx/wx.h>
#include <math.h>
#define _USE_MATH_DEFINES
#include <wx/datetime.h>

#include "../status.h"
#include "../utils/iutils.h"
#include "../utils/xml/xml.h"

#include "robot.h"

Robot::Robot() {
	mCommLink = NULL;
	mBatteryLifeTime = 0; // Default to unlimited
	mSensorMask = SENSOR_ALL; // Default to all sensors enabled
}

Robot::~Robot() {

}

wxInt8 Robot::AttemptConnect() {
	wxInt8 ret = 0;
	while(true) {
		if(!(ret = Connect()))
			break;
		if(iMessageBox("Failed to connect to robot. Do you want to re-try?", "Connecting", wxYES_NO) != wxYES)
			break;
	}
	NewBatteries();
	return ret;
}

wxInt8 Robot::Connect() {
	StatusLine("Connecting to " + RobotFactory::GetDesc(GetID()) + "...");
	wxInt8 ret;
	if((ret = mCommLink->Connect()))
		return ret;
	StatusLine("Connection established.");
	SetChanged();
	return 0;
}

wxInt8 Robot::Disconnect() {
	if(mCommLink->GetConnected()) {
		float speeds[] = { 0.0f, 0.0f, 0.0f };
		SetMotors(speeds);
	}
	return mCommLink->Disconnect();
}

wxInt32 Robot::BatteryTime() {
	if(mBatteryLifeTime == 0) return -1;
	return (wxInt32)(mBatteryLifeTime - (wxDateTime::Now().GetTicks() - mBatteryStart));
}

bool Robot::BatteriesOK() {
	if(mBatteryLifeTime == 0) return true;
	if(wxDateTime::Now().GetTicks() - mBatteryStart > mBatteryLifeTime)
		return false;
	return true;
}

wxInt8 Robot::ChangeBatteries() {
	Disconnect();
	iMessageBox("Please change the batteries of the robot and press OK.", "Change batteries", wxOK);
	return AttemptConnect();
}

void Robot::NewBatteries() {
	SetChanged();
	mBatteryStart = wxDateTime::Now().GetTicks();
}

void Robot::StartLog(Logger *logger) {
	logger->WriteString("%ROBOT - " + RobotFactory::GetDesc(GetID()) + "\n");
	LogLegenda(logger);
}

void Robot::EndLog(Logger *logger) {}

void Robot::Paint(wxDC& dc, wxPoint& position, wxSize& size) {
	/* Draw robot name */
	wxFont font = dc.GetFont();
	font.SetWeight(wxBOLD);
	dc.SetFont(font);
	dc.DrawText(RobotFactory::GetDesc(GetID()), position.x + 10, position.y + 15);
	font.SetWeight(wxNORMAL);
	dc.SetFont(font);
	
	/* Draw sensor data */
	position.x += 10;
	position.y += 50;
	size.SetHeight(size.GetHeight() - 35);
	size.SetWidth(size.GetWidth() - 20);
	PaintSensorData(dc, position, size);
}

/* --- XML --- */
void Robot::Load (XMLElement *myself, XMLLoader *loader) {
	mSensorMask = myself->GetAttribute("sensormask")->GetIntValue();
}

XMLElement* Robot::SaveSelf () {
	XMLElement *result = new XMLElement("robot");
	result->AddAttribute(new XMLAttribute("sensormask",(wxInt32)mSensorMask));
	return result;
}

void Robot::SaveChilds (XMLSaver *saver) {

}

void Robot::SetParentName (const wxString &parentName) {
}

wxString Robot::GetXMLName () { return GetShortXMLName(); }
wxString Robot::GetShortXMLName() { return Robot::GetStaticXMLType(); }

