/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _ROBOT_H
#define _ROBOT_H

/** SensorMask definitions, see docs/sensors.txt */
#define SENSOR_MOTOR_SPEED		1
#define SENSOR_DISTANCE			2
#define SENSOR_AMBIENT			4
#define SENSOR_VISION			8
#define SENSOR_ANALOG			16
#define SENSOR_FLOOR			32
#define SENSOR_DELTA_POSITION	64

#define SENSOR_ALL 				0xFFFFFFFF
#define ADD_SENSOR(mask)		mSensorMask |= (mask);
#define DEL_SENSOR(mask)		mSensorMask &= (~mask);

#define SENSOR_ENABLED(mask, sensorMask)	((mask & sensorMask) == sensorMask)

/** Includes */
#include "../utils/observable.h"
#include "../utils/logger.h"
#include "../comm/commlink.h"
#include "../utils/xml/xml.h"


/** 
 * Abstract class, providing basic facilities for control of the robot.
 */
class Robot : public Observable, public ILoggable, public XMLSerializable
{
public:
	Robot();
	virtual ~Robot();

	/** 
	 * Initialize buffers and related variables. Called by RobotFactory
	 * immediately after creating the instance.
	 */
	virtual void Initialize() { }
	
	
	/* ---------------- ROBOT CONTROL -------------------- */

	/** Try to connect to robot, ask for user whether to re-try or not
	 *  in case of failure.
	 *  @return Error code, zero if connected.
	 */
	wxInt8 AttemptConnect();
	
	/** 
	 * Connect to the robot using the current CommLink. 
	 * @return Error code, zero if connected.
	 */
	virtual wxInt8 Connect();

	/** 
	 * Disconnect from the robot. 
	 * * @return Error code, zero if no problems occurred.
	 */
	virtual wxInt8 Disconnect();

	/** 
	 * Reset position of the robot to its initial position
	 */
	virtual wxInt8 ResetPos() { return 0; };

	/**
	 * Enable or disable a LED on the Robot.
	 * @param num The number of the LED.
	 * @param enable True to enable, false to disable.
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 SetLED(wxUint8 num, bool enable) =0;

	/** 
	 * Read and store the sensor data. 
	 * @return Error code, zero if ok.
	 */
	virtual wxInt8 ReadSensors() =0;

	/** 
	 * Set the motor speeds to the values given.
	 * @param speeds Three floats between -1.0f and 1.0f. (Less than 3 may be used.)
	 */
	virtual wxInt8 SetMotors(float *speeds) =0;

	/**
	 * Use axis and button states of the joystick to control the robot.
	 * @param jsButtons Joystick button states.
	 * @param jsAxes Joystick axis states.
	 */
	virtual wxInt8 ManualControl(int *jsButtons, float *jsAxes) { return 0; }
	
	/* ---------------- BATTERIES -------------------- */
	
	/**
	 * @return Number of minutes battery time left. Negative if unlimited.
	 */
	wxInt32 BatteryTime();
	/** Returns false if batteries should be changed. */
	virtual bool BatteriesOK();
	/** Change the batteries of the robot. */
	wxInt8 ChangeBatteries();
	/** Indicate to robot that batteries have just been changed. */
	void NewBatteries();

	
	/* ---------------- ACCESSORS -------------------- */

	/** Returns true if currently connected to the robot. */
	bool GetConnected() { return mCommLink->GetConnected(); }

	/** Set the CommLink to be used to establish connection with robot. */
	void SetCommLink(CommLink *commLink) { mCommLink = commLink; }
	/** Returns the CommLink used to establish connection with robot. */
	CommLink* GetCommLink() { return mCommLink; }

	/** Sets the SensorMask. */
	void SetSensorMask(wxUint32 sensorMask) { mSensorMask = sensorMask; }
	/** Returns the SensorMask. */
	wxUint32 GetSensorMask() { return mSensorMask; }
	/** Enables or disables the specified sensor. */
	void SetSensorEnabled(wxUint32 sensorMask, bool enable) { if(enable) ADD_SENSOR(sensorMask) else DEL_SENSOR(sensorMask) }
	/** Returns true if the specified sensor(mask) is enabled, false otherwise. */
	bool GetSensorEnabled(wxUint32 sensorMask) { return (mSensorMask & sensorMask) == sensorMask; }

	/* ---------------- CONFIGURE ------------------ */

	/** Redefine this method to ask user for robot configuration. */
	virtual void Configure() {}

	/* ---------------- LOGGING -------------------- */

	/** Called when a Log is started. */
	void StartLog(Logger *logger);

	/** Has to be implemented by extensions to write 'legenda' of log file. */
	virtual void LogLegenda(Logger *logger) =0;
	
	/** Called when a Log is ended. */
	void EndLog(Logger *logger);

	/* ---------------- PAINTING -------------------- */

	/** Paint the robot. */
	void Paint(wxDC& dc, wxPoint& position, wxSize& size);

	/* ------------------- XML ---------------------- */

	/* XML stuff */
	virtual void			Load (XMLElement *myself, XMLLoader *loader);
	virtual void			SaveChilds (XMLSaver *saver);
	virtual void			SetParentName (const wxString &parentName);
	virtual XMLElement*		SaveSelf ();
	virtual wxString		GetXMLName ();
	virtual wxString		GetShortXMLName();
	static wxString			GetStaticXMLType() { return "robot"; }
	
	/* ---------------- REGISTER STUFF ------------------ */
	void SetID(int id) {mID = id;}
	int GetID() {return mID;}

protected:
	/** 
	 * Paint the sensor data, which is robot dependent. 
	 * @param dc Device Context to use.
	 * @param position Upperleft corner of drawing area.
	 * @param size Size of drawing area.
	 */
	virtual void PaintSensorData(wxDC& dc, wxPoint& position, wxSize& size) =0;
	
	CommLink *mCommLink; /**< CommLink used to establish connection with robot. */
	time_t mBatteryStart; /**< The time (in seconds) where the life of this battery started. */
	time_t mBatteryLifeTime; /**< Life time (in seconds) of one battery (or set of batteries). */
	wxUint32 mSensorMask; /**< Mask used to determine which sensors are used, different for each Robot. */
	
	int mID;
};

#endif /* _ROBOT_H */
