/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _FACTORYENGINE_H
#define _FACTORYENGINE_H

#include <wx/wx.h>

#include <vector>
#ifdef WIN32
	#include <wx/msw/winundef.h>
#endif

//#include "../status.h"





/**
 * This class implements an engine for a factory. It maintains a list
 * of creator function associated with ID, permitting to create any
 * subclass of BaseItem that has previously been registered.
 */
template<class BaseItem, class ItemID> class FactoryEngine
{
public:

	/* ---------------- TYPES DEFINITIONS -------------------- */
	
	/** Creator function type. */
	typedef BaseItem* (*ItemCreatorFunc)();
	
	
	/* ---------------- CTOR / DTOR -------------------- */
	
	/** Constructor */
	FactoryEngine() {}
	
	/** Destructor */
	virtual ~FactoryEngine()
	{	
		for(size_t i = 0; i < mEntries.size(); i++)
			delete mEntries[i];
	}
	
	
	/* ---------------- REGISTERATION -------------------- */


	/**
	 * Register an item. A macro should be provided for simplification, and the
	 * class FactoryEnging::ItemCreator can be used to automatically provide
	 * a creator function.
	 */
	void RegisterItem(ItemID id, ItemCreatorFunc creator, const char* desc, const char* xmlName)
	{
		// should check for preexisting id.
		ItemData* data = new ItemData;
		data->id = id;
		data->creator = creator;
		data->description = desc;
		data->xmlName = xmlName;
		mEntries.push_back(data);
	}
	
	
	/** This class is used to automatically generate the static creator function for registered items. */
	template<class Item> class ItemCreator
	{
	public:
		static BaseItem*	CreateItem() {return new Item;}
	};
	
	
	/* ---------------- CREATION -------------------- */
	
	/**
	 * Create an item based on its ID.
	 */
	BaseItem* CreateItemWithID(ItemID id)
	{
		wxUint32 index;
		if(GetIndexByID(id, index))
			return mEntries[index]->creator();
		else
			return NULL;
	}
	
	/**
	 * Create an item based on its index.
	 */
	BaseItem* CreateItemWithIndex(wxUint32 index) {return mEntries[index]->creator();}
	
	
	/* ---------------- ACCESSORS -------------------- */

	/** Return the number of registered item. */
	wxUint32 GetEntryNumber() {return mEntries.size();}
	
	/**
	 * Get the description of an item.
	 * @param id ID of the item.
	 * @return Description of the item.
	 */
	wxString GetDescriptionByID(ItemID id)
	{
		wxUint32 index;
		if(GetIndexByID(id, index))
			return mEntries[index]->description;
		else
			return "";
	}
	
	/**
	 * Get the description of an item.
	 * @param index Index of the item.
	 * @return Description of the item.
	 */
	wxString GetDescriptionByIndex(wxUint32 index) {return mEntries[index]->description;}
	
	/** Convert an item index to its ID. */
	ItemID GetIDByIndex(wxUint32 index) {return mEntries[index]->id;}
	
	/**
	 * Convert an item id to its index.
	 * @param id The ID of the item.
	 * @param index On output, the index.
	 * @return Return true on success.
	 */
	bool GetIndexByID(ItemID id, wxUint32 &index)
	{
		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->id == id)
			{
				index = i;
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Get the ID of an item.
	 * @param xmlName The XML name of the item.
	 * @param id On output, the ID.
	 * @return Return true on success.
	 */
	bool GetIDByXMLName(wxString xmlName, ItemID &id)
	{

		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->xmlName == xmlName)
			{
				id = mEntries[i]->id;
				return true;
			}
		}
		
		return false;
	}
	
	/**
	 * Get the ID of an item.
	 * @param xmlName The XML name of the item.
	 * @param index On output, index of the item.
	 * @return Return true on success.
	 */
	bool GetIndexByXMLName(wxString xmlName, wxUint32 &index)
	{
		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->xmlName == xmlName)
			{
				index = i;
				return true;
			}
		}
		
		return false;
	}
		

protected:

	/** Internal data structure. */
	struct ItemData
	{
		ItemID id;
		ItemCreatorFunc creator;
		wxString description;
		wxString xmlName;
	};
	

	/** List of registered items. */
	std::vector<ItemData*> mEntries;
};


typedef std::pair<int, int> DualID;


/**
 * This subclass specialize FactoryEngine to handle registration and creation
 * of items that must be identified by a pair of ID. In particular, functions
 * are added to access items by a so-called sub-index. A sub-index, is the index
 * of a particular item by considering ONLY items that have the SAME secondary ID.
 * For example, if the third item (index 2) has a secondary ID of XX, just like the
 * second item but different from the first item, the sub-index of the third item is
 * 1, because there is only one item before that has the same secondary ID.
 */
template<class BaseItem> class DualIDFactoryEngine : virtual public FactoryEngine<BaseItem, DualID>
{
public:

	/** Constructor. */
	DualIDFactoryEngine() {}
	
	/** Desctructor. */
	virtual ~DualIDFactoryEngine() {}
	
	/**
	 * Create an item.
	 * @param subIndex Sub-index of the item.
	 * @param secID Secondary index to consider.
	 * @return The created item.
	 */
	BaseItem* CreateItemWithSubIndex(wxUint32 subIndex, int secID)
	{
		wxUint32 index;
		if(GetIndexBySubIndex(subIndex, secID, index))
			return CreateItemWithIndex(index);
		else
			return NULL;
	}
	
	/** Returns the number of entry for a particular secondary ID. */
	wxUint32 GetEntryNumber(int secID)
	{
		wxUint32 num = 0;
		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->id.second == secID)
				++num;
		}
		return num;
	}
	
	/**
	 * Get an item's description.
	 * @param subIndex Sub-index of the item.
	 * @param secID Secondary ID to consider.
	 * @return The description.
	 */
	wxString GetDescriptionBySubIndex(wxUint32 subIndex, int secID)
	{
		wxUint32 index;
		if(GetIndexBySubIndex(subIndex, secID, index))
			return GetDescriptionByIndex(index);
		else
			return "";
	}
	
	/** Returns an ID from a sub-index and secondary ID. */
	DualID GetIDBySubIndex(wxUint32 subIndex, int secID)
	{
		wxUint32 index;
		if(GetIndexBySubIndex(subIndex, secID, index))
			return GetIDByIndex(index);
		else
			return std::make_pair(0, 0);
	}
	
	/**
	 * Get the sub-index of an item.
	 * @param id Item ID to consider.
	 * @param subIndex On output: the sub-index.
	 * @return Return true on success.
	 */
	bool GetSubIndexByID(DualID id, wxUint32 &subIndex)
	{
		wxUint32 si = 0;
		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->id.second == id.second)
			{
				if(mEntries[i]-> id == id)
				{
					subIndex = si;
					return true;
				}
				
				++si;
			}
		}
		return false;
	}
	
	/**
	 * Get an absolute index of an item.
	 * @param subIndex The item sub-index.
	 * @param secID The secondary ID to consider.
	 * @param index On output: the absolute index.
	 * @return Return true on success.
	 */
	bool GetIndexBySubIndex(wxUint32 subIndex, int secID, wxUint32 &index)
	{
		for(size_t i = 0; i < mEntries.size(); i++)
		{
			if(mEntries[i]->id.second == secID)
			{
				if(subIndex-- == 0)
				{
					index = i;
					return true;
				}
			}
		}
		
		return false;
	}

};


#endif

