/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <math.h>

#include "laplacefilter.h"

LaplaceFilter::LaplaceFilter(wxUint8 size, wxUint8 dimension, float *laplace) {
	mSize = size;
	if(!laplace) {
		mLaplace = new float[3];
		mLaplace[0] = -0.5f;
		mLaplace[1] =  1.0f;
		mLaplace[2] = -0.5f;
		mDimension  =  3;
	} else {
		mLaplace = laplace;
		mDimension = dimension;
	}

	mOutput = new float[mSize];
	mTemp = new float[mSize];
	for(wxUint8 i=0; i<mSize; i++) {
		mOutput[i] = 0.0f;
		mTemp[i] = 0.0f;
	}
}

LaplaceFilter::~LaplaceFilter() {
	delete [] mLaplace;
	delete [] mOutput;
	delete [] mTemp;
}

const float *LaplaceFilter::Apply(const wxByte *input) {
	/* Translate bytes to floats in temporary array */
	for(wxUint8 i=0; i<mSize; i++)
		mTemp[i] = input[i] / 255.0f;
	return Apply(mTemp);
}

const float *LaplaceFilter::Apply(const float *input) {
	wxUint8 i, j;
	float tmp;
	wxUint8 offset = mDimension / 2;

	/* Convolution  */
	for(i=offset; i < mSize - offset; i++) {
		tmp = 0.0f;
		for(j=0; j < mDimension; j++)
			tmp += input[i - offset + j] * mLaplace[j];
		mOutput[i] = tmp;
	}

	/* Copy the first value of the convolved image into the first cells */
	tmp = mOutput[offset];
	for(i=0; i < offset; i++)
		mOutput[i] = tmp;

	/* Copy the last value of the convolved image into the remaining cells */
	tmp = mOutput[mSize - offset - 1];
	for(i=mSize - offset; i < mSize; i++)
		mOutput[i] = tmp;

	/* Make negative contrasts positive */
	for(i=0; i < mSize; i++)
		mOutput[i] = fabs(mOutput[i]);

	return mOutput;
}
