/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

#ifndef _XMLARRAY_H
#define _XMLARRAY_H

/** Class to convert arrays of arbitrary dimensions from/to XML :
  * Accepted componenttypes :
  * - int,float,double,boolean,byte and their respective encapsulator classes,
  *   Strings and objects that accept XMLSerializable
 */

#include <wx/dynarray.h>
#include "xmlserializable.h"

class XMLAttribute;
class XMLSerializableArray;
class InstantiationAssistant;
class wxString;

WX_DEFINE_ARRAY(void*, PointerArray);
WX_DEFINE_ARRAY(XMLSerializable*, XMLSerializableArray);
WX_DEFINE_ARRAY(XMLAttribute*, XMLAttributeArray);

class XMLArray : public XMLSerializable
{ 
public:
	/** Make a new XMLArray (constructor for loading)
	  * @param parentName The XML name of the object that is going to save
	  *        this array.
	  * @returns The new XMLArray
	 */
	XMLArray();

	/** Empty constructor for loading 
	* @param parentName The XML name of the object that is going to load
	*        this array.
	*  @returns A new XMLArray which can be used for loading
	*/
	XMLArray(wxString *pName);

	/** Constructor for loading 
	* @param parentName The XML name of the object that is going to load
	*        this array.
	* @param assistant The IntantiationAssistant to use when creating new
			objects (this parameter is only meaningful for XMLArrays that
			directly or indirectly contain XMLSerializable objects.)
	*  @returns A new XMLArray which can be used for loading
	*/
	XMLArray(wxString *parentName, InstantiationAssistant *assistant);

	/** Make a new XMLArray (constructor for saving)
	* @param input The array,Vector or LinkedList to save
	* @param parentName The XML name of the object that is going to save
	*        this array.
	* @returns The new XMLArray
	*/
	XMLArray(void *input, wxString *parentName);

	/** Make a new XMLArray (constructor for saving)
	* @param input The array,Vector or LinkedList to save
	* @param parentName The XML name of the object that is going to save
	*        this array.
	* @param alias A short description of what this array contains. Mainly
	*        useful for debugging.
	* @returns The new XMLArray
	*/
	XMLArray(void *input, wxString *parentName, wxString *alias);

	/** Return the result array,Vector or LinkedList as an object. The XMLArray
	  *  has to be loaded first.
	  * @returns The array which has been parsed
	*/
	void*			GetResult();

	/** Return the result array,Vector or LinkedList as an object. The XMLArray
	  * has to be loaded first. First checks if the result class equals the
	  * parameter class.
	  * @param expectedClass The class with which the result has to match
	  * @throws XMLInvalidInputException If the XMLArray was not loaded, if
	  *         there is an internal problem or if the result class doesn't 
	  *         match the parameter class.
	  * @returns The array which has been parsed
	*/
	void* GetResult (wxUint32 expectedType);

	PointerArray* GetResultPointerArray();

	/** This method tries to guess the component type of the data array. If it 
	  * cannot determine the component type, then it assumes that all components
	  * are XMLSerializable. This assumption may be wrong, but it doesn't hurt
	  * since the saver later checks all components individually.
	  * @returns The component type of the data (the result is not 100% reliable)
	 */
	wxUint32 GetComponentType ();

	/** @returns The instantiation assistant of this XMLArray, or null if doesn't have one */
	InstantiationAssistant* GetInstantiationAssistant ();

	/** Sets the InstantiationAssistant of this XMLArray
	  * @param The new InstantiationAssistant
	 */
	void SetInstantiationAssistant (InstantiationAssistant *assistant);

	virtual wxString	GetXMLName();
	virtual void		SetParentName (const wxString &parentName);

	/** Save an arbitrary array/vector/LinkedList
	* @param array An array to save (Vectors and LinkedLists are also accepted)
	* @param parent The object which is trying to save this array
	* @param saver The XMLSaver to which the array has to be saved
	* @param description A short description of the array
	*/
	static void SaveArray (void* array, XMLSerializable *parent, XMLSaver *saver, wxString *description);

  
	/** Save an arbitrary array/vector/LinkedList
	  * @param array An array to save (Vectors and LinkedLists are also accepted)
	  * @param parent The object which is trying to save this array
	  * @param saver The XMLSaver to which the array has to be saved
	 */
	static void SaveArray (void* array, XMLSerializable *parent, XMLSaver *saver);

	/** Load an arbitrary array. Do not check the result.
	  * @param parent The object that is loading this array
	  * @param loader The XMLLoader to load this array from
	  * @returns The resulting array,vector or LinkedList
	 */
	static void* LoadArray(XMLSerializable *parent, XMLLoader *loader);

	/** Load an arbitrary array. Check if the result matches the expected class.
	  * @param parent The object that is loading this array
	  * @param loader The XMLLoader to load this array from
	  * @param expectedClass The expected class of the result
	  * @returns The resulting array,vector or LinkedList
	 */
	static void* LoadArray(XMLSerializable *parent, XMLLoader *loader, wxUint32 expectedType);

	/** Load an arbitrary array. Instantiate child objects via the 
	  * parameter InstantiationAssistant
	  * @param parent The object that is loading this array
	  * @param loader The XMLLoader to load this array from
	  * @param assistant The InstantiationAssistant to use
	  * @returns The resulting array,vector or LinkedList
	 */
	static void* LoadArray(XMLSerializable *parent, XMLLoader *loader, InstantiationAssistant *assistant);
	
	virtual XMLElement*	SaveSelf ();
	virtual void		SaveChilds(XMLSaver *saver);

protected:	
	// The data itself
/*
	void *data;
	Class dataClass;
*/
	wxUint32 containerType;
	// Name of parent
	wxString *parentName;
	// Short description of what this array contains. Mainly for debugging and error checking.
	wxString *alias;
	// InstantiationAssistant for creating objects in XMLArrayElements
	InstantiationAssistant *watson;
  
	const static wxUint32	lookUpTableLen=13;
	/*char *lookUpTable[] = { "I", "d", "F", "Z",
							"Integer", "Double",
								"Float",
								"Bool",
								"String",
								"L",
								"[",
								"B",
								"Byte"
							};*/
};

// Component types
/*enum wxInt32 {
		INT=1, 
		DOUBLE,
		FLOAT,
		BOOLEAN,
		INT_CLASS,
		DOUBLE_CLASS,
		FLOAT_CLASS,
		BOOLEAN_CLASS,
		STRING,
		XMLSERIALIZABLE,
		OTHER_ARRAY,
		BYTE,
		BYTE_CLASS
};*/

/*enum wxInt32 {
  		CONTAINER_ARRAY=0,
		CONTAINER_VECTOR=1,
		CONTAINER_LINKEDLIST=2
};*/

#endif
