/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

#include "wx/wxprec.h"
#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

#include <wx/string.h>
#include <wx/dynarray.h>
#include <wx/tokenzr.h>
#include <wx/arrimpl.cpp>
#include "xmlelement.h"
#include "xmlattribute.h"

XMLElement::XMLElement (const wxString &this_name) { 
	mName = new wxString(this_name);
	mAttributes=new XMLAttributeArray();
}

XMLElement::XMLElement (const wxString &this_name, XMLAttributeArray *attList) {
	mName = new wxString(this_name);
	mAttributes=new XMLAttributeArray();

	for (wxUint32 t=0; t<attList->GetCount(); t++)
		mAttributes->Add(attList->Item(t));
}

XMLElement::~XMLElement() {
	for(wxUint32 i=0; i<mAttributes->GetCount(); i++)
		delete mAttributes->Item(i);
	delete mAttributes;
	delete mName;
}

/* you -dont- delete */
wxString XMLElement::GetName () {
	return *mName;
}

/* you -dont- delete */
/*wxString* XMLElement::GetNameP () {
	return mName;
}*/
	  
void XMLElement::SetName(const wxString &this_name) {
	delete mName;
	mName = new wxString(this_name);
}

void XMLElement::AddAttribute (XMLAttribute* attribute){
	mAttributes->Add(attribute);
}

void XMLElement::RemoveAttribute (const wxString &name) { 
	mAttributes->RemoveAt(GetAttributeIndex(name));
}

wxInt32 XMLElement::GetAttributeIndex (const wxString &name) {
	wxString *pName = new wxString(name);
	for (wxUint32 t=0; t<mAttributes->GetCount(); t++) {
		wxString *aName = new wxString(((XMLAttribute *) (mAttributes->Item(t)))->GetName());
		if(pName->IsSameAs( aName->c_str() )) {
			delete pName; pName = NULL;
			delete aName; aName = NULL;
			return t;
		}
		delete aName; aName = NULL;
	}
	delete pName;
	return -1;
}

XMLAttribute* XMLElement::GetAttribute (const wxString &name) {
	return (XMLAttribute *)(mAttributes->Item(XMLElement::GetAttributeIndex(name)));
}

wxString XMLElement::ToString () {
	return "XML element named " + *mName;
}

void XMLElement::RemoveAllAttributes () {
	for(wxUint32 i=0; i<mAttributes->GetCount(); i++)
		delete mAttributes->Item(i);
	delete mAttributes;
	mAttributes=new XMLAttributeArray();
}

XMLAttributeArray* XMLElement::GetAttributesArray () {
	return mAttributes;
}

void XMLElement::SetAttribute (XMLAttribute* attribute) {
	if(wxNOT_FOUND != mAttributes->Index(attribute) ) 
		mAttributes->Remove(attribute);
	mAttributes->Add(attribute);
}

wxString* XMLElement::GetOpenTag () { 
	wxString result = "<"+(*mName);
	XMLAttributeArray* attlist = GetAttributesArray();
	for (wxUint32 t=0; t<attlist->GetCount(); t++)
		result += " "+(attlist->Item(t))->ToString();
	result += ">";
	return new wxString(result);
}  
	  
wxString* XMLElement::GetCloseTag() {
	return new wxString("</"+*mName+">");
}  

/* you delete result */
XMLElement* XMLElement::Parse(const wxString &string){
	wxASSERT_MSG(!string.StartsWith ("</"), "XMLElement.parse cannot handle close tags: "+string);
	wxASSERT_MSG((string.StartsWith("<") && string.Last()=='>'),"Invalid XML tag (no < or > ) :"+string);

	wxString attName	= "";
	wxString attValue	= ""; 
	
	wxStringTokenizer *sp=NULL, *eq=NULL;
	sp = new wxStringTokenizer(string.Mid(1)," >");
	wxASSERT_MSG(sp->HasMoreTokens(), "XML tag without valid name "+string);

	XMLElement* result=new XMLElement(sp->NextToken());
	while (sp->HasMoreTokens()) {
		wxString curSp = sp->NextToken();
		eq = new wxStringTokenizer(curSp,"=");
		wxASSERT_MSG(eq->CountTokens()==2, "Wrong attribute in XML tag :"+string);

		attName = eq->NextToken();
		attValue = eq->NextToken();
		wxASSERT_MSG(attValue.StartsWith("\""),"Wrong attribute value in XML tag (no quotes?) :"+attValue+"/"+string);

		wxString *value = new wxString();
		*value = attValue.AfterFirst('\"');
		*value = value->BeforeLast('\"');
		result->AddAttribute(new XMLAttribute(attName,*value,false));

		delete value; delete eq;
	}
	delete sp;
	return result;
}
