/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

#ifndef _XMLLOADER_H
#define _XMLLOADER_H

/** This class can restore the state of the program from an XML file. It also
  * provides methods which XMLSerializable objects can use to retrieve
  * their child objects.
 */

#include "xmlarray.h"

class XMLStack;
class XMLReader;
class XMLElement;
class XMLSerializable;

class XMLLoader {
public:
	/** Make a new XMLLoader which loads from a file
	  * @param file The file to load from
	 */     
	XMLLoader(const wxString &file);

	~XMLLoader();

	/** Load an XMLSerializable from the XML file
	  * @param parent The parent of this XMLSerializable
	  * @param object The XMLSerializable object itself
	 */
	void			Load (XMLSerializable *parent, XMLSerializable *object);

	/** Close this XMLLoader */
	void			Close();
  
	/** Load an XMLSerializableArray from the XML file
	  * @param parent The parent of the XMLSerializables
	  * @param o The enumeration of XML serializables
	 */
	void			LoadEnumeration (XMLSerializable *parent, XMLSerializableArray *a);  
  
	/** Get the next element from the file
	  * @param parent The XMLSerializable which is asking for the next element
	  * @return The next XMLElement
	 */    
	XMLElement*		GetNextElement(XMLSerializable *parent);
  
	/** Checks if an XMLSerializable can get another element from the file
	  * @param parent The XMLSerializable that wants to request the next element
	  * @return A boolean value which indicates if another element can be fetched.
	 */
	bool			HasNextElement(XMLSerializable *parent);
  
	/** Get the next element from the file if it has a specific name
	  * @param parent The XMLSerializable which is asking for the next element
	  * @param name The name which the next element has to have
	  * @return The next XMLElement
	 */    
	XMLElement*		GetNextElement(XMLSerializable *parent, const wxString &name);

	/** Checks if an XMLSerializable can get an element with a specific
	  * name from the file.
	  * @param parent The XMLSerializable that wants to request the next element
	  * @param name The requested name
	  * @return A boolean value which indicates if an element with such a name
	  *         can be fetched.
	 */
	bool			HasNextElement (XMLSerializable *parent, const wxString &name);
  
	/** @return The name of the XMLElement in the buffer, regardless of its
	  *         parent. If the buffer is empty, then the return value is "".
	 */
	wxString		GetNextElementName ();
  
protected:
	/** Fill the nextElement buffer if it is empty */
	void			LookAhead ();

	/** Tranfer the contents of the nextElement buffer to a XMLSerializable
	  * and fill the buffer.
	  * @return The XMLElement in the buffer
	 */
	XMLElement*		PullNextElement();

	XMLReader		*mReader;
	XMLElement		*mNextElement;
	XMLStack		*mStack;
};

#endif
