/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/

#include "wx/wxprec.h"
#ifndef WX_PRECOMP
	#include "wx/wx.h"
#endif

#include <wx/string.h>

#include "xmlstack.h"
#include "xmlutils.h"
#include "xmlserializable.h"
#include "xmlelement.h"

/** This class is used by the XMLLoader and the XMLSaver to remember where
  * the parser is saving/loading in the XML tree.
 */

/** Make a new XMLStack */
XMLStack::XMLStack () {
	mStringStack = new wxArrayString();
	mStringStack->Alloc(20);
	mBranchName = NULL;
	UpdateBranchName();
}

XMLStack::~XMLStack() {
	delete mStringStack;
	/*if(!mBranchName==NULL)
		delete mBranchName;*/
	mBranchName = NULL;
}

/** Standard stack push method */
void XMLStack::Push(const wxString x) {
	mStringStack->Add(x);
    //UpdateBranchName();
}

/** Standard stack pop method */
wxString XMLStack::Pop() {
	wxString last = mStringStack->Last();
	mStringStack->Remove(Size()-1);
    //UpdateBranchName();
    return last;
}

wxString XMLStack::Peek(const wxUint32 i) {
	return mStringStack->Item(i);
}

bool XMLStack::IsEmpty() {
	return Size()==0;
}  
  
/** @return the complete dot separated branchname  */
wxString* XMLStack::GetBranchName () { return mBranchName; }
  
/** Check if the branch name equals a certain XML tag name
  * @param s The full XML name 
  * @return A boolean indicating if the branch name equals the tag name
 */
bool XMLStack::IsCurrentParent (const wxString &s) {
	wxString *gNameB = XMLUtils::GetGenericName(*mBranchName);
	wxString *gNameS = XMLUtils::GetGenericName(s);
	bool result = (s.IsEmpty() && IsEmpty())
			|| mBranchName->IsSameAs(s.c_str())
			|| gNameB->IsSameAs(*gNameS);
	delete gNameB;
	delete gNameS;
	return result;
}

/** Check if a certain branch name is currenly at control in the stack
  * @param x The XMLSerializable
  * @return A boolean indicating if the XMLSerializable is at the top of the
  *         stack.
 */
bool XMLStack::IsCurrentParent (XMLSerializable *x) {
	return x==NULL ? IsCurrentParent("") : IsCurrentParent(x->GetXMLName());
}
  
/** Check if the branch name begins with a certain XML name
  * @param s The full XML name 
  * @return A boolean indicating if the branch name begins with the tag name
 */
bool XMLStack::IsARootParent (const wxString &s) {
	wxString *gNameB = XMLUtils::GetGenericName(*mBranchName, false);
	wxString *gNameS = XMLUtils::GetGenericName(s,false);
	bool result = (s.IsEmpty() && IsEmpty()) || s.Length()==0 
			|| mBranchName->StartsWith(s.c_str())
			|| gNameB->StartsWith(gNameS->c_str());
	delete gNameB;
	delete gNameS;
	return result;

}

/** Check if a certain branch name is at the root of the stack
  * @param x The XMLSerializable
  * @return A boolean indicating if the XMLSerializable is at the root of the
  *         stack.
 */
bool XMLStack::IsARootParent (XMLSerializable *x) {
	return x==NULL ? IsARootParent("") : IsARootParent(x->GetXMLName());
}

wxUint32 XMLStack::Size() {
	return mStringStack->GetCount();
}
 
/** Internal method to update the branch name if the stack has changed */
void XMLStack::UpdateBranchName () {
	wxString res = "";
	/* waarschijnlijk alleen maar wxStrings op deze stack, ipv XMLElements */
	if(Size()>0) {
		for(wxUint32 i=0; i<Size(); i++) {
			wxString elem = Peek(i);
			res += elem;
			//(&res)->Append(*elem);
			res += '.';
			//(&res)->Append(".");
		}
		res = (&res)->BeforeLast('.');
	}
	delete mBranchName;
	mBranchName = &res;
}
