/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#include <wx/wx.h>
#include <wx/file.h>
#include <wx/textfile.h>
#include <math.h>

#include "../status.h"
#include "../utils/randomtools.h"

#include "genome.h"

Genome::Genome(wxUint16 length, wxInt16 populationNum, wxInt16 individualNum) {
	mLength = length;
	mDna = new wxUint8[mLength];
	
	//Initialize randomly
	for(wxUint16 i=0; i<mLength; i++)
		mDna[i] = RandomTools::Flip(0.5);

	
	/* A fair (??) way to initialize randomly the genome. The problem is the following, e.g. synapse existence:
	*
	* As the sum of the bits equals the (mean) connectivity, and because the sum of random number is not anymore a random number (!), 
	* the connectivity of any generated networks is always equal to about 0.5, and the population standard deviation to the mean connectivity is really low.
	* Therefore in order to allow individual with lower (or higher) connectivity to be tested, (and perhaps to increase the standard deviation), the method below is proposed.
	*
	* The same comment hold for many parameters: obviously thresholds, sign...
	*
	* Anyway the random generator should be also checked.
	* 16 bits version, but 32 bits should be even better*/

	/*
	wxUint16 bitsNum=16;
	wxUint16 rest=mLength % bitsNum;
	wxUint16 i,j,k;
	wxUint16 number;
	wxUint16 max=pow(2,16)-1;
	//wxASSERT_MSG(0,"warning: bit string is generated with number");
	for ( i=0; i<mLength-bitsNum; i+=bitsNum){
		number=RandomTools::RandInt16(max);
		for ( j=0; j<bitsNum; j++){
			mDna[i+j]= number & 1 ;
			number >>=1;
		}
	}
	for ( k=i+j-bitsNum+1; k <mLength; k++){
		number=RandomTools::RandInt16(max);
		mDna[k]= number & 1 ;
		number >>=1;
	}*/
	
	
	
	/* Another possible approach.
	
	Let's say we want to sample the search space better by randomly choosing for each individual the mean
	number of "1" in the genome.
	*/
	
	/*
	float mean = rand() / (float) RAND_MAX;
	for(int i = 0; i < mLength; i++)
		mDna[i] = RandomTools::Flip(mean);
	*/

	mPopulationNum = populationNum;
	mIndividualNum = individualNum;
}

Genome::Genome(wxUint16 length, wxUint8 *dna, wxInt16 populationNum, wxInt16 individualNum) {
	mLength = length;
	mDna = dna;	
	mPopulationNum=populationNum;
	mIndividualNum=individualNum;
}

Genome::Genome(Genome *genome ,wxInt16 populationNum, wxInt16 individualNum) {
	wxUint8 *dna = genome->GetDna();
	mLength = genome->GetLength();
	mDna = new wxUint8[mLength];
	mPopulationNum=populationNum;
	mIndividualNum=individualNum;

	for(wxUint16 i=0; i<mLength; i++)
		mDna[i] = dna[i];
}

Genome::~Genome() {
	delete [] mDna;
}

wxInt8 Genome::DecodeInt(wxUint32& value, wxUint16 start, wxUint16 length) {
	wxUint32 base = 1;

	// Check if there will be no out of bound
	if(start+length > mLength)
		return 1;
		
	value = 0;
	for(int i=start; i<(start+length); i++) {
		if(mDna[i])
			value += base;
		base *= 2;
	}
	return 0;
}

wxInt8 Genome::DecodeFloat(float &value, wxUint16 start, wxUint16 length) {
	wxUint32 base = 1;
	wxUint16 finish = start + length;

	// Check if there will be no out of bound
	if(finish > mLength)
		return 1;
		
	value = 0.0f;
	for(int i=start; i<finish; i++) {
		if(mDna[i])
			value += base;
		base *= 2;
	}
	value /= base;
	return 0;
}

void Genome::SetLength(wxUint16 length) {
	// delete the last bitstring
	delete [] mDna;

	// set the new length
	mLength=length;
	// create the new bitstring
	mDna = new wxUint8[mLength];

	// Initialize randomly
	for(wxUint16 i=0; i<mLength; i++)
		mDna[i] = RandomTools::Flip(0.5);
}

wxInt8 Genome::WriteToFile(wxFile *file) {
	file->Write(wxString::Format("%d ", mLength));
	for(wxUint16 i=0; i<mLength; i++)
		file->Write(wxString::Format("%d ", mDna[i]));
	file->Write("\n", 1);
	return 0;
}

Genome* Genome::ReadFromFile(wxTextFile *file, wxUint16 curGenomeLength) {
	long unsigned temp;
	wxUint16 length;
	wxInt16 pos, idx;
	wxUint8 *dna;
	
	wxString line = file->GetNextLine();
	if(!line.Mid(0, pos = line.Index(' ')).ToULong(&temp)) {
		StatusLine("Genome - Couldn't read genome, no genome length available.");
		return NULL;
	}
	pos++;
	length = (wxUint16)temp;
	if(curGenomeLength != length) {
		StatusLine(wxString::Format("Genome - Genome length doesn't match current settings! (Current settings: %d, saved population: %d.)", curGenomeLength, length));
		return NULL;
	}
	
	dna = new wxUint8[length];
	
	for(int i=0; i<length; i++) {
		idx = line.Mid(pos).Index(' ');
		if(!line.Mid(pos, idx == wxNOT_FOUND ? wxSTRING_MAXLEN : idx).ToULong(&temp)) {
			StatusLine("Genome - Couldn't read genome, not enough bit values available.");
			delete dna;
			return NULL;
		}
		pos += idx + 1;
		dna[i] = (wxUint8)temp;
	}

	return new Genome(length, dna);
}
