/****************************************************************************
  goevo -- application for evolutionary robotics with neural controllers
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne &
  Adaptive Intelligence Lab, University Utrecht.

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    goevo; http://asl.epfl.ch/resources/evo/goevo.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne & 
    Adaptive Intelligence Lab, University Utrecht.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

  See docs/readme.txt and docs/gpl.txt
******************************************************************************/


#ifndef _GENOME_H
#define _GENOME_H

class wxTextFile;
class wxFile;

/**
 * A simple genome, consisting of one bitstring.
 * @see Individual
 * @see IReproduce
 * @see Truncation
 * @see IGenOperator
 * @see Crossover
 * @see Mutation
 */
class Genome
{
public:

	/**
	 * Genome constructor
	 * @param length The length of the bitstring.
	 */
	Genome(wxUint16 length, wxInt16 populationNum = -1, wxInt16 individualNum = -1);

	/**
	 * Genome constructor
	 * @param length The length of the bitstring.
	 * @param dna The bitstring.
	 */
	Genome(wxUint16 length, wxUint8 *dna, wxInt16 populationNum = -1, wxInt16 individualNum = -1);


	/**
	 * Genome copy constructor
	 * @param genome The genome to be copied.
	 */
	Genome(Genome *genome, wxInt16 populationNum = -1, wxInt16 individualNum = -1);

	/**
	 * Genome destructor
	 */
	~Genome();

	/* -------------------------------- ACCESSORS ---------------------------- */
	
	/**
	 * Accessor method.
	 * @return The bitstring of this genome.
	 */
	wxUint8 *GetDna() { return mDna; }

	/**
	 * Accessor method.
	 * @param dna The bitstring to link to the bitstring of this genome.
	 */
	void SetDna(wxUint8 *dna) { mDna = dna; }

	/**
	 * Accessor method.
	 * @param position is the position of the wanted element in the bitstring.
	 * @return The DNA value.
	 */
	wxUint8 GetValue(wxUint16 position) { return mDna[position]; }

	/**
	 * Accessor method
	 * @param position The position of the wanted element in the bitstring.
	 * @param value The value.
	 */
	void SetValue(wxUint16 position, wxUint8 value) { mDna[position] = value; }

	/**
	 * Decodes an integer from the bitstring.
	 * @param value Integer to store result in.
	 * @param start The position where begining to read the bitstring.
	 * @param length The length of the bistring which must be read.
	 * @return Error code (zero if ok).
	 */
	wxInt8 DecodeInt(wxUint32 &value, wxUint16 start, wxUint16 length);

	/**
	 * Decodes a float (between 0.0 and 1.0) from the bitstring.
	 * @param value Float to store result in.
	 * @param start The position where begining to read the bitstring.
	 * @param length The length which must be read from the bitstring.
	 * @return Error code (zero if ok).
	 */
	wxInt8 DecodeFloat(float &value, wxUint16 start, wxUint16 length);

	/**
	 * Accessor method.
	 * @return The length value.
	 */
	wxUint16 GetLength() { return mLength; }

	/**
	 * Modify the length of the bitstring.
	 * @param length The new length.
	 */
	void SetLength(wxUint16 length);
	
	/**
	 * Accessor method.
	 * @return The individual number.
	 */
	wxInt16 GetIndividualNum() {return mIndividualNum;}
	/**
	 * Accessor method.
	 * @return The population number.
	 */
	wxInt16 GetPopulationNum() {return mPopulationNum;}

	/* -------------------------------- I/O ---------------------------- */

	/** Writes genome to file. */
	wxInt8 WriteToFile(wxFile *file);

	static Genome* ReadFromFile(wxTextFile *file, wxUint16 curGenomeLength);

protected:
	wxUint8 *mDna; /**< Array of bits: the bitstring. */
	wxUint16 mLength; /**< Length of the bitstring. */
	wxInt16 mPopulationNum; /**<The genome belongs to the "mPopulationNum" population number>**/
	wxInt16 mIndividualNum; /**<The genome belongs to the "mPopulationNum" individual number>**/
};

#endif /* _GENOME_H */
